program CSocket;

{$APPTYPE CONSOLE}

uses
  Crt, Windows, SysUtils,
{$IFDEF WDOSX}
  WDosSock,
{$ENDIF WDOSX}
{$IFDEF WIN}
  WinSock,
{$ENDIF WIN}
  Sockets;

type
  TClientSocket = class
  private
    TcpClient: TTcpClient;
    procedure TcpClientConnect(Sender: TObject);
    procedure TcpClientError(Sender: TObject; SocketError: Integer);
  public
    class procedure Main;
    constructor Create;
    destructor Destroy; override;
    procedure Run;
  end;

constructor TClientSocket.Create;
begin
  inherited;
  // Create socket to connect to the Server
  TcpClient := TTcpClient.Create(nil);
  TcpClient.OnConnect := TcpClientConnect;
  TcpClient.OnError := TcpClientError;
end;

destructor TClientSocket.Destroy;
begin
  TcpClient.Free;
  inherited;
end;

class procedure TClientSocket.Main;
var
  ClientSocket: TClientSocket;
begin
  ClientSocket := TClientSocket.Create;
  try
    ClientSocket.Run;
  finally
    ClientSocket.Free;
  end;
end;

procedure TClientSocket.Run;
var
  Ch: Char;
begin
  TcpClient.RemoteHost := ParamStr(1);
  TcpClient.RemotePort := ParamStr(2);

  Write('Connecting to ' + TcpClient.RemoteHost + ':'
      + TcpClient.RemotePort + '... ');
  TcpClient.Open;

  if not TcpClient.Connected then Exit;
  Writeln('Ok');

  // Loop until we are disconnected. When the user types something,
  // we send it to the server. If the server sends something, we
  // show it on screen.
  while True do
  begin

    // Check if we are still online
    if TcpClient.ConnectionAborted then Break;

    // Any data to read from the server?
    while TcpClient.WaitForData do
    begin
      if TcpClient.ReceiveBuf(Ch, SizeOf(Ch)) > 0 then
        Write(Ch);
    end;

    // Something to send?
    if KeyPressed then
    begin
      Ch := ReadKey;
      if Ch = #27 then Break;
      if Ch = #13 then
      begin
        Ch := #10;
        TcpClient.SendBuf(Ch, SizeOf(Ch));
        Ch := #13;
      end;
      TcpClient.SendBuf(Ch, SizeOf(Ch));
    end;
    UpdateNetwork;
  end;

  TcpClient.Close;
end;

procedure TClientSocket.TcpClientConnect(Sender: TObject);
{$IFDEF WDOSX}
var
  ConnectionTimeout: Integer;
{$ENDIF WDOSX}
begin
{$IFDEF WDOSX}
  // Set timeout to 10 seconds
  ConnectionTimeout := 10000;
  SetSockOpt(TcpClient.Handle, SOL_SOCKET, SO_CONNTIMEOUT,
      ConnectionTimeout, SizeOf(ConnectionTimeout));
{$ENDIF WDOSX}
end;

procedure TClientSocket.TcpClientError(Sender: TObject; SocketError: Integer);
begin
  // write socket error number
  Writeln(Format('TcpClient Error: %d', [SocketError]));
end;

begin
  if (ParamCount <> 2) then
  begin
    WriteLn('Usage:   CSocket [IP | HostName] [Port]');
    WriteLn('Example: CSocket smtp.fibertel.com.ar 25');
    Exit;
  end;

  TClientSocket.Main;
end.

