program HttpServ;

{$APPTYPE CONSOLE}

uses
  Crt,
  SysUtils,
  Sockets,
  Threads,
{$IFDEF WDOSX}
  WDosInterrupts,
{$ENDIF}
  WebFS in 'WEBFS.PAS',
  WebSrv;

var
  MustRestart: Boolean;
  QuitServer: Boolean;

procedure HomePage(Server: TWebServer; Request: TWebRequest; Response:
  TWebResponse);
begin
  Response.Content :=
    '<HTML><TITLE>Dynamic Page</TITLE>' +
    '<H1>This is a Dynamic Page</H1>' +
    'Today it''s ' + DateTimeToStr(Now) + '<br>' +
    '</HTML>';
end;

procedure WebAdmin(Server: TWebServer; Request: TWebRequest; Response:
  TWebResponse);
begin
  Response.Content :=
    '<HTML><TITLE>Web Administration</TITLE>' +
    '<H1>Web Administration</H1>' +
    '<a href="/restart">Restart the Server</a><br />' +
    '<a href="/quit">Quit the Server</a>' +
    '</HTML>';
end;

procedure DoReset(Server: TWebServer; Request: TWebRequest; Response:
  TWebResponse);
begin
  QuitServer := True;
  MustRestart := True;
  Response.Content :=
    '<HTML><TITLE>Restart Server</TITLE>' +
    '<H1>Restart Server</H1>' +
    'Server is restarting...<BR><BR>' +
    '<a href="/">Go back to Home Page</a>' +
    '</HTML>';
end;

procedure DoQuit(Server: TWebServer; Request: TWebRequest; Response:
  TWebResponse);
begin
  QuitServer := True;
  MustRestart := False;
  Response.Content :=
    '<HTML><TITLE>Quit Server</TITLE>' +
    '<H1>Quit Server</H1>' +
    'HTTP Server is ending... <BR><BR>' +
    '<a href="/">Go back to Home Page</a>' +
    '</HTML>';
end;

procedure Run;
var
  Server: TWebServer;
begin
  // init reboot flag
  MustRestart := False;

  // Bring up a Web Server
  Write('Starting Web Server... ');
  Server := TWebServer.Create;
  try
    Server.RemoteHost := ''; // INADDR_ANY;
{$IFDEF WIN}
    Server.LocalPort := '8080';
{$ENDIF}
    Server.RootPath := '.\';
    Server.DefaultPage := '/pages/index.htm';

    Server.Start;
    if not Server.Active then
    begin
      WriteLn('Error starting Web Server.');
      Halt(1);
    end;
    WriteLn('Started.');
    Writeln('Port: ' + Server.LocalPort);
    Writeln;
    Writeln('Press <Esc> to quit ...');

    // Add Web Server Actions
    Server.AddAction('/index', HomePage);
    Server.AddAction('/webadmin', WebAdmin);
    Server.AddAction('/restart', DoReset);
    Server.AddAction('/quit', DoQuit);
    StartWebFS(Server, '/filesystem');

    // Loop until user presses <Esc>
    QuitServer := False;
    repeat
      UpdateNetwork;
    until QuitServer or (KeyPressed and (ReadKey = #27));
  finally
    Sleep(500);
    Server.Free;
  end;
end;

begin
  try
    Run;
{$IFDEF WDOSX}
    if MustRestart then Reboot(True);
{$ENDIF}
  except
    on e: Exception do
      Writeln(e.Message);
  end;
end.


