unit Unit1;

interface

{$IFDEF VCL}
uses
  Classes, Controls, StdCtrls, Menus, Forms, ExtCtrls, Dialogs;
{$ELSE}
uses
  SysUtils, Classes, QGraphics, QControls,
  QForms, QStdCtrls, QMenus, QDialogs, QExtCtrls,
  Threads, MyThreads;
{$ENDIF}

type
  TForm1 = class(TForm)
    lbxMessages: TListBox;
    btnThreadOne: TButton;
    btnThreadTwo: TButton;
    btnTerminate: TButton;
    btnClear: TButton;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnThreadOneClick(Sender: TObject);
    procedure btnThreadTwoClick(Sender: TObject);
    procedure btnTerminateClick(Sender: TObject);
    procedure lbxMessagesDblClick(Sender: TObject);
    procedure btnClearClick(Sender: TObject);
  private
    { Private declarations }
    ThreadOne: TThreadOne;
    ThreadTwo: TThreadTwo;
    ThreadOneMore: TThreadOne;
    procedure ThreadTerminate(Sender: TObject);
  public
    procedure LogMessage(const Text: string);
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

procedure TForm1.FormCreate(Sender: TObject);
begin
  // create suspended
  ThreadOne := TThreadOne.Create(True);
  ThreadOne.Name := 'Thread1';
  ThreadOne.OnTerminate := ThreadTerminate;

  // create not suspended
  ThreadTwo := TThreadTwo.Create(False);
  ThreadTwo.OnTerminate := ThreadTerminate;

  // test free on terminate
  ThreadOneMore := TThreadOne.Create(False);
  ThreadOneMore.FreeOnTerminate := True;
  ThreadOneMore.Name := 'Thread1.1';
  ThreadOneMore.OnTerminate := ThreadTerminate;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  ThreadOne.Free;
  ThreadTwo.Free;
end;

procedure TForm1.LogMessage(const Text: string);
begin
  lbxMessages.Items.Add(Text);
end;

procedure TForm1.ThreadTerminate(Sender: TObject);
begin
  ShowMessage(Sender.ClassName + ' terminated');
end;

procedure TForm1.btnThreadOneClick(Sender: TObject);
begin
  // test Suspend and Resume
  if ThreadOne.Suspended then
  begin
    ThreadOne.Resume;
    btnThreadOne.Caption := 'Suspend Thread&1';
  end
  else
  begin
    ThreadOne.Suspend;
    btnThreadOne.Caption := 'Resume Thread&1';
  end;
end;

procedure TForm1.btnThreadTwoClick(Sender: TObject);
begin
  // test Suspend and Resume
  if ThreadTwo.Suspended then
  begin
    ThreadTwo.Resume;
    btnThreadTwo.Caption := 'Suspend Thread&2';
  end
  else
  begin
    ThreadTwo.Suspend;
    btnThreadTwo.Caption := 'Resume Thread&2';
  end;
end;

procedure TForm1.btnTerminateClick(Sender: TObject);
begin
  // test Terminate with ThreadTwo
  ThreadTwo.Terminate;
end;

procedure TForm1.lbxMessagesDblClick(Sender: TObject);
begin
  lbxMessages.Clear;
end;

procedure TForm1.btnClearClick(Sender: TObject);
begin
  lbxMessagesDblClick(Self);
end;

end.
