unit MyThreads;

interface

uses
  SysUtils, Classes, Threads, SyncObjs;

type
  TThreadOne = class(TThread)
  private
    FName: string;
  protected
    procedure Execute; override;
  public
    property Name: string read FName write FName;
  end;

  TThreadTwo = class(TThread)
  private
    { Private declarations }
  protected
    procedure Execute; override;
  end;

var
  CriticalSection: TCriticalSection;

implementation

procedure DoCritical;
begin
  Writeln('Task ', GetCurrentTask, ' about to enter the Critical Section');
  CriticalSection.Enter;
  Writeln('Critical Section entered by Task ', GetCurrentTask);
  Sleep(50);
  CriticalSection.Leave;
  Writeln('Critical Section left by Task ', GetCurrentTask);
end;

{ ThreadOne }

procedure TThreadOne.Execute;
var
  P: Pointer;
  I: Integer;
begin
  for I := 1 to 500 do
  begin
    if Terminated then Break;
    GetMem(P, 30000);
    WriteLn(Name);
    Sleep(1000);
    DoCritical;
    FreeMem(P);
  end;
  ReturnValue := 42;
end;

{ ThreadTwo }

procedure TThreadTwo.Execute;
var
  P: Pointer;
begin
  repeat
    GetMem(P, 30000);
    WriteLn('Thread 2');
    Sleep(1000);
    FillChar(P^, 30000, 0);
    FreeMem(P);
    DoCritical;
  until Terminated;
end;

initialization
  CriticalSection := TCriticalSection.Create;

finalization
  CriticalSection.Free;

end.
