program ABS1;

{$APPTYPE CONSOLE}

{
  This program demonstrates how to avoid a compiler error with Delphi 6
  and deprcated absolute keyword.

  In older Delphi 2-5 versions there was an easy way for access of DOS
  real mode memory (first 1 MB from 0000:0000 - F000:FFFF) with WDosX
  by use of the deprecated absolute keyword. So your code was similar
  to:

  var
    ShiftState: Byte absolute $00417;
                     ^^^^^^^^
  This usage of keyword "absolute" is dropped in Delphi 6 onwards for
  compatibility with Kylix and position independent code (PIC) in
  Linux shared objects (.so). For more informations refer to
  Delphi and Kylix online help.

  Until now there are two workarounds to fix older Delphi WDosX code.
  Both versions use a byte pointer, pointing at the desired memory
  address.

  - procedure WritelnShiftState uses this pointer direct
    (so you need to use the pointer dereferencer '^') ,
  - function ShiftState wraps the pointer inside a function
    (so you can only read the byte value),
}


uses
  SysUtils;

// <-- here you must use the modified SysUtils from DWPL 2.0

procedure WritelnShiftState;
const
  ShiftState: PByte = PByte($00417);
var
  B: Byte;
begin
  B := ShiftState^;
  Writeln(B);
end;

function ShiftState: Byte;
begin
  Result := PByte($00417)^;
end;

begin
  Writeln(ShiftState);
  WritelnShiftState;
end.
