program MVM1;

{$APPTYPE CONSOLE}

{
  This program demonstrates how to avoid a compiler error with Delphi 6
  while using Mem[$xxx] as var parameter e.g. with the Move function.

  In older Delphi 2-5 versions there was an easy way for access of DOS
  real mode memory (first 1 MB from 0000:0000 - F000:FFFF) with WDosX
  by use of the deprecated absolute keyword. There you declared
  an global array:

  var
    Mem: array[$00000..$FFFFF] of Byte absolute 0;
                                       ^^^^^^^^
  This usage of keyword "absolute" is dropped in Delphi 6 onwards for
  compatibility with Kylix and position independent code (PIC) in
  Linux shared objects (.so). For more informations refer to
  Delphi and Kylix online help.

  The Mem array is now a global instance with a default property. But
  this is not the same to the old Mem array, e.g. you can not us
  an element as a var parameter.

  Until now there are two known workarounds to fix older
  Delphi WDosX code.

  1. This version uses a pointer, pointing to
     the desired memory address, so you need to use the
     pointer dereferencer '^').

  2. Second versions also uses pointers, but defines a pointer to a
     memory array. So you only need to add two characters after
     Mem (Mem -> MemP^) with your old Delphi 5 code.
     Tip from Laszlo Agocs.
}

uses
  SysUtils; // <-- here you must use the modified SysUtils from DWPL 2.0

type
  PMem = ^TMem;
  TMem = array[0..$0FFFFF] of Byte;

const
  MemP: PMem = Pointer(0);

var
  B: Byte;
begin
  // old, D2-5:
  // Move(Mem[$00417], B, 1);

  // new 1) D6-up:
  Move(Pointer($00417)^, B, 1);
  Writeln(B);

  // new 2) D6-up:
  Move(MemP^[$00417], B, 1);
  Writeln(B);
end.
