
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Library (NetCLX)           }
{ FtpSrv.pas                                            }
{                                                       }
{ for Delphi 5, 6, 7                                    }
{ Version 2.6.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   www:      http://www.dwp42.org                      }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/demos/rtl/wdosx/sockets/ftpsrv/FtpSrv.pas,v 1.5 2003/12/01 20:43:20 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: FtpSrv.pas,v $
  Revision 1.5  2003/12/01 20:43:20  iwache
  Call to function BlockWriteEx replaced by use of file streams
  for sending and receiving files. This fix the bug with D5 and
  speeds up data transfer for file upload.

  Revision 1.4  2003/10/22 20:12:19  iwache
  Conditional compilation settings for compatibilty
  with Delphi 5 added by use of the JEDI.INC include file.

  ----------------------------------------------------------------------------
}
unit FtpSrv;

{$INCLUDE ..\..\..\..\..\source\rtl\sys\JEDI.INC}
{$I-}

interface

uses
  SysUtils, Classes, WDosSock, WDosIP, WDosNetwork, WDosUtil,
  WDosClock, Threads;

type
  TFTPServer = class;
  TFTPConnectionFlags = (ftpBinary);
  TFTPConnectionStatus = (stBanner, stWaitingUser, stWaitingPassword,
    stWaitingCommand, stSendingDir, stSendingList, stReceivingFile,
    stSendingFile, stDone);

  TFTPConnection = class
  private
    FCanWrite: Boolean;
    FWorkTask: Cardinal;
    FServer: TFTPServer;
    FCtrlSocket: Integer; // Control Connection
    FDataSocket: Integer; // Data Connection
    FCloseTimeout: Int64;
    FLastComm: string[96];
    FUserName: AnsiString;
    FRootPath: AnsiString;
    FSearchRec: TSearchRec;
    FCurrentPath: AnsiString;
    FDataPort: TSockAddr; // Remote port information
    FRemotePeer: TSockAddr; // Remote peer information
    FStorFileStream: TFileStream;
    FRetrFileStream: TFileStream;
    FCurrentFileName: AnsiString;
    FStatus: TFTPConnectionStatus;
    FFlags: set of TFTPConnectionFlags;
    procedure ProcessControlLine(Line: AnsiString);
    procedure AdjustPath(var Path: AnsiString);
    function UnixFileName(Path: AnsiString): AnsiString;
  protected
    procedure Activity;
    procedure SendBanner;
    procedure SendDirBlock;
    procedure SendListBlock;
    procedure SendFileBlock;
    procedure ReceiveFileBlock;
    procedure CheckControlSocket;
    function OpenDataConnection: Boolean;
    procedure SendResponse(Response: AnsiString);
    //
    procedure Proc_Pwd(Params: AnsiString);
    procedure Proc_Cwd(Params: AnsiString);
    procedure Proc_Mkd(Params: AnsiString);
    procedure Proc_Rmd(Params: AnsiString);
    procedure Proc_Help(Params: AnsiString);
    procedure Proc_SiteHelp(Params: AnsiString);
    procedure Proc_User(Params: AnsiString);
    procedure Proc_Pass(Params: AnsiString);
    procedure Proc_Port(Params: AnsiString);
    procedure Proc_Type(Params: AnsiString);
    procedure Proc_RnFr(Params: AnsiString);
    procedure Proc_Rnto(Params: AnsiString);
    procedure Proc_List(Params: AnsiString);
    procedure Proc_Nlst(Params: AnsiString);
    procedure Proc_Stor(Params: AnsiString);
    procedure Proc_Dele(Params: AnsiString);
    procedure Proc_Retr(Params: AnsiString);
    procedure Proc_Site(Params: AnsiString);
    procedure Proc_Size(Params: AnsiString);
    procedure Proc_Quit(Params: AnsiString);
    procedure Proc_Abor(Params: AnsiString);
  public
    constructor Create(Server: TFTPServer; Socket: Integer; RemotePeer:
      TSockAddr; RootPath: AnsiString);
    destructor Destroy; override;
    procedure Work;

    property Status: TFTPConnectionStatus read FStatus;
  end;

  TFTPServer = class
  private
    FSocket: Integer;
    FUserList: TList;
    FConnections: TList;
    FWorkTask: Cardinal;
    FRootPath: AnsiString;
  protected
    procedure AddConnection(Socket: Integer; Addr: TSockAddr);
    function ValidateUser(UserName, Password: AnsiString; var CanWrite:
      Boolean): Boolean;
  public
    constructor Create;
    destructor Destroy; override;
    function Start(Address: TIPAddress; Port: WORD; RootPath: AnsiString):
      Boolean;
    procedure AddUser(UserName, Password: AnsiString; CanWrite: Boolean);
    procedure Work;
  end;

implementation

const
  CLOSE_TIMEOUT = 60000 * 5;
  LIST_ITEMCOUNT = 10;

  // 100
  MSG_ALREADYCONNECT = '125 Already connected. Starting Transfer.';
  MSG_CONNECTING = '150 Opening data connection.';

  // 200
  MSG_NOOP = '200 Ok.';
  MSG_SYST = '215 DOS WDOSX/Delphi.';
  MSG_SIZE = '213 %d';
  MSG_PORTOK = '200 PORT command successful.';
  MSG_MKDOK = '200 Directory %s created.';
  MSG_TYPEOK = '200 Type set to %s.';
  MSG_BANNER =
    '220 DWPL Delphi WDOSX/DOS FTP Server ready. September 2001, Diego De Marco';
  MSG_BYE = '221 Goodbye.';
  MSG_ENDDIR = '226 Transfer completed.';
  MSG_TXOK = '226 Transfer completed.';
  MSG_RXOK = '226 Transfer completed.';
  MSG_ABORT = '226 Operation aborted.';
  MSG_LOGGED = '230-DWPL Delphi WDOSX/DOS FTP Server by Diego De Marco, September 2001'
    + CRLF +
    '230-Original version by Xon Electronics SRL (www.xonelectronics.it).' + CRLF
      +
    '230-Comments/bugs to www.dpsautomation.com.' + CRLF +
    '230 User %s logged in.';
  MSG_DELEOK = '250 File %s deleted.';
  MSG_RMDOK = '250 Directory %s removed.';
  MSG_RNTOOK = '250 File %s renamed to %s.';
  MSG_PWDOK = '257 "%s" is current directory (%d bytes free).';

  // 300
  MSG_NEEDPASS = '331 User %s accepted. Password needed.';
  MSG_RNFROK = '350 File %s exists. Issue RNTO next.';

  // 400
  MSG_LOWMEM = '421 Insufficient resources. Please try again later.';
  MSG_CANTCONNECT = '425 Cannot open data connection.';
  MSG_CONNBROKEN = '446 Connection closed. Transfer aborted.';

  // 500
  MSG_BADCMD = '500 Unknown command "%s".';
  MSG_BADTYPE = '501 Unknown data type "%s".';
  MSG_BADPORT = '501 Bad PORT command. Usage: PORT n,n,n,n, n,n.';
  MSG_NOTIMPLEMENTED = '502 Command "%s" not implemented yet.';
  MSG_USERFIRST = '503 You are not logged in. Use USER command first.';
  MSG_NOTLOGGED =
    '530 You are not logged in. Use USER and PASS commands first.';
  MSG_ALREADYLOGGED = '550 Already Logged-in. Disconnect first.';
  MSG_BADLOGIN = '550 Bau username or password. User %s cannot log in.';
  MSG_NOPERM = '550 Access denied.';
  MSG_CANTOPEN = '550 Cannot open %s.';
  MSG_BADDELE = '550 Cannot delete file %s.';
  MSG_BADRNTO = '550 Cannot rename %s to %s.';
  MSG_WRITEERR = '552 Error writing file %s. Transfer aborted.';
  MSG_NODIR = '553 Path %s does not exist or cannot be accessed.';
  MSG_CANTMAKE = '553 Cannot create %s.';

  // Site
  MSG_SITE_HELP_HELP = '214 Syntax: SITE HELP [ <sp> <string>]';

type
  PUSerEntry = ^TUSerEntry;
  TUserEntry = record
    UserName: ShortString;
    Password: ShortString;
    CanWrite: Boolean;
  end;

var
  ServerList: TList;

procedure FTPServerTask(Param: Pointer); forward;
procedure FTPConnectionTask(Param: Pointer); forward;

{ TFTPConnection }

constructor TFTPConnection.Create(Server: TFTPServer; Socket: Integer;
  RemotePeer: TSockAddr; RootPath: AnsiString);
begin
  inherited Create;
  FServer := Server;
  FFlags := [];
  FLastComm := '';
  FCtrlSocket := Socket;
  FDataSocket := 0;
  FRemotePeer := RemotePeer;
  FCloseTimeout := GetMSCounter + CLOSE_TIMEOUT;
  FUserName := '';
  FillChar(FDataPort, SizeOf(FDataPort), 0);
  FRootPath := GoodDir(RootPath);
  FCurrentPath := GoodDir(FRootPath);
  FCurrentFileName := '';
  FStatus := stBanner;
  FWorkTask := CreateTask(FTPConnectionTask, Self, 32768);
end;

destructor TFTPConnection.Destroy;
begin
  if (FStatus <> stDone) then
    SendResponse(MSG_BYE);
  if FDataSocket <> 0 then
    CloseSocket(FDataSocket);
  if FCtrlSocket <> 0 then
    CloseSocket(FCtrlSocket);
  FRetrFileStream.Free;
  FStorFileStream.Free;
  TerminateTask(FWorkTask);
  inherited Destroy;
end;

procedure TFTPConnection.Work;
begin
  case FStatus of
    stDone: Exit;
    stBanner: SendBanner;
    stWaitingUser: ;
    stWaitingPassword: ;
    stWaitingCommand: ;
    stSendingDir: SendDirBlock;
    stSendingList: SendListBlock;
    stReceivingFile: ReceiveFileBlock;
    stSendingFile: SendFileBlock;
  end;
  // Check for inactivity Timeout, and test connections
  if GetMSCounter > FCloseTimeout then
  begin
    FStatus := stDone;
    //		WriteLn('Closed due to inactivity.');
  end
  else
  begin
    CheckControlSocket;
  end;
end;

procedure TFTPConnection.SendBanner;
begin
  SendResponse(MSG_BANNER);
  FStatus := stWaitingUser;
end;

procedure TFTPConnection.SendDirBlock;
var
  i: integer;
  DT: TDateTime;
  Line: AnsiString;
begin
  Activity;
  Line := '';
  for i := 1 to LIST_ITEMCOUNT do
  begin
    if (FSearchRec.Name <> '.') and (FSearchRec.Name <> '..') then
    begin
      // Send Directory Entry
      DT := FileDateToDateTime(FSearchRec.Time);
      if (FSearchRec.Attr and faDirectory) = faDirectory then
        Line := Line + 'd'
      else
        Line := Line + '-';
{$IFDEF COMPILER6_UP}
  {$WARN SYMBOL_PLATFORM OFF}
{$ENDIF}
      if FSearchRec.Attr and faReadOnly = faReadOnly then
        Line := Line + 'r--r--r--'
      else
        Line := Line + 'rw-rw-rw-';
{$IFDEF COMPILER6_UP}
  {$WARN SYMBOL_PLATFORM ON}
{$ENDIF}
      Line := Line +
        '   1 owner    group    ' +
        PadL(IntToStr(FSearchRec.Size), 12, ' ') + ' ' +
        ShortMonthNames[Month(DT)] + ' ' +
        PadL(IntToStr(Day(DT)), 2, ' ') + ' ' +
        PadL(IntToStr(Hour(DT)), 2, ' ') + ':' +
        PadL(IntToStr(Minute(DT)), 2, '0') + ' ' +
        LowerCase(FSearchRec.Name) +
        CRLF;
    end;
    // Search for Next Entry
    if (FStatus <> stSendingDir) or (FindNext(FSearchRec) <> 0) then
    begin
      // No more entries to list, or connection lost
      FindClose(FSearchRec);
      if stSendingDir = stSendingDir then
      begin
        // End of list
        FStatus := stWaitingCommand;
        Send(FDataSocket, Line[1], Length(Line), 0);
        SendResponse(MSG_ENDDIR);
      end;
      if (FDataSocket <> FCtrlSocket) then
      begin
        CloseSocket(FDataSocket);
        FDataSocket := 0;
      end;
      Exit;
    end;
  end;
  Send(FDataSocket, Line[1], Length(Line), 0);
end;

procedure TFTPConnection.SendListBlock;
var
  I: Integer;
  Line: AnsiString;
begin
  Activity;
  Line := '';
  for i := 1 to LIST_ITEMCOUNT do
  begin
    if (FSearchRec.Name <> '.') and ((FSearchRec.Attr and faDirectory) = 0) then
    begin
      // Send List Entry.
      Line := Line + LowerCase(FSearchRec.Name) + CRLF;
    end;
    // Search for Next entry
    if (FStatus <> stSendingList) or (FindNext(FSearchRec) <> 0) then
    begin
      // No more entries to list, or we are disconnected.
      FindClose(FSearchRec);
      if FStatus = stSendingList then
      begin
        // End of list
        Send(FDataSocket, Line[1], Length(Line), 0);
        SendResponse(MSG_ENDDIR);
        FStatus := stWaitingCommand;
      end;
      if (FDataSocket <> FCtrlSocket) then
      begin
        CloseSocket(FDataSocket);
        FDataSocket := 0;
      end;
      Exit;
    end;
  end;
  Send(FDataSocket, Line[1], Length(Line), 0);
end;

procedure TFTPConnection.SendFileBlock;
var
  Len: LongInt;
  Close: Boolean;
  Buffer: array[0..8192] of Char;
begin
  Activity;
  Close := False;
  Len := FRetrFileStream.Read(Buffer, SizeOf(Buffer));
  if Send(FDataSocket, Buffer, Len, 0) < Len then
  begin
    SendResponse(MSG_CONNBROKEN);
    Close := True;
  end
  else
  begin
    if Len < SizeOf(Buffer) then
    begin
      SendResponse(MSG_TXOK);
      Close := True;
    end;
  end;
  if Close then
  begin
    FreeAndNil(FRetrFileStream);
    FStatus := stWaitingCommand;
    if FDataSocket <> FCtrlSocket then
    begin
      CloseSocket(FDataSocket);
      FDataSocket := 0;
    end;
  end;
end;

procedure TFTPConnection.ReceiveFileBlock;
var
  Len: LongInt;
  Close: Boolean;
  Buffer: AnsiString;
begin
  Close := False;
  ioctlsocket(FDataSocket, FIONREAD, @Len);
  if Len > 0 then
  begin
    Activity;
    SetLength(Buffer, Len);
    SetLength(Buffer, Recv(FDataSocket, Buffer[1], Len, 0));
    if FStorFileStream.Write(Buffer[1], Len) <> Len then
    begin
      // file stream write error
      FreeAndNil(FStorFileStream);
      DeleteFile(FCurrentFileName);
      SendResponse(Format(MSG_WRITEERR, [UnixFileName(FCurrentFileName)]));
      Close := True;
    end;
  end
  else
  begin
    // Queda algo ms por leer, o el cliente cerr el Socket?
    if recv(FDataSocket, buffer, 0, 0) < 0 then
    begin
      FreeAndNil(FStorFileStream);
      SendResponse(MSG_RXOK);
      Close := True;
    end;
  end;
  if Close then
  begin
    FCurrentFileName := '';
    if FDataSocket <> FCtrlSocket then
    begin
      CloseSocket(FDataSocket);
      FDataSocket := 0;
    end;
    FStatus := stWaitingCommand;
  end;
end;

procedure TFTPConnection.CheckControlSocket;
var
  Len: LongInt;
  Line: AnsiString;
begin
  // Still Connected?
  if Recv(FCtrlSocket, Len, 0, 0) < 0 then
  begin
    //		WriteLn('Closed due to error in Recv()');
    FStatus := stDone;
    Exit;
  end;
  // Any commands to read?
  ioctlsocket(FCtrlSocket, FIONREAD, @len);
  if Len > 0 then
  begin
    Activity;
    SetLength(Line, Len);
    Len := recv(FCtrlSocket, Line[1], Len, 0);
    SetLength(Line, Len);
    ProcessControlLine(Line);
  end;
end;

procedure TFTPConnection.SendResponse(Response: AnsiString);
begin
  if FCtrlSocket = 0 then
    Exit;
  Response := Response + CRLF;
  //  WriteLn(Response);
  Send(FCtrlSocket, Response[1], Length(Response), 0);
end;

procedure TFTPConnection.ProcessControlLine(Line: AnsiString);
var
  i: Integer;
  Cmd, Params: AnsiString;
begin
  //	WriteLn(Line);
   // Separate Command and Params
  i := Pos(' ', Line);
  if i = 0 then
  begin
    Cmd := UpperCase(Trim(Line));
    Params := '';
  end
  else
  begin
    Cmd := UpperCase(Trim(Copy(Line, 1, i - 1)));
    Params := Trim(Copy(Line, i + 1, Length(Line)));
  end;

  // Check Command
  if Cmd = 'ABOR' then
    Proc_Abor(Params)
  else if Cmd = 'CDUP' then
    Proc_Cwd('..')
  else if Cmd = 'CWD' then
    Proc_Cwd(Params)
  else if Cmd = 'DELE' then
    Proc_Dele(Params)
  else if Cmd = 'HELP' then
    Proc_Help(Params)
  else if Cmd = 'LIST' then
    Proc_List(Params)
  else if Cmd = 'MKD' then
    Proc_Mkd(Params)
  else if Cmd = 'NLST' then
    Proc_Nlst(Params)
  else if Cmd = 'NOOP' then
    SendResponse(MSG_NOOP)
  else if Cmd = 'PASS' then
    Proc_Pass(Params)
  else if Cmd = 'PORT' then
    Proc_Port(Params)
  else if Cmd = 'PWD' then
    Proc_Pwd(Params)
  else if Cmd = 'QUIT' then
    Proc_Quit(Params)
  else if Cmd = 'RETR' then
    Proc_Retr(Params)
  else if Cmd = 'RMD' then
    Proc_Rmd(Params)
  else if Cmd = 'RNFR' then
    Proc_RnFr(Params)
  else if Cmd = 'RNTO' then
    Proc_Rnto(Params)
  else if Cmd = 'SITE' then
    Proc_Site(Params)
  else if Cmd = 'SIZE' then
    Proc_Size(Params)
  else if Cmd = 'STOR' then
    Proc_Stor(Params)
  else if Cmd = 'SYST' then
    SendResponse(MSG_SYST)
  else if Cmd = 'TYPE' then
    Proc_Type(Params)
  else if Cmd = 'USER' then
    Proc_User(Params)
  else if Cmd = 'XCWD' then
    Proc_Cwd(Params)
  else if Cmd = 'XMKD' then
    Proc_Mkd(Params)
  else if Cmd = 'XPWD' then
    Proc_Pwd(Params)
  else if Cmd = 'XRMD' then
    Proc_Rmd(Params)
  else
    SendResponse(Format(MSG_NOTIMPLEMENTED, [Cmd]));
end;

procedure TFTPConnection.Activity;
begin
  FCloseTimeout := GetMSCounter + CLOSE_TIMEOUT;
end;

procedure TFTPConnection.AdjustPath(var Path: AnsiString);
var
  i: Integer;
begin
  while (Length(Path) > 1) and (Path[Length(Path)] in ['\', '/']) do
    SetLength(Path, Length(Path) - 1);
  for i := 1 to Length(Path) do
    if Path[i] = '/' then
      Path[i] := '\';
  if Copy(Path, 1, 1) = '\' then
    Path := GoodDir(FRootPath) + Copy(Path, 2, Length(Path))
  else
  begin
    Path := GoodDir(FCurrentPath) + Path;
  end;
  Path := LowerCase(Path);
  Path := ExpandFileName(Path);
end;

function TFTPConnection.UnixFileName(Path: AnsiString): AnsiString;
var
  i: integer;
begin
  Path := LowerCase(Path);
  if copy(Path, 2, 1) = ':' then
    Delete(Path, 1, 2);
  for i := 1 to Length(Path) do
  begin
    if Path[i] = '\' then
      Path[i] := '/';
    if (Length(Path) > 1) and (Copy(Path, Length(Path), 1) = '/') then
      Delete(Path, Length(Path), 1);
  end;
  Result := Path;
end;

function TFTPConnection.OpenDataConnection: Boolean;
begin
  if FDataSocket <> 0 then
  begin
    // Already connected
    SendResponse(MSG_ALREADYCONNECT);
    Result := True;
  end
  else if FDataPort.sin_addr.S_addr = 0 then
  begin
    // No PORT issued
    SendResponse(MSG_ALREADYCONNECT);
    FDataSocket := FCtrlSocket;
    Result := True;
  end
  else
  begin
    // Port ISSUED. Connect.
    SendResponse(MSG_CONNECTING);
    FDataPort.sin_family := AF_INET;
    FDataSocket := Socket(AF_INET, SOCK_STREAM, 0);
    Result := (Connect(FDataSocket, FDataPort, SizeOf(FDataPort)) = 0);
    if not Result then
    begin
      SendResponse(MSG_CANTCONNECT);
      CloseSocket(FDataSocket);
      FDataSocket := 0;
    end;
  end;
end;

procedure TFTPConnection.Proc_Pwd(Params: AnsiString);
var
  Drive: Integer;
begin
  if FStatus < stWaitingCommand then
  begin
    SendResponse(MSG_NOTLOGGED);
  end
  else
  begin
    if Pos(':', FCurrentPath) > 0 then
      Drive := Ord(UpCase(FCurrentPath[1])) - 64
    else
    begin
      Drive := 0;
    end;
    SendResponse(Format(MSG_PWDOK, [UnixFilename(FCurrentPath),
      DiskFree(Drive)]));
  end;
end;

procedure TFTPConnection.Proc_Cwd(Params: AnsiString);
begin
  if FStatus < stWaitingCommand then
  begin
    SendResponse(MSG_NOTLOGGED);
  end
  else
  begin
    AdjustPath(Params);
    if DirectoryExists(Params) then
    begin
      FCurrentPath := Params;
      Proc_Pwd('');
    end
    else
    begin
      SendResponse(Format(MSG_NODIR, [UnixFileName(Params)]));
    end;
  end;
end;

procedure TFTPConnection.Proc_Mkd(Params: AnsiString);
begin
  if FStatus < stWaitingCommand then
  begin
    SendResponse(MSG_NOTLOGGED);
  end
  else
  begin
    AdjustPath(Params);
    if CreateDir(Params) then
      SendResponse(Format(MSG_MKDOK, [UnixFileName(Params)]))
    else
    begin
      SendResponse(Format(MSG_CANTMAKE, [UnixFileName(Params)]));
    end;
  end;
end;

procedure TFTPConnection.Proc_Rmd(Params: AnsiString);
begin
  if FStatus < stWaitingCommand then
  begin
    SendResponse(MSG_NOTLOGGED);
  end
  else
  begin
    AdjustPath(Params);
    if RemoveDir(Params) then
      SendResponse(Format(MSG_RMDOK, [UnixFileName(Params)]))
    else
    begin
      SendResponse(Format(MSG_NODIR, [UnixFileName(Params)]));
    end;
  end;
end;

procedure TFTPConnection.Proc_Help(Params: AnsiString);
begin
  SendResponse(
    '214-Valid commands are:' + CRLF +
    '   ABOR' + CRLF +
    '   CDUP' + CRLF +
    '   CWD' + CRLF +
    '   DELE' + CRLF +
    '   HELP' + CRLF +
    '   LIST' + CRLF +
    '   MKD' + CRLF +
    '   NLST' + CRLF +
    '   NOOP' + CRLF +
    '   PASS' + CRLF +
    '   PORT' + CRLF +
    '   PWD' + CRLF +
    '   QUIT' + CRLF +
    '   RETR' + CRLF +
    '   RMD' + CRLF +
    '   RNFR' + CRLF +
    '   RNTO' + CRLF +
    '   SITE' + CRLF +
    '   SIZE' + CRLF +
    '   STOR' + CRLF +
    '   SYST' + CRLF +
    '   TYPE' + CRLF +
    '   USER' + CRLF +
    '   XCWD' + CRLF +
    '   XMKD' + CRLF +
    '   XPWD' + CRLF +
    '   XRMD' + CRLF +
    '214  HELP command successful');
end;

procedure TFTPConnection.Proc_SiteHelp(Params: AnsiString);
begin
  Params := UpperCase(Trim(Params));
  if Params = '' then
  begin
    SendResponse(
      '214-Valid commands are:' + CRLF +
      '   HELP' + CRLF +
      '214  HELP command successful');
  end
  else if Params = 'HELP' then
  begin
    SendResponse(MSG_SITE_HELP_HELP);
  end
  else
  begin
    SendResponse(Format(MSG_NOTIMPLEMENTED, [Params]));
  end;
end;

procedure TFTPConnection.Proc_User(Params: AnsiString);
begin
  if FStatus = stWaitingUser then
  begin
    FUserName := UpperCase(Params);
    SendResponse(Format(MSG_NEEDPASS, [FUserName]));
    FStatus := stWaitingPassword;
  end
  else
  begin
    SendResponse(MSG_ALREADYLOGGED);
  end;
end;

procedure TFTPConnection.Proc_Pass(Params: AnsiString);
begin
  if FStatus < stWaitingPassword then
    SendResponse(MSG_USERFIRST)
  else if FStatus > stWaitingPassword then
    SendResponse(MSG_ALREADYLOGGED)
  else if not FServer.ValidateUser(FUserName, Params, FCanWrite) then
  begin
    SendResponse(Format(MSG_BADLOGIN, [FUserName]));
    FStatus := stWaitingUser;
  end
  else
  begin
    SendResponse(Format(MSG_LOGGED, [FUserName]));
    FStatus := stWaitingCommand;
  end;
end;

procedure TFTPConnection.Proc_Port(Params: AnsiString);
begin
  // The incomming string is something like "AAA,BBB,CCC,DDD,EEE,FFF"
  // where AAA..DDD is the IP Address of the remote client, and
  // EEE and FFF represent the High and Low bytes of the client's port.
  FDataPort.sin_addr.S_addr := IP_StringToAddr(
    ParseParamByNumber(Params, 1, ',') + '.' +
    ParseParamByNumber(Params, 2, ',') + '.' +
    ParseParamByNumber(Params, 3, ',') + '.' +
    ParseParamByNumber(Params, 4, ',')
    );
  FDataPort.sin_port := Swap(Word(
    IVal(ParseParamByNumber(Params, 5, ',')) * $100 +
    IVal(ParseParamByNumber(Params, 6, ','))
    ));
  //	WriteLn('Port: ', FDataPort.port);
  if (FDataPort.sin_addr.S_addr <> 0) and (FDataPort.sin_port <> 0) then
  begin
    if FDataSocket <> 0 then
      CloseSocket(FDataSocket);
    FDataSocket := 0;
    SendResponse(MSG_PORTOK);
  end
  else
  begin
    SendResponse(MSG_BADPORT);
  end;
end;

procedure TFTPConnection.Proc_Type(Params: AnsiString);
begin
  Params := Uppercase(Params);
  if (Params = 'BINARY') or (Params = 'IMAGE') or (Params = 'I') then
    Include(FFlags, ftpBinary)
  else if (Params = 'ASCII') or (Params = 'A') or (Params = 'TEXT') then
    Exclude(FFlags, ftpBinary)
  else
  begin
    SendResponse(Format(MSG_BADTYPE, [Params]));
    Exit
  end;
  SendResponse(Format(MSG_TYPEOK, [Params]));
end;

procedure TFTPConnection.Proc_RnFr(Params: AnsiString);
begin
  if FStatus < stWaitingCommand then
  begin
    SendResponse(MSG_NOTLOGGED);
  end
  else
  begin
    AdjustPath(Params);
    if FileExists(Params) then
    begin
      SendResponse(MSG_RNFROK);
      FCurrentFileName := Params;
    end
    else
    begin
      SendResponse(Format(MSG_CANTOPEN, [UnixFileName(Params)]));
      FCurrentFileName := '';
    end;
  end;
end;

procedure TFTPConnection.Proc_Rnto(Params: AnsiString);
begin
  if FStatus < stWaitingCommand then
  begin
    SendResponse(MSG_NOTLOGGED);
  end
  else
  begin
    AdjustPath(Params);
    if FileExists(FCurrentFileName)
      and RenameFile(FCurrentFileName, Params) then
      SendResponse(Format(MSG_RNTOOK,
        [UnixFileName(FCurrentFileName), UnixFileName(Params)]))
    else
    begin
      SendResponse(Format(MSG_BADRNTO,
        [UnixFileName(FCurrentFileName), UnixFileName(Params)]));
    end;
  end;
end;

procedure TFTPConnection.Proc_List(Params: AnsiString);
begin
  if FStatus < stWaitingCommand then
  begin
    SendResponse(MSG_NOTLOGGED);
  end
  else
  begin
    if (Params = '') or (Params = '*') then
      Params := '*.*';
    AdjustPath(Params);
{$IFDEF COMPILER6_UP}
  {$WARN SYMBOL_PLATFORM OFF}
{$ENDIF}
    if FindFirst(Params, faAnyFile - faVolumeID - faHidden, FSearchRec) = 0 then
    begin
{$IFDEF COMPILER6_UP}
  {$WARN SYMBOL_PLATFORM ON}
{$ENDIF}
      if OpenDataConnection then
        FStatus := stSendingDir;
    end
    else
    begin
      FindClose(FSearchRec);
      SendResponse(Format(MSG_NODIR, [UnixFileName(Params)]));
    end;
  end;
end;

procedure TFTPConnection.Proc_Nlst(Params: AnsiString);
begin
  if FStatus < stWaitingCommand then
  begin
    SendResponse(MSG_NOTLOGGED);
  end
  else
  begin
    if (Params = '') or (Params = '*') then
      Params := '*.*';
    AdjustPath(Params);
{$IFDEF COMPILER6_UP}
  {$WARN SYMBOL_PLATFORM OFF}
{$ENDIF}
    if FindFirst(Params, faAnyFile - faVolumeID - faHidden, FSearchRec) = 0 then
    begin
{$IFDEF COMPILER6_UP}
  {$WARN SYMBOL_PLATFORM ON}
{$ENDIF}
      if OpenDataConnection then
        FStatus := stSendingList;
    end
    else
    begin
      FindClose(FSearchRec);
      SendResponse(Format(MSG_NODIR, [UnixFileName(Params)]));
    end;
  end;
end;

procedure TFTPConnection.Proc_Stor(Params: AnsiString);
var
  FileCreated: Boolean;
begin
  if FStatus < stWaitingCommand then
  begin
    SendResponse(MSG_NOTLOGGED);
  end
  else
  begin
    AdjustPath(Params);
    FCurrentFileName := Params;
    FileCreated := False;
    try
      FStorFileStream := TFileStream.Create(FCurrentFileName, fmCreate);
      FileCreated := True;
    except
      // do nothing
    end;
    if not FileCreated then
    begin
      SendResponse(Format(MSG_CANTMAKE, [UnixFileName(Params)]));
      FCurrentFileName := '';
    end
    else
    begin
      if OpenDataConnection then
      begin
        FStatus := stReceivingFile;
      end
      else
      begin
        FreeAndNil(FStorFileStream);
        DeleteFile(FCurrentFileName);
        FCurrentFileName := '';
      end;
    end;
  end;
end;

procedure TFTPConnection.Proc_Dele(Params: AnsiString);
begin
  if FStatus < stWaitingCommand then
  begin
    SendResponse(MSG_NOTLOGGED);
  end
  else
  begin
    AdjustPath(Params);
    if DeleteFile(Params) then
      SendResponse(Format(MSG_DELEOK, [UnixFileName(Params)]))
    else
    begin
      SendResponse(Format(MSG_BADDELE, [UnixFileName(Params)]));
    end;
  end;
end;

procedure TFTPConnection.Proc_Retr(Params: AnsiString);
var
  FileOpened: Boolean;
begin
  if FStatus < stWaitingCommand then
  begin
    SendResponse(MSG_NOTLOGGED);
  end
  else
  begin
    AdjustPath(Params);

    FileOpened := False;
    try
      FRetrFileStream := TFileStream.Create(Params, fmOpenRead);
      FileOpened := True;
    except
      SendResponse(Format(MSG_CANTOPEN, [UnixFileName(Params)]))
    end;

    if FileOpened then
    begin
      if OpenDataConnection then
      begin
        FStatus := stSendingFile;
        FCurrentFileName := Params;
      end
      else
      begin
        FreeAndNil(FRetrFileStream);
      end;
    end;
  end;
end;

procedure TFTPConnection.Proc_Site(Params: AnsiString);
var
  Cmd: AnsiString;
begin
  if FStatus < stWaitingCommand then
  begin
    SendResponse(MSG_NOTLOGGED);
  end
  else
  begin
    Cmd := UpperCase(ParseParamByNumber(Params, 1, ' '));
    if (Cmd = 'HELP') or (Cmd = '') then
    begin
      Params := Copy(Params, Pos(UpperCase(Params), ' SITE ') + 6,
        Length(Params));
      Proc_SiteHelp(Params);
    end
    else
    begin
      SendResponse(Format(MSG_NOTIMPLEMENTED, ['SITE ' + Params]));
    end;
  end;
end;

procedure TFTPConnection.Proc_Size(Params: AnsiString);
var
  h: Integer;
  nSize: LongInt;
begin
  if FStatus < stWaitingCommand then
  begin
    SendResponse(MSG_NOTLOGGED);
  end
  else
  begin
    AdjustPath(Params);
    h := FileOpen(Params, fmOpenRead);
    if h < 0 then
    begin
      SendResponse(Format(MSG_CANTOPEN, [UnixFileName(Params)]));
    end
    else
    begin
      nSize := FileSeek(h, 0, fsFromEnd);
      FileClose(h);
      SendResponse(Format(MSG_SIZE, [nSize]));
    end;
  end;
end;

procedure TFTPConnection.Proc_Quit(Params: AnsiString);
begin
  if FStatus <> stDone then
  begin
    SendResponse(MSG_BYE);
    FStatus := stDone;
  end;
end;

procedure TFTPConnection.Proc_Abor(Params: AnsiString);
begin
  // Close Data Socket
  SendResponse(MSG_ABORT);
  if (FDataSocket <> 0) then
  begin
    if (FDataSocket <> FCtrlSocket) then
      CloseSocket(FDataSocket);
    FDataSocket := 0;
  end;
  // Close any outgoing file
  if FStatus = stSendingFile then
    FreeAndNil(FRetrFileStream);
  // Delete any incomming file
  if FStatus = stReceivingFile then
  begin
    FreeAndNil(FStorFileStream);
    DeleteFile(FCurrentFileName);
  end;
  // Done
  FStatus := stWaitingCommand;
end;

{ TFTPServer }

constructor TFTPServer.Create;
begin
  inherited Create;
  FUserList := TList.Create;
  FConnections := TList.Create;
  FSocket := 0;
  GetDir(0, FRootPath);
  ServerList.Add(Self);
end;

destructor TFTPServer.Destroy;
var
  P: PUserEntry;
  Conn: TFTPConnection;
begin
  TerminateTask(FWorkTask);
  ServerList.Remove(Self);
  if FSocket > 0 then
    CloseSocket(FSocket);
  while FConnections.Count > 0 do
  begin
    Conn := TFTPConnection(FConnections[0]);
    Conn.Free;
    FConnections.Delete(0);
  end;
  FConnections.Free;
  // Liberamos la lista de Usuarios
  while FUserList.Count > 0 do
  begin
    P := FUserList[0];
    Dispose(P);
    FUserList.Delete(0);
  end;
  FUserList.Free;
end;

procedure TFTPServer.AddUser(UserName, Password: AnsiString; CanWrite: Boolean);
var
  P: PUserEntry;
begin
  New(P);
  P^.UserName := UserName;
  P^.Password := Password;
  P^.CanWrite := CanWrite;
  FUserList.Add(P);
end;

function TFTPServer.ValidateUser(UserName, Password: AnsiString; var CanWrite:
  Boolean): Boolean;
var
  i: integer;
  P: PUserEntry;
begin
  // If there are no users defined, access is always allowed.
  if FUserList.Count = 0 then
  begin
    Result := True;
    CanWrite := True;
    Exit;
  end;
  // There are one or more users.
  for i := 0 to FUserList.Count - 1 do
  begin
    P := FUserList[i];
    if AnsiCompareText(P^.UserName, UserName) = 0 then
    begin
      if P^.Password = Password then
      begin
        CanWrite := P^.CanWrite;
        Result := True;
        Exit;
      end;
    end;
  end;
  Result := False;
end;

function TFTPServer.Start(Address: TIPAddress; Port: WORD; RootPath:
  AnsiString): Boolean;
var
  Addr: TSockAddr;
begin
  Result := False;
  if FSocket > 0 then
    Exit;
  // Open Server Socket to accept incoming connections
  Addr.sin_family := PF_INET;
  Addr.sin_port := Port;
  Addr.sin_addr.S_addr := Address;
  FSocket := Socket(PF_INET, SOCK_STREAM, IPPROTO_IP);
  if (FSocket <= 0) or (Bind(FSocket, Addr, SizeOf(Addr)) <> 0)
    or (Listen(FSocket, 10) <> 0) then
  begin
    if FSocket <> 0 then
      CloseSocket(FSocket);
    FSocket := 0;
    Exit;
  end;
  // Done
  FWorkTask := CreateTask(FTPServerTask, Self, 32768);
  FRootPath := RootPath;
  Result := True;
end;

procedure TFTPServer.AddConnection(Socket: Integer; Addr: TSockAddr);
var
  Conn: TFTPConnection;
begin
  Conn := TFTPConnection.Create(Self, Socket, Addr, FRootPath);
  FConnections.Add(Conn);
end;

procedure TFTPServer.Work;
var
  i: Integer;
  HSock, L: LongInt;
  Addr: TSockAddr;
  Conn: TFTPConnection;
begin
  // Look for a new incomming connection
  L := SizeOf(Addr);
  HSock := Accept(FSocket, Addr, L);
  if HSock > 0 then
    AddConnection(HSock, Addr);
  // Check for connections to destroy
  i := 0;
  while i < FConnections.Count do
  begin
    Conn := FConnections[i];
    if Conn.Status = stDone then
    begin
      Conn.Free;
      FConnections.Delete(i);
    end
    else
    begin
      // Work with this connection
   //			Conn.Work;
      Inc(i);
    end;
  end;
end;

procedure FTPServerTask(Param: Pointer);
begin
  repeat
    Net_Work_Low;
    TFTPServer(Param).Work;
    Yield;
  until False;
end;

procedure FTPConnectionTask(Param: Pointer);
begin
  repeat
    TFTPConnection(Param).Work;
    Yield;
  until False;
end;

initialization
  // Creamos la lista de Servers
  ServerList := TList.Create;
finalization
  // Quitamos los Servers
  while ServerList.Count > 0 do
  begin
    TFTPServer(ServerList[0]).Free;
    ServerList.Delete(0);
  end;
  // Eliminamos la lista de servers
  ServerList.Free;
end.

