program Ping;

{$APPTYPE CONSOLE}

uses
  Crt,
  SysUtils,
  WDosSock,
  WDosICMP,
  WDosClock,
  Threads,
  SocketInfos in '..\common\SocketInfos.pas';

type
  TPingThread = class(TThread)
  private
    { Private declarations }
    FDest: TInAddr;
  protected
    constructor Create(Dest: TInAddr);
    procedure Execute; override;
  end;

constructor TPingThread.Create(Dest: TInAddr);
begin
  inherited Create(False);
  FDest := Dest;
end;

procedure TPingThread.Execute;
var
  Start: Int64;
begin
  repeat
    Start := GetMSCounter;
    if ICMP_Ping(FDest.S_addr, 2000) = 0 then
    begin
      WriteLn('Response received from ', inet_ntoa(FDest),
        ' Time <= ', GetMSCounter - Start, ' ms.');
      Sleep(1000);
    end
    else
    begin
      WriteLn('Timeout');
      Sleep(1000);
    end;
    Sleep(0);
  until Terminated;
end;

var
  PingThread: TPingThread;
  HostEnt: PHostEnt;
  Hostname: string;
  Dest: TInAddr;
begin
  // Check Params
  if ParamCount <> 1 then
  begin
    WriteLn('Usage: PING destination');
    WriteLn('       Where destination is an IP Address or Host Name.');
    Halt(1);
  end;

  // Destination is a name or an IP address?
  Dest.S_addr := inet_addr(PChar(ParamStr(1)));
  if Dest.S_addr = INADDR_ANY then
  begin

    // Doesn't seem to be an IP. Let's try to resolve it with DNS.
    Hostname := ParamStr(1);
    HostEnt := gethostbyname(PChar(Hostname));
    Dest.S_addr := PInAddr(HostEnt^.h_addr^)^.S_addr;
    if Dest.S_addr = INADDR_NONE then
    begin
      WriteLn('Could not resolve host name "' + ParamStr(1) + '".');
      Halt(2);
    end;
  end else Hostname := '';

  // create Ping thread
  PingThread := TPingThread.Create(Dest);
  try
    // display all current socket infos
    DisplaySocketInfos;

    Writeln(Format('Pinging to %s [%s]. Press <Esc> to cancel ...',
        [Hostname, inet_ntoa(Dest)]));

    // Main Loop (ends when <Esc> is pressed)
    repeat
      UpdateNetwork;
    until KeyPressed and (ReadKey = #27);
  finally
    PingThread.Free;
  end;
end.

