
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (tools)       }
{ RunStubit.dpr                                         }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2002 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

program RunStubit;

{
  Purpose of this helper program is a better cooperation of
  Delphi IDE and WDosX Stubit for handling long file names
  and different working directory of the produced exe file.

  $Header: /cvsroot/dwpl/dwpl/demos/runstubit/RunStubit.dpr,v 1.1 2003/07/23 16:09:09 lagocs Exp $
  ----------------------------------------------------------------------------

  $Log: RunStubit.dpr,v $
  Revision 1.1  2003/07/23 16:09:09  lagocs
  from now on everyone can create a clean dwpl folder tree for creating
  file releases!

  Revision 1.2  2003/01/31 21:20:23  iwache
  ParamCount test extended to [1..3] mandatory parameters.
  This is useful for call stubit with some options settings
  like -m_float/-nowfse/-extract. But these settings are currently
  not propagated to stubit by RunStubit.

  ----------------------------------------------------------------------------
}

uses
  Windows,
  ShellAPI;

const
  PathDelim  = '\';
  DriveDelim = ':';
  PathSep    = ';';

type
  { MultiByte Character Set (MBCS) byte type }
  TMbcsByteType = (mbSingleByte, mbLeadByte, mbTrailByte);

var
  LeadBytes: set of Char = [];

procedure Launch(const FileName, Parameters, StartDir: string);
var
  ShellInfo: TShellExecuteInfo;
  HProcess: THandle;
begin
  // Initialize the ShellInfo record
  FillChar(ShellInfo, SizeOf(TShellExecuteInfo), 0);
  ShellInfo.cbSize := SizeOf(TShellExecuteInfo);
  ShellInfo.fMask := SEE_MASK_NOCLOSEPROCESS;
  ShellInfo.Wnd := HWnd_Desktop;
  ShellInfo.lpVerb := 'Open';
  ShellInfo.lpFile := PChar(FileName);
  ShellInfo.lpParameters := PChar(Parameters);
  ShellInfo.lpDirectory := PChar(StartDir);
  ShellInfo.nShow := SW_NORMAL;

  if ShellExecuteEx(@ShellInfo) then
  begin
    HProcess := ShellInfo.hProcess;
    WaitForSingleObject(HProcess, Infinite);
  end;
end;

function ByteType(const S: string; Index: Integer): TMbcsByteType;
begin
  Result := mbSingleByte;
end;

function StrScan(const Str: PChar; Chr: Char): PChar; assembler;
asm
        PUSH    EDI
        PUSH    EAX
        MOV     EDI,Str
        MOV     ECX,$FFFFFFFF
        XOR     AL,AL
        REPNE   SCASB
        NOT     ECX
        POP     EDI
        MOV     AL,Chr
        REPNE   SCASB
        MOV     EAX,0
        JNE     @@1
        MOV     EAX,EDI
        DEC     EAX
@@1:    POP     EDI
end;

function LastDelimiter(const Delimiters, S: string): Integer;
var
  P: PChar;
begin
  Result := Length(S);
  P := PChar(Delimiters);
  while Result > 0 do
  begin
    if (S[Result] <> #0) and (StrScan(P, S[Result]) <> nil) then
      if (ByteType(S, Result) = mbTrailByte) then
        Dec(Result)
      else
        Exit;
    Dec(Result);
  end;
end;

function ExtractFileDir(const FileName: string): string;
var
  I: Integer;
begin
  I := LastDelimiter(PathDelim + DriveDelim,Filename);
  if (I > 1) and (FileName[I] = PathDelim) and
    (((FileName[I - 1] <> PathDelim) and (FileName[I-1] <> DriveDelim)) or
    (ByteType(FileName, I-1) = mbTrailByte)) then Dec(I);
  while (ByteType(FileName, I-1) = mbTrailByte) and (I > 0) do Dec(I);
  Result := Copy(FileName, 1, I);
end;

function ExtractFileName(const FileName: string): string;
var
  I: Integer;
begin
  I := LastDelimiter(PathDelim + DriveDelim, FileName);
  Result := Copy(FileName, I + 1, MaxInt);
end;

var
  ShortPathName: string;
  LongPathName: string;
  StartDir: string;
  Parameters: string;
begin
  if not ParamCount in [1..3] then Exit;
  SetLength(ShortPathName, 255);
  LongPathName := ParamStr(1);
  GetShortPathName(PChar(LongPathName), PChar(ShortPathName),
      Length(ShortPathName));
  ShortPathName := PChar(ShortPathName);
  StartDir := ExtractFileDir(ShortPathName);
  Parameters := ExtractFileName(ShortPathName);
  Launch('stubit.exe', Parameters, StartDir);
end.

