
{*******************************************************}
{                                                       }
{ WCLX DWP Cross Platform Component library             }
{ Vi Tutorial 3C Project                                }
{                                                       }
{ for Delphi 4, 5, 6, 7                                 }
{ Version 2.6.0                                         }
{ Copyright (c) 2003 by Team-DWP                        }
{ e-mail: team-dwp@dwp42.org                            }
{ www: http://www.dwp42.org                             }
{                                                       }
{ The initial developer of the code is                  }
{   Oleg A. Bachin, bachin@cis-kiev.com                 }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Turbo Vision 2.0 Demo                             }
{     Copyright (c) 1992 by Borland International       }
{     www: http://www.borland.com                       }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/demos/turbovision/editors/Edt1.dpr,v 1.2 2003/11/19 19:11:10 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: Edt1.dpr,v $
  Revision 1.2  2003/11/19 19:11:10  iwache
  Extended to test the new implemented
  file open and change directory dialogs.
  Also demonstates the changes to ViEditors.pas
  to improve the work with Vi only mode.

  Revision 1.1  2003/11/06 20:29:00  iwache
  Demo project for ViEditors with direct Vi layer access added.

  ----------------------------------------------------------------------------
}

program Edt1;

{$APPTYPE CONSOLE}

uses
  SysUtils, Classes, TutorConst, ViDrivers, ViTypes, ViClasses, ViScreens,
  ViViews, ViMenus, ViApp, ViUtils, ViDialogs, ViMsgBox, ViStdDlg,
  ViEditors, ViColorSel;

type
  TTutorApp = class(TViApplication)
  private
    FFileOpenHistory: TStrings;
    FChDirHistory: TStrings;
    procedure DoAboutBox;
    procedure ShowClipboard;
    procedure ChangeDir;
    procedure OpenFileExecute;
  public
    constructor Create;
    destructor Destroy; override;
    procedure HandleEvent(var Event: TEvent); override;
    procedure InitMenuBar; override;
    procedure InitStatusLine; override;
    procedure NewWindow(const FileName: string);
    procedure DoColorSelect;
  end;

constructor TTutorApp.Create;
begin
  inherited Create;
  FFileOpenHistory := TStringList.Create;
  FChDirHistory := TStringList.Create;
 
  DisableCommands([cmOrderWin, cmStockWin, cmSupplierWin]);
end;

destructor TTutorApp.Destroy;
begin
  FChDirHistory.Free;
  FFileOpenHistory.Free;
  inherited Destroy;
end;

procedure TTutorApp.DoAboutBox;
begin
  MessageBox(#3'Turbo Vision Tutorial Application'#13 +
    #3'Copyright 1992'#13#3'Borland International',
    mfInformation or mfOKButton);
end;

procedure TTutorApp.HandleEvent(var Event: TEvent);
begin
  inherited HandleEvent(Event);
  if Event.What = evCommand then
  begin
    case Event.Command of
      cmChangeDir:
        begin
          ChangeDir;
          ClearEvent(Event);
        end;
      cmNew:
        begin
          NewWindow('');
          ClearEvent(Event);
        end;
      cmOpen:
        begin
          OpenFileExecute;
          ClearEvent(Event);
        end;
      cmOptionsVideo:
        begin
          SetScreenMode(ViScreen.Mode xor smFont8x8);
          ClearEvent(Event);
        end;
      cmAbout:
        begin
          DoAboutBox;
          ClearEvent(Event);
        end;
      cmShowClipboard:
        begin
          ShowClipboard;
          ClearEvent(Event);
        end;
      cmColorSelect:
        begin
          DoColorSelect;
          ClearEvent(Event);
        end;
      cmSaveAs:
        begin
          ViDesktop.Current.PutEvent(Event);
          ClearEvent(Event);
        end;
    end;
  end;
end;

procedure TTutorApp.InitMenuBar;
var
  R: TRect;
begin
  GetExtent(R);
  R.B.Y := R.A.Y + 1;
  ViMenuBar := TViMenuBar.Create(R, NewMenu(
    NewSubMenu('~F~ile', hcNoContext, NewMenu(
    StdFileMenuItems(nil)),
    NewSubMenu('~E~dit', hcNoContext, NewMenu(
    StdEditMenuItems(
    nil)),
    NewSubMenu('~W~indow', hcNoContext, NewMenu(
    StdWindowMenuItems(nil)),
    NewSubMenu('~H~elp', hcNoContext, NewMenu(
    NewItem('~A~bout...', '', kbNoKey, cmAbout, hcNoContext,
    NewLine(
    NewItem('~C~olor settings...', '', kbNoKey, cmColorSelect, hcNoContext,
    NewLine(
    NewItem('~T~oggle screen', '', kbNoKey, cmOptionsVideo, hcNoContext,
    NewLine(
    NewItem('Show ~C~lipboard', '', kbNoKey, cmShowClipboard, hcNoContext,
    nil)))))))),
    nil))))
    ), True);
end;

procedure TTutorApp.InitStatusLine;
var
  R: TRect;
begin
  GetExtent(R);
  R.A.Y := R.B.Y - 1;
  ViStatusLine := TViStatusLine.Create(R,
    NewStatusDef(0, $EFFF,
    NewStatusKey('~F3~ Open', kbF3, cmOpen,
    NewStatusKey('~F4~ New', kbF4, cmNew,
    NewStatusKey('~Alt+F3~ Close', kbAltF3, cmClose,
    StdStatusKeys(nil)))),
    NewStatusDef($F000, $FFFF,
    NewStatusKey('~F6~ Next', kbF6, cmOrderNext,
    NewStatusKey('~Shift+F6~ Prev', kbShiftF6, cmOrderPrev,
    StdStatusKeys(nil))), nil)));
end;

procedure TTutorApp.NewWindow(const FileName: string);
var
  R: TRect;
begin
  R.A.X := 0;
  R.A.Y := 0;
  R.B.X := Trunc(ViScreen.Width / 2) - 1;
  R.B.Y := Trunc(ViScreen.Height / 2) - 1;
  ViDesktop.Insert(TViEditWindow.Create(R, FileName, 0));
end;

procedure TTutorApp.ShowClipboard;
begin
  {
    if not Assigned( Clipboard.Owner ) then
      if Assigned( ViDesktop ) then
        ViDesktop.Insert( Clipboard );
  }
  Clipboard.Show;
end;

procedure TTutorApp.DoColorSelect;
var
  D: TViColorDialog;
begin
  D := TViColorDialog.Create(ViApplication.Palette,
    ColorGroup('Desktop', DesktopColorItems(nil),
    ColorGroup('Menu', MenuColorItems(nil),
    ColorGroup('Dialog boxes', DialogColorItems(dpGrayDialog, nil),
    nil))));
  try
    if ExecuteDialog(D, nil) <> cmCancel then
    begin
      ViApplication.Palette := D.Pal;
      Redraw;
    end;
  finally
    D.Free;
  end;
end;

procedure TTutorApp.ChangeDir;
var
  ChDirDialog: TViChDirDialog;
begin
  ChDirDialog := TViChDirDialog.Create('Select current Directory', cdNormal,
      FChDirHistory);
  try
//    ChDirDialog.DirName := 'C:\';
    if ExecuteDialog(ChDirDialog, nil) <> cmCancel then
    begin
//      MessageBox(ChDirDialog.DirName, mfInformation + mfOkButton);
      FChDirHistory.Add(ChDirDialog.DirName);
      SetCurrentDir(ChDirDialog.DirName);
    end;
  finally
    ChDirDialog.Free;
  end;
end;

procedure TTutorApp.OpenFileExecute;
var
  FileDialog: TViFileDialog;
  TheFile: string;
const
  FDOptions: Word = fdOKButton or fdOpenButton;
begin
  TheFile := '*.*';
  FileDialog := TViFileDialog.Create(TheFile, 'Open file', 'File ~n~ame',
      FDOptions, FFileOpenHistory);
  try
    if ExecuteDialog(FileDialog, nil) <> cmCancel then
      NewWindow(FileDialog.GetFileName);
  finally
    FileDialog.Free;
  end;
end;


var
  TutorApp: TTutorApp;

begin
  TutorApp := TTutorApp.Create;
  try
    try
      TutorApp.Run;
    except
      on E: Exception do
        WriteLn(E.Message);
    end;
  finally
    TutorApp.Free;
  end;
end.

