
{*******************************************************}
{                                                       }
{ WCLX DWP Cross Platform Component library             }
{ Vi Tutorial 3C Project                                }
{                                                       }
{ for Delphi 5, 6, 7                                    }
{ Version 2.6.0                                         }
{ Copyright (c) 2003 by Team-DWP                        }
{ e-mail: team-dwp@dwp42.org                            }
{ www: http://www.dwp42.org                             }
{                                                       }
{ The initial developer of the code is                  }
{   Immo Wache, immo.wache@dwp42.org                    }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Turbo Vision 2.0 Demo                             }
{     Copyright (c) 1992 by Borland International       }
{     www: http://www.borland.com                       }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/demos/turbovision/tvi2/TutorApp.pas,v 1.2 2003/11/19 19:09:51 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: TutorApp.pas,v $
  Revision 1.2  2003/11/19 19:09:51  iwache
  Copyright, Header and Log tags added.
  Extended to test the new implemented
  file open and change directory dialogs.

  ----------------------------------------------------------------------------
}
unit TutorApp;

interface

uses
  SysUtils, Classes, ViTypes, ViUtils, ViClasses, ViViews, ViApp, ViScreens,
  ViMenus, ViDialogs, ViStdDlg, ViMsgBox;

type
  TTutorApplication = class(TViApplication)
  private
    FFileOpenHistory: TStrings;
    FChDirHistory: TStrings;
    procedure DoAboutBox;
    procedure NewWindow;
    procedure OpenWindow;
    procedure ChangeDir;
  public
    constructor Create;
    destructor Destroy; override;
    procedure HandleEvent(var Event: TEvent); override;
    procedure InitMenuBar; override;
    procedure InitStatusLine; override;
  end;

implementation

uses
  TutorConsts;

{ TTutorApplication }

constructor TTutorApplication.Create;
begin
  inherited;
  FFileOpenHistory := TStringList.Create;
  FChDirHistory := TStringList.Create;
end;

destructor TTutorApplication.Destroy;
begin
  FChDirHistory.Free;
  FFileOpenHistory.Free;
  inherited;
end;

procedure TTutorApplication.InitMenuBar;
var
  R: TRect;
begin
  GetExtent(R);
  R.B.Y := R.A.Y + 1;
  ViMenuBar := TViMenuBar.Create(R, NewMenu(
    NewSubMenu('~F~ile', hcNoContext, NewMenu(
      StdFileMenuItems(nil)),
    NewSubMenu('~E~dit', hcNoContext, NewMenu(
      StdEditMenuItems(
      NewLine(
      NewItem('~S~how clipboard', '', kbNoKey, cmClipShow, hcNoContext,
      nil)))),
    NewSubMenu('~O~rders', hcNoContext, NewMenu(
      NewItem('~N~ew', 'F9', kbF9, cmOrderNew, hcNoContext,
      NewItem('~S~ave', '', kbNoKey, cmOrderSave, hcNoContext,
      NewLine(
      NewItem('Next', 'PgDn', kbPgDn, cmOrderNext, hcNoContext,
      NewItem('Prev', 'PgUp', kbPgUp, cmOrderPrev, hcNoContext,
      nil)))))),
    NewSubMenu('O~p~tions', hcNoContext, NewMenu(
      NewItem('~T~oggle video', '', kbNoKey, cmOptionsVideo, hcNoContext,
      NewItem('~S~ave desktop', '', kbNoKey, cmOptionsSave, hcNoContext,
      NewItem('~L~oad desktop', '', kbNoKey, cmOptionsLoad, hcNoContext,
      nil)))),
    NewSubMenu('~W~indow', hcNoContext, NewMenu(
      NewItem('Orders', '', kbNoKey, cmOrderWin, hcNoContext,
      NewItem('Stock items', '', kbNoKey, cmStockWin, hcNoContext,
      NewItem('Suppliers', '', kbNoKey, cmSupplierWin, hcNoContext,
      NewLine(
      StdWindowMenuItems(nil)))))),
    NewSubMenu('~H~elp', hcNoContext, NewMenu(
      NewItem('~A~bout...', '', kbNoKey, cmAbout, hcNoContext,
      nil)),
    nil))))))
  ), True); // Attention: must be True for ViMode
end;

procedure TTutorApplication.InitStatusLine;
var
  R: TRect;
begin
  GetExtent(R);
  R.A.Y := R.B.Y - 1;
  ViStatusLine := TViStatusLine.Create(R,
    NewStatusDef(0, $EFFF,
      NewStatusKey('~F3~ Open', kbF3, cmOpen,
      NewStatusKey('~F4~ New', kbF4, cmNew,
      NewStatusKey('~Alt+F3~ Close', kbAltF3, cmClose,
      StdStatusKeys(nil)))),
    NewStatusDef($F000, $FFFF,
      NewStatusKey('~F6~ Next', kbF6, cmOrderNext,
      NewStatusKey('~Shift+F6~ Prev', kbShiftF6, cmOrderPrev,
      StdStatusKeys(nil))), nil)));
end;

procedure TTutorApplication.DoAboutBox;
begin
  MessageBox(#3'Turbo Vision Tutorial Application'#13 +
    #3'Copyright 1992'#13#3'Borland International',
    mfInformation or mfOKButton);
end;

procedure TTutorApplication.HandleEvent(var Event: TEvent);
begin
  inherited HandleEvent(Event);
  if Event.What = evCommand then
  begin
    case Event.Command of
      cmChangeDir:
        begin
          ChangeDir;
          ClearEvent(Event);
        end;
      cmNew:
        begin
          NewWindow;
          ClearEvent(Event);
        end;
      cmOpen:
        begin
          OpenWindow;
          ClearEvent(Event);
        end;
      cmOptionsVideo:
        begin
          SetScreenMode(ViScreen.Mode xor smFont8x8);
          ClearEvent(Event);
        end;
      cmAbout:
        begin
          DoAboutBox;
          ClearEvent(Event);
        end;
    end;
  end;
end;

procedure TTutorApplication.NewWindow;
var
  R: TRect;
  TheWindow: TViWindow;
begin
  RectAssign(R, 0, 0, 60, 20);
  TheWindow := TViWindow.Create(R, 'Window', wnNoNumber);
  InsertWindow(TheWindow);
end;

procedure TTutorApplication.OpenWindow;
var
  R: TRect;
  FileDialog: TViFileDialog;
  TheFile: string;
const
  FDOptions: Word = fdOKButton or fdOpenButton;
begin
  TheFile := '*.*';
  FileDialog := TViFileDialog.Create(TheFile, 'Open file', 'File ~n~ame',
      FDOptions, FFileOpenHistory);
  try
    if ExecuteDialog(FileDialog, nil) <> cmCancel then
    begin
      RectAssign(R, 0, 0, 75, 20);
      InsertWindow(TViWindow.Create(R, FileDialog.GetFileName, wnNoNumber));
    end;
  finally
    FileDialog.Free;
  end;
end;

procedure TTutorApplication.ChangeDir;
var
  ChDirDialog: TViChDirDialog;
begin
  ChDirDialog := TViChDirDialog.Create('Select Directory', cdNormal,
      FChDirHistory);
  try
    ChDirDialog.DirName := 'C:\';
    if ExecuteDialog(ChDirDialog, nil) <> cmCancel then
    begin
      MessageBox(ChDirDialog.DirName, mfInformation + mfOkButton);
      FChDirHistory.Add(ChDirDialog.DirName);
    end;
  finally
    ChDirDialog.Free;
  end;
end;


end.
