
{*******************************************************}
{                                                       }
{ DWPL Delphi WDosX and Cross Platform Library          }
{ (automation)                                          }
{ ModbusCtrls.pas                                       }
{                                                       }
{ for Delphi 5, 6, 7                                    }
{ Version 2.6.0                                         }
{ Copyright (c) 2000-2003 by Team-DWP                   }
{ e-mail: team-dwp@dwp42.org                            }
{ www: http://www.dwp42.org                             }
{                                                       }
{ The initial developer of the code is                  }
{   Immo Wache, immo.wache@dwp42.org                    }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ Licensees holding a valid Team-DWP No-Nonsense        }
{ License for this Software may use this file in        }
{ accordance with such license, which appears in        }
{ the file license.txt that came with this software.    }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/automation/ModbusCtrls.pas,v 1.2 2004/03/30 20:33:03 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: ModbusCtrls.pas,v $
  Revision 1.2  2004/03/30 20:33:03  iwache
  Class TModbusClientThread modified:
    - New procedure ConfirmPacket added. Call of this method
      sets internal state machine back to mcsIdle.
    - Property Confirmed removed.
    - Property State changed from type TModbusClientThreadState to
      TModbusClientState.
  Component TModbusTcpClient modified:
    - Property State: TModbusClientState added.

  Revision 1.1  2003/11/06 20:26:49  iwache
  Units for Modbus/TCP access to automation added.

  ----------------------------------------------------------------------------
}

unit ModbusCtrls;

interface

uses
  Windows, SysUtils, Classes,
{$IFDEF DOS}
  Threads,
{$ENDIF}
  Sockets;

const
  { Modbus/TCP Port }
  MODBUS_TCP_PORT = 502;

  { Modbus functions }
  READ_COILS = 1;
  READ_INPUT_DISCRETES = 2;
  READ_MULTIPLE_REGISTERS = 3;
  READ_INPUT_REGISTERS = 4;
  WRITE_COIL = 5;
  WRITE_SINGLE_REGISTER = 6;
  READ_EXCEPTION_STATUS = 7;
  GET_COMM_EVENT_COUNTER = 11;
  FORCE_MULTIPLE_COILS = 15;
  WRITE_MULTIPLE_REGISTERS = 16;
  READ_WRITE_REGISTERS = 23;
  EXCEPTION_FIRST = $80;
  EXCEPTION_LAST = $FF;

  DConnectTime = 1000;

type
  EModbusError = class(Exception);

  TReceiveState = (rvsHeader, rvsBody);

  TModbusReadCoilsEvent = procedure(Sender: TObject; TransactionId: Word;
    UnitId: Byte; Data: array of Byte) of object;
  TModbusReadInputDiscretesEvent = procedure(Sender: TObject;
    TransactionId: Word; UnitId: Byte; Data: array of Byte) of object;
  TModbusReadMultipleRegistersEvent = procedure(Sender: TObject;
    TransactionId: Word; UnitId: Byte; Data: array of Word) of object;
  TModbusReadInputRegistersEvent = procedure(Sender: TObject;
    TransactionId: Word; UnitId: Byte; Data: array of Word) of object;
  TModbusWriteCoilEvent = procedure(Sender: TObject; TransactionId: Word;
    UnitId: Byte; RefNum: Word; Value: Boolean) of object;
  TModbusWriteSingleRegisterEvent = procedure(Sender: TObject;
    TransactionId: Word; UnitId: Byte; RefNum, Value: Word) of object;
  TModbusReadExceptionStatusEvent = procedure(Sender: TObject;
    TransactionId: Word; UnitId: Byte; Status: Byte) of object;
  TModbusForceMultipleCoilsEvent = procedure(Sender: TObject;
    TransactionId: Word; UnitId: Byte; RefNum, BitCount: Word) of object;
  TModbusWriteMultipleRegistersEvent = procedure(Sender: TObject;
    TransactionId: Word; UnitId: Byte; RefNum, WordCount: Word) of object;
  TModbusGetCommEventCounterEvent = procedure(Sender: TObject;
    TransactionId: Word; UnitId: Byte; Status, Count: Word) of object;
  TModbusReadWriteRegistersEvent = TModbusReadMultipleRegistersEvent;
  TModbusResponseEvent = procedure(Sender: TObject; TransactionId: Word;
    UnitId, FunctionCode: Byte; Data: array of Byte) of object;
  TModbusExceptionEvent = procedure(Sender: TObject; TransactionId: Word;
    UnitId, FunctionCode, ExceptionCode: Byte) of object;

  TModbusClientState = (mcsIdle, mcsSend, mcsReceive);

  TModbusClientThread = class(TThread)
  private
    FTcpClient: TTcpClient;
    FRemoteHost: string;
    FRemotePort: string;
    FBuffer: array of Byte;
    FSocketError: Integer;
    FOnTcpClientConnect: TSocketNotifyEvent;
    FOnTcpClientDisconnect: TSocketNotifyEvent;
    FOnTcpClientConnectionAborted: TSocketNotifyEvent;
    FOnTcpClientError: TSocketErrorEvent;
    FOnTcpClientReceive: TSocketDataEvent;
    FStream: TStream;
    FState: TModbusClientState;
    FConnectTime: Integer;
    procedure ClientConnect;
    procedure ClientConnectionAborted;
    procedure ClientError;
    procedure ClientReceive;
    procedure DoTcpClientConnect(Sender: TObject);
    procedure DoTcpClientDisconnect(Sender: TObject);
    procedure DoConnectionAborted;
    procedure DoTcpClientError(Sender: TObject; SocketError: Integer);
    procedure DoTcpClientReceive(Sender: TObject; Buf: PChar; var DataLen: Integer);
  protected
    procedure Execute; override;
    function SendBuf(var Buf; BufSize: Integer): Integer;
    function SendStream(Stream: TStream): Integer;
  public
    constructor Create;
    destructor Destroy; override;
    procedure ConfirmPacket;
    property RemoteHost: string read FRemoteHost write FRemoteHost;
    property RemotePort: string read FRemotePort write FRemotePort;
    property State: TModbusClientState read FState;
    property ConnectTime: Integer read FConnectTime write FConnectTime;
    property OnTcpClientConnect: TSocketNotifyEvent read FOnTcpClientConnect write
        FOnTcpClientConnect;
    property OnTcpClientDisconnect: TSocketNotifyEvent read FOnTcpClientDisconnect
        write FOnTcpClientDisconnect;
    property OnTcpClientError: TSocketErrorEvent read FOnTcpClientError write
        FOnTcpClientError;
    property OnTcpClientReceive: TSocketDataEvent read FOnTcpClientReceive write
        FOnTcpClientReceive;
    property OnTcpClientConnectionAborted: TSocketNotifyEvent read
        FOnTcpClientConnectionAborted write FOnTcpClientConnectionAborted;
  end;

  TModbusTcpClient = class(TComponent)
  private
    FClientThread: TModbusClientThread;
    FReceiveState: TReceiveState;
    FReceiveStream: TMemoryStream;
    FTransactionId: Word;
    FProtocolId: Word;
    FLengthField: Word;
    FUnitId: Byte;
    FFunctionCode: Byte;
    FOnException: TModbusExceptionEvent;
    FOnForceMultipleCoils: TModbusForceMultipleCoilsEvent;
    FOnGetCommEventCounter: TModbusGetCommEventCounterEvent;
    FOnReadCoils: TModbusReadCoilsEvent;
    FOnReadExceptionStatus: TModbusReadExceptionStatusEvent;
    FOnReadInputDiscretes: TModbusReadInputDiscretesEvent;
    FOnReadInputRegisters: TModbusReadInputRegistersEvent;
    FOnReadMultipleRegisters: TModbusReadMultipleRegistersEvent;
    FOnResponse: TModbusResponseEvent;
    FOnWriteCoil: TModbusWriteCoilEvent;
    FOnWriteMultipleRegisters: TModbusWriteMultipleRegistersEvent;
    FOnWriteSingleRegister: TModbusWriteSingleRegisterEvent;
    FOnSocketError: TSocketErrorEvent;
    FOnSocketConnect: TSocketNotifyEvent;
    FOnSocketDisconnect: TSocketNotifyEvent;
    FOnSocketConnectionAborted: TSocketNotifyEvent;
    FOnReadWriteRegisters: TModbusReadWriteRegistersEvent;
    procedure DoForceMultipleCoils;
    procedure DoGetCommEventCounter;
    procedure DoReadCoils;
    procedure DoReadExceptionStatus;
    procedure DoReadInputDiscretes;
    procedure DoReadInputRegisters;
    procedure DoReadMultipleRegisters;
    procedure DoWriteCoil;
    procedure DoWriteMultipleRegisters;
    procedure DoWriteSingleRegister;
    procedure DoException;
    procedure DoResponse;
    procedure DoSocketConnect(Sender: TObject);
    procedure DoSocketDisconnect(Sender: TObject);
    procedure DoSocketConnectionAborted(Sender: TObject);
    procedure DoSocketError(Sender: TObject; SocketError: Integer);
    procedure DoTcpClientReceive(Sender: TObject; Buf: PChar; var DataLen: Integer);
    procedure DoReadWriteRegisters;
    function GetRemotePort: string;
    procedure SetRemotePort(const Value: string);
    function GetRemoteHost: string;
    procedure SetRemoteHost(const Value: string);
    function GetConnectTime: Integer;
    procedure SetConnectTime(Value: Integer);
    function GetState: TModbusClientState;
  protected
    procedure ProcessReceiveStream(Stream: TStream); virtual;
    procedure HandleModbusPaket; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function SendFunctionArray(Data: array of Byte): Integer;
    function SendFunctionStream(Stream: TStream): Boolean;
    function ReadCoils(TransactionId, RefNum, BitCount: Word;
      UnitId: Byte = 1): Boolean;
    function ReadInputDiscretes(TransactionId, RefNum, BitCount: Word;
      UnitId: Byte = 1): Boolean;
    function ReadMultipleRegisters(TransactionId, RefNum, WordCount: Word;
      UnitId: Byte = 1): Boolean;
    function ReadInputRegisters(TransactionId, RefNum, WordCount: Word;
      UnitId: Byte = 1): Boolean;
    function WriteCoil(TransactionId, RefNum: Word; Value: Boolean;
      UnitId: Byte = 1): Boolean;
    function WriteSingleRegister(TransactionId, RefNum, Value: Word;
      UnitId: Byte = 1): Boolean;
    function ForceMultipleCoils(TransactionId, RefNum, BitCount: Word;
      Data: array of Byte; UnitId: Byte = 1): Boolean;
    function WriteMultipleRegisters(TransactionId, RefNum, WordCount: Word;
      Data: array of Word; UnitId: Byte = 1): Boolean;
    function ReadWriteRegisters(TransactionId, ReadRefNum, ReadWordCount, 
        WriteRefNum, WriteWordCount: Word; Data: array of Word; UnitId: Byte): 
        Boolean;
    function ReadExceptionStatus(TransactionId: Word;
      UnitId: Byte = 1): Boolean;
    function GetCommEventCounter(TransactionId: Word;
      UnitId: Byte = 1): Boolean;
    property State: TModbusClientState read GetState;
  published
    property RemoteHost: string read GetRemoteHost write SetRemoteHost;
    property RemotePort: string read GetRemotePort write SetRemotePort;
    property OnSocketConnect: TSocketNotifyEvent read FOnSocketConnect write
        FOnSocketConnect;
    property OnSocketDisconnect: TSocketNotifyEvent read FOnSocketDisconnect
        write FOnSocketDisconnect;
    property OnSocketConnectionAborted: TSocketNotifyEvent read
        FOnSocketConnectionAborted write FOnSocketConnectionAborted;
    property OnSocketError: TSocketErrorEvent read FOnSocketError write
        FOnSocketError;
    property OnReadCoils: TModbusReadCoilsEvent read FOnReadCoils
        write FOnReadCoils;
    property OnReadInputDiscretes: TModbusReadInputDiscretesEvent
        read FOnReadInputDiscretes write FOnReadInputDiscretes;
    property OnReadMultipleRegisters: TModbusReadMultipleRegistersEvent
        read FOnReadMultipleRegisters write FOnReadMultipleRegisters;
    property OnReadInputRegisters: TModbusReadInputRegistersEvent
        read FOnReadInputRegisters write FOnReadInputRegisters;
    property OnWriteCoil: TModbusWriteCoilEvent read FOnWriteCoil
        write FOnWriteCoil;
    property OnWriteSingleRegister: TModbusWriteSingleRegisterEvent
        read FOnWriteSingleRegister write FOnWriteSingleRegister;
    property OnReadExceptionStatus: TModbusReadExceptionStatusEvent
        read FOnReadExceptionStatus write FOnReadExceptionStatus;
    property OnForceMultipleCoils: TModbusForceMultipleCoilsEvent
        read FOnForceMultipleCoils write FOnForceMultipleCoils;
    property OnWriteMultipleRegisters: TModbusWriteMultipleRegistersEvent
        read FOnWriteMultipleRegisters write FOnWriteMultipleRegisters;
    property OnReadWriteRegisters: TModbusReadWriteRegistersEvent
        read FOnReadWriteRegisters write FOnReadWriteRegisters;
    property OnGetCommEventCounter: TModbusGetCommEventCounterEvent
        read FOnGetCommEventCounter write FOnGetCommEventCounter;
    property OnException: TModbusExceptionEvent read FOnException
        write FOnException;
    property OnResponse: TModbusResponseEvent read FOnResponse
        write FOnResponse;
    property ConnectTime: Integer read GetConnectTime write SetConnectTime
        default DConnectTime;
  end;

function ModbusFunctionToStr(ModbusFunction: Byte): string;
function ModbusExceptionToStr(ExceptionCode: Byte): string;

implementation

uses
{$IFDEF DOS}
  {$IFDEF WDOSX}
  WDosSock,
  WDosClock,
  {$ENDIF}
  {$IFDEF WIN}
  WinSock,
  {$ENDIF}
{$ELSE}
  WinSock,
{$ENDIF}
  Math,
  ModbusConst;

const
  HEADER_OFFSET = 8;

{
 ***************************** common functions ********************************
}

function GetTimeCounter: Integer;
begin
{$IFDEF DOS}
  {$IFDEF WDOSX}
  Result := GetMSCounter;
  {$ENDIF}
  {$IFDEF WIN}
  Result := GetTickCount;
  {$ENDIF}
{$ELSE}
  Result := GetTickCount;
{$ENDIF}
end;

function ModbusFunctionToStr(ModbusFunction: Byte): string;

const
  SModbusFunction: array[0..23] of string = (
    SUnknown, SReadCoils,
    SReadInputDiscretes, SReadMultipleRegisters,
    SReadInputRegisters, SWriteCoil,
    SWriteSingleRegister, SReadExceptionStatus,
    SUnknown, SUnknown,
    SUnknown, SGetCommEventCounter,
    SUnknown, SUnknown,
    SUnknown, SForceMultipleCoils,
    SWriteMultipleRegisters, SUnknown,
    SUnknown, SUnknown,
    SUnknown, SUnknown,
    SUnknown, SReadWriteRegisters);

begin
  if ModbusFunction > High(SModbusFunction) then
    ModbusFunction := 0;
  Result := SModbusFunction[ModbusFunction];
end;

function ModbusExceptionToStr(ExceptionCode: Byte): string;

const
  saExceptionCode: array[0..11] of string = (
    SExceptUnknown, SExceptIllegalFunction,
    SExceptIllegalDataAddress, SExceptIllegalDataValue,
    SExceptIllegalResponseLength, SExceptAcknowledge,
    SExceptSlaveDeviceBusy, SExceptNegativeAcknowledge,
    SExceptMemoryParityError, SExceptUnknown,
    SExceptGatewayPathUnavailable, SExceptGatewayTargetDeviceFailed);

begin
  if ExceptionCode > High(saExceptionCode) then
    ExceptionCode := 0;
  Result := saExceptionCode[ExceptionCode];
end;


{
 ***************************** TModbusTcpClient ********************************
}

constructor TModbusClientThread.Create;
begin
  inherited Create(True);
  FRemotePort := IntToStr(MODBUS_TCP_PORT);
  FState := mcsIdle;
  FTcpClient := TTcpClient.Create(nil);
  FTcpClient.OnConnect := DoTcpClientConnect;
  FTcpClient.OnDisconnect := DoTcpClientDisconnect;
  FTcpClient.OnError := DoTcpClientError;
  FTcpClient.OnReceive := DoTcpClientReceive;
  FStream := TMemoryStream.Create;
end;

destructor TModbusClientThread.Destroy;
begin
  Terminate;
  WaitFor;
  FStream.Free;
  FTcpClient.Free;
  inherited Destroy;
end;

procedure TModbusClientThread.ClientConnect;
begin
  OnTcpClientConnect(FTcpClient);
end;

procedure TModbusClientThread.DoTcpClientConnect(Sender: TObject);
begin
  if Assigned(FOnTcpClientConnect) then
    Synchronize(ClientConnect);
end;

procedure TModbusClientThread.ClientConnectionAborted;
begin
  OnTcpClientConnectionAborted(FTcpClient);
end;

procedure TModbusClientThread.DoConnectionAborted;
begin
  if Assigned(FOnTcpClientConnectionAborted) then
    Synchronize(ClientConnectionAborted);
end;

procedure TModbusClientThread.DoTcpClientDisconnect(Sender: TObject);
begin
  if Assigned(FOnTcpClientDisconnect) then
    OnTcpClientDisconnect(Sender);
end;

procedure TModbusClientThread.ClientError;
begin
  OnTcpClientError(FTcpClient, FSocketError);
end;

procedure TModbusClientThread.DoTcpClientError(Sender: TObject; SocketError:
    Integer);
begin
  if Assigned(FOnTcpClientError) then
  begin
    FSocketError := SocketError;
    Synchronize(ClientError);
  end;
end;

procedure TModbusClientThread.ClientReceive;
var
  Len: Integer;
begin
  Len := Length(FBuffer);
  OnTcpClientReceive(FTcpClient, PChar(@FBuffer[0]), Len);
end;

procedure TModbusClientThread.DoTcpClientReceive(Sender: TObject; Buf: PChar;
    var DataLen: Integer);
begin
  if Assigned(FOnTcpClientReceive) then
  begin
    SetLength(FBuffer, DataLen);
    Move(Buf[0], FBuffer[0], DataLen);
    Synchronize(ClientReceive);
  end;
end;

procedure TModbusClientThread.Execute;
var
  Buffer: array[0..255] of Byte;
  TimeOut: Boolean;
  StartTime: Integer;
begin
  // initialize StartTime
  StartTime := GetTimeCounter;

  while not Terminated do
  begin
    // determine TimeOut
    TimeOut := (FTcpClient.Connected
        and ((GetTimeCounter - StartTime) > ConnectTime));

    case State of
    mcsIdle:
      begin
        if TimeOut then
          FTcpClient.Disconnect;
      end;
    mcsSend:
      begin
        // set Host and Port of the socket
        // this closes automatic the open socket
        FTcpClient.RemoteHost := FRemoteHost;
        FTcpClient.RemotePort := FRemotePort;

        // connect the socket and send the stream
        if FTcpClient.Connect then
        begin
          FStream.Position := 0;
          FTcpClient.SendStream(FStream);
          FState := mcsReceive;
          // reset TimeOut timer
          StartTime := GetTimeCounter;
        end
        else
          FState := mcsIdle;
      end;
    mcsReceive:
      begin
{$IFDEF DOS}
        // Check if we are still online
        if FTcpClient.ConnectionAborted then
        begin
          DoConnectionAborted;
          FTcpClient.Disconnect;
          FState := mcsIdle;
          Continue;
        end;

        // Any data to read from the Modbus server?
        if FTcpClient.WaitForData then
        begin
          FTcpClient.ReceiveBuf(Buffer, SizeOf(Buffer), 0);
          // reset TimeOut timer
          StartTime := GetTimeCounter;
        end
        else
          if TimeOut then
          begin
            FState := mcsIdle;
            DoConnectionAborted;
            FTcpClient.Disconnect;
            FState := mcsIdle;
          end;
{$ELSE}
        // wait max. ConnectTime for answer from the Modbus server
        if FTcpClient.WaitForData(ConnectTime) then
          FTcpClient.ReceiveBuf(Buffer, SizeOf(Buffer), 0)
        else
        begin
            FState := mcsIdle;
            DoConnectionAborted;
            FTcpClient.Disconnect;
            FState := mcsIdle;
        end;
{$ENDIF}
      end;
    end; // case
    Sleep(0);
  end;
  FTcpClient.Disconnect;
end;

function TModbusClientThread.SendBuf(var Buf; BufSize: Integer): Integer;
begin
  if State <> mcsIdle then
    raise EModbusError.Create(sModbusTcpClientNotIdle);
  FStream.Size := 0;
  Result := FStream.Write(Buf, BufSize);
  FState := mcsSend;
end;

function TModbusClientThread.SendStream(Stream: TStream): Integer;
begin
  if State <> mcsIdle then
    raise EModbusError.Create(sModbusTcpClientNotIdle);
  FStream.Size := 0;
  Result := FStream.CopyFrom(Stream, 0);
  FState := mcsSend;
end;

procedure TModbusClientThread.ConfirmPacket;
begin
  if FState = mcsReceive then
    FState := mcsIdle;
end;


{
 ***************************** TModbusTcpClient ********************************
}

constructor TModbusTcpClient.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FReceiveStream := TMemoryStream.Create;

  // create modbus client thread
  FClientThread := TModbusClientThread.Create;
  FClientThread.OnTcpClientConnect := DoSocketConnect;
  FClientThread.OnTcpClientDisconnect := DoSocketDisconnect;
  FClientThread.OnTcpClientConnectionAborted := DoSocketConnectionAborted;
  FClientThread.OnTcpClientError := DoSocketError;
  FClientThread.OnTcpClientReceive := DoTcpClientReceive;

  // initialize published properties
  RemotePort := IntToStr(MODBUS_TCP_PORT);
  ConnectTime := DConnectTime;

  // start modbus client thread
  FClientThread.Resume;
end;

destructor TModbusTcpClient.Destroy;
begin
  FClientThread.Free;
  FReceiveStream.Free;
  inherited Destroy;
end;

procedure TModbusTcpClient.DoReadCoils;
var
  Len: Byte;
  Data: array of Byte;
begin
  if Assigned(OnReadCoils) then
  begin
    FReceiveStream.ReadBuffer(Len, SizeOf(Len));
    SetLength(Data, Len);
    FReceiveStream.ReadBuffer(Data[0], Len);
    OnReadCoils(Self, FTransactionId, FUnitId, Data);
  end
  else
    DoResponse;
end;

procedure TModbusTcpClient.DoReadInputDiscretes;
var
  Len: Byte;
  Data: array of Byte;
begin
  if Assigned(OnReadInputDiscretes) then
  begin
    FReceiveStream.ReadBuffer(Len, SizeOf(Len));
    SetLength(Data, Len);
    FReceiveStream.ReadBuffer(Data[0], Len);
    OnReadInputDiscretes(Self, FTransactionId, FUnitId, Data);
  end
  else
    DoResponse;
end;

procedure TModbusTcpClient.DoReadMultipleRegisters;
var
  ByteLen: Byte;
  Len: Byte;
  Data: array of Word;
  I: Integer;
begin
  if Assigned(OnReadMultipleRegisters) then
  begin
    FReceiveStream.ReadBuffer(ByteLen, SizeOf(ByteLen));
    Len := ByteLen div 2;
    SetLength(Data, Len);
    FReceiveStream.ReadBuffer(Data[0], ByteLen);
    for I := Low(Data) to High(Data) do
      Data[I] := ntohs(Data[I]);
    OnReadMultipleRegisters(Self, FTransactionId, FUnitId, Data);
  end
  else
    DoResponse;
end;

procedure TModbusTcpClient.DoReadInputRegisters;
var
  ByteLen: Byte;
  Len: Byte;
  Data: array of Word;
  I: Integer;
begin
  if Assigned(OnReadInputRegisters) then
  begin
    FReceiveStream.ReadBuffer(ByteLen, SizeOf(ByteLen));
    Len := ByteLen div 2;
    SetLength(Data, Len);
    FReceiveStream.ReadBuffer(Data[0], ByteLen);
    for I := Low(Data) to High(Data) do
      Data[I] := ntohs(Data[I]);
    OnReadInputRegisters(Self, FTransactionId, FUnitId, Data);
  end
  else
    DoResponse;
end;

procedure TModbusTcpClient.DoWriteCoil;
var
  RefNum: Word;
  bValue: Byte;
  Value: Boolean;
begin
  if Assigned(OnWriteCoil) then
  begin
    FReceiveStream.ReadBuffer(RefNum, SizeOf(RefNum));
    FReceiveStream.ReadBuffer(bValue, SizeOf(bValue));
    RefNum := ntohs(RefNum);
    Value := (bValue = $FF);
    OnWriteCoil(Self, FTransactionId, FUnitId, RefNum, Value);
  end
  else
    DoResponse;
end;

procedure TModbusTcpClient.DoWriteSingleRegister;
var
  RefNum: Word;
  Value: Word;
begin
  if Assigned(OnWriteSingleRegister) then
  begin
    FReceiveStream.ReadBuffer(RefNum, SizeOf(RefNum));
    FReceiveStream.ReadBuffer(Value, SizeOf(Value));
    RefNum := ntohs(RefNum);
    Value := ntohs(Value);
    OnWriteSingleRegister(Self, FTransactionId, FUnitId, RefNum, Value);
  end
  else
    DoResponse;
end;

procedure TModbusTcpClient.DoReadExceptionStatus;
var
  Status: Byte;
begin
  if Assigned(OnReadExceptionStatus) then
  begin
    FReceiveStream.ReadBuffer(Status, SizeOf(Status));
    OnReadExceptionStatus(Self, FTransactionId, FUnitId, Status);
  end
  else
    DoResponse;
end;

procedure TModbusTcpClient.DoForceMultipleCoils;
var
  RefNum: Word;
  BitCount: Word;
begin
  if Assigned(OnForceMultipleCoils) then
  begin
    FReceiveStream.ReadBuffer(RefNum, SizeOf(RefNum));
    FReceiveStream.ReadBuffer(BitCount, SizeOf(BitCount));
    RefNum := ntohs(RefNum);
    BitCount := ntohs(BitCount);
    OnForceMultipleCoils(Self, FTransactionId, FUnitId, RefNum, BitCount);
  end
  else
    DoResponse;
end;

procedure TModbusTcpClient.DoWriteMultipleRegisters;
var
  RefNum: Word;
  WordCount: Word;
begin
  if Assigned(OnWriteMultipleRegisters) then
  begin
    FReceiveStream.ReadBuffer(RefNum, SizeOf(RefNum));
    FReceiveStream.ReadBuffer(WordCount, SizeOf(WordCount));
    RefNum := ntohs(RefNum);
    WordCount := ntohs(WordCount);
    OnWriteMultipleRegisters(Self, FTransactionId, FUnitId, RefNum, WordCount);
  end
  else
    DoResponse;
end;

procedure TModbusTcpClient.DoReadWriteRegisters;
var
  ByteLen: Byte;
  Len: Byte;
  Data: array of Word;
  I: Integer;
begin
  if Assigned(OnReadWriteRegisters) then
  begin
    FReceiveStream.ReadBuffer(ByteLen, SizeOf(ByteLen));
    Len := ByteLen div 2;
    SetLength(Data, Len);
    FReceiveStream.ReadBuffer(Data[0], ByteLen);
    for I := Low(Data) to High(Data) do
      Data[I] := ntohs(Data[I]);
    OnReadWriteRegisters(Self, FTransactionId, FUnitId, Data);
  end
  else
    DoResponse;
end;

procedure TModbusTcpClient.DoGetCommEventCounter;
var
  Status: Word;
  Count: Word;
begin
  if Assigned(OnGetCommEventCounter) then
  begin
    FReceiveStream.ReadBuffer(Status, SizeOf(Status));
    FReceiveStream.ReadBuffer(Count, SizeOf(Count));
    Status := ntohs(Status);
    Count := ntohs(Count);
    OnGetCommEventCounter(Self, FTransactionId, FUnitId, Status, Count);
  end
  else
    DoResponse;
end;

procedure TModbusTcpClient.DoException;
var
  FunctionCode: Byte;
  ExceptionCode: Byte;
begin
  if Assigned(OnException) then
  begin
    FunctionCode := FFunctionCode and $7F;
    FReceiveStream.ReadBuffer(ExceptionCode, SizeOf(ExceptionCode));
    OnException(Self, FTransactionId, FUnitId, FunctionCode, ExceptionCode);
  end;
end;

procedure TModbusTcpClient.DoResponse;
var
  Len: Integer;
  Data: array of Byte;
begin
  if Assigned(OnResponse) then
  begin
    Len := FReceiveStream.Size;
    SetLength(Data, Len);
    FReceiveStream.ReadBuffer(Data[0], Len);
    OnResponse(Self, FTransactionId, FUnitId, FFunctionCode, Data);
  end;
end;

procedure TModbusTcpClient.HandleModbusPaket;
begin
  case FFunctionCode of
    READ_COILS: DoReadCoils;
    READ_INPUT_DISCRETES: DoReadInputDiscretes;
    READ_MULTIPLE_REGISTERS: DoReadMultipleRegisters;
    READ_INPUT_REGISTERS: DoReadInputRegisters;
    WRITE_COIL: DoWriteCoil;
    WRITE_SINGLE_REGISTER: DoWriteSingleRegister;
    READ_EXCEPTION_STATUS: DoReadExceptionStatus;
    FORCE_MULTIPLE_COILS: DoForceMultipleCoils;
    WRITE_MULTIPLE_REGISTERS: DoWriteMultipleRegisters;
    GET_COMM_EVENT_COUNTER: DoGetCommEventCounter;
    READ_WRITE_REGISTERS: DoReadWriteRegisters;
    EXCEPTION_FIRST..EXCEPTION_LAST: DoException;
  else
    DoResponse;
  end; { case }
end;

procedure TModbusTcpClient.ProcessReceiveStream(Stream: TStream);
var
  HeaderLen: Integer;
  Len: Integer;
begin
  Len := Stream.Size;
  while Len > 0 do
    case FReceiveState of
      rvsHeader:
        begin
          // we do reading the header
          HeaderLen := HEADER_OFFSET - FReceiveStream.Size;
          if HeaderLen > Len then
            HeaderLen := Len;
          FReceiveStream.CopyFrom(Stream, HeaderLen);
          if FReceiveStream.Size = HEADER_OFFSET then
          begin
            // we got the whole header
            FReceiveStream.Position := 0;
            // read the header values
            FReceiveStream.ReadBuffer(FTransactionId, SizeOf(FTransactionId));
            FReceiveStream.ReadBuffer(FProtocolId, SizeOf(FProtocolId));
            FReceiveStream.ReadBuffer(FLengthField, SizeOf(FLengthField));
            FReceiveStream.ReadBuffer(FUnitId, SizeOf(FUnitId));
            FReceiveStream.ReadBuffer(FFunctionCode, SizeOf(FFunctionCode));
            FTransactionId := ntohs(FTransactionId);
            FProtocolId := ntohs(FProtocolId);
            FLengthField := ntohs(FLengthField);
            FReceiveStream.Clear;

            // check for invaild packet headers
            if (FProtocolId <> 0) or (FLengthField <= 0) then
            begin
              // drop the rest of packet data
              Len := 0;
              Continue;
            end;
            // go to next state
            FReceiveState := rvsBody;
          end;
          // calculate rest of stream
          with Stream do
            Len := Size - Position;
        end;
      rvsBody:
        begin
          HeaderLen := FLengthField - 2;
          if HeaderLen > Len then
            HeaderLen := Len;
          FReceiveStream.CopyFrom(Stream, HeaderLen);
          if FReceiveStream.Size = FLengthField - 2 then
          begin
            // we got the whole body
            FReceiveStream.Position := 0;
            // confirm the received packet
            FClientThread.ConfirmPacket;
            // handle received modbus paket
            try
              HandleModbusPaket;
            except
              // do nothing
            end;
            // clear receive stream
            FReceiveStream.Clear;
            // go to init state
            FReceiveState := rvsHeader;
          end;
          // drop rest of stream (fix buggy answers from Modbus/TCP servers)
          Len := 0;
        end;
    end;
end;

procedure TModbusTcpClient.DoTcpClientReceive(Sender: TObject; Buf: PChar; var
    DataLen: Integer);
var
  Stream: TMemoryStream;
begin
  Stream := TMemoryStream.Create;
  try
    Stream.WriteBuffer(Buf[0], DataLen);
    Stream.Position := 0;
    try
      ProcessReceiveStream(Stream);
    except
      FReceiveState := rvsHeader;
      FReceiveStream.Clear;
      raise;
    end;
  finally
    Stream.Free;
  end;
end;

function TModbusTcpClient.SendFunctionArray(Data: array of Byte): Integer;
begin
  Result := FClientThread.SendBuf(Data[0], Length(Data));
end;

function TModbusTcpClient.SendFunctionStream(Stream: TStream): Boolean;
begin
  Result := (FClientThread.SendStream(Stream) = Stream.Size);
end;

function TModbusTcpClient.ReadCoils(TransactionId, RefNum, BitCount: Word;
  UnitId: Byte): Boolean;
var
  Stream: TMemoryStream;
  ProtocolId: Word;
  LengthField: Word;
  FunctionCode: Byte;
begin
  // check range
  if not InRange(BitCount, 1, 2000) then
    raise EModbusError.CreateFmt('Bit count %d not in (1...2000)', [BitCount]);

  Stream := TMemoryStream.Create;
  try
    TransactionId := htons(TransactionId);
    ProtocolId := htons(0);
    LengthField := htons(6);
    FunctionCode := READ_COILS;
    RefNum := htons(RefNum);
    BitCount := htons(BitCount);
    Stream.WriteBuffer(TransactionId, SizeOf(TransactionId));
    Stream.WriteBuffer(ProtocolId, SizeOf(ProtocolId));
    Stream.WriteBuffer(LengthField, SizeOf(LengthField));
    Stream.WriteBuffer(UnitId, SizeOf(UnitId));
    Stream.WriteBuffer(FunctionCode, SizeOf(FunctionCode));
    Stream.WriteBuffer(RefNum, SizeOf(RefNum));
    Stream.WriteBuffer(BitCount, SizeOf(BitCount));
    Stream.Position := 0;
    Result := SendFunctionStream(Stream);
  finally
    Stream.Free;
  end;
end;

function TModbusTcpClient.ReadInputDiscretes(TransactionId, RefNum, BitCount:
  Word; UnitId: Byte): Boolean;
var
  Stream: TMemoryStream;
  ProtocolId: Word;
  LengthField: Word;
  FunctionCode: Byte;
begin
  // check range
  if not InRange(BitCount, 1, 2000) then
    raise EModbusError.CreateFmt('Bit count %d not in (1...2000)', [BitCount]);

  Stream := TMemoryStream.Create;
  try
    TransactionId := htons(TransactionId);
    ProtocolId := htons(0);
    LengthField := htons(6);
    FunctionCode := READ_INPUT_DISCRETES;
    RefNum := htons(RefNum);
    BitCount := htons(BitCount);
    Stream.WriteBuffer(TransactionId, SizeOf(TransactionId));
    Stream.WriteBuffer(ProtocolId, SizeOf(ProtocolId));
    Stream.WriteBuffer(LengthField, SizeOf(LengthField));
    Stream.WriteBuffer(UnitId, SizeOf(UnitId));
    Stream.WriteBuffer(FunctionCode, SizeOf(FunctionCode));
    Stream.WriteBuffer(RefNum, SizeOf(RefNum));
    Stream.WriteBuffer(BitCount, SizeOf(BitCount));
    Stream.Position := 0;
    Result := SendFunctionStream(Stream);
  finally
    Stream.Free;
  end;
end;

function TModbusTcpClient.ReadMultipleRegisters(TransactionId, RefNum,
  WordCount: Word; UnitId: Byte): Boolean;
var
  Stream: TMemoryStream;
  ProtocolId: Word;
  LengthField: Word;
  FunctionCode: Byte;
begin
  // check range
  if not InRange(WordCount, 1, 125) then
    raise EModbusError.CreateFmt('Word count %d not in (1...125)', [WordCount]);

  Stream := TMemoryStream.Create;
  try
    TransactionId := htons(TransactionId);
    ProtocolId := htons(0);
    LengthField := htons(6);
    FunctionCode := READ_MULTIPLE_REGISTERS;
    RefNum := htons(RefNum);
    WordCount := htons(WordCount);
    Stream.WriteBuffer(TransactionId, SizeOf(TransactionId));
    Stream.WriteBuffer(ProtocolId, SizeOf(ProtocolId));
    Stream.WriteBuffer(LengthField, SizeOf(LengthField));
    Stream.WriteBuffer(UnitId, SizeOf(UnitId));
    Stream.WriteBuffer(FunctionCode, SizeOf(FunctionCode));
    Stream.WriteBuffer(RefNum, SizeOf(RefNum));
    Stream.WriteBuffer(WordCount, SizeOf(WordCount));
    Stream.Position := 0;
    Result := SendFunctionStream(Stream);
  finally
    Stream.Free;
  end;
end;

function TModbusTcpClient.ReadInputRegisters(TransactionId, RefNum, WordCount:
  Word; UnitId: Byte): Boolean;
var
  Stream: TMemoryStream;
  ProtocolId: Word;
  LengthField: Word;
  FunctionCode: Byte;
begin
  // check range
  if not InRange(WordCount, 1, 125) then
    raise EModbusError.CreateFmt('Word count %d not in (1...125)', [WordCount]);

  Stream := TMemoryStream.Create;
  try
    TransactionId := htons(TransactionId);
    ProtocolId := htons(0);
    LengthField := htons(6);
    FunctionCode := READ_INPUT_REGISTERS;
    RefNum := htons(RefNum);
    WordCount := htons(WordCount);
    Stream.WriteBuffer(TransactionId, SizeOf(TransactionId));
    Stream.WriteBuffer(ProtocolId, SizeOf(ProtocolId));
    Stream.WriteBuffer(LengthField, SizeOf(LengthField));
    Stream.WriteBuffer(UnitId, SizeOf(UnitId));
    Stream.WriteBuffer(FunctionCode, SizeOf(FunctionCode));
    Stream.WriteBuffer(RefNum, SizeOf(RefNum));
    Stream.WriteBuffer(WordCount, SizeOf(WordCount));
    Stream.Position := 0;
    Result := SendFunctionStream(Stream);
  finally
    Stream.Free;
  end;
end;

function TModbusTcpClient.WriteCoil(TransactionId, RefNum: Word; Value: Boolean;
  UnitId: Byte): Boolean;
var
  Stream: TMemoryStream;
  ProtocolId: Word;
  LengthField: Word;
  FunctionCode: Byte;
  bValue, bZero: Byte;
begin
  Stream := TMemoryStream.Create;
  try
    TransactionId := htons(TransactionId);
    ProtocolId := htons(0);
    LengthField := htons(6);
    FunctionCode := WRITE_COIL;
    RefNum := htons(RefNum);
    if Value then
      bValue := $FF
    else
      bValue := $00;
    bZero := 0;
    Stream.WriteBuffer(TransactionId, SizeOf(TransactionId));
    Stream.WriteBuffer(ProtocolId, SizeOf(ProtocolId));
    Stream.WriteBuffer(LengthField, SizeOf(LengthField));
    Stream.WriteBuffer(UnitId, SizeOf(UnitId));
    Stream.WriteBuffer(FunctionCode, SizeOf(FunctionCode));
    Stream.WriteBuffer(RefNum, SizeOf(RefNum));
    Stream.WriteBuffer(bValue, SizeOf(bValue));
    Stream.WriteBuffer(bZero, SizeOf(bZero));
    Stream.Position := 0;
    Result := SendFunctionStream(Stream);
  finally
    Stream.Free;
  end;
end;

function TModbusTcpClient.WriteSingleRegister(TransactionId, RefNum, Value:
  Word; UnitId: Byte): Boolean;
var
  Stream: TMemoryStream;
  ProtocolId: Word;
  LengthField: Word;
  FunctionCode: Byte;
begin
  Stream := TMemoryStream.Create;
  try
    TransactionId := htons(TransactionId);
    ProtocolId := htons(0);
    LengthField := htons(6);
    FunctionCode := WRITE_SINGLE_REGISTER;
    RefNum := htons(RefNum);
    Value := htons(Value);
    Stream.WriteBuffer(TransactionId, SizeOf(TransactionId));
    Stream.WriteBuffer(ProtocolId, SizeOf(ProtocolId));
    Stream.WriteBuffer(LengthField, SizeOf(LengthField));
    Stream.WriteBuffer(UnitId, SizeOf(UnitId));
    Stream.WriteBuffer(FunctionCode, SizeOf(FunctionCode));
    Stream.WriteBuffer(RefNum, SizeOf(RefNum));
    Stream.WriteBuffer(Value, SizeOf(Value));
    Stream.Position := 0;
    Result := SendFunctionStream(Stream);
  finally
    Stream.Free;
  end;
end;

function TModbusTcpClient.ReadExceptionStatus(TransactionId: Word;
  UnitId: Byte): Boolean;
var
  Stream: TMemoryStream;
  ProtocolId: Word;
  LengthField: Word;
  FunctionCode: Byte;
begin
  Stream := TMemoryStream.Create;
  try
    TransactionId := htons(TransactionId);
    ProtocolId := htons(0);
    LengthField := htons(2);
    FunctionCode := READ_EXCEPTION_STATUS;
    Stream.WriteBuffer(TransactionId, SizeOf(TransactionId));
    Stream.WriteBuffer(ProtocolId, SizeOf(ProtocolId));
    Stream.WriteBuffer(LengthField, SizeOf(LengthField));
    Stream.WriteBuffer(UnitId, SizeOf(UnitId));
    Stream.WriteBuffer(FunctionCode, SizeOf(FunctionCode));
    Stream.Position := 0;
    Result := SendFunctionStream(Stream);
  finally
    Stream.Free;
  end;
end;

function TModbusTcpClient.GetCommEventCounter(TransactionId: Word;
  UnitId: Byte): Boolean;
var
  Stream: TMemoryStream;
  ProtocolId: Word;
  LengthField: Word;
  FunctionCode: Byte;
begin
  Stream := TMemoryStream.Create;
  try
    TransactionId := htons(TransactionId);
    ProtocolId := htons(0);
    LengthField := htons(2);
    FunctionCode := GET_COMM_EVENT_COUNTER;
    Stream.WriteBuffer(TransactionId, SizeOf(TransactionId));
    Stream.WriteBuffer(ProtocolId, SizeOf(ProtocolId));
    Stream.WriteBuffer(LengthField, SizeOf(LengthField));
    Stream.WriteBuffer(UnitId, SizeOf(UnitId));
    Stream.WriteBuffer(FunctionCode, SizeOf(FunctionCode));
    Stream.Position := 0;
    Result := SendFunctionStream(Stream);
  finally
    Stream.Free;
  end;
end;

function TModbusTcpClient.ForceMultipleCoils(TransactionId, RefNum, BitCount:
  Word; Data: array of Byte; UnitId: Byte): Boolean;
var
  Stream: TMemoryStream;
  ProtocolId: Word;
  LengthField: Word;
  FunctionCode: Byte;
  ByteCount: Byte;
begin
  if not InRange(BitCount, 1, 800) then
    raise EModbusError.CreateFmt('Bit count %d not in (1...800)', [BitCount]);

  Stream := TMemoryStream.Create;
  try
    ByteCount := Length(Data);
    TransactionId := htons(TransactionId);
    ProtocolId := htons(0);
    LengthField := ByteCount + 7;
    LengthField := htons(LengthField);
    FunctionCode := FORCE_MULTIPLE_COILS;
    RefNum := htons(RefNum);
    BitCount := htons(BitCount);
    Stream.WriteBuffer(TransactionId, SizeOf(TransactionId));
    Stream.WriteBuffer(ProtocolId, SizeOf(ProtocolId));
    Stream.WriteBuffer(LengthField, SizeOf(LengthField));
    Stream.WriteBuffer(UnitId, SizeOf(UnitId));
    Stream.WriteBuffer(FunctionCode, SizeOf(FunctionCode));
    Stream.WriteBuffer(RefNum, SizeOf(RefNum));
    Stream.WriteBuffer(BitCount, SizeOf(BitCount));
    Stream.WriteBuffer(ByteCount, SizeOf(ByteCount));
    Stream.WriteBuffer(Data[0], ByteCount);
    Stream.Position := 0;
    Result := SendFunctionStream(Stream);
  finally
    Stream.Free;
  end;
end;

function TModbusTcpClient.WriteMultipleRegisters(TransactionId, RefNum,
  WordCount: Word; Data: array of Word; UnitId: Byte): Boolean;
var
  Stream: TMemoryStream;
  ProtocolId: Word;
  LengthField: Word;
  FunctionCode: Byte;
  ByteCount: Byte;
  I: Integer;
  Value: Word;
begin
  // check range
  if not InRange(WordCount, 1, 100) then
    raise EModbusError.CreateFmt('Word count %d not in (1...100)', [WordCount]);

  Stream := TMemoryStream.Create;
  try
    ByteCount := Length(Data) * 2;
    TransactionId := htons(TransactionId);
    ProtocolId := htons(0);
    LengthField := ByteCount + 7;
    LengthField := htons(LengthField);
    FunctionCode := WRITE_MULTIPLE_REGISTERS;
    RefNum := htons(RefNum);
    WordCount := htons(WordCount);
    Stream.WriteBuffer(TransactionId, SizeOf(TransactionId));
    Stream.WriteBuffer(ProtocolId, SizeOf(ProtocolId));
    Stream.WriteBuffer(LengthField, SizeOf(LengthField));
    Stream.WriteBuffer(UnitId, SizeOf(UnitId));
    Stream.WriteBuffer(FunctionCode, SizeOf(FunctionCode));
    Stream.WriteBuffer(RefNum, SizeOf(RefNum));
    Stream.WriteBuffer(WordCount, SizeOf(WordCount));
    Stream.WriteBuffer(ByteCount, SizeOf(ByteCount));
    for I := Low(Data) to High(Data) do
    begin
      Value := htons(Data[I]);
      Stream.WriteBuffer(Value, SizeOf(Value));
    end;
    Stream.Position := 0;
    Result := SendFunctionStream(Stream);
  finally
    Stream.Free;
  end;
end;

function TModbusTcpClient.ReadWriteRegisters(TransactionId, ReadRefNum, 
    ReadWordCount, WriteRefNum, WriteWordCount: Word; Data: array of Word;
    UnitId: Byte): Boolean;
var
  Stream: TMemoryStream;
  ProtocolId: Word;
  LengthField: Word;
  FunctionCode: Byte;
  WriteByteCount: Byte;
  I: Integer;
  Value: Word;
begin
  // check range
  if not InRange(ReadWordCount, 1, 125) then
    raise EModbusError.CreateFmt('Word count for read %d not in (1...125)',
        [ReadWordCount]);
  if not InRange(WriteWordCount, 1, 100) then
    raise EModbusError.CreateFmt('Word count for write %d not in (1...100)',
        [WriteWordCount]);

  Stream := TMemoryStream.Create;
  try
    WriteByteCount := Length(Data) * 2;
    TransactionId := htons(TransactionId);
    ProtocolId := htons(0);
    LengthField := WriteByteCount + 11;
    LengthField := htons(LengthField);
    FunctionCode := READ_WRITE_REGISTERS;
    ReadRefNum := htons(ReadRefNum);
    ReadWordCount := htons(ReadWordCount);
    WriteRefNum := htons(WriteRefNum);
    WriteWordCount := htons(WriteWordCount);
    Stream.WriteBuffer(TransactionId, SizeOf(TransactionId));
    Stream.WriteBuffer(ProtocolId, SizeOf(ProtocolId));
    Stream.WriteBuffer(LengthField, SizeOf(LengthField));
    Stream.WriteBuffer(UnitId, SizeOf(UnitId));
    Stream.WriteBuffer(FunctionCode, SizeOf(FunctionCode));
    Stream.WriteBuffer(ReadRefNum, SizeOf(ReadRefNum));
    Stream.WriteBuffer(ReadWordCount, SizeOf(ReadWordCount));
    Stream.WriteBuffer(WriteRefNum, SizeOf(WriteRefNum));
    Stream.WriteBuffer(WriteWordCount, SizeOf(WriteWordCount));
    Stream.WriteBuffer(WriteByteCount, SizeOf(WriteByteCount));
    for I := Low(Data) to High(Data) do
    begin
      Value := htons(Data[I]);
      Stream.WriteBuffer(Value, SizeOf(Value));
    end;
    Stream.Position := 0;
    Result := SendFunctionStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TModbusTcpClient.DoSocketConnect(Sender: TObject);
begin
  if Assigned(FOnSocketConnect) then
    OnSocketConnect(Sender);
end;

procedure TModbusTcpClient.DoSocketDisconnect(Sender: TObject);
begin
  if Assigned(FOnSocketDisconnect) then
    OnSocketDisconnect(Sender);
end;

procedure TModbusTcpClient.DoSocketConnectionAborted(Sender: TObject);
begin
  if Assigned(FOnSocketConnectionAborted) then
    OnSocketConnectionAborted(Sender);
end;

procedure TModbusTcpClient.DoSocketError(Sender: TObject; SocketError: Integer);
begin
  if Assigned(FOnSocketError) then
    OnSocketError(Sender, SocketError);
end;

function TModbusTcpClient.GetRemotePort: string;
begin
  Result := FClientThread.RemotePort;
end;

procedure TModbusTcpClient.SetRemotePort(const Value: string);
begin
  FClientThread.RemotePort := Value;
end;

function TModbusTcpClient.GetRemoteHost: string;
begin
  Result := FClientThread.RemoteHost;
end;

procedure TModbusTcpClient.SetRemoteHost(const Value: string);
begin
  FClientThread.RemoteHost := Value;
end;

function TModbusTcpClient.GetConnectTime: Integer;
begin
  Result := FClientThread.ConnectTime;
end;

procedure TModbusTcpClient.SetConnectTime(Value: Integer);
begin
  FClientThread.ConnectTime := Value;
end;

function TModbusTcpClient.GetState: TModbusClientState;
begin
  Result := FClientThread.State;
end;


end.

