
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Visual Library (clx)       }
{ QStdCtrls.pas                                         }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Kylix and Delphi Cross-Platform Runtime Library   }
{     Copyright (C) 1996, 2001                          }
{     Borland Software Corporation                      }
{     www: http://www.borland.com                       }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}

{ *************************************************************************** }
{                                                                             }
{ Kylix and Delphi Cross-Platform Visual Component Library                    }
{                                                                             }
{ Copyright (c) 2000, 2001 Borland Software Corporation                       }
{                                                                             }
{ This file may be distributed and/or modified under the terms of the GNU     }
{ General Public License (GPL) version 2 as published by the Free Software    }
{ Foundation and appearing at http://www.borland.com/kylix/gpl.html.          }
{                                                                             }
{ Licensees holding a valid Borland No-Nonsense License for this Software may }
{ use this file in accordance with such license, which appears in the file    }
{ license.txt that came with this Software.                                   }
{                                                                             }
{ *************************************************************************** }
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/QStdCtrls.pas,v 1.11 2004/03/30 20:27:39 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: QStdCtrls.pas,v $
  Revision 1.11  2004/03/30 20:27:39  iwache
  TCustomMemo and TMemoStrings implemented and connected
  with WQ layer.

  Revision 1.10  2003/03/03 16:04:43  iwache
  procedure ExceptionHandler in initialization section added

  Revision 1.9  2003/01/25 11:21:16  iwache
  DWPL-Bugs-673813:
  Description:
  The dblclick event of TEdit and TListBox is invalid.
  The ReadOnly property of TEdit is invalid.

  Solution for Doubleclick:
  DblClick Event handler added in TWidgetControl.EventFilter.
  TCustomListBox.DblClick modified to fit behaviour with
  the TV-ListBox on doubleclick. An other solution would
  be fit behaiour of TV-ListBox to the VisualCLX one. But this
  would need much more work and tests.

  Solution for ReadOnly property of TEdit:
  TViInputLine added two new properties 'ReadOnly' and 'Text'.
  TViInputLine.HandleEvent and TViInputLine.SetData modified
  to handle ReadOnly property right.

  WQ.pas: modified for both bugs.
  ViTypes.pas: event code evMouseDbl  = $0800; added.
  This is a dummy TV event code.

  Revision 1.8  2003/01/22 20:52:33  iwache
  2'nd Fix for dwpl-Bug #667896 :
  Description:
  Buttons do not handle Enter...
  Solution:
  function TButton.NeedKey implemented and some modifications
  done to VisualCLX, WQ-Layer and TV-Layer.

  Revision 1.7  2003/01/20 22:18:30  iwache
  Fix for dwpl-Bug #667896 :
  Description:
  Buttons do not handle Enter, it is simply ignored, even if
  the focus is on a button (not to mention `default' buttons)
  (this applies to all DWPL versions up to v2.2.0 b3)
  Function-Test:
  Enter function tested (property Default) - OK
  ESC function tested (property Escape) - OK
  Space bar don't works with focused buttons - Error
  Solution:
  Simulate a KeyUp event for space key on buttons.
  This behaviour is different to Windows CLX where
  the event fires only once when the space key fires
  a keyup event. Unfortunately we don't have any keyup
  events with DOS and WDosX :-(

  Revision 1.6  2002/10/27 22:25:02  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
unit QStdCtrls;

{$R-,T-,H+,X+}

interface

uses
  WQ, Types, SysUtils, Classes, QTypes, QControls,
  QForms, QMenus, QGraphics, QActnList;

type
  TCustomGroupBox = class(TFrameControl)
  private
    FCaption: TCaption;
    function QGroupBox: TQGroupBox;
  protected
    procedure AlignControls(AControl: TControl; var Rect: TRect); override;
    procedure CreateWidget; override;
    procedure FontChanged; override;
    function GetText: TCaption; override;
    procedure InitWidget; override;
    procedure SetText(const Value: TCaption); override;
    function GetAlignment: TAlignment;
    procedure SetAlignment(const Value: TAlignment);
    function WantKey(Key: Integer; Shift: TShiftState; const KeyText: WideString): Boolean; override;
    function WidgetFlags: Integer; override;
    property Alignment: TAlignment read GetAlignment write SetAlignment default taLeftJustify;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Invalidate; override;
  end;

  TGroupBox = class(TCustomGroupBox)
  published
    property Align default alNone;
    property Alignment;
    property Anchors;
    property Bitmap;
    property BorderStyle default bsEtched;
    property Caption;
    property Color default clBackground;
    property Constraints;
    property DragMode;
    property Enabled;
    property Font;
    property Masked default False;
    property ParentColor default True;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint default False;
    property TabOrder;
    property TabStop default False;
    property Visible;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

type
  TLCDMode = (lcdHexadecimal, lcdDecimal, lcdOctal, lcdBinary);
  TLCDSegmentStyle = (ssOutline, ssFilled, ssFlat);
  TLCDPointSize = (ptSmall, ptLarge);
  TNumDigits = 0..99;

type
  TTextLayout = (tlTop, tlCenter, tlBottom);

  TCustomLabel = class(TFrameControl)
  private
    FAutoSize: Boolean;
    FFocusControl: TWidgetControl;
    FShowAccel: Boolean;
    procedure DoAutoSize;
    procedure GetSize(var AWidth, AHeight: Integer);
    function GetAlignment: TAlignment;
    procedure SetAlignment(const Value: TAlignment);
    procedure SetFocusControl(const Value: TWidgetControl);
    procedure SetLayout(const Value: TTextLayout);
    procedure SetWordWrap(const Value: Boolean);
    function GetLayout: TTextLayout;
    function GetWordWrap: Boolean;
    function GetTransparent: Boolean;
    procedure SetTransparent(const Value: Boolean);
    procedure SetShowAccel(const Value: Boolean);
    function QLabel: TQLabel;
  protected
    function GetText: TCaption; override;
    procedure SetText(const Value: TCaption); override;
    procedure CreateWidget; override;
    procedure FontChanged; override;
    function GetLabelText: WideString; virtual;
    procedure InitWidget; override;
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
    procedure SetAutoSize(Value: Boolean); virtual;
    procedure TextChanged; override;
    function WantKey(Key: Integer; Shift: TShiftState;
      const KeyText: WideString): Boolean; override;
    property Alignment: TAlignment read GetAlignment write SetAlignment
      default taLeftJustify;
    property AutoSize: Boolean read FAutoSize write SetAutoSize default True;
    property BorderStyle default bsNone;
    property Caption: TCaption read GetText write SetText;
    property FocusControl: TWidgetControl read FFocusControl write SetFocusControl;
    property Layout: TTextLayout read GetLayout write SetLayout default tlTop;
    property ParentColor default True;
    property ShowAccelChar: Boolean read FShowAccel write SetShowAccel default True;
    property ShowHint default False;
    property Transparent: Boolean read GetTransparent write SetTransparent default False;
    property WordWrap: Boolean read GetWordWrap write SetWordWrap default False;
  public
    constructor Create(AOwner: TComponent); override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
  end;

  TLabel = class(TCustomLabel)
  published
    property Align;
    property Alignment;
    property Anchors;
    property AutoSize;
    property Bitmap;
    property BorderStyle;
    property Caption; 
    property Color;
    property Constraints;
    property DragMode;
    property Enabled;
    property FocusControl;
    property Font;
    property Masked;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property ShowAccelChar;
    property ShowHint;  
    property Transparent;
    property Layout;
    property Visible;
    property WordWrap;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

  TEditCharCase = (ecNormal, ecUpperCase, ecLowerCase);
  TEchoMode = (emNormal, emNone, emPassword);

  TCustomEdit = class(TWidgetControl)
  private
    FOnChange: TNotifyEvent;
    FOnReturnPressed: TNotifyEvent;
    FAutoSize: Boolean;
    FHideSelection: Boolean;
    FIgnoreEvents: Boolean;
    FMaxLength: Integer;
    FCharCase: TEditCharCase;
    function FontHeight: Integer;
    function GetAlignment: TAlignment;
    function GetBorderStyle: TControlBorderStyle;
    function GetModified: Boolean;
(*
    procedure ReturnPressedHook; cdecl;
*)
    procedure SetAlignment(const Value: TAlignment);
    procedure SetBorderStyle(const Value: TControlBorderStyle);
    procedure SetMaxLength(const Value: Integer);
    procedure SetModified(const Value: Boolean);
    procedure SetOnReturnPressed(const Value: TNotifyEvent);
    procedure TextChangedHook(Sender: TObject);
    function GetAutoSelect: Boolean;
    procedure SetAutoSelect(const Value: Boolean);
    function GetReadOnly: Boolean;
    procedure SetReadOnly(const Value: Boolean);
    procedure SetAutoSize(const Value: Boolean);
    procedure SetHideSelection(const Value: Boolean);
    procedure UpdateHeight;
    procedure SetEchoMode(Value: TEchoMode);
    function GetEchoMode: TEchoMode;
  protected
    procedure Change; virtual;
    procedure CreateWidget; override;
    procedure FontChanged; override;
    function GetSelLength: Integer; virtual;
    function GetSelStart: Integer; virtual;
    function GetSelText: WideString; virtual;
    function GetText: TCaption; override;
    procedure HookEvents; override;
    procedure InitWidget; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure Loaded; override;
    procedure SetSelLength(Value: Integer); virtual;
    procedure SetSelStart(Value: Integer); virtual;
    procedure SetSelText(const Value: WideString); virtual;
    procedure SetText(const Value: TCaption); override;
    function QLineEdit: TQLineEdit;
    function EventFilter(Sender: TObject; var Event: TEvent): Boolean;
        override;
    property Alignment: TAlignment read GetAlignment write SetAlignment default taLeftJustify;
    property AutoSelect: Boolean read GetAutoSelect write SetAutoSelect default True;
    property AutoSize: Boolean read FAutoSize write SetAutoSize default True;
    property BorderStyle: TControlBorderStyle read GetBorderStyle write SetBorderStyle default bsSingle;
    property CharCase: TEditCharCase read FCharCase write FCharCase default
        ecNormal;
    property Color default clBase;
    property DragMode default dmManual;
    property EchoMode: TEchoMode read GetEchoMode write SetEchoMode default
        emNormal;
    property HideSelection: Boolean read FHideSelection write SetHideSelection default True;
    property MaxLength: Integer read FMaxLength write SetMaxLength default -1;
    property ParentColor default False;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly default False;
    property ShowHint default False;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnReturnPressed: TNotifyEvent read FOnReturnPressed write SetOnReturnPressed;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Clear; virtual;
    procedure ClearSelection;
    procedure CopyToClipboard;
    procedure CutToClipboard; virtual;
    function GetSelTextBuf(Buffer: PChar; BufSize: Integer): Integer;
    procedure PasteFromClipboard; virtual;
    procedure SelectAll;
    procedure SetSelTextBuf(Buffer: PChar);
    property Modified: Boolean read GetModified write SetModified;
    property SelLength: Integer read GetSelLength write SetSelLength;
    property SelStart: Integer read GetSelStart write SetSelStart;
    property SelText: WideString read GetSelText write SetSelText;
    property Text: TCaption read GetText write SetText;
  published
    property TabStop default True;
  end;

  TEdit = class(TCustomEdit)
  published
    property Anchors;
    property AutoSelect;
    property AutoSize;
    property Alignment;
    property BorderStyle;
    property CharCase;
    property Color;
    property Constraints;
    property DragMode;
    property EchoMode;
    property Enabled;
    property Font;
    property HideSelection;
    property MaxLength;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Text;
    property Visible;
    property OnChange;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyString;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnReturnPressed;
    property OnStartDrag;
  end;

  TCaretPos = record
    Line,
    Col: Integer;
  end;

  TMemoSelection = record
    case Integer of
      0: (Line1, Col1, Line2, Col2: Integer);
      1: (TopLeft, BottomRight: TCaretPos);
  end;

  TWrapMode = (wmNone, wmWidth, wmPixel, wmColumn);
  TWrapBreak = (wbWhitespace, wbAnywhere);
  TScrollStyle = (ssNone, ssHorizontal, ssVertical, ssBoth, ssAutoHorizontal,
    ssAutoVertical, ssAutoBoth);

  TCustomMemo = class(TFrameControl)
  private
    FLines: TStrings;
    FOnReturnPressed: TNotifyEvent;
    FOnChange: TNotifyEvent;
    FWantTabs: Boolean;
    FWantReturns: Boolean;
    FScrollBars: TScrollStyle;
    FWrapMode: TWrapMode;
    FWordWrap: Boolean;
    FMaxLength: Integer;
    function CaretPosToPos(CaretPos: TCaretPos): Integer;
    function GetAlignment: TAlignment;
    function GetCaretPos: TCaretPos;
    function GetHMargin: Integer;
    function GetModified: Boolean;
    function GetReadOnly: Boolean;
    function GetSelection: TMemoSelection;
    function GetTextWidth(LineNum: Integer): Integer;
    function GetWrapBreak: TWrapBreak;
    function GetWrapValue: Integer;
    procedure InternalSetWrapMode(NewMode: TWrapMode);
    function PosToCaretPos(Pos: Integer): TCaretPos;
    procedure ReturnPressedHook(Sender: TObject);
    procedure TextChangedHook(Sender: TObject);
    procedure CMKeyDown(var Msg: TCMKeyDown); message CM_KEYDOWN;
  protected
    procedure Change; virtual;
    procedure CreateWidget; override;
    procedure FontChanged; override;
    function GetSelLength: Integer; virtual;
    function GetSelStart: Integer; virtual;
    function GetSelText: WideString; virtual;
    function GetText: TCaption; override;
    procedure HookEvents; override;
    procedure InitWidget; override;
    function LineLength(Index: Integer): Integer;
    procedure Loaded; override;
    function NeedKey(Key: Integer; Shift: TShiftState;
      const KeyText: WideString): Boolean; override;
    procedure ReturnPressed; virtual;
    procedure SetText(const Value: TCaption); override;
    procedure SetSelText(const Value: WideString); virtual;
    procedure SetSelStart(const Value: Integer); virtual;
    procedure SetSelLength(const Value: Integer); virtual;
    procedure SetModified(const Value: Boolean); virtual;
    procedure SetReadOnly(const Value: Boolean); virtual;
    procedure SetCaretPos(const Value: TCaretPos);
    procedure SetHMargin(const Value: Integer);
    procedure SetLines(const Value: TStrings);
    procedure SetScrollBars(const Value: TScrollStyle);
    procedure SetWordWrap(const Value: Boolean);
    procedure SetMaxLength(const Value: Integer);
    procedure SetAlignment(const Value: TAlignment);
    procedure SetWrapMode(const Value: TWrapMode);
    procedure SetWrapValue(const Value: Integer);
    procedure SetWrapBreak(const Value: TWrapBreak);
    function QMultiLineEdit: TQMultiLineEdit;
    property ParentColor default False;
    property ScrollBars: TScrollStyle read FScrollBars write SetScrollBars default ssNone;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnReturnPressed: TNotifyEvent read FOnReturnPressed write FOnReturnPressed;

    property ReadOnly: Boolean read GetReadOnly write SetReadOnly default False;
    property HMargin: Integer read GetHMargin write SetHMargin default 3;
    property Alignment: TAlignment read GetAlignment write SetAlignment default taLeftJustify;
    property BorderStyle default bsSunken3d;
    property Lines: TStrings read FLines write SetLines;
    property MaxLength: Integer read FMaxLength write SetMaxLength default -1;
    property ShowHint default False;
    property TabStop default True;
    property WantReturns: Boolean read FWantReturns write FWantReturns default True;
    property WantTabs: Boolean read FWantTabs write FWantTabs default False;
    property WordWrap: Boolean read FWordWrap write SetWordWrap default True;
    property WrapAtValue: Integer read GetWrapValue write SetWrapValue default -1;
    property WrapBreak: TWrapBreak read GetWrapBreak write SetWrapBreak default wbWhitespace;
    property WrapMode: TWrapMode read FWrapMode write SetWrapMode default wmWidth;
  public
    procedure Append(const Text: WideString);
    function AtEOF: Boolean;
    procedure Clear; virtual;
    procedure ClearSelection;
    procedure UnSelect;
    procedure CopyToClipboard;
    procedure CutToClipboard;
    function GetTextBuf(Buffer: PChar; BufSize: Integer): Integer; // depricated
    function GetTextLen: Integer; // depricated
    function HasSelection: Boolean;
    procedure Insert(const Text: WideString; Mark: Boolean = False); overload;
    procedure Insert(const Text: WideString; Column, Row: Integer;
      Mark: Boolean = False); overload;
    procedure PasteFromClipboard;
    procedure SelectAll;
    property CaretPos: TCaretPos read GetCaretPos write SetCaretPos;
    property Modified: Boolean read GetModified write SetModified;
    property Selection: TMemoSelection read GetSelection;
    property SelLength: Integer read GetSelLength write SetSelLength;
    property SelStart: Integer read GetSelStart write SetSelStart;
    property SelText: WideString read GetSelText write SetSelText;
    property Text: TCaption read GetText write SetText;
    property TextWidth[LineNum: Integer]: Integer read GetTextWidth;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

  TMemo = class(TCustomMemo)
  published
    property Align;
    property Alignment;
    property Anchors;
    property BorderStyle;
    property Color;
    property Constraints;
    property DragMode;
    property Enabled;
    property Font;
    property HMargin;
    property Lines;
    property MaxLength;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ScrollBars;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property WantReturns;
    property WantTabs;
    property WordWrap;
    property WrapAtValue;
    property WrapBreak;
    property WrapMode;
    property OnChange;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyString;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

  TComboBoxStyle = (csDropDown, csDropDownList, csOwnerDrawFixed,
    csOwnerDrawVariable);

  TComboBoxInsertMode = (ciNone, ciTop, ciCurrent, ciBottom, ciAfterCurrent,
    ciBeforeCurrent);

  TOwnerDrawState = set of (odSelected, odGrayed, odDisabled, odChecked,
    odFocused, odDefault, odHotLight, odInactive, odNoAccel, odNoFocusRect,
    odReserved1, odReserved2, odComboBoxEdit);

  TComboBoxHighlightEvent = procedure(Sender: TObject; Index: Integer) of object;

  TDrawItemEvent = procedure(Sender: TObject; Index: Integer;
    Rect: TRect; State: TOwnerDrawState; var Handled: Boolean) of object;

  TMeasureItemEvent = procedure(Control: TWinControl; Index: Integer;
    var Height: Integer) of object;

  TCustomComboBox = class(TWidgetControl)
  private
    FCanvas: TCanvas;
    FItems: TStrings;
    FMaxItems: Integer;
    FMaxLength: Integer;
    FStyle: TComboBoxStyle;
    FItemHeight: Integer;
    FJustDropped: Boolean;
    FMemStream: TMemoryStream;
    FOnChange: TNotifyEvent;
    FOnHighlighted: TComboBoxHighlightEvent;
    FOnSelect: TNotifyEvent;
    FCharCase: TEditCharCase;
    FOnCloseUp: TNotifyEvent;
    FOnDrawItem: TDrawItemEvent;
    FOnMeasureItem: TMeasureItemEvent;
    FOnDropDown: TNotifyEvent;
    function AdjustHeight: Integer;
    procedure ChangeHook(Sender: TObject);
    function GetAutoComplete: Boolean;
    function GetDropDownCount: Integer;
    function GetItemIndex: Integer;
(*
    procedure HighlightedHook(Index: Integer); cdecl;
*)
    procedure SelectedHook(Sender: TObject);
    procedure SetAutoComplete(const Value: Boolean);
    procedure SetDropDownCount(const Value: Integer);
    procedure SetItemIndex(const Value: Integer);
    procedure SetItems(const Value: TStrings);
    procedure SetMaxItems(const Value: Integer);
    function GetInsertMode: TComboBoxInsertMode;
    procedure SetInsertMode(const Value: TComboBoxInsertMode);
    procedure SetDuplicates(const Value: TDuplicates);
    function GetDuplicates: TDuplicates;
    function GetDroppedDown: Boolean;
    procedure SetDroppedDown(const Value: Boolean);
    function GetSelLength: Integer;
    function GetSelStart: Integer;
    function GetSelText: WideString;
    procedure SetSelLength(const Value: Integer);
    procedure SetSelStart(const Value: Integer);
    procedure SetSelText(const Value: WideString);
(*
    procedure ItemMeasureHook(index: Integer; var height, width: Integer); cdecl;
*)
    procedure SetSorted(const Value: Boolean);
    function GetSorted: Boolean;
    procedure SetItemHeight(const Value: Integer);
    function GetItemHeight: Integer;
(*
    procedure DoDropDown;
    procedure DoCloseUp;
*)
    procedure CMKeyDown(var Msg: TCMKeyDown); message CM_KEYDOWN;
    procedure SetMaxLength(const Value: Integer);
  protected
    procedure Change; dynamic;
    procedure Click; override;
    procedure CloseUp; dynamic;
    procedure CreateWidget; override;
    procedure CursorChanged; override;
    procedure HookEvents; override;
    procedure DestroyWidget; override;
    function DrawItem(Index: Integer; Rect: TRect;
      State: TOwnerDrawState): Boolean; virtual;
    procedure DoEnter; override;
    procedure DoExit; override;
    procedure MeasureItem(Control: TWinControl; Index: Integer;
      var Height, Width: Integer); virtual;
    procedure DropDown; dynamic;
    procedure FontChanged; override;
    procedure Highlighted(Index: Integer); dynamic;
    procedure InitWidget; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure Loaded; override;
    procedure PaletteChanged(Sender: TObject); override;
    function GetText: TCaption; override;
    procedure SetText(const Value: TCaption); override;
    procedure RestoreWidgetState; override;
    procedure SaveWidgetState; override;
    procedure Select; dynamic;
    procedure SetStyle(Value: TComboBoxStyle); virtual;
    function QOpenComboBox: TQOpenComboBox;
    property AutoComplete: Boolean read GetAutoComplete write SetAutoComplete default False;
    property DropDownCount: Integer read GetDropDownCount write SetDropDownCount default 8;
    property Duplicates: TDuplicates read GetDuplicates write SetDuplicates default dupAccept;
    property InsertMode: TComboBoxInsertMode read GetInsertMode write SetInsertMode default ciNone;
    property ItemHeight: Integer read GetItemHeight write SetItemHeight default 16;
    property MaxItems: Integer read FMaxItems write SetMaxItems default -1;
    property MaxLength: Integer read FMaxLength write SetMaxLength default -1;
    property ParentColor default False;
    property Sorted: Boolean read GetSorted write SetSorted default False;
    property Style: TComboBoxStyle read FStyle write SetStyle default csDropDown;
    property Text: TCaption read GetText write SetText;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnCloseUp: TNotifyEvent read FOnCloseUp write FOnCloseUp;
    property OnDropDown: TNotifyEvent read FOnDropDown write FOnDropDown;
    property OnHighlighted: TComboBoxHighlightEvent read FOnHighlighted write FOnHighlighted;
    property OnSelect: TNotifyEvent read FOnSelect write FOnSelect;
    property OnDrawItem: TDrawItemEvent read FOnDrawItem write FOnDrawItem;
    property OnMeasureItem: TMeasureItemEvent read FOnMeasureItem write FOnMeasureItem;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Clear;
    function Focused: Boolean; override;
    procedure SelectAll;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure Sort;
    property Canvas: TCanvas read FCanvas;
    property DroppedDown: Boolean read GetDroppedDown write SetDroppedDown;
    property Items: TStrings read FItems write SetItems;
    property ItemIndex: Integer read GetItemIndex write SetItemIndex default -1;
    property SelLength: Integer read GetSelLength write SetSelLength;
    property SelStart: Integer read GetSelStart write SetSelStart;
    property SelText: WideString read GetSelText write SetSelText;
  published
    property CharCase: TEditCharCase read FCharCase write FCharCase default ecNormal;
    property TabStop default True;
  end;

  TComboBox = class(TCustomComboBox)
  published
    property Style; { Must be published before Items }
    property Anchors;
    property AutoComplete;
    property CharCase;
    property Color default clBase;
    property Constraints;
    property DragMode;
    property DropDownCount;
    property Duplicates;
    property Enabled;
    property Font;
    property InsertMode;
    property ItemHeight;
    property Items;
    property ItemIndex;
    property MaxItems;
    property MaxLength;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint default False;
    property Sorted;
    property TabOrder;
    property TabStop;
    property Text;
    property Visible;
    property OnChange;
    property OnClick;
    property OnCloseUp;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawItem;
    property OnDropDown;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnHighlighted;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyString;
    property OnKeyUp;
    property OnMeasureItem;
    property OnSelect;
    property OnStartDrag;
  end;

{ TButtonControl }

  TButtonControl = class;

  TToggleState = (tsOff, tsNoChange, tsOn);
  TToggledEvent = procedure(Sender: TObject; Toggled: Boolean) of object;
  TStateChangedEvent = procedure(Sender: TObject; State: TToggleState) of object;

  TButtonActionLink = class(TWidgetControlActionLink)
  protected
    FClient: TButtonControl;
    procedure AssignClient(AClient: TObject); override;
    function IsCheckedLinked: Boolean; override;
    procedure SetChecked(Value: Boolean); override;
  end;

  TButtonActionLinkClass = class of TButtonActionLink;

  TButtonControl = class(TWidgetControl)
  private
    FClicksDisabled: Boolean;
    FAccelChar: Char;
    FOnStateChanged: TStateChangedEvent;
    function IsCheckedStored: Boolean;
(*
    procedure StateChangedHook(State: Integer); cdecl;
*)
    function GetAutoRepeat: Boolean;
    function GetDown: Boolean;
    function GetState: TToggleState;
    procedure SetAutoRepeat(const Value: Boolean);
    procedure SetDown(const Value: Boolean);
    function IsToggleButton: Boolean;
    function QButton: TQButton;
  protected
    function GetText: TCaption; override;
    procedure SetText(const Value: TCaption); override;
    procedure ActionChange(Sender: TObject; CheckDefaults: Boolean); override;
    function GetActionLinkClass: TControlActionLinkClass; override;
    function GetChecked: Boolean; virtual;
    procedure HookEvents; override;
    procedure SetChecked(Value: Boolean); virtual;
    procedure StateChanged(State: TToggleState); dynamic;
    function WantKey(Key: Integer; Shift: TShiftState;
      const KeyText: WideString): Boolean; override;
    property Caption: TCaption read GetText write SetText;
    property Checked: Boolean read GetChecked write SetChecked stored IsCheckedStored default False;
    property Down: Boolean read GetDown write SetDown;
    property State: TToggleState read GetState;
    property AutoRepeat: Boolean read GetAutoRepeat write SetAutoRepeat default False;
    property ClicksDisabled: Boolean read FClicksDisabled write FClicksDisabled;
    property OnStateChanged: TStateChangedEvent read FOnStateChanged write FOnStateChanged;
  public
    constructor Create(AOwner: TComponent); override;
    function AccelChar: Char;
    procedure AnimateClick;
    procedure Toggle; virtual;
  end;

  TButton = class(TButtonControl)
  private
    FCancel: Boolean;
    FDefault: Boolean;
    FModalResult: TModalResult;
    FOnToggled: TToggledEvent;
    function GetToggleButton: Boolean;
    procedure SetOnToggled(const Value: TToggledEvent);
    procedure SetToggleButton(const Value: Boolean);
(*
    procedure ToggleHook(IsToggled: Boolean); cdecl;
*)
    procedure SetDefault(const Value: Boolean);
  protected
    procedure BeginAutoDrag; override;
    procedure CreateWidget; override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    function NeedKey(Key: Integer; Shift: TShiftState;
      const KeyText: WideString): Boolean; override;
    procedure Toggled(IsToggled: Boolean); dynamic;
    function WantKey(Key: Integer; Shift: TShiftState;
      const KeyText: WideString): Boolean; override;
    function GetClientRect: TRect; override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Click; override;
    property Down;
  published
    property Action;
    property Anchors;
    property Bitmap;
    property Cancel: Boolean read FCancel write FCancel default False;
    property Caption;
    property Color default clBtnFace;
    property Constraints;
    property Default: Boolean read FDefault write SetDefault default False;
    property DragMode;
    property Enabled;
    property Font;
    property ModalResult: TModalResult read FModalResult write FModalResult default mrNone;
    property ParentColor default False;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint default False;
    property TabOrder;
    property TabStop default True;
    property ToggleButton: Boolean read GetToggleButton write SetToggleButton default False;
    property Visible;
    property OnClick;
    property OnContextPopup;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyString;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
    property OnStateChanged;
    property OnToggled: TToggledEvent read FOnToggled write SetOnToggled;
  end;

  TCheckBoxState = (cbUnchecked, cbChecked, cbGrayed);

  TCustomCheckBox = class(TButtonControl)
  private
    FAllowGrayed: Boolean;
    FState: TCheckBoxState;
    procedure SetState(Value: TCheckBoxState);
    procedure SetAllowGrayed(const Value: Boolean);
    function QCheckBox: TQCheckBox;
    procedure ClickedHook(Sender: TObject);
  protected
    procedure CreateWidget; override;
    function GetChecked: Boolean; override;
    procedure SetChecked(Value: Boolean); override;
    procedure StateChanged(State: TToggleState); override;
    function WantKey(Key: Integer; Shift: TShiftState;
      const KeyText: WideString): Boolean; override;
    procedure HookEvents; override;
    property AllowGrayed: Boolean read FAllowGrayed write SetAllowGrayed default False;
    property State: TCheckBoxState read FState write SetState default cbUnchecked;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Toggle; override;
  published
    property TabStop default True;
  end;

  TCheckBox = class(TCustomCheckBox)
  published
    property Action;
    property AllowGrayed;
    property Anchors;
    property Bitmap;
    property Caption;
    property Checked;
    property Color default clBackground;
    property Constraints;
    property DragMode;
    property Enabled;
    property Font;
    property Masked default False;
    property ParentColor default True;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint default False;
    property State;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnClick;
    property OnContextPopup;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyString;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

  TRadioButton = class(TButtonControl)
  private
    procedure ClickedHook(Sender: TObject);
    procedure TurnSiblingsOff;
  protected
    procedure CreateWidget; override;
    function GetChecked: Boolean; override;
    procedure HookEvents; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X,
      Y: Integer); override;
    procedure SetChecked(Value: Boolean); override;
    function WantKey(Key: Integer; Shift: TShiftState;
      const KeyText: WideString): Boolean; override;
    function QRadioButton: TQRadioButton;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Click; override;
  published
    property Action;
    property Anchors;
    property Bitmap;
    property Caption;
    property Checked;
    property Color default clBackground;
    property Constraints;
    property DragMode;
    property Enabled;
    property Font;
    property Masked default False;
    property ParentColor default True;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint default False;
    property TabOrder;
    property TabStop default True;
    property Visible;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyString;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

(*
const
  QEventType_LBClick = QEventType(Integer(QEventType_ClxBase) + $20);
*)
type
  TListBoxLayout = (llFixedNumber, llFitToWidth, llVariable);
  TListBoxStyle = (lbStandard, lbOwnerDrawFixed, lbOwnerDrawVariable);


  TCustomListBox = class(TFrameControl)
  private
    FDeleting: Boolean;
    FItems: TStrings;
    { FTabWidth: Integer; }
    FCanvas: TCanvas;
    FStyle: TListBoxStyle;
    FOnDrawItem: TDrawItemEvent;
    FOnMeasureItem: TMeasureItemEvent;
    FItemHeight: Integer;
    FSaveItems: TStringList;
    FExtendedSelect: Boolean;
    FMultiSelect: Boolean;
    FClicking: Boolean;
(*
    procedure DrawEmptyFocus;
*)
    function GetColumns: Integer;
    function GetItemIndex: Integer;
    function GetSelCount: Integer;
    function GetSelected(Index: Integer): Boolean;
    function GetTopIndex: Integer;
    procedure SetColumns(const Value: Integer);
    procedure SetExtendedSelect(const Value: Boolean);
    procedure SetItemIndex(const Value: Integer);
    procedure SetItems(const Value: TStrings);
    procedure SetMultiSelect(const Value: Boolean);
    procedure SetSelected(Index: Integer; const Value: Boolean);
    procedure SetSorted(const Value: Boolean);
    procedure SetTopIndex(const Value: Integer);
    function GetRows: Integer;
    procedure SetRows(const Value: Integer);
    function GetColumnLayout: TListBoxLayout;
    function GetRowLayout: TListBoxLayout;
    procedure SetColumnLayout(const Value: TListBoxLayout);
    procedure SetRowLayout(const Value: TListBoxLayout);
(*
    procedure ItemMeasureHook(index: Integer; var height, width: Integer); cdecl;
*)
    procedure SetStyle(const Value: TListBoxStyle);
    function GetSorted: Boolean;
    procedure SetItemHeight(const Value: Integer);
    procedure SelChangeHook(Sender: TObject);
    procedure AdjustHeight;
  protected
    procedure Changed; virtual;
    procedure CreateWidget; override;
    procedure DblClick; override;
    procedure DeleteString(Index: Integer); dynamic;
    function DrawItem(Index: Integer; Rect: TRect;
      State: TOwnerDrawState): Boolean; virtual;
    procedure EnabledChanged; override;
    procedure MeasureItem(Control: TWinControl; Index: Integer;
      var Height, Width: Integer); virtual;
    procedure FontChanged; override;
    function GetItemData(Index: Integer): LongInt; dynamic;
    procedure InitWidget; override;
    procedure HookEvents; override;
    procedure Loaded; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure ResetContent; dynamic;
    procedure RestoreWidgetState; override;
    procedure SaveWidgetState; override;
    procedure SetItemData(Index: Integer; AData: LongInt); dynamic;
    procedure WidgetDestroyed; override;
    function QListBox: TQListBox;
    property BorderStyle default bsSunken3d;
    property Columns: Integer read GetColumns write SetColumns default 1;
    property ColumnLayout: TListBoxLayout read GetColumnLayout write SetColumnLayout default llFixedNumber;
    property ExtendedSelect: Boolean read FExtendedSelect write SetExtendedSelect default True;
    property ItemHeight: Integer read FItemHeight write SetItemHeight default 16;
    property MultiSelect: Boolean read FMultiSelect write SetMultiSelect default False;
    property ParentColor default False;
    property RowLayout: TListBoxLayout read GetRowLayout write SetRowLayout default llVariable;
    property Rows: Integer read GetRows write SetRows default 1;
    property Sorted: Boolean read GetSorted write SetSorted default False;
    property Style: TListBoxStyle read FStyle write SetStyle default lbStandard;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Clear;
    procedure ClearSelection;
    function ItemAtPos(Pos: Types.TPoint; Existing: Boolean): Integer;
    function ItemRect(Index: Integer): TRect;
    function ItemVisible(Index: Integer): Boolean;
    procedure MakeCurrentVisible;
    procedure Sort;
    procedure Update; override;
    function VisibleItemCount: Integer;
    property Canvas: TCanvas read FCanvas;
    property Items: TStrings read FItems write SetItems;
    property ItemIndex: Integer read GetItemIndex write SetItemIndex;
    property SelCount: Integer read GetSelCount;
    property Selected[Index: Integer]: Boolean read GetSelected write SetSelected;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    property TopIndex: Integer read GetTopIndex write SetTopIndex;
    property OnDrawItem: TDrawItemEvent read FOnDrawItem write FOnDrawItem;
    property OnMeasureItem: TMeasureItemEvent read FOnMeasureItem write FOnMeasureItem;
  published
    property TabStop default True;
  end;

  TListBox = class(TCustomListBox)
  published
    property Style; { Must be published before Items }
    property Align;
    property Anchors;
    property BorderStyle;
    property Color default clBase;
    property Columns;
    property ColumnLayout;
    property Constraints;
    property DragMode;
    property Enabled;
    property ExtendedSelect;
    property Font;
    property ItemHeight;
    property Items;
    property MultiSelect;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property RowLayout;
    property Rows;
    property ShowHint default False;
    property Sorted;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnDrawItem;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyString;
    property OnKeyUp;
    property OnMeasureItem;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

  TScrollCode = (scLineUp, scLineDown, scPageUp, scPageDown, scPosition,
    scTrack, scTop, scBottom, scEndScroll);

  TScrollEvent = procedure(Sender: TObject; ScrollCode: TScrollCode;
    var ScrollPos: Integer) of object;

  TScrollBar = class(TWidgetControl)
  private
    FParamsChanging: Boolean;
    FKind: TScrollBarKind;
    FOnChange: TNotifyEvent;
    FOnScroll: TScrollEvent;
    procedure DoScroll(Code: TScrollCode);
    procedure ScrollHook(Sender: TObject; PartCode: Integer);
(*
    procedure NextLineHook; cdecl;
    procedure NextPageHook; cdecl;
    procedure PrevLineHook; cdecl;
    procedure PrevPageHook; cdecl;
*)
    procedure SetKind(Value: TScrollBarKind);
    procedure SetMax(Value: Integer);
    procedure SetMin(Value: Integer);
    procedure SetPosition(Value: Integer);
    function GetTracking: Boolean;
    procedure SetTracking(const Value: Boolean);
(*
    procedure SliderPressedHook; cdecl;
    procedure SliderReleasedHook; cdecl;
*)
    procedure ValueChangedHook(Sender: TObject);
    function GetPosition: Integer;
    function GetLargeChange: TScrollBarInc;
    function GetMax: Integer;
    function GetMin: Integer;
    function GetSmallChange: TScrollBarInc;
    procedure SetLargeChange(const Value: TScrollBarInc);
    procedure SetSmallChange(const Value: TScrollBarInc);
    procedure SetPageSize(Value: Integer);
    function GetPageSize: Integer;
  protected
    procedure Change; dynamic;
    procedure CreateWidget; override;
    procedure HookEvents; override;
    procedure InitWidget; override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure Loaded; override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure Scroll(ScrollCode: TScrollCode; var ScrollPos: Integer); dynamic;
    procedure TabStopChanged; override;
    function QScrollBar: TQScrollBar;
  public
    constructor Create(AOwner: TComponent); override;
    procedure SetParams(APosition, AMin, AMax: Integer);
  published
    property Align default alNone;
    property Anchors;
    property Bitmap;
    property Color default clBackground;
    property Constraints;
    property DragMode;
    property Enabled;
    property Kind: TScrollBarKind read FKind write SetKind default sbHorizontal;
    property LargeChange: TScrollBarInc read GetLargeChange write SetLargeChange default 1;
    property Max: Integer read GetMax write SetMax default 100;
    property Min: Integer read GetMin write SetMin default 0;
    property ParentColor default True;
    property ParentShowHint;
    property PopupMenu;
    property Position: Integer read GetPosition write SetPosition default 0;
    property ShowHint default False;
    property SmallChange: TScrollBarInc read GetSmallChange write SetSmallChange default 1;
    property TabOrder;
    property TabStop default True;
    property Track: Boolean read GetTracking write SetTracking default True;
    property Visible;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnContextPopup;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyString;
    property OnKeyUp;
    property OnScroll: TScrollEvent read FOnScroll write FOnScroll;
    property OnStartDrag;
    property PageSize: Integer read GetPageSize write SetPageSize stored False;
  end;

implementation

uses
  SysConst, QConsts(*, ViTypes*);

(*
const
  AlignmentFlags: array[TAlignment] of Integer =
    (Integer(AlignmentFlags_AlignLeft), Integer(AlignmentFlags_AlignRight),
     Integer(AlignmentFlags_AlignHCenter));

  AlignmentFlags_AlignmentMask = $FFF8;
  AlignmentFlags_LayoutMask = $FFC7;

  ClxListBoxItem_Selected = $01;
  ClxListBoxItem_Grayed = $02;
  ClxListBoxItem_Focused = $04;

type
  QClxListItem_paint_Event = procedure(listBox: QListBoxH; p: QPainterH; Index: Integer; Rect: PRect;
    Flags: Integer; var Handled: Boolean) of object cdecl;
  QClxListItem_measure_Event = procedure(index: Integer; var height,
    width: Integer) of object cdecl;
*)

type
  TCustomListBoxStrings = class(TStrings)
  private
    FSorted: Boolean;
    FPutting: Boolean;
    FOwner: TWidgetControl;
    procedure SetSorted(const Value: Boolean);
    function QListBox: TQListBox; virtual; abstract;
  protected
    function Get(Index: Integer): string; override;
    function GetCount: Integer; override;
    procedure Put(Index: Integer; const S: string); override;
    property Sorted: Boolean read FSorted write SetSorted;
  public
    function Add(const S: string): Integer; override;
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    procedure Exchange(Index1, Index2: Integer); override;
    procedure Insert(Index: Integer; const S: string); override;
    procedure Move(CurIndex, NewIndex: Integer); override;
    procedure Sort;
  end;

  TRowState = record
    RowLayout, ColLayout: TListBoxLayout;
    Rows, Cols: Integer;
  end;

  TListBoxStrings = class(TCustomListBoxStrings)
  private
    FRowState: TRowState;
    function QListBox: TQListBox; override;
  protected
    function GetObject(Index: Integer): TObject; override;
    procedure PutObject(Index: Integer; AObject: TObject); override;
    procedure SetUpdateState(Updating: Boolean); override;
  end;

  TComboBoxStrings = class(TCustomListBoxStrings)
  private
    FDuplicates: TDuplicates;
    procedure SetDuplicates(const Value: TDuplicates);
    function QListBox: TQListBox; override;
  protected
    function GetObject(Index: Integer): TObject; override;
    procedure PutObject(Index: Integer; AObject: TObject); override;
  public
    constructor Create; virtual;
    procedure Insert(Index: Integer; const S: string); override;
    property Duplicates: TDuplicates read FDuplicates write SetDuplicates;
  end;

{ TCustomListBoxStrings }

function TCustomListBoxStrings.Add(const S: string): Integer;
begin
  if not Sorted or (csRecreating in FOwner.ControlState) then
    Result := GetCount
  else begin
    { Can't do this on the Qt side since the ComboBox secretly creates its own
      listbox, so adding to a sorted ListBox, ComboBox is gonna be kind of slow...
      recommend turning off Sorted and calling Sort as needed. }
    Result := 0;
    while (Result < GetCount) and (AnsiCompareStr(S, Strings[Result]) > 0) do
      Inc(Result);
  end;
  Insert(Result, S);
end;

procedure TCustomListBoxStrings.Clear;
begin
(*
  QListBox_clear(ListBoxHandle);
*)
  QListBox.Clear;
end;

procedure TCustomListBoxStrings.Delete(Index: Integer);
begin
  TCustomListBox(FOwner).FDeleting := True;
  try
(*
    QListBox_removeItem(ListBoxHandle, Index);
*)
    QListBox.RemoveItem(Index);
  finally
    TCustomListBox(FOwner).FDeleting := False
  end;
end;

procedure TCustomListBoxStrings.Exchange(Index1, Index2: Integer);
var
  TempData: TObject;
  TempString: string;
begin
  TempString := Strings[Index1];
  TempData := GetObject(Index1);
  Strings[Index1] := Strings[Index2];
  PutObject(Index1, GetObject(Index2));
  Strings[Index2] := TempString;
  PutObject(Index2, TempData);
end;

function TCustomListBoxStrings.Get(Index: Integer): string;
begin
(*
  Item := QListBox_item(ListBoxHandle, Index);
  if Item = nil then Error(SListIndexError, Index);
  QListBoxItem_text(Item, PWideString(@WItem));
*)
  if (Index < 0) or (Index >= Count) then
    Error(SListIndexError, Index);
  Result := QListBox.GetItem(Index);
end;

function TCustomListBoxStrings.GetCount: Integer;
begin
(*
  Result := QListBox_count(ListBoxHandle);
*)
  Result := QListBox.Count;
end;

procedure TCustomListBoxStrings.Insert(Index: Integer; const S: string);
var
  Count: Integer;
begin
(*
  Item := QClxListBoxItem_create(PWideString(@WS), DrawHooks, nil);
  if Index >= GetCount then Index := -1;
  QListBox_insertItem(ListBoxHandle, Item, Index);
*)
  Count := GetCount;
  if Index > Count then Index := Count;
  QListBox.InsertItem(S, Index);
end;

procedure TCustomListBoxStrings.Move(CurIndex, NewIndex: Integer);
var
  TempData: TObject;
  TempString: string;
begin
  if CurIndex <> NewIndex then
  begin
    TempString := Get(CurIndex);
    TempData := GetObject(CurIndex);
    Delete(CurIndex);
    InsertObject(NewIndex, TempString, TempData);
  end;
end;

procedure TCustomListBoxStrings.Put(Index: Integer; const S: string);
var
  CurIndex: Integer;
begin
  FPutting := True;
  try
    CurIndex := QListBox.CurrentItem;
    inherited Put(Index, S);
    QListBox.CurrentItem := CurIndex;
  finally
    FPutting := False;
  end;
end;

procedure TCustomListBoxStrings.SetSorted(const Value: Boolean);
begin
  if FSorted <> Value then
  begin
    FSorted := Value;
    Sort;
  end;
end;

procedure TCustomListBoxStrings.Sort;
begin
(*
  QListBox_sort(ListBoxHandle, True);
  { QWidget_update and QWidget_repaint have no effect here }
  QListBox_triggerUpdate(ListBoxHandle, False);
*)
  QListBox.Sort;
end;


{ TListBoxStrings }

function TListBoxStrings.GetObject(Index: Integer): TObject;
begin
  Result := TObject(TCustomListBox(FOwner).GetItemData(Index));
end;

procedure TListBoxStrings.PutObject(Index: Integer; AObject: TObject);
begin
  TCustomListBox(FOwner).SetItemData(Index, Integer(AObject));
end;

procedure TListBoxStrings.SetUpdateState(Updating: Boolean);
begin
  if Updating then
  begin
    FRowState.RowLayout := TCustomListBox(FOwner).GetRowLayout;
    FRowState.ColLayout := TCustomListBox(FOwner).GetColumnLayout;
    FRowState.Rows := TCustomListBox(FOwner).Rows;
    FRowState.Cols := TCustomListBox(FOwner).Columns;
  end
  else begin
    if FRowState.RowLayout = llFixedNumber then
      TCustomListBox(FOwner).Rows := FRowState.Rows
    else
      TCustomListBox(FOwner).RowLayout := FRowState.RowLayout;
      
    if FRowState.ColLayout = llFixedNumber then
      TCustomListBox(FOwner).Columns := FRowState.Cols
    else
      TCustomListBox(FOwner).ColumnLayout := FRowState.ColLayout;
  end;
  inherited SetUpdateState(Updating);
end;

function TListBoxStrings.QListBox: TQListBox;
begin
  Result := FOwner.Handle as TQListBox;
end;


{ TComboBoxStrings }

procedure TComboBoxStrings.Insert(Index: Integer; const S: string);
begin
  if Cardinal(GetCount) >= Cardinal(TCustomComboBox(FOwner).MaxItems) then
    Exit;
  if IndexOf(S) > -1 then
    case Duplicates of
      dupIgnore: Exit;
      dupError: Error(@SDuplicateString, 0);
    end;
  inherited Insert(Index, S);
end;

procedure TComboBoxStrings.SetDuplicates(const Value: TDuplicates);
begin
  if Duplicates <> Value then
  begin
    FDuplicates := Value;
(*
    QComboBox_setDuplicatesEnabled(TCustomComboBox(FOwner).Handle, FDuplicates = dupAccept);
*)
  end;
end;

constructor TComboBoxStrings.Create;
begin
  inherited Create;
  FDuplicates := dupAccept;
end;

function TComboBoxStrings.GetObject(Index: Integer): TObject;
begin
  Result := nil;
(*
  Item := QClxListBoxItemH(QListBox_item(ListBoxHandle, Index));
  if Item <> nil then
    Result := TObject(QClxListBoxItem_getData(Item));
*)
end;

procedure TComboBoxStrings.PutObject(Index: Integer; AObject: TObject);
begin
(*
  Item := QClxListBoxItemH(QListBox_item(ListBoxHandle, Index));
  if Item <> nil then
    QClxListBoxItem_setData(Item, Pointer(AObject));
*)
end;

function TComboBoxStrings.QListBox: TQListBox;
begin
  Result := FOwner.Handle as TQListBox;
end;


{ TCustomGroupBox }

procedure TCustomGroupBox.AlignControls(AControl: TControl;
  var Rect: TRect);
begin
(*
  QFrame_contentsRect(Handle, @Rect);
*)
  if Caption <> '' then
    Inc(Rect.Top, Font.Height div 2);
  inherited AlignControls(AControl, Rect);
end;

constructor TCustomGroupBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := [csAcceptsControls, csCaptureMouse, csClickEvents,
    csSetCaption, csDoubleClicks, csReplicatable];
  Palette.ColorRole := crBackground;
  Palette.TextColorRole := crForeground;
  TabStop := False;
  Width := 185;
  Height := 105;
end;

procedure TCustomGroupBox.CreateWidget;
begin
(*
  FHandle := QGroupBox_create(ParentWidget, nil);
*)
  FHandle := TQGroupBox.Create(ParentWidget);
end;

procedure TCustomGroupBox.FontChanged;
begin
  inherited FontChanged;
  Invalidate;
end;

function TCustomGroupBox.GetText: TCaption;
begin
  Result := FCaption;
end;

procedure TCustomGroupBox.Invalidate;
begin
  inherited;
  InvalidateRect(ClientRect, True);
end;


procedure TCustomGroupBox.SetText(const Value: TCaption);
var
  NeedAlign: Boolean;
begin
  NeedAlign := ((Value = '') and (Caption <> '')) or
    ((Value <> '') and (Caption = ''));
  if Caption <> Value then
  begin
    FCaption := Value;
(*
    QGroupBox_setTitle(Handle, PWideString(@FCaption));
*)
    QGroupBox.Caption := FCaption;
    if NeedAlign then Realign;
    Invalidate;
    TextChanged;
  end;
end;

procedure TCustomGroupBox.SetAlignment(const Value: TAlignment);
begin
  if Value <> Alignment then
  begin
(*
    QGroupBox_setAlignment(Handle, AlignmentFlags[Value]);
*)
    Invalidate;
  end;
end;

function TCustomGroupBox.GetAlignment: TAlignment;
begin
(*
  Result := QAlignmentToTAlignment(QGroupBox_alignment(Handle));
*)
  Result := taLeftJustify;
end;

procedure TCustomGroupBox.InitWidget;
begin
  inherited InitWidget;
  BorderStyle := bsEtched;
end;

function TCustomGroupBox.WantKey(Key: Integer; Shift: TShiftState;
  const KeyText: WideString): Boolean;
var
  AltKey: Integer;
begin
  if (Shift = [ssAlt]) then
  begin
    AltKey := TranslateAltKeyToKey(Key);
    if IsAccel(AltKey, Caption) and CanFocus then
    begin
      SelectFirst;
      Result := True;
      Exit;
    end
  end;
  Result := inherited WantKey(Key, Shift, KeyText);
end;

function TCustomGroupBox.WidgetFlags: Integer;
begin
(*
  Result := inherited WidgetFlags or Integer(WidgetFlags_WRepaintNoErase);
*)
  Result := inherited WidgetFlags;
end;

function TCustomGroupBox.QGroupBox: TQGroupBox;
begin
  Result := Handle as TQGroupBox;
end;


{ TCustomLabel }

constructor TCustomLabel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csNoFocus];
  Palette.ColorRole := crBackground;
  Palette.TextColorRole := crForeground;
  TabStop := False;
  FAutoSize := True;
  FShowAccel := True;
end;

procedure TCustomLabel.CreateWidget;
begin
  FHandle := TQLabel.Create(ParentWidget);
  QLabel.WFlags := WidgetFlags;

(*
  FHandle := QLabel_create(ParentWidget, nil, WidgetFlags);
  Hooks := QWidget_hook_create(FHandle);
*)
end;

procedure TCustomLabel.DoAutoSize;
var
  AWidth,
  AHeight: Integer;
begin
  if FAutoSize then
  begin
    AWidth := Width;
    AHeight := Height;
    GetSize(AWidth, AHeight);

//    AWidth := Width;
//    AHeight := 16;
    SetBounds(Left, Top, AWidth, AHeight);
  end;
end;

procedure TCustomLabel.GetSize(var AWidth, AHeight: Integer);
(*
const
  WrapFlag: array[Boolean] of Integer =
    (0, Integer(AlignmentFlags_WordBreak));
var
  FM: QFontMetricsH;
  Str: WideString;
  R: TRect;
*)
var
  Dummy: Integer;
begin
(*
  FM := QFontMetrics_create(Font.Handle);
  try
    Str := GetText;
    QFontMetrics_boundingRect(FM, @R, 0, 0, AWidth, MaxInt,
    QLabel_alignment(Handle) or WrapFlag[WordWrap],
      @Str, -1, 0, nil, nil);
    if Align in [alLeft, alRight, alNone, alCustom] then
      AWidth := R.Right - R.Left;
    if Align in [alBottom, alTop, alNone, alCustom] then
      AHeight := R.Bottom - R.Top;
  finally
    QFontMetrics_destroy(FM);
  end;
*)
  QLabel.GetGeometry(Dummy, Dummy, AWidth, AHeight);
  AWidth := Length(StripHotKey(GetText)) * 8;
end;

procedure TCustomLabel.FontChanged;
begin
  inherited FontChanged;
  DoAutoSize;
end;

procedure TCustomLabel.TextChanged;
begin
  inherited TextChanged;
  if not (csLoading in ComponentState) then
    DoAutoSize;
end;

function TCustomLabel.GetAlignment: TAlignment;
begin
(*
  Result := QAlignmentToTAlignment(QLabel_alignment(Handle));
*)
  Result := taLeftJustify;
end;

function TCustomLabel.GetText: TCaption;
begin
  Result := QLabel.Caption;
(*
  QLabel_text(Handle, PWideString(@Result));
*)
end;

function TCustomLabel.GetLabelText: WideString;
begin
  Result := Caption; //For compatibility. Not used by underlying Qt Label.
end;

function TCustomLabel.GetLayout: TTextLayout;
(*
var
  Value: Integer;
*)
begin
(*
  Value := QLabel_alignment(Handle);
  if Value and Integer(AlignmentFlags_AlignTop) <> 0 then
    Result := tlTop
  else
    if Value and Integer(AlignmentFlags_AlignBottom) <> 0 then
      Result := tlBottom
    else
      Result := tlCenter; { default }
*)
  Result := tlCenter;
end;

function TCustomLabel.GetWordWrap: Boolean;
begin
(*
  Result := QLabel_alignment(Handle) and Integer(AlignmentFlags_WordBreak) <> 0;
*)
  Result := False;
end;

procedure TCustomLabel.InitWidget;
(*
var
  AlignFlags: Integer;
*)
begin
  inherited InitWidget;
  BorderStyle := bsNone;
  Layout := tlTop;
(*
  AlignFlags := QLabel_alignment(Handle);
  QLabel_setAlignment(Handle, AlignFlags or Integer(AlignmentFlags_ShowPrefix)
    or Integer(AlignmentFlags_AlignTop) or Integer(AlignmentFlags_AlignLeft));
  QWidget_setFocusPolicy(Handle, QWidgetFocusPolicy_NoFocus);
*)
end;

procedure TCustomLabel.Loaded;
begin
  inherited Loaded;
  if Align = alNone then SetAutoSize(FAutoSize);
end;

procedure TCustomLabel.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
end;

procedure TCustomLabel.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  Update;
end;

procedure TCustomLabel.SetAlignment(const Value: TAlignment);
(*
var
  AlignFlags: Word;
*)
begin
  if Alignment <> Value then
  begin
(*
    AlignFlags := QLabel_alignment(Handle);
    AlignFlags := AlignFlags and AlignmentFlags_AlignmentMask or AlignmentFlags[Value];
    QLabel_setAlignment(Handle, AlignFlags);
*)
  end;
end;

procedure TCustomLabel.SetAutoSize(Value: Boolean);
begin
  FAutoSize := Value;
  DoAutoSize;
end;

procedure TCustomLabel.SetShowAccel(const Value: Boolean);
(*
var
  Flags: Integer;
*)  
begin
  if FShowAccel <> Value then
  begin
    FShowAccel := Value;
(*
    Flags := QLabel_alignment(Handle);
    if FShowAccel then
      Flags := Flags or Integer(AlignmentFlags_ShowPrefix)
    else
      Flags := Flags and not Integer(AlignmentFlags_ShowPrefix);
    QLabel_setAlignment(Handle, Flags);
*)
  end;
end;

procedure TCustomLabel.SetText(const Value: TCaption);
begin
  if Caption <> Value then
  begin
    QLabel.Caption := Value;
(*
    QLabel_setText(Handle, PWideString(@Value));
    if Assigned(FFocusControl) then
      QLabel_setBuddy(Handle, FFocusControl.Handle);
*)
    TextChanged;
  end;
end;

procedure TCustomLabel.SetFocusControl(const Value: TWidgetControl);
begin
  if Value <> FFocusControl then
  begin
(*
    if Value <> nil then
      Buddy := Value.Handle
    else
      Buddy := nil;
    QLabel_setBuddy(Handle, Buddy);
*)
    FFocusControl := Value;
    { Note: Qt will force the ShowPrefix alignment flag when a buddy is assigned,
      no matter what flags are passed to QLabel::setAlignment }
  end;
end;

procedure TCustomLabel.SetLayout(const Value: TTextLayout);
(*
const
  LayoutFlags: array[TTextLayout] of Integer =
    (Integer(AlignmentFlags_AlignTop), Integer(AlignmentFlags_AlignVCenter),
     Integer(AlignmentFlags_AlignBottom));
var
  AlignFlags: Word;
*)
begin
  if Layout <> Value then
  begin;
(*
    AlignFlags := QLabel_alignment(Handle);
    AlignFlags := AlignFlags and AlignmentFlags_LayoutMask;
    QLabel_setAlignment(Handle, AlignFlags or LayoutFlags[Value]);
*)
  end;
end;

procedure TCustomLabel.SetWordWrap(const Value: Boolean);
(*
const
  WordWrapFlags: array[Boolean] of Integer = (0, Integer(AlignmentFlags_WordBreak));
var
  Flags: Word;
*)
begin
  if WordWrap = Value then Exit;
(*
  Flags := QLabel_alignment(Handle) and ($FFFF - Integer(AlignmentFlags_WordBreak));
  QLabel_setAlignment(Handle, Flags or WordWrapFlags[Value]);
*)
end;

function TCustomLabel.GetTransparent: Boolean;
begin
  Result := Masked;
end;

procedure TCustomLabel.SetTransparent(const Value: Boolean);
begin
  Masked := Value;
end;

function TCustomLabel.WantKey(Key: Integer; Shift: TShiftState;
  const KeyText: WideString): Boolean;
begin
  if (Shift = [ssAlt]) then
  begin
    Key := TranslateAltKeyToKey(Key);
    if IsAccel(Key, Caption) and Assigned(FocusControl) then
    begin
      FocusControl.SetFocus;
      Result := True;
      Exit;
    end;
  end;
  Result := inherited WantKey(Key, Shift, KeyText);
end;

function TCustomLabel.QLabel: TQLabel;
begin
  Result := Handle as TQLabel;
end;

{ TCustomEdit }

procedure TCustomEdit.Clear;
begin
  QLineEdit.Clear;
end;

procedure TCustomEdit.ClearSelection;
begin
  QLineEdit.ClearSelection;
end;

procedure TCustomEdit.CopyToClipboard;
begin
  QLineEdit.Copy;
end;

constructor TCustomEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Palette.ColorRole := crBase;
  Palette.TextColorRole := crText;
  FAutoSize := True;
  FMaxLength := -1;
  FHideSelection := True;
  ParentColor := False;
  TabStop := True;
  InputKeys := [ikChars, ikEdit, ikArrows];
end;

procedure TCustomEdit.CreateWidget;
begin
  FHandle := TQLineEdit.Create(ParentWidget);
  QLineEdit.WFlags := WidgetFlags;
(*
  FHandle := QClxLineEdit_create(ParentWidget, nil);
  Hooks := QLineEdit_hook_create(Handle);
*)
end;

procedure TCustomEdit.CutToClipboard;
begin
  QLineEdit.Cut;
end;

procedure TCustomEdit.TextChangedHook(Sender: TObject);
begin
  try
(*
    QClxLineEdit_resetSelection(Handle);
*)
    Modified := True;
    Change;
  except
    Application.HandleException(Self);
  end;
end;

function TCustomEdit.GetAlignment: TAlignment;
begin
(*
  Result := QAlignmentToTAlignment(QLineEdit_alignment(Handle));
*)
  Result := taLeftJustify;
end;

function TCustomEdit.GetBorderStyle: TControlBorderStyle;
begin
(*
  if QLineEdit_frame(Handle) then
    Result := bsSingle
  else
    Result := bsNone;
*)
  Result := bsNone;
end;

function TCustomEdit.GetModified: Boolean;
begin
  Result := QLineEdit.Edited;
end;

function TCustomEdit.GetSelLength: Integer;
begin
  Result := Length(SelText);
end;

function TCustomEdit.GetSelStart: Integer;
begin
  Result := QLineEdit.SelStart;
end;

function TCustomEdit.GetSelText: WideString;
begin
  Result := QLineEdit.MarkedText;
end;

function TCustomEdit.GetSelTextBuf(Buffer: PChar; BufSize: Integer): Integer;
var
  ASelText: AnsiString;
begin
  { Provided for compatibility. Causes unnecessary copying and a Unicode
    to MBCS translation }
  ASelText := SelText;
  StrLCopy(Buffer, PChar(Pointer(ASelText)), BufSize - 1);
  Result := StrLen(Buffer);
end;

function TCustomEdit.GetText: TCaption;
begin
  Result := QLineEdit.Caption;
end;

procedure TCustomEdit.PasteFromClipboard;
var
  Value: string;
  WValue: WideString;
begin
  if not ReadOnly then
  begin
    Value := Clipboard.AsText;
    if CharCase = ecUpperCase then
      AnsiStrUpper(PChar(Value))
    else if CharCase = ecLowerCase then
      AnsiStrLower(PChar(Value));
    WValue := Value;
(*
    QLineEdit.Caption := Value;
    QLineEdit.ClearSelection;
*)
    SelText := Value;
  end;
end;

procedure TCustomEdit.SelectAll;
begin
  QLineEdit.SelectAll;
end;

procedure TCustomEdit.SetAlignment(const Value: TAlignment);
begin
(*
  if Value <> Alignment then QLineEdit_setAlignment(Handle, AlignmentFlags[Value]);
*)
end;

procedure TCustomEdit.SetBorderStyle(const Value: TControlBorderStyle);
begin
(*
  if BorderStyle <> Value then QLineEdit_setFrame(Handle, Boolean(Value));
*)
end;

procedure TCustomEdit.SetMaxLength(const Value: Integer);
begin
  if FMaxLength <> Value then
  begin
    if Value < 0 then
    begin
      FMaxLength := -1;
      QLineEdit.MaxLength := leMaxLength;
    end
    else begin
      FMaxLength := Value;
      QLineEdit.MaxLength := FMaxLength;
    end;
  end;
end;

procedure TCustomEdit.SetModified(const Value: Boolean);
begin
  if Modified <> Value then
    QLineEdit.Edited := Value;
end;

procedure TCustomEdit.SetOnReturnPressed(const Value: TNotifyEvent);
(*
var
  Method: TMethod;
*)
begin
  FOnReturnPressed := Value;
(*
  if Assigned(FOnReturnPressed) then
    QLineEdit_returnPressed_Event(Method) := ReturnPressedHook
  else
    Method := NullHook;
  QLineEdit_hook_hook_returnPressed(QLineEdit_hookH(Hooks), Method);
*)
end;

procedure TCustomEdit.SetSelLength(Value: Integer);
begin
(*
  { Workaround for Qt bug when SelLength < 0 and SelStart < Abs(SelLength) }
  if SelStart + Value < 0 then Value := -SelStart;
  QClxLineEdit_setSelection(Handle, SelStart, Value);
*)
  QLineEdit.SetSelection(SelStart, Value);
  Invalidate;  
end;

procedure TCustomEdit.SetSelStart(Value: Integer);
begin
  if SelStart <> Value then
    QLineEdit.SetSelStart(Value);
end;

procedure TCustomEdit.SetSelText(const Value: WideString);
begin
  QLineEdit.MarkedText := Value;
end;

procedure TCustomEdit.SetSelTextBuf(Buffer: PChar);
begin
  SetSelText(Buffer);
end;

procedure TCustomEdit.SetText(const Value: TCaption);
begin
  if Text <> Value then
  begin
    QLineEdit.Caption := Value;
    TextChanged;
  end;
end;

procedure TCustomEdit.UpdateHeight;
begin
  if AutoSize then SetBounds(Left, Top, Width, FontHeight);
end;

procedure TCustomEdit.Change;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

(*
procedure TCustomEdit.ReturnPressedHook;
begin
  if Assigned(FOnReturnPressed) then
  try
    FOnReturnPressed(Self);
  except
    Application.HandleException(Self);
  end;
end;
*)

function TCustomEdit.GetAutoSelect: Boolean;
begin
  Result := QLineEdit.AutoSelect;
end;

procedure TCustomEdit.SetAutoSelect(const Value: Boolean);
begin
  if AutoSelect <> Value then
    QLineEdit.AutoSelect := Value;
end;

function TCustomEdit.FontHeight: Integer;
begin
(*
  FM := QFontMetrics_create(Font.Handle);
  try
    QWidget_FontMetrics(Handle, FM);
    Result := QFontMetrics_height(FM) + 8;
  finally
    QFontMetrics_destroy(FM);
  end;
*)
  Result := 16;
end;
                       
procedure TCustomEdit.HookEvents;
begin
  inherited HookEvents;
(*
  QLineEdit_textChanged_Event(Method) := TextChangedHook;
  QLineEdit_hook_hook_textChanged(QLineEdit_hookH(Hooks), Method);
*)
  QLineEdit.OnTextChanged := TextChangedHook;
end;

procedure TCustomEdit.InitWidget;
begin
  inherited InitWidget;
  AutoSelect := True;
  SelStart := 0;
  UpdateHeight;
end;

procedure TCustomEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);
  if (Shift - [ssCtrl, ssAlt] = Shift) and (Key >= $20) and (Key <= $FF)
  and ((MaxLength > -1) and (Length(Text) >= MaxLength) and (SelLength = 0)) then
  begin
    Key := 0;
    Beep;
  end;
end;

procedure TCustomEdit.Loaded;
begin
  inherited Loaded;
  UpdateHeight;
  QLineEdit.Edited := False;
end;

function TCustomEdit.GetReadOnly: Boolean;
begin
  Result := QLineEdit.ReadOnly;
end;

procedure TCustomEdit.SetReadOnly(const Value: Boolean);
begin
  if ReadOnly <> Value then
    QLineEdit.ReadOnly := Value;
end;

procedure TCustomEdit.SetAutoSize(const Value: Boolean);
begin
  if FAutoSize <> Value then
  begin
    FAutoSize := Value;
    UpdateHeight;
  end;
end;

procedure TCustomEdit.SetHideSelection(const Value: Boolean);
begin
  if HideSelection <> Value then
  begin
    FHideSelection := Value;
    QLineEdit.HideSelection := Value;
  end;
end;

procedure TCustomEdit.FontChanged;
begin
  inherited FontChanged;
  UpdateHeight;
end;

function TCustomEdit.QLineEdit: TQLineEdit;
begin
  Result := Handle as TQLineEdit;
end;

function TCustomEdit.EventFilter(Sender: TObject; var Event: TEvent):
    Boolean;

  function LowCase(Ch: Char): Char;
  begin
    Result := Ch;
    case Result of
      'A'..'Z':  Inc(Result, Ord('a') - Ord('A'));
    end;
  end;

  procedure ForwardEvent;
  begin
    FIgnoreEvents := True;
    try
(*
      QApplication_sendEvent(Sender, Event);
      inherited EventFilter(Sender, Event);
      Result := True;
*)
      Result := inherited EventFilter(Sender, Event);

    finally
      FIgnoreEvents := False;
    end;
  end;

var
  Key: Word;
begin
  if FIgnoreEvents then
  begin
    Result := False;
    Exit;
  end;

  case Event.What of
    Q_EVMouseDown:
      if (mbRightButton and Event.Buttons) <> 0 then
        Result := inherited EventFilter(Sender, Event)
      else
        ForwardEvent;
    Q_EVMouseUp,
    Q_EVKeyUp:
      begin
        Key := Event.KeyCode;
        if (Key <> Key_Tab) and (Key <> Key_Backtab) then
          ForwardEvent;
      end;
    Q_EVKeyDown:
      begin
        if not (csDesigning in ComponentState) then
        begin
          Key := Event.KeyCode;
          Result := True;
          case Key of
            Key_CtrlV, Key_ShiftInsert:
              PasteFromClipboard;
            Key_CtrlX, Key_ShiftDelete:
              CutToClipboard;
            Key_CtrlC:
              CopyToClipboard;
          else
            Result := False;
          end;
          if not Result and (CharCase <> ecNormal) then
          begin
            if CharCase = ecUpperCase then
              Event.CharCode := UpCase(Event.CharCode)
            else
              Event.CharCode := LowCase(Event.CharCode);
//            Result := Event.CharCode = #0;
          end;
          if not Result then Result := inherited EventFilter(Sender, Event);
        end
      end;
  else
    Result := inherited EventFilter(Sender, Event);
  end;
end;

procedure TCustomEdit.SetEchoMode(Value: TEchoMode);
begin
  if EchoMode <> Value then
    QLineEdit.EchoMode := Ord(Value);
end;

function TCustomEdit.GetEchoMode: TEchoMode;
begin
  Result := TEchoMode(QLineEdit.EchoMode);
end;

{ TButtonControl }

function TButtonControl.AccelChar: Char;
begin
  Result := FAccelChar;
end;

procedure TButtonControl.AnimateClick;
begin
(*
  QButton_animateClick(Handle);
*)
  Click;
end;

constructor TButtonControl.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Palette.ColorRole := crButton;
  Palette.TextColorRole := crButtonText;
end;

function TButtonControl.GetAutoRepeat: Boolean;
begin
(*
  Result := QButton_autoRepeat(Handle);
*)
  Result := False;
end;

function TButtonControl.GetChecked: Boolean;
begin
  Result := False;
end;

function TButtonControl.GetDown: Boolean;
begin
  if IsToggleButton then
    Result := State = tsOn
  else
(*
    Result := QButton_isDown(Handle);
*)
    Result := False;
end;

function TButtonControl.GetState: TToggleState;
begin
(*
  Result := TToggleState(QButton_state(Handle));
*)
  Result := tsOff;
end;

function TButtonControl.GetText: TCaption;
begin
  Result := QButton.Caption;
(*
  QButton_text(Handle, PWideString(@Result));
*)
end;

procedure TButtonControl.HookEvents;
(*
var
  Method: TMethod;
*)
begin
  inherited HookEvents;
(*
  QButton_stateChanged_Event(Method) := StateChangedHook;
  QButton_hook_hook_stateChanged(QButton_hookH(Hooks), Method);
*)
end;

function TButtonControl.IsCheckedStored: Boolean;
begin
  Result := (ActionLink = nil) or not TButtonActionLink(ActionLink).IsCheckedLinked;
end;

procedure TButtonControl.SetAutoRepeat(const Value: Boolean);
begin
(*
  if AutoRepeat <> Value then QButton_setAutoRepeat(Handle, Value);
*)
end;

procedure TButtonControl.SetChecked(Value: Boolean);
begin
end;

procedure TButtonControl.SetDown(const Value: Boolean);
begin
  if (Down <> Value) and IsToggleButton then
  begin
    if ((State = tsOff) and Value) or ((State = tsOn) and not Value) then
(*
      QButton_toggle(Handle)
*)
  end;
end;

procedure TButtonControl.SetText(const Value: TCaption);
begin
  if Caption <> Value then
  begin
    QButton.Caption := Value;
(*
    QButton_setText(Handle, PWideString(@Value));
    FAccelChar := Char(QButton_accel(Handle));
    QButton_setAccel(Handle, 0);
*)
    TextChanged;
  end;
end;

(*
procedure TButtonControl.StateChangedHook(State: Integer);
begin
  try
    StateChanged(TToggleState(State));
  except
    Application.HandleException(Self);
  end;
end;
*)

procedure TButtonControl.Toggle;
begin
(*
  QButton_toggle(Handle);
*)
end;

procedure TButtonControl.ActionChange(Sender: TObject; CheckDefaults: Boolean);
begin
  inherited ActionChange(Sender, CheckDefaults);
  if Sender is TCustomAction then
    with TCustomAction(Sender) do
    begin
      if not CheckDefaults or (Self.Checked = False) then
        Self.Checked := Checked;
    end;
end;

function TButtonControl.GetActionLinkClass: TControlActionLinkClass;
begin
  Result := TButtonActionLink;
end;

procedure TButtonControl.StateChanged(State: TToggleState);
begin
  if Assigned(FOnStateChanged) then FOnStateChanged(Self, State);
end;

function TButtonControl.IsToggleButton: Boolean;
begin
(*
  Result := QButton_isToggleButton(Handle);
*)
  Result := False;
end;

function TButtonControl.WantKey(Key: Integer; Shift: TShiftState;
  const KeyText: WideString): Boolean;
begin
  Result := IsAccel(Key, Caption) and Enabled;
  if Result then Click else
    Result := inherited WantKey(Key, Shift, KeyText);
end;

function TButtonControl.QButton: TQButton;
begin
  Result := Handle as TQButton;
end;


{ TButton }

procedure TButton.BeginAutoDrag;
begin
  BeginDrag(False, Mouse.DragThreshold);
end;

procedure TButton.Click;
var
  Form: TCustomForm;
begin
  Form := GetParentForm(Self);
  if Form <> nil then Form.ModalResult := ModalResult;
  inherited Click;
end;

procedure TButton.CreateWidget;
begin
  FHandle := TQPushButton.Create(ParentWidget);
  QButton.WFlags := WidgetFlags;
(*
  FHandle := QPushButton_create(ParentWidget, nil);
  Hooks := QButton_hook_create(Handle);
*)
end;

function TButton.GetToggleButton: Boolean;
begin
(*
  Result := QButton_isToggleButton(Handle);
*)
  Result := False;
end;

procedure TButton.KeyUp(var Key: Word; Shift: TShiftState);
begin
  inherited KeyUp(Key, Shift);
  if (Key = Key_Space) and (Shift - [ssAlt, ssShift] = Shift) then
    Click;
end;

function TButton.NeedKey(Key: Integer; Shift: TShiftState;
  const KeyText: WideString): Boolean;
begin
  Result := ((Key = Key_Enter) or (Key = Key_Return)) and (Shift = []);
  if Result then Click;
end;

procedure TButton.SetDefault(const Value: Boolean);
begin
  if Default <> Value then
  begin
    FDefault := Value;
    (QButton as TQPushButton).SetDefault(Value);
  end;
end;

(*
procedure TButton.ToggleHook(IsToggled: Boolean);
begin
  try
    Toggled(IsToggled);
  except
    Application.HandleException(Self);
  end;
end;
*)

procedure TButton.Toggled(IsToggled: Boolean);
begin
  if Assigned(FOnToggled) then FOnToggled(Self, IsToggled);
end;

procedure TButton.SetToggleButton(const Value: Boolean);
begin
(*
  if ToggleButton <> Value then QPushButton_setToggleButton(Handle, Value);
*)
end;

procedure TButton.SetOnToggled(const Value: TToggledEvent);
(*
var
  Method: TMethod;
*)
begin
  FOnToggled := Value;
(*
  if Assigned(FOnToggled) then
    QButton_toggled_Event(Method) := ToggleHook
  else
    Method := NullHook;
  QButton_hook_hook_Toggled(QButton_hookH(Hooks), Method);
*)
end;

constructor TButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle - [csDoubleClicks];
  ParentColor := False;
  TabStop := True;
  Width := 75;
  Height := 25;
end;

function TButton.WantKey(Key: Integer; Shift: TShiftState;
  const KeyText: WideString): Boolean;
var
  AltKey: Integer;
begin
  Result := inherited WantKey(Key, Shift, KeyText);
  if not Result then
  begin
    if (Shift = [ssAlt]) then
    begin
      AltKey := TranslateAltKeyToKey(Key);
      if IsAccel(AltKey, Caption) then
      begin
        Click;
        Result := True;
      end;
    end
    else
      case Key of
        (*Key_Return,*) Key_Enter:
          begin
            Result := FDefault and Enabled and Visible;
            if Result then Click;
          end;
        Key_Escape:
          begin
            Result := FCancel and Enabled and Visible;
            if Result then Click;
          end;
      end;
  end;
end;

function TButton.GetClientRect: TRect;
var
  ViRect: WQ.TRect;
begin
  ViRect := (QButton as TQPushButton).GetClientRect;
  with ViRect do
  begin
    Result.Left := A.X;
    Result.Top := A.Y;
    Result.Right := B.X;
    Result.Bottom := B.Y;
  end;
end;


{ TCustomComboBox }

function TCustomComboBox.AdjustHeight: Integer;
begin
(*
  if Style in [csOwnerDrawFixed, csOwnerDrawVariable] then
    Result := FItemHeight + 6
  else
  begin
    FM := QFontMetrics_create(Font.Handle);
    try
      QWidget_FontMetrics(Handle, FM);
      Result := QFontMetrics_height(FM) + 8;
    finally
      QFontMetrics_destroy(FM);
    end;
  end;
*)
  Result := 16;
end;

procedure TCustomComboBox.Change;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TCustomComboBox.Click;
var
  NewText: WideString;
begin
  if DroppedDown then
  begin
    if FJustDropped then
      FJustDropped := False
    else begin
(*
      QListBox_currentText(ListBoxHandle, PWideString(@NewText));
*)
      Text := NewText;
      SelectAll;
    end;
  end;
  inherited Click;
end;

procedure TCustomComboBox.ChangeHook(Sender: TObject);
begin
  try
    Change;
  except
    Application.HandleException(Self);
  end;
end;

procedure TCustomComboBox.Clear;
begin
  FItems.Clear;
end;

constructor TCustomComboBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FItems := TComboBoxStrings.Create;
  TComboBoxStrings(FItems).FOwner := Self;
  FCanvas := TControlCanvas.Create;
  TControlCanvas(FCanvas).Control := Self;
  FStyle := csDropDown;
  FMaxItems := -1;
  Palette.ColorRole := crBase;
  Palette.TextColorRole := crText;
  FItemHeight := 16;
  Width := 140;
  ParentColor := False;
  TabStop := True;
  FMaxLength := -1;
  InputKeys := [ikChars, ikArrows, ikNav];
end;

procedure TCustomComboBox.CreateWidget;
begin
  FHandle := TQOpenComboBox.Create(ParentWidget);
  QOpenComboBox.WFlags := WidgetFlags;
(*
  FHandle := QOpenComboBox_create(FStyle = csDropDown, ParentWidget, nil);
  FListBoxDrawHooks := QClxListBoxHooks_create;
  Hooks := QComboBox_hook_create(Handle);
*)
end;

procedure TCustomComboBox.CursorChanged;
begin
  inherited CursorChanged;
(*
  if not QComboBox_editable(Handle) and (EditHandle <> nil) then
    QWidget_setCursor(EditHandle, Screen.Cursors[Cursor]);
  if ListBoxHandle <> nil then
     QWidget_setCursor(ListboxHandle, Screen.Cursors[Cursor]);
*)
end;

procedure TCustomComboBox.HookEvents;
(*
var
  Method: TMethod;
*)
begin
  inherited HookEvents;
  QOpenComboBox.OnTextChanged := ChangeHook;
  QOpenComboBox.OnChanged := SelectedHook;
(*
  QComboBox_textChanged_Event(Method) := ChangeHook;
  QComboBox_hook_hook_textChanged(QComboBox_hookH(Hooks), Method);
  QComboBox_activated_Event(Method) := SelectedHook;
  QComboBox_hook_hook_activated(QComboBox_hookH(Hooks), Method);
  QComboBox_highlighted_Event(Method) := HighlightedHook;
  QComboBox_hook_hook_highlighted(QComboBox_hookH(Hooks), Method);
  QClxListItem_measure_Event(Method) := ItemMeasureHook;
  QClxListBoxHooks_hook_measure(FListBoxDrawHooks, Method);
  QClxListItem_paint_Event(Method) := ItemPaintHook;
  QClxListBoxHooks_hook_paint(FListBoxDrawHooks, Method);
*)
end;

destructor TCustomComboBox.Destroy;
begin
  FItems.Free;
  FCanvas.Free;
  inherited Destroy;
end;

procedure TCustomComboBox.DropDown;
begin
(*
  QOpenComboBox_popup(Handle);
*)
end;

function TCustomComboBox.Focused: Boolean;
begin
(*
  Result := QWidget_hasFocus(Handle) or ((EditHandle <> nil) and QWidget_hasFocus(EditHandle))
    or QWidget_hasFocus(ListBoxHandle);
*)
  Result := inherited Focused;
end;

procedure TCustomComboBox.FontChanged;
begin
  inherited FontChanged;
  Height := AdjustHeight;
end;

function TCustomComboBox.GetAutoComplete: Boolean;
begin
(*
  Result := QComboBox_autoCompletion(Handle);
*)
  Result := False;
end;

function TCustomComboBox.GetDropDownCount: Integer;
begin
(*
  Result := QComboBox_sizeLimit(Handle);
*)
  Result := 0;
end;

function TCustomComboBox.GetDuplicates: TDuplicates;
begin
  Result := TComboBoxStrings(Items).Duplicates;
end;

function TCustomComboBox.GetDroppedDown: Boolean;
begin
(*
  Result := (ListBoxHandle <> nil) and (QWidget_isVisible(ListBoxHandle));
*)
  Result := False;
end;

function TCustomComboBox.GetInsertMode: TComboBoxInsertMode;
begin
(*
  Result := TComboBoxInsertMode(QComboBox_insertionPolicy(Handle));
*)
  Result := ciNone;
end;

function TCustomComboBox.GetItemIndex: Integer;
begin
(*
  Result := QComboBox_currentItem(Handle);
*)
  Result := QOpenComboBox.CurrentItem;
end;

function TCustomComboBox.GetSelLength: Integer;
begin
  Result := Length(GetSelText);
end;

function TCustomComboBox.GetSelStart: Integer;
begin
(*
  if EditHandle <> nil then
    Result := QLineEdit_cursorPosition(EditHandle)
  else
    Result := -1;
*)
  Result := QOpenComboBox.SelStart;
end;

function TCustomComboBox.GetSelText: WideString;
begin
(*
  if EditHandle <> nil then
    QLineEdit_markedText(EditHandle, PWideString(@Result))
  else
    Result := '';
*)
  Result := QOpenComboBox.MarkedText;
end;

function TCustomComboBox.GetText: TCaption;
begin
(*
  QComboBox_currentText(Handle, PWideString(@Result));
*)
  Result := QOpenComboBox.Caption;
end;

(*
procedure TCustomComboBox.HighlightedHook(Index: Integer);
begin
  try
    FJustDropped := False;
    if DroppedDown then
      Highlighted(Index);
  except
    Application.HandleException(Self);
  end;
end;
*)

procedure TCustomComboBox.InitWidget;
begin
  inherited InitWidget;
  InsertMode := ciNone;
(*
  QComboBox_setDuplicatesEnabled(Handle, Duplicates = dupAccept);
  QComboBox_setSizeLimit(Handle, 8);
  EditHandle;
  ListBoxHandle;
*)
  ItemIndex := -1;
end;

procedure TCustomComboBox.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);
(*
  case Key of
    Key_Return, Key_Enter:
      begin
        if (MaxItems > -1) and (Items.Count >= MaxItems) then Key := 0;
        if DroppedDown then
        begin
          Key := 0;
          CloseUp;
        end;
      end;
    Key_Escape:
      CloseUp;
  end;
*)
end;

procedure TCustomComboBox.KeyUp(var Key: Word; Shift: TShiftState);
begin
(*
  case Key of
    Key_Down, Key_Up, Key_PageUp, Key_PageDown, Key_End, Key_Home:
      if DroppedDown then Click;
  end;
*)
  inherited KeyUp(Key, Shift);
end;

procedure TCustomComboBox.Loaded;
begin
  inherited;
  Height := AdjustHeight;
end;

procedure TCustomComboBox.CMKeyDown(var Msg: TCMKeyDown);
begin
(*
  if (InsertMode <> ciNone) and ((Msg.Key = Key_Enter)
    or (Msg.Key = Key_Return)) then
    Exit
  else
*)
    inherited;
end;

procedure TCustomComboBox.PaletteChanged(Sender: TObject);
begin
  inherited PaletteChanged(Sender);
(*
  if EditHandle <> nil then
    QWidget_setPalette(EditHandle, (Sender as TPalette).Handle, True);
  if ListBoxHandle <> nil then
    QWidget_setPalette(ListBoxHandle, (Sender as TPalette).Handle, True);
*)
end;

procedure TCustomComboBox.SelectedHook(Sender: TObject);
begin
  try
    Click;
    Select;
  except
    Application.HandleException(Self);
  end;
end;

procedure TCustomComboBox.Select;
begin
  if Assigned(FOnSelect) then FOnSelect(Self);
end;

procedure TCustomComboBox.SetAutoComplete(const Value: Boolean);
begin
(*
  if AutoComplete <> Value then QComboBox_setAutoCompletion(Handle, Value);
*)
end;

procedure TCustomComboBox.SetDropDownCount(const Value: Integer);
begin
(*
  if DropDownCount <> Value then QComboBox_setSizeLimit(Handle, Value);
*)
end;

procedure TCustomComboBox.SetDuplicates(const Value: TDuplicates);
begin
  TComboBoxStrings(Items).Duplicates := Value;
end;

procedure TCustomComboBox.SetInsertMode(const Value: TComboBoxInsertMode);
begin
  if InsertMode <> Value then
(*
    QComboBox_setInsertionPolicy(Handle, QComboBoxPolicy(Value));
*)
end;

procedure TCustomComboBox.SetDroppedDown(const Value: Boolean);
begin
  if Value <> GetDroppedDown then DropDown else CloseUp;
end;

procedure TCustomComboBox.SetItemIndex(const Value: Integer);
begin
(*
  QComboBox_setCurrentItem(Handle, Value);
*)
  QOpenComboBox.CurrentItem := Value;
end;

procedure TCustomComboBox.SetItems(const Value: TStrings);
begin
  FItems.Assign(Value);
end;

procedure TCustomComboBox.SetMaxItems(const Value: Integer);
begin
  if MaxItems <> Value then
  begin
    FMaxItems := Value;
(*
    if Value >= 0 then
      QComboBox_setMaxCount(Handle, FMaxItems)
    else
      QComboBox_setMaxCount(Handle, MaxInt);
*)
  end;
end;

procedure TCustomComboBox.SetMaxLength(const Value: Integer);
begin
  if FMaxLength <> Value then
  begin
    if Value < 0 then
      FMaxLength := -1
    else
      FMaxLength := Value;
(*
    if EditHandle <> nil then
    begin
      if FMaxLength > -1 then
        QLineEdit_setMaxLength(EditHandle, FMaxLength)
      else
        QLineEdit_setMaxLength(EditHandle, MaxInt);
    end;
*)
    if FMaxLength > -1 then
      QOpenComboBox.MaxLength := FMaxLength
    else
      QOpenComboBox.MaxLength := MaxInt;
  end;
end;

procedure TCustomComboBox.SetSelLength(const Value: Integer);
begin
(*
  if EditHandle <> nil then
    QLineEdit_setSelection(EditHandle, SelStart, Value);
*)
  QOpenComboBox.SetSelection(SelStart, Value);
end;

procedure TCustomComboBox.SetSelStart(const Value: Integer);
begin
(*
  if EditHandle <> nil then
    QLineEdit_setCursorPosition(EditHandle, Value);
*)
  QOpenComboBox.SetSelStart(Value);
end;

procedure TCustomComboBox.SetSelText(const Value: WideString);
begin
(*
  if EditHandle <> nil then
    QLineEdit_insert(EditHandle, PWideString(@Value));
*)
  QOpenComboBox.MarkedText := Value;
end;

procedure TCustomComboBox.SetStyle(Value: TComboBoxStyle);
(*
var
  Method: TMethod;
*)  
begin
  if FStyle <> Value then
  begin
    FStyle := Value;
(*
    RecreateWidget;
*)
    Height := AdjustHeight;
    if FStyle = csDropDownList then
      InputKeys := InputKeys - [ikEdit]
    else
      InputKeys := InputKeys + [ikEdit];
    case FStyle of
      csDropDown, csDropDownList:
        begin
(*
          Method := NullHook;
          QClxListBoxHooks_hook_paint(FListBoxDrawHooks, Method);
*)
          QOpenComboBox.Style := Integer(FStyle);
        end;
      csOwnerDrawFixed, csOwnerDrawVariable:
        begin
(*
          QClxListItem_paint_Event(Method) := ItemPaintHook;
          QClxListBoxHooks_hook_paint(FListBoxDrawHooks, Method);
*)
          QOpenComboBox.Style := Integer(FStyle);
        end;
    end;
  end;
end;

procedure TCustomComboBox.DestroyWidget;
begin
(*
  if Assigned(FEditHook) then
    QLineEdit_hook_destroy(FEditHook);
  if Assigned(FListHook) then
    QListBox_hook_destroy(FListHook);
  if Assigned(FEditHandle) then
    QClxObjectMap_remove(FEditHandle);
  FEditHook := nil;
  FListHook := nil;
  FListBoxDrawHooks := nil;
  FEditHandle := nil;
  FListBoxHandle := nil;
*)
  inherited DestroyWidget;
end;

procedure TCustomComboBox.SelectAll;
begin
(*
  if EditHandle <> nil then
    QLineEdit_selectAll(EditHandle);
*)
  QOpenComboBox.SelectAll;
end;

procedure TCustomComboBox.SetText(const Value: TCaption);
var
  ValueIndex: integer;
begin
  if Text <> Value then
  begin
    ValueIndex := Items.IndexOf(Value);
    if ValueIndex >= 0 then
    begin
      QOpenComboBox.CurrentItem := ValueIndex;
    end else
    begin
      QOpenComboBox.Caption := Value;
    end;
    TextChanged;
  end;
end;

procedure TCustomComboBox.RestoreWidgetState;
begin
  inherited RestoreWidgetState;
  if FMemStream <> nil then
  begin
    FMemStream.Position := 0;
    FMemStream.ReadComponent(Self);
    FreeAndNil(FMemStream);
  end;
end;

procedure TCustomComboBox.SaveWidgetState;
begin
  inherited SaveWidgetState;
  FMemStream := TMemoryStream.Create;
  FMemStream.WriteComponent(Self);
end;

procedure TCustomComboBox.Highlighted(Index: Integer);
begin
  if Assigned(FOnHighlighted) then FOnHighlighted(Self, Index);
end;

(*
procedure TCustomComboBox.ItemMeasureHook(index: Integer;
  var height, width: Integer);
begin
  if FStyle in [csOwnerDrawVariable] then
  try
    height := FItemHeight;
    MeasureItem(Self, index, height, width);
  except
    Application.HandleException(Self);
  end;
end;
*)

function TCustomComboBox.DrawItem(Index: Integer; Rect: TRect;
  State: TOwnerDrawState): Boolean;
begin
  Result := Assigned(FOnDrawItem);
  if Result then
  begin
    if odSelected in State then
    begin
      Canvas.Brush.Color := clHighlight;
      Canvas.FillRect(Rect);
    end;
    FOnDrawItem(Self, Index, Rect, State, Result);
  end
end;

procedure TCustomComboBox.DoEnter;
begin
(*
  if EditHandle <> nil then
    SelectAll;
*)
  SelectAll;
  inherited DoEnter;
end;

procedure TCustomComboBox.DoExit;
begin
(*
  if (Style <> csDropDownList) and (EditHandle <> nil) then
    QLineEdit_deselect(EditHandle);
*)
//  QOpenComboBox.Deselect;
  inherited DoExit;
end;

procedure TCustomComboBox.MeasureItem(Control: TWinControl;
  Index: Integer; var Height, Width: Integer);
begin
  if Assigned(FOnMeasureItem) then FOnMeasureItem(Control, Index, Height);
end;

procedure TCustomComboBox.SetSorted(const Value: Boolean);
begin
  if Sorted <> Value then
    TComboBoxStrings(FItems).Sorted := Value;
end;

procedure TCustomComboBox.Sort;
begin
  TComboBoxStrings(Items).Sort;
end;

function TCustomComboBox.GetSorted: Boolean;
begin
  Result := TComboBoxStrings(FItems).Sorted;
end;

procedure TCustomComboBox.SetItemHeight(const Value: Integer);
begin
  if FItemHeight <> Value then
  begin
    FItemHeight := Value;
(*
    RecreateWidget;
*)
    Height := AdjustHeight;
  end;
end;

function TCustomCombobox.GetItemHeight: Integer;
begin
  if FStyle in [csOwnerDrawFixed, csOwnerDrawVariable] then
    Result := FItemHeight
  else
    // The choice of 6 in the statement below is based of the value in AdjustHeight
    Result := AdjustHeight - 6;
end;

(*
procedure TCustomComboBox.DoCloseUp;
begin
  FJustDropped := False;
  if Assigned(FOnCloseUp) then FOnCloseUp(Self);
end;
*)

procedure TCustomComboBox.CloseUp;
begin
(*
  QWidget_hide(ListBoxHandle);
*)
end;

(*
procedure TCustomComboBox.DoDropDown;
var
  Size: TSize;
  I: Integer;
begin
  if Assigned(FOnDropDown) then FOnDropDown(Self);
  QListBox_sizeHint(ListBoxHandle, @Size);
  Size.cx := Width;
  if Items.Count > DropDownCount then
    I := DropDownCount
  else
    I := Items.Count;
  Size.cy := I * ItemHeight;
  Inc(Size.cy, QFrame_frameWidth(ListBoxHandle) * 2);
  QWidget_resize(ListBoxHandle, @Size);
  if Items.IndexOf(Text) < 0 then
    QListbox_clearSelection(ListBoxHandle);
  Application.CancelHint;
end;
*)

procedure TCustomComboBox.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  AHeight := AdjustHeight;
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
end;

function TCustomComboBox.QOpenComboBox: TQOpenComboBox;
begin
  Result := Handle as TQOpenComboBox;
end;


{ TCustomCheckBox }

constructor TCustomCheckBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Palette.TextColorRole := crForeground;
  Palette.ColorRole := crBackground;
  TabStop := True;
  ControlStyle := ControlStyle - [csClickEvents];
end;

procedure TCustomCheckBox.CreateWidget;
begin
  FHandle := TQCheckBox.Create(ParentWidget);
  QCheckBox.WFlags := WidgetFlags;
end;

function TCustomCheckBox.GetChecked: Boolean;
begin
  Result := QCheckBox.IsChecked;
end;

procedure TCustomCheckBox.SetAllowGrayed(const Value: Boolean);
begin
  if FAllowGrayed <> Value then
  begin
    FAllowGrayed := Value;
    QCheckBox.SetTriState(Value);
    if State = cbGrayed then
      QCheckBox.SetNoChange;
  end;
end;

procedure TCustomCheckBox.SetChecked(Value: Boolean);
begin
  if Value then State := cbChecked else State := cbUnchecked;
end;

procedure TCustomCheckBox.SetState(Value: TCheckBoxState);
begin
  if FState <> Value then
  begin
    FState := Value;
    case Value of
      cbChecked: QCheckBox.SetChecked(True);
      cbUnChecked: QCheckBox.SetChecked(False);
      cbGrayed: if FAllowGrayed then QCheckBox.SetNoChange;
    end;
  end;
end;

procedure TCustomCheckBox.StateChanged(State: TToggleState);
begin
  case State of
    tsOff: FState := cbUnchecked;
    tsNoChange: FState := cbGrayed;
    tsOn: FState := cbChecked;
  end;
  inherited StateChanged(State);
  Click;
end;

procedure TCustomCheckBox.Toggle;
begin
  case State of
    cbUnchecked:
      if AllowGrayed then State := cbGrayed else State := cbChecked;
    cbGrayed: State := cbChecked;
    cbChecked: State := cbUnchecked;
  end;
end;

function TCustomCheckBox.WantKey(Key: Integer;
  Shift: TShiftState; const KeyText: WideString): Boolean;
begin
  if (Shift = [ssAlt]) then
  begin
    Key := TranslateAltKeyToKey(Key);
    if IsAccel(Key, Caption) and CanFocus then
    begin
      SetFocus;
      if Focused then Toggle;
      Result := True;
      Exit;
    end
  end;
  Result := inherited WantKey(Key, Shift, KeyText);
end;

function TCustomCheckBox.QCheckBox: TQCheckBox;
begin
  Result := Handle as TQCheckBox;
end;

procedure TCustomCheckBox.ClickedHook;
begin
  try
    StateChanged(TToggleState(QCheckBox.State));
  except
    Application.HandleException(Self);
  end;
end;

procedure TCustomCheckBox.HookEvents;
begin
  inherited;
  QButton.OnClicked := ClickedHook;
end;


{ TCustomListBox }

type
  TSelectionMode = (qsmSingle, qsmMulti, qsmExtended);
const
  SelectionFlags: array[Boolean, Boolean] of TSelectionMode =
    ((qsmSingle, qsmSingle), (qsmMulti, qsmExtended));

procedure TCustomListBox.Clear;
begin
  FItems.Clear;
end;

procedure TCustomListBox.ClearSelection;
begin
(*
  QListBox_clearSelection(Handle);
*)
end;

constructor TCustomListBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FStyle := lbStandard;
(*
  FDrawHooks := QClxListBoxHooks_create;
*)
  FItems := TListBoxStrings.Create;
  FDeleting := False;
  TListBoxStrings(FItems).FOwner := Self;
  FCanvas := TControlCanvas.Create;
  TControlCanvas(FCanvas).Control := Self;
  Palette.ColorRole := crBase;
  Palette.TextColorRole := crText;
  ExtendedSelect := True;
  Width := 121;
  Height := 97;
  FItemHeight := 16;
  ParentColor := False;
  TabStop := True;
  InputKeys := [ikChars, ikArrows, ikNav];
end;

procedure TCustomListBox.Changed;
begin
end;

procedure TCustomListBox.CreateWidget;
begin
  FHandle := TQListBox.Create(ParentWidget);
  QListBox.WFlags := WidgetFlags;
(*
  FHandle := QListBox_create(ParentWidget, nil, 0);
  Hooks := QListBox_hook_create(Handle);
  FViewportHook := QWidget_hook_create(QScrollView_viewport(Handle));
*)
end;

destructor TCustomListBox.Destroy;
begin
  FItems.Free;
  FCanvas.Free;
(*
  QClxListBoxHooks_destroy(FDrawHooks);
*)
  inherited Destroy;
end;

procedure TCustomListBox.DblClick;
(*
var
  MousePos: TPoint;
*)  
begin
(*
  MousePos := ScreenToClient(Mouse.CursorPos);
  if ItemAtPos(MousePos, True) > -1 then
*)  
    inherited DblClick;
end;

(*
procedure TCustomListBox.DrawEmptyFocus;
begin
  if (Items.Count = 0) and Focused then
    FCanvas.DrawFocusRect(Rect(0, 0, ClientWidth - QFrame_frameWidth(Handle)*2,
      ItemHeight));
end;
*)

procedure TCustomListBox.EnabledChanged;
begin
  inherited EnabledChanged;
  //Force Qt to use the correct brush for disabled painting.
  ColorChanged;
  Repaint;
end;

function TCustomListBox.GetColumnLayout: TListBoxLayout;
begin
(*
  Result := TListBoxLayout(QListBox_columnMode(Handle));
*)
  Result := llFixedNumber;
end;

function TCustomListBox.GetColumns: Integer;
begin
(*
  Result := QListBox_numColumns(Handle);
*)
  Result := 0;
end;

function TCustomListBox.GetItemIndex: Integer;
begin
(*
  Result := QListBox_currentItem(Handle);
  if not QListBox_isSelected(Handle, QListBox_item(Handle, Result))
    then Result := -1;
*)
  Result := QListBox.CurrentItem;
end;

function TCustomListBox.GetRowLayout: TListBoxLayout;
begin
(*
  Result := TListBoxLayout(QListBox_rowMode(Handle));
*)
  Result := llFixedNumber;
end;

function TCustomListBox.GetRows: Integer;
begin
(*
  Result := QListBox_numRows(Handle);
*)
  Result := 0;
end;

function TCustomListBox.GetSelCount: Integer;
var
  I: Integer;
begin
  Result := 0;
  I := 0;
  while (I < Items.Count) do
  begin
    if Selected[I] then Inc(Result);
    Inc(I);
  end;
end;

function TCustomListBox.GetSelected(Index: Integer): Boolean;
begin
  if (Index < 0) or (Index >= Items.Count) then
    TList.Error(SListIndexError, Index);
(*
  Result := QListBox_isSelected(Handle, Index);
*)
  Result := False;
end;

function TCustomListBox.GetTopIndex: Integer;
begin
(*
  Result := QListBox_topItem(Handle);
*)
  Result := 0;
end;

function TCustomListBox.ItemAtPos(Pos: TPoint; Existing: Boolean): Integer;
var
  P: TPoint;
begin
  P := Types.TPoint(Point(Pos.X, Pos.Y));
(*
  Result := QListBox_index(Handle, QListBox_itemAt(Handle, @P));
*)
  Result := -1;
  if (Result = -1) and not Existing then Result := Items.Count;
end;

function TCustomListBox.ItemRect(Index: Integer): TRect;
begin
(*
  Item := QListBox_item(Handle, Index);
  QListBox_itemRect(Handle, @Result, Item);
*)
end;

function TCustomListBox.ItemVisible(Index: Integer): Boolean;
begin
(*
  Result := QListBox_itemVisible(Handle, Index);
*)
  Result := False;
end;

procedure TCustomListBox.Loaded;
begin
  inherited Loaded;
  //Force QListBox to have at least one column. Qt doesn't.
(*
  QListBox_setColumnMode(Handle, Columns);
*)
end;

procedure TCustomListBox.MakeCurrentVisible;
begin
(*
  QListBox_ensureCurrentVisible(Handle);
*)
end;

procedure TCustomListBox.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
(*
  QListBox_triggerUpdate(Handle, True);
*)
end;

procedure TCustomListBox.SetColumnLayout(const Value: TListBoxLayout);
begin
  if ColumnLayout <> Value then
(*
    QListBox_setColumnMode(Handle, QListBoxLayoutMode(Value));
*)
end;

procedure TCustomListBox.SetColumns(const Value: Integer);
begin
  if Columns <> Value then
(*
    QListBox_setColumnMode(Handle, Value)
*)
end;

procedure TCustomListBox.SetExtendedSelect(const Value: Boolean);
begin
  if FExtendedSelect <> Value then
  begin
    FExtendedSelect := Value;
(*
    QListBox_setSelectionMode(Handle, QListBoxSelectionMode(SelectionFlags[MultiSelect, Value]));
*)
  end;
end;

procedure TCustomListBox.SetItemIndex(const Value: Integer);
begin
  if ItemIndex <> Value then
    if Value < 0 then ClearSelection
    else QListBox.CurrentItem := Value;
(*
    else QListBox_setSelected(Handle, Value, True);
*)
end;

procedure TCustomListBox.SetItems(const Value: TStrings);
begin
  FItems.Assign(Value);
end;

procedure TCustomListBox.SetMultiSelect(const Value: Boolean);
begin
  if FMultiSelect <> Value then
  begin
    FMultiSelect := Value;
(*
    QListBox_setSelectionMode(Handle,
      QListBoxSelectionMode(SelectionFlags[MultiSelect, ExtendedSelect]));
*)
  end;
end;

procedure TCustomListBox.SetRowLayout(const Value: TListBoxLayout);
begin
  if RowLayout <> Value then
(*
    QListBox_setRowMode(Handle, QListBoxLayoutMode(Value));
*)
end;  

procedure TCustomListBox.SetRows(const Value: Integer);
begin
  if Rows <> Value then
(*
    QListBox_setRowMode(Handle, Value);
*)
end;

procedure TCustomListBox.SetSelected(Index: Integer; const Value: Boolean);
begin
(*
  QListBox_setSelected(Handle, Index, Value);
*)
end;

procedure TCustomListBox.SetSorted(const Value: Boolean);
begin
  if Sorted <> Value then
    TListBoxStrings(FItems).Sorted := Value;
end;

procedure TCustomListBox.SetTopIndex(const Value: Integer);
begin
(*
  if TopIndex <> Value then QListBox_setTopItem(Handle, Value);
*)
end;

function TCustomListBox.VisibleItemCount: Integer;
begin
(*
  Result := QListBox_numItemsVisible(Handle);
*)
  Result := 0;
end;

procedure TCustomListBox.HookEvents;
(*
var
  Method: TMethod;
*)  
begin
  inherited HookEvents;
(*
  TEventFilterMethod(Method) := ViewportEventFilter;
  Qt_hook_hook_events(FViewportHook, Method);

  QListBox_selectionChanged_Event(Method) := SelChangeHook;
  QListBox_hook_hook_selectionChanged(QListBox_hookH(Hooks), Method);

  QClxListItem_measure_Event(Method) := ItemMeasureHook;
  QClxListBoxHooks_hook_measure(FDrawHooks, Method);
  QClxListItem_paint_Event(Method) := ItemPaintHook;
  QClxListBoxHooks_hook_paint(FDrawHooks, Method);
*)
  QListBox.OnChanged := SelChangeHook;
end;

(*
procedure TCustomListBox.ItemMeasureHook(index: Integer;
  var height, width: Integer);
begin
  try
    height := FItemHeight;
//    if Style = lbOwnerDrawVariable then
      MeasureItem(Self, index, FItemHeight, width)
  except
    Application.HandleException(Self);
  end;
end;
*)

procedure TCustomListBox.SetStyle(const Value: TListBoxStyle);
begin
  if FStyle <> Value then
  begin
    FStyle := Value;
    RecreateWidget;
    AdjustHeight;
  end;
end;

function TCustomListBox.DrawItem(Index: Integer; Rect: TRect;
  State: TOwnerDrawState): Boolean;
begin
  Result := (Style in [lbOwnerDrawFixed, lbOwnerDrawVariable]) and
    Assigned(FOnDrawItem);
  if Result then FOnDrawItem(Self, Index, Rect, State, Result)
end;

procedure TCustomListBox.MeasureItem(Control: TWinControl;
  Index: Integer; var Height, Width: Integer);
begin
  if (Style = lbOwnerDrawVariable) and Assigned(FOnMeasureItem) then
    FOnMeasureItem(Control, Index, Height);
end;

procedure TCustomListBox.Sort;
begin
(*
  QListBox_sort(Handle, True);
  { QWidget_update and QWidget_repaint have no effect here }
  QListBox_triggerUpdate(Handle, False);
*)
  QListBox.Sort;
end;

function TCustomListBox.GetSorted: Boolean;
begin
  Result := TListBoxStrings(FItems).Sorted;
end;

procedure TCustomListBox.SetItemHeight(const Value: Integer);
begin
  if FStyle in [lbOwnerDrawFixed, lbOwnerDrawVariable] then
  begin
    if FItemHeight <> Value then
    begin
      FItemHeight := Value;
      RecreateWidget;
    end else
    AdjustHeight;
  end;
end;

procedure TCustomListBox.RestoreWidgetState;
begin
  inherited;
  FItems.Assign(FSaveItems);
  FontChanged;
  FSaveItems.Free;
end;

procedure TCustomListBox.SaveWidgetState;
begin
  inherited;
  FSaveItems := TStringList.Create;
  FSaveItems.Assign(FItems);
end;

function TCustomListBox.GetItemData(Index: Integer): LongInt;
begin
(*
  Result := -1;
  Item := QClxListBoxItemH(QListBox_item(Handle, Index));
  if Item <> nil then
    Result := Longint(QClxListBoxItem_getData(Item));
*)
  Result := QListBox.Data[Index];
end;

procedure TCustomListBox.SetItemData(Index, AData: Integer);
begin
(*
  Item := QClxListBoxItemH(QListBox_item(Handle, Index));
  if Item <> nil then
    QClxListBoxItem_setData(Item, Pointer(AData));
*)
  QListBox.Data[Index] := AData;
end;

procedure TCustomListBox.SelChangeHook(Sender: TObject);
begin
(*
  try
    { Don't call click if deleting an item }
    if not TCustomListBoxStrings(Items).FPutting and (not FClicking) and
      (not FDeleting) then
      //Click;
      QApplication_postEvent(Handle, QCustomEvent_create(QEventType_LBClick, Self));
    Changed;
  except
    Application.HandleException(Self);
  end;
*)
  { Don't call click if deleting an item }
  if not TCustomListBoxStrings(Items).FPutting and (not FClicking) and
      (not FDeleting) then Click;
  Changed;
end;

procedure TCustomListBox.ResetContent;
begin
  FItems.Clear;
end;

procedure TCustomListBox.DeleteString(Index: Integer);
begin
  FItems.Delete(Index);
end;

procedure TCustomListBox.Update;
begin
  if HandleAllocated then
(*
    QListBox_triggerUpdate(Handle, False);
*)                       
    QListBox.Update;
  inherited;
end;

procedure TCustomListBox.MouseDown(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  //Rely on Qt's SelectionChanged() signal for single clicks.
  if (Button <> mbLeft) or (ItemAtPos(Point(X, Y), True) <> ItemIndex) then
    ControlState := ControlState - [csClicked];
  inherited MouseDown(Button, Shift, X, Y);
end;

procedure TCustomListBox.FontChanged;
begin
  inherited FontChanged;
  AdjustHeight;
end;

procedure TCustomListBox.AdjustHeight;
begin
  if not (Style in [lbOwnerDrawFixed, lbOwnerDrawVariable]) then
  begin
(*
    FM := QFontMetrics_create(Font.Handle);
    try
      QWidget_FontMetrics(Handle, FM);
      FItemHeight := QFontMetrics_height(FM);
    finally
      QFontMetrics_destroy(FM);
    end;
*)
    FItemHeight := 16;
  end;
end;

procedure TCustomListBox.InitWidget;
begin
  inherited InitWidget;
(*
  FViewportHandle := QScrollView_viewport(Handle);
  QClxObjectMap_add(FViewportHandle, Integer(Self));
  BorderStyle := bsSunken3d;
  QListBox_setColumnMode(Handle, 1)
*)
end;

procedure TCustomListBox.WidgetDestroyed;
begin
(*
  QClxObjectMap_remove(FViewportHandle);
*)
  inherited WidgetDestroyed;
end;

function TCustomListBox.QListBox: TQListBox;
begin
  Result := Handle as TQListBox;
end;


{ TButtonActionLink }

procedure TButtonActionLink.AssignClient(AClient: TObject);
begin
  inherited AssignClient(AClient);
  FClient := AClient as TButtonControl;
end;

function TButtonActionLink.IsCheckedLinked: Boolean;
begin
  Result := inherited IsCheckedLinked and
    (FClient.Checked = (Action as TCustomAction).Checked);
end;

procedure TButtonActionLink.SetChecked(Value: Boolean);
begin
  if IsCheckedLinked then
  begin
    FClient.ClicksDisabled := True;
    try
      FClient.Checked := Value;
    finally
      FClient.ClicksDisabled := False;
    end;
  end;
end;

type
  TMemoStrings = class(TStrings)
  private
    Memo: TCustomMemo;
    FMultipleAdd: Boolean;
    FChanged: Boolean;
  protected
    procedure Changed;
    function Get(Index: Integer): string; override;
    function GetCount: Integer; override;
    function GetTextStr: string; override;
    procedure SetTextStr(const Value: string); override;
    procedure SetUpdateState(Updating: Boolean); override;
  public
    function Add(const S: string): Integer; override;
    procedure AddStrings(Strings: TStrings); override;
    procedure Assign(Source: TPersistent); override;
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    procedure Insert(Index: Integer; const S: string); override;
    procedure LoadFromFile(const Filename: string); override;
  end;

{ TMemoStrings }

function CopyRemoveCR(Source: string): string;
var
  Count: Integer;
  BufPtr, BufPos, BufEnd: PChar;
  CRFound: Boolean;
  CRChar: Char;
  Stream: TStringStream;
begin
  CRFound := False;
  CRChar := #13;
  BufPtr := @Source[1];
  BufPos := BufPtr;
  Count := Length(Source);
  BufEnd := BufPtr + Count;
  Stream := TStringStream.Create('');
  try
    if Count > 0 then
    begin
      while BufPos < BufEnd do
      begin
        if BufPos^ = #13 then
        begin
          if (BufPos < BufEnd - 1) and ((BufPos + 1)^ = #10) then
          begin
            Stream.WriteBuffer(BufPtr^, BufPos - BufPtr);
            BufPtr := BufPos + 1;
          end else if BufPos = BufEnd - 1 then
            CRFound := True;
        end else if CRFound then
        begin
          if BufPos^ <> #10 then
            Stream.WriteBuffer(CRChar, 1);
          CRFound := False;
        end;
        Inc(BufPos);
      end;
      Stream.WriteBuffer(BufPtr^, BufPos - BufPtr);
    end;
    Result := Stream.DataString;
  finally
    Stream.Free;
  end;
end;

function CopyAddCR(Source: string): string;
var
  Count: Integer;
  BufPtr, BufPos, BufEnd: PChar;
  CRFound: Boolean;
  CRChar: Char;
  Stream: TStringStream;
begin
  CRFound := False;
  CRChar := #13;
  BufPtr := @Source[1];
  BufPos := BufPtr;
  Count := Length(Source);
  BufEnd := BufPtr + Count;
  Stream := TStringStream.Create('');
  try
    if Count > 0 then
    begin
      while BufPos < BufEnd do
      begin
        if BufPos^ = #13 then
        begin
          if (BufPos < BufEnd - 1) and ((BufPos + 1)^ = #10) then
          begin
            Stream.WriteBuffer(BufPtr^, BufPos - BufPtr + 2);
            Inc(BufPos);
            BufPtr := BufPos + 1;
          end else if BufPos = BufEnd - 1 then
            CRFound := True;
        end else if CRFound then
          CRFound := False
        else if BufPos^ = #10 then
        begin
          Stream.WriteBuffer(BufPtr^, BufPos - BufPtr);
          Stream.WriteBuffer(CRChar, 1);
          BufPtr := BufPos;
        end;
        Inc(BufPos);
      end;
      Stream.WriteBuffer(BufPtr^, BufPos - BufPtr);
    end;
    Result := Stream.DataString;
  finally
    Stream.Free;
  end;
end;

function TMemoStrings.Add(const S: string): Integer;
(*
var
  CaretPos: TCaretPos;
*)
begin
  Memo.QMultiLineEdit.Append(S);
  Result := Count - 1;
(*
  Insert(-1, S);
  //Workaround for QMultiLineEdit bug when inserting a first, blank line.
  if FMultipleAdd and (S = '') and (Count = 0) then Insert(-1, S);
  Result := Count - 1;
  CaretPos.Line := Result;
  CaretPos.Col := 0;
  if not (csLoading in Memo.ComponentState) then
    Memo.SetCaretPos(CaretPos);
*)
end;

procedure TMemoStrings.AddStrings(Strings: TStrings);
begin
  FMultipleAdd := True;
  try
    inherited AddStrings(Strings);
  finally
    FMultipleAdd := False;
  end;
end;

procedure TMemoStrings.Assign(Source: TPersistent);
var
  I, L, Size, Cnt: Integer;
  P: PChar;
  S, LB, NewText: string;
begin
  if Source is TStrings then
  with TStrings(Source) do
  begin
    Cnt := Count;
    Size := 0;
    LB := sLineBreak;
    for I := 0 to Cnt - 1 do Inc(Size, Length(Strings[I]) + Length(LB));
    SetString(NewText, nil, Size);
    P := Pointer(NewText);
    for I := 0 to Count - 1 do
    begin
      S := Strings[I];
      L := Length(S);
      if L <> 0 then
      begin
        System.Move(Pointer(S)^, P^, L);
        Inc(P, L);
      end;
      L := Length(LB);
      if L <> 0 then
      begin
        System.Move(Pointer(LB)^, P^, L);
        Inc(P, L);
      end;
    end;
    SetTextStr(NewText);
  end;
end;

procedure TMemoStrings.Changed;
begin
  Memo.Change;
end;

procedure TMemoStrings.Clear;
begin
(*
  QMultiLineEdit_clear(Memo.Handle);
*)
  Memo.QMultiLineEdit.Clear;
  Memo.Modified := False;
end;

procedure TMemoStrings.Delete(Index: Integer);
begin
(*
  QMultiLineEdit_removeLine(Memo.Handle, Index);
*)
  Memo.QMultiLineEdit.RemoveLine(Index);
  Changed;
end;

function TMemoStrings.Get(Index: Integer): string;
var
  WLine: WideString;
begin
(*
  QMultiLineEdit_textLine(Memo.Handle, PWideString(@WLine), Index);
*)
  WLine := Memo.QMultiLineEdit.TextLine(Index);
  Result := WLine;
end;

function TMemoStrings.GetCount: Integer;
begin
(*
  Result := QMultiLineEdit_numLines(Memo.Handle);
*)
  Result := Memo.QMultiLineEdit.NumLines;
(*
  if (Result > 0) and (Get(Result-1) = '') then
    Dec(Result);
*)
end;

function TMemoStrings.GetTextStr: string;
begin
  Result := Memo.Text;
(*
  //an empty QMultiLineEdit *always* has a linefeed in it.
  if Result = #10 then
    Result := '';
*)
end;

procedure TMemoStrings.Insert(Index: Integer; const S: string);
(*
var
  CaretPos: TCaretPos;
  WS: WideString;
*)
begin
(*
  WS := S;
  QMultiLineEdit_insertLine(Memo.Handle, PWideString(@WS), Index);
*)
  Memo.QMultiLineEdit.InsertLine(Index, S);
(*
  //Work around QMultiLineEdit bugs.
  CaretPos.Line := Index + 1;
  CaretPos.Col := 0;
  Memo.SetCaretPos(CaretPos);
  if Pos(#10, S) > -1 then Memo.Repaint;
*)
  Changed;
end;

procedure TMemoStrings.LoadFromFile(const Filename: string);
begin
  inherited LoadFromFile(Filename);
  Memo.Modified := False;
end;

procedure TMemoStrings.SetTextStr(const Value: string);
(*
var
  WTemp: WideString;
*)
begin
  Clear;
  Memo.QMultiLineEdit.SetText(Value);
(*
  WTemp := CopyRemoveCR(Value);
  QMultiLineEdit_setText(Memo.Handle, PWideString(@WTemp));
*)
end;

procedure TMemoStrings.SetUpdateState(Updating: Boolean);
begin
  inherited SetUpdateState(Updating);
  if not Updating and FChanged then Changed;
  FChanged := False;
end;

{ TCustomMemo }

procedure TCustomMemo.Append(const Text: WideString);
begin
(*
  QMultiLineEdit_append(Handle, PWideString(@Text));
*)
  QMultiLineEdit.Append(Text);
end;

function TCustomMemo.AtEOF: Boolean;
begin
(*
  Result := QMultiLineEdit_atEnd(Handle);
*)
  Result := QMultiLineEdit.AtEnd;
end;

procedure TCustomMemo.Change;
begin
  Modified := True;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TCustomMemo.Clear;
begin
  FLines.Clear;
end;

procedure TCustomMemo.UnSelect;
begin
(*
  QMultiLineEdit_deselect(Handle);
*)
  QMultiLineEdit.Deselect;
end;

procedure TCustomMemo.ClearSelection;
begin
(*
  if HasSelection then QMultiLineEdit_insert(Handle, nil);
*)
  if HasSelection then QMultiLineEdit.Insert('');
end;

procedure TCustomMemo.CopyToClipboard;
begin
(*
  QMultiLineEdit_copy(Handle);
*)
  QMultiLineEdit.Copy;
end;

constructor TCustomMemo.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FLines := TMemoStrings.Create;
  TMemoStrings(FLines).Memo := Self;
  FWantReturns := True;
  FWantTabs := False;
  FMaxLength := -1;
  FWordWrap := True;
  FWrapMode := wmWidth;
  FScrollBars := ssNone;
  Palette.ColorRole := crBase;
  Palette.TextColorRole := crText;
  ParentColor := False;
  TabStop := True;
  InputKeys := [ikArrows, ikChars, ikNav, ikEdit];
  Height := 89;
  Width := 185;
end;

procedure TCustomMemo.CreateWidget;
begin
  FHandle := TQMultiLineEdit.Create(ParentWidget);
(*
  FHandle := QOpenMultiLineEdit_create(ParentWidget, nil);
  Hooks := QMultiLineEdit_hook_create(Handle);
*)
end;

procedure TCustomMemo.CMKeyDown(var Msg: TCMKeyDown);
begin
  if ((Msg.Key = Key_Return) or (Msg.Key = Key_Enter)) then
  begin
    Msg.Handled := not FWantReturns;
    if Msg.Handled then inherited;
  end
  else
    inherited;
end;

procedure TCustomMemo.CutToClipboard;
begin
(*
  QMultiLineEdit_cut(Handle);
*)
  QMultiLineEdit.Cut;
end;

destructor TCustomMemo.Destroy;
begin
  FLines.Free;
  inherited Destroy;
end;

function TCustomMemo.GetCaretPos: TCaretPos;
begin
(*
  QMultiLineEdit_cursorPosition(Handle, @Result.Line, @Result.Col);
*)
  QMultiLineEdit.CursorPosition(Result.Line, Result.Col);
end;

function TCustomMemo.GetHMargin: Integer;
begin
(*
  Result := QMultiLineEdit_hMargin(Handle);
*)
  Result := 0;
end;

function TCustomMemo.GetModified: Boolean;
begin
(*
  Result := QMultiLineEdit_edited(Handle);
*)
  Result := QMultiLineEdit.Edited;
end;

function TCustomMemo.GetReadOnly: Boolean;
begin
(*
  Result := QMultiLineEdit_isReadOnly(Handle);
*)
  Result := QMultiLineEdit.IsReadOnly;
end;

function TCustomMemo.GetSelection: TMemoSelection;
begin
(*
  QOpenMultiLineEdit_getMarkedRegion(Handle, @Result.Line1, @Result.Col1,
    @Result.Line2, @Result.Col2);
*)
  QMultiLineEdit.GetMarkedRegion(Result.Line1, Result.Col1,
      Result.Line2, Result.Col2);
end;

function TCustomMemo.GetSelLength: Integer;
begin
  Result := Length(SelText);
end;

function TCustomMemo.GetSelStart: Integer;
var
  Caret: TCaretPos;
begin
  if not HasSelection then
    Caret := CaretPos
  else
    Caret := Selection.TopLeft;
  Result := CaretPosToPos(Caret);
end;

function TCustomMemo.GetSelText: WideString;
begin
(*
  QOpenMultiLineEdit_MarkedText(Handle, PWideString(@Result));
*)
  Result := QMultiLineEdit.MarkedText;
end;

function TCustomMemo.GetText: TCaption;
begin
(*
  QOpenMultiLineEdit_getText(Handle, PWideString(@Result));
  if (Length(Result) > 0) and (Result[Length(Result)] = #10) then
    Delete(Result, Length(Result), 1);
  Result := CopyAddCR(Result);
*)
  Result := QMultiLineEdit.GetText;
end;

function TCustomMemo.GetTextWidth(LineNum: Integer): Integer;
begin
(*
  Result := QOpenMultiLineEdit_textWidth(Handle, LineNum);
*)
  Result := QMultiLineEdit.TextWidth(LineNum);
end;

procedure TCustomMemo.Insert(const Text: WideString; Mark: Boolean = False);
begin
(*
  QOpenMultiLineEdit_insert(Handle, PWideString(@Text), Mark);
*)
  QMultiLineEdit.Insert(Text, Mark);
end;

function TCustomMemo.HasSelection: Boolean;
begin
(*
  Result := QOpenMultiLineEdit_hasMarkedText(Handle);
*)
  Result := QMultiLineEdit.HasMarkedText;
end;

procedure TCustomMemo.Insert(const Text: WideString; Column, Row: Integer;
  Mark: Boolean = False);
begin
(*
  QMultiLineEdit_insertAt(Handle, PWideString(@Text), Column, Row, Mark);
*)
  QMultiLineEdit.InsertAt(Text, Column, Row, Mark);
end;

procedure TCustomMemo.PasteFromClipboard;
begin
(*
  QMultiLineEdit_paste(Handle);
*)
  QMultiLineEdit.Paste;
end;

procedure TCustomMemo.SelectAll;
begin
(*
  QMultiLineEdit_selectAll(Handle);
*)
  QMultiLineEdit.SelectAll;
end;

procedure TCustomMemo.SetCaretPos(const Value: TCaretPos);
begin
(*
  QMultiLineEdit_setCursorPosition(Handle, Value.Line, Value.Col, False);
*)
  QMultiLineEdit.SetCursorPosition(Value.Line, Value.Col);
end;

procedure TCustomMemo.SetHMargin(const Value: Integer);
begin
  if Value <> HMargin then
(*
    QMultiLineEdit_setHMargin(Handle, Value);
*)
end;

procedure TCustomMemo.SetLines(const Value: TStrings);
begin
  FLines.Assign(Value);
end;

procedure TCustomMemo.SetModified(const Value: Boolean);
begin
(*
  if Value <> Modified then QMultiLineEdit_setEdited(Handle, Value);
*)
  if Value <> Modified then QMultiLineEdit.SetEdited(Value);
end;

procedure TCustomMemo.SetReadOnly(const Value: Boolean);
begin
(*
  if Value <> ReadOnly then QMultiLineEdit_setReadOnly(Handle, Value);
*)
  if Value <> ReadOnly then QMultiLineEdit.SetReadOnly(Value);
end;

procedure TCustomMemo.SetText(const Value: TCaption);
begin
(*
  if (Length(Value) <> QOpenMultiLineEdit_textLength(Handle)) or (Value <> Text) then
*)
  begin
    TMemoStrings(Lines).SetTextStr(Value);
    TextChanged;
  end;
end;

procedure TCustomMemo.SetSelStart(const Value: Integer);
var
  Chars, Line, LLen: Integer;
  Pos: TCaretPos;
begin
  Chars := 0;
  Line := -1;
  repeat
    Inc(Line);
    LLen := LineLength(Line);
    Chars := Chars + LLen;
  until Chars > Value;
  Pos.Line := Line;
  Pos.Col := Value - (Chars - LLen) ;
  SetCaretPos(Pos);
end;

procedure TCustomMemo.SetSelLength(const Value: Integer);
var
  SelEnd: TCaretPos;
begin
  SelEnd := PosToCaretPos(GetSelStart + Value);
(*
  QMultiLineEdit_setCursorPosition(Handle, SelEnd.Line, SelEnd.Col, True);
*)
end;

procedure TCustomMemo.ReturnPressedHook;
begin
  try
    ReturnPressed;
  except
    Application.HandleException(Self);
  end;
end;

procedure TCustomMemo.TextChangedHook;
begin
  try
    if TMemoStrings(Lines).UpdateCount = 0 then
      Change
    else
      TMemoStrings(Lines).FChanged := True;
  except
    Application.HandleException(Self);
  end;
end;

function TCustomMemo.LineLength(Index: Integer): Integer;
begin
  { Qt returns length only up to the CR (not LF), even
    though QMultiLineEdit_text returns embedded CRLF's. }
(*
  Result := QOpenMultiLineEdit_lineLength(Handle, Index) + 1; { add the LF }
*)
  Result := QMultiLineEdit.LineLength(Index);
end;

procedure TCustomMemo.Loaded;
begin
  inherited Loaded;
(*
  QMultiLineEdit_setEdited(Handle, False);
*)
  QMultiLineEdit.SetEdited(False);
end;

function TCustomMemo.CaretPosToPos(CaretPos: TCaretPos): Integer;
var
  L, E: Integer;
begin
  L := 0;
  E := CaretPos.Line;
  Result := CaretPos.Col;
  while L < E do
  begin
    Inc(Result, LineLength(L));
    Inc(L);
  end;
end;

function TCustomMemo.PosToCaretPos(Pos: Integer): TCaretPos;
var
  Len, Offset: Integer;
begin
  Len := 0;
  Offset := 0;
  Result.Line := -1;
  repeat
    Inc(Offset, Len);
    Inc(Result.Line);
    if Result.Line < Lines.Count then
      Len := LineLength(Result.Line);
  until (Offset + Len >= Pos) or (Result.Line >= Lines.Count);
  Result.Col := Pos - Offset;
end;

procedure TCustomMemo.ReturnPressed;
begin
  if Assigned(FOnReturnPressed) then FOnReturnPressed(Self);
end;

procedure TCustomMemo.FontChanged;
begin
(*
  QWidget_setFont(Handle, Font.Handle);
*)
  inherited FontChanged;
end;

procedure TCustomMemo.SetWordWrap(const Value: Boolean);
begin
  if WordWrap <> Value then
  begin
    FWordWrap := Value;
    if FWordWrap then
      InternalSetWrapMode(FWrapMode)
    else
      InternalSetWrapMode(wmNone);
  end;
end;

procedure TCustomMemo.InitWidget;
begin
  inherited InitWidget;
  InternalSetWrapMode(wmWidth);
  WrapBreak := wbWhitespace;
(*
  QMultiLineEdit_setEdited(Handle, False);
*)
  QMultiLineEdit.SetEdited(False);
  WrapAtValue := -1;
  BorderStyle := bsSunken3d;
  ScrollBars := ssNone;
end;

function TCustomMemo.GetTextLen: Integer;
begin
  Result := Length(Text);
end;

function TCustomMemo.GetTextBuf(Buffer: PChar; BufSize: Integer): Integer;
var
  LText: string;
begin
  { Provided for compatibility. Causes unnecessary copying. }
  LText := Text;
  StrLCopy(Buffer, PChar(LText), BufSize-1);
  Result := StrLen(Buffer);
end;

procedure TCustomMemo.SetMaxLength(const Value: Integer);
var
  S: string;
  CPos: TCaretPos;
begin
  if FMaxLength <> Value then
  begin
    if Value < 0 then
      FMaxLength := -1
    else
      FMaxLength := Value;

    //Qt documentation says MaxLength truncates text, but it doesn't
(*
    if (FMaxLength > -1) and (QOpenMultiLineEdit_textLength(Handle) > FMaxLength) then
*)
    begin
      S := Text;
      CPos := CaretPos;
      SetLength(S, FMaxLength);
      Text := S;
      CaretPos := CPos;
    end;
(*
    QMultiLineEdit_setMaxLength(Handle, FMaxLength);
*)
  end;
end;

function TCustomMemo.GetAlignment: TAlignment;
begin
(*
  Result := QAlignmentToTAlignment(QMultiLineEdit_alignment(Handle));
*)
  Result := taLeftJustify;
end;

procedure TCustomMemo.SetAlignment(const Value: TAlignment);
var
  S: WideString;
  Workaround: Boolean;
  CPos: TCaretPos;
begin
  if Value <> Alignment then
  begin
    Workaround := Alignment = taLeftJustify;
    if Workaround then
    begin
      CPos := CaretPos;
      S := Text;
      Lines.Clear;
    end;
(*
    QMultiLineEdit_setAlignment(Handle, AlignmentFlags[Value]);
*)
    if Workaround then
    begin
      TMemoStrings(Lines).SetTextStr(S);
      CaretPos := CPos;
    end;
  end;
end;

procedure TCustomMemo.SetSelText(const Value: WideString);
begin
(*
  QOpenMultiLineEdit_insert(Handle, PWideString(@Value), True);
*)
  QMultiLineEdit.Insert(Value, True);
end;

procedure TCustomMemo.SetScrollBars(const Value: TScrollStyle);
(*
const
  ScrollFlags: array[TScrollStyle] of Cardinal = (
    0,                                          //ssNone
    Cardinal(QTableFlags_hScrollBar),           //ssHorizontal
    Cardinal(QTableFlags_vScrollBar),           //ssVertical
    Cardinal(QTableFlags_hScrollBar)
      or Cardinal(QTableFlags_vScrollBar),      //ssBoth
    Cardinal(QTableFlags_autoHScrollBar),       //ssAutoHorizontal
    Cardinal(QTableFlags_autoVScrollBar),       //ssAutoVertical
    Cardinal(QTableFlags_autoHScrollBar)
      or Cardinal(QTableFlags_autoVScrollBar)   //ssAutoBoth
  );

  ScrollFlagMask = $0000000F;
*)

begin
  FScrollBars := Value;
(*
  QOpenMultiLineEdit_clearTableFlags(Handle, ScrollFlagMask);
  QOpenMultiLineEdit_setTableFlags(Handle, ScrollFlags[FScrollBars]);
*)
end;

procedure TCustomMemo.SetWrapMode(const Value: TWrapMode);
begin
  if WrapMode <> Value then
  begin
    FWrapMode := Value;
    if FWordWrap then
      InternalSetWrapMode(FWrapMode);
  end;
end;

function TCustomMemo.GetWrapValue: Integer;
begin
(*
  Result := QMultiLineEdit_wrapColumnOrWidth(Handle);
*)
  Result := 0;
end;

procedure TCustomMemo.SetWrapValue(const Value: Integer);
begin
  if WrapAtValue <> Value then
(*
    QMultiLineEdit_setWrapColumnOrWidth(Handle, Value);
*)
end;

procedure TCustomMemo.InternalSetWrapMode(NewMode: TWrapMode);
begin
(*
  case NewMode of
    wmNone:
      QMultiLineEdit_setWordWrap(Handle, QMultiLineEditWordWrap_NoWrap);
    wmWidth:
      QMultiLineEdit_setWordWrap(Handle, QMultiLineEditWordWrap_WidgetWidth);
    wmPixel:
      QMultiLineEdit_setWordWrap(Handle, QMultiLineEditWordWrap_FixedPixelWidth);
    wmColumn:
      QMultiLineEdit_setWordWrap(Handle, QMultiLineEditWordWrap_FixedColumnWidth);
  end;
*)
  SetScrollBars(FScrollBars);
end;

function TCustomMemo.GetWrapBreak: TWrapBreak;
begin
(*
  Result := TWrapBreak(QMultiLineEdit_wrapPolicy(Handle));
*)
  Result := wbWhitespace;
end;

procedure TCustomMemo.SetWrapBreak(const Value: TWrapBreak);
begin
(*
  if WrapBreak <> Value then
    case Value of
      wbWhitespace:
        QMultiLineEdit_setWrapPolicy(Handle, QMultiLineEditWrapPolicy_AtWhiteSpace);
      wbAnywhere:
        QMultiLineEdit_setWrapPolicy(Handle, QMultiLineEditWrapPolicy_Anywhere);
    end;
*)
end;

function TCustomMemo.NeedKey(Key: Integer; Shift: TShiftState;
  const KeyText: WideString): Boolean;
begin
  Result := (((Key = Key_Tab) or (Key = Key_BackTab)) and FWantTabs)
    or inherited NeedKey(Key, Shift, KeyText);
end;

procedure TCustomMemo.HookEvents;
(*
var
  Method: TMethod;
*)
begin
  inherited;
(*
  QMultiLineEdit_textChanged_Event(Method) := TextChangedHook;
  QMultiLineEdit_hook_hook_textChanged(QMultiLineEdit_hookH(Hooks), Method);
  QMultiLineEdit_returnPressed_Event(Method) := ReturnPressedHook;
  QMultiLineEdit_hook_hook_returnPressed(QMultiLineEdit_hookH(Hooks), Method);
*)
  QMultiLineEdit.OnChanged := TextChangedHook;
  QMultiLineEdit.OnReturnPressed := ReturnPressedHook;
end;

function TCustomMemo.QMultiLineEdit: TQMultiLineEdit;
begin
  Result := Handle as TQMultiLineEdit;
end;


{ TRadioButton }

procedure TRadioButton.TurnSiblingsOff;
var
  I: Integer;
  Sibling: TControl;
begin
  if Parent <> nil then
    with Parent do
      for I := 0 to ControlCount - 1 do
      begin
        Sibling := Controls[I];
        if (Sibling <> Self) and (Sibling is TRadioButton) then
          TRadioButton(Sibling).SetChecked(False);
      end;
  if not QRadioButton.IsChecked then
    QRadioButton.SetChecked(True);
end;

procedure TRadioButton.Click;
begin
  TabStop := Checked;
  TurnSiblingsOff;
  inherited Click;
end;

procedure TRadioButton.ClickedHook;
begin
  try
    Click;
  except
    Application.HandleException(Self);
  end;
end;

constructor TRadioButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Width := 113;
  Height := 17;
  ControlStyle := [csSetCaption, csDoubleClicks];
  Palette.ColorRole := crBackground;
  Palette.TextColorRole := crForeground;
  TabStop := True;
end;

procedure TRadioButton.CreateWidget;
begin
  FHandle := TQRadioButton.Create(ParentWidget);
  QRadioButton.WFlags := WidgetFlags;
(*
  FHandle := QRadioButton_create(ParentWidget, nil);
  Hooks := QButton_hook_create(Handle);
*)
end;

function TRadioButton.GetChecked: Boolean;
begin
  Result := QRadioButton.IsChecked;
end;

procedure TRadioButton.HookEvents;
(*
var
  Method: TMethod;
*)  
begin
  inherited;
  QButton.OnClicked := ClickedHook;

(*
  QButton_clicked_Event(Method) := ClickedHook;
  QButton_hook_hook_Clicked(QButton_hookH(Hooks), Method);
*)
end;

procedure TRadioButton.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  if not (ssDouble in Shift) then
    inherited MouseDown(Button, Shift, X, Y)
  else if Button = mbLeft then
    DblClick;
end;

procedure TRadioButton.SetChecked(Value: Boolean);
begin
  TabStop := Value;
  if Checked <> Value then
  begin
    if HandleAllocated then
      QRadioButton.SetChecked(Value);
    if Value then
    begin
      TurnSiblingsOff;
      { inherited Changed; }
      if not ClicksDisabled then Click;
    end;
  end;
end;

function TRadioButton.WantKey(Key: Integer; Shift: TShiftState;
  const KeyText: WideString): Boolean;
begin
  if (Shift = [ssAlt]) then
  begin
    Key := TranslateAltKeyToKey(Key);
    if IsAccel(Key, Caption) and CanFocus then
    begin
      SetFocus;
      Checked := True;
      Result := True;
      Exit;
    end
  end;
  Result := inherited WantKey(Key, Shift, KeyText);
end;

function TRadioButton.QRadioButton: TQRadioButton;
begin
  Result := Handle as TQRadioButton;
end;


{ TScrollBar }

procedure TScrollBar.Change;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

constructor TScrollBar.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Palette.ColorRole := crButton;
  Palette.TextColorRole := crButtonText;
  FKind := sbHorizontal;
  Width := 121;
  Height := 18;
  TabStop := True;
  ControlStyle := [csFramed, csDoubleClicks, csOpaque];
  InputKeys := [ikArrows];
end;

procedure TScrollBar.CreateWidget;
begin
  FHandle := TQScrollBar.Create(ParentWidget);
  QScrollBar.WFlags := WidgetFlags;
(*
  FHandle := QScrollBar_create(ParentWidget, nil);
  Hooks := QScrollBar_hook_create(FHandle);
*)
end;

procedure TScrollBar.DoScroll(Code: TScrollCode);
var
  NewPos,
  ScrollPos: Integer;
begin
  if FParamsChanging then Exit;
  NewPos := Position;
  case Code of
    scPageDown:
      Inc(NewPos, LargeChange);
    scPageUp:
      Dec(NewPos, LargeChange);
    scLineDown:
      Inc(NewPos, SmallChange);
    scLineUp:
      Dec(NewPos, SmallChange);
  end;
  if NewPos > Max then NewPos := Max;
  if NewPos < Min then NewPos := Min;
  ScrollPos := NewPos;
  Scroll(Code, ScrollPos);
  if ScrollPos <> NewPos then
  begin
    case Code of
      scPageDown:
        Dec(ScrollPos, LargeChange);
      scPageUp:
        Inc(ScrollPos, LargeChange);
      scLineDown:
        Dec(ScrollPos, SmallChange);
      scLineUp:
        Inc(ScrollPos, SmallChange);
    end;
    Position := ScrollPos;
  end;
end;

function TScrollBar.GetLargeChange: TScrollBarInc;
begin
  Result := QScrollBar.PageStep;
(*
  Result := QRangeControl_pageStep(RangeControl);
*)
end;

function TScrollBar.GetMax: Integer;
begin
  Result := QScrollBar.MaxValue;
(*
  Result := QRangeControl_maxValue(RangeControl);
*)
end;

function TScrollBar.GetMin: Integer;
begin
  Result := QScrollBar.MinValue;
(*
  Result := QRangeControl_minValue(RangeControl);
*)
end;

function TScrollBar.GetPosition: Integer;
begin
  Result := QScrollBar.Value;
(*
  Result := QRangeControl_value(RangeControl);
*)
end;

function TScrollBar.GetSmallChange: TScrollBarInc;
begin
  Result := QScrollBar.LineStep;
(*
  Result := QRangeControl_lineStep(RangeControl);
*)
end;

function TScrollBar.GetTracking: Boolean;
begin
(*
  Result := QScrollBar_tracking(Handle);
*)
  Result := False;
end;

procedure TScrollBar.HookEvents;
begin
  QScrollBar.OnChange := ValueChangedHook;
  QScrollBar.OnScroll := ScrollHook;
(*
  QScrollBar_hook_hook_valueChanged(QScrollBar_hookH(Hooks), Method);
  QScrollBar_sliderPressed_Event(Method) := SliderPressedHook;
  QScrollBar_hook_hook_sliderPressed(QScrollBar_hookH(Hooks), Method);
  QScrollBar_sliderReleased_Event(Method) := SliderReleasedHook;
  QScrollBar_hook_hook_sliderReleased(QScrollBar_hookH(Hooks), Method);
  QScrollBar_nextLine_Event(Method) := NextLineHook;
  QScrollBar_hook_hook_nextLine(QScrollBar_hookH(Hooks), Method);
  QScrollBar_prevLine_Event(Method) := PrevLineHook;
  QScrollBar_hook_hook_prevLine(QScrollBar_hookH(Hooks), Method);
  QScrollBar_nextPage_Event(Method) := NextPageHook;
  QScrollBar_hook_hook_nextPage(QScrollBar_hookH(Hooks), Method);
  QScrollBar_prevPage_Event(Method) := PrevPageHook;
  QScrollBar_hook_hook_prevPage(QScrollBar_hookH(Hooks), Method);
*)
  inherited HookEvents;
end;

procedure TScrollBar.InitWidget;
begin
  inherited InitWidget;
  SmallChange := 1;
  LargeChange := 1;
  Max := 100;
  Min := 0;
  Track := True;
  QScrollBar.SetOrientation(TOrientation(sbHorizontal));
(*
  QScrollBar_setOrientation(Handle, Orientation(sbHorizontal));
*)
end;

procedure TScrollBar.KeyUp(var Key: Word; Shift: TShiftState);
begin
  inherited KeyUp(Key, Shift);
(*
  case Key of
    Key_Left, Key_Right, Key_Up, Key_Down, Key_PageUp,
    Key_PageDown, Key_Home, Key_End:
      if Enabled and Visible then DoScroll(scEndScroll);
  end;
*)
end;

procedure TScrollBar.Loaded;
begin
  inherited;
  QScrollBar.SetOrientation(TOrientation(Kind));
(*
  QScrollBar_setOrientation(Handle, Orientation(Kind));
*)
end;

procedure TScrollBar.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  if Enabled and Visible then DoScroll(scEndScroll);
end;

procedure TScrollBar.ScrollHook(Sender: TObject; PartCode: Integer);

const ScrollCode: array[0..8] of TScrollCode = (
    scLineUp, scLineDown, scPageUp, scPageDown,
    scLineUp, scLineDown, scPageUp, scPageDown,
    scPosition);

//    (scLineUp, scLineDown, scPageUp, scPageDown,
//    scPosition, scTrack, scTop, scBottom, scEndScroll);

begin
  try
    DoScroll(ScrollCode[PartCode]);
    if Position = Max then DoScroll(scBottom)
    else if Position = Min then DoScroll(scTop);
  except
    Application.HandleException(Self);
  end;
end;

(*
procedure TScrollBar.NextLineHook;
begin
  try
    DoScroll(scLineDown);
    if Position = Max then DoScroll(scBottom);
  except
    Application.HandleException(Self);
  end;
end;

procedure TScrollBar.NextPageHook;
begin
  try
    DoScroll(scPageDown);
    if Position = Max then DoScroll(scBottom);
  except
    Application.HandleException(Self);
  end;
end;

procedure TScrollBar.PrevLineHook;
begin
  try
    DoScroll(scLineUp);
    if Position = Min then DoScroll(scTop);
  except
    Application.HandleException(Self);
  end;
end;

procedure TScrollBar.PrevPageHook;
begin
  try
    DoScroll(scPageUp);
    if Position = Min then DoScroll(scTop);
  except
    Application.HandleException(Self);
  end;
end;
*)

procedure TScrollBar.Scroll(ScrollCode: TScrollCode;
  var ScrollPos: Integer);
begin
  if Assigned(FOnScroll) then FOnScroll(Self, ScrollCode, ScrollPos);
end;

procedure TScrollBar.SetKind(Value: TScrollBarKind);
begin
  { TScrollBarKind maps directly to Qt::Orientation. This could change! }
  if Kind <> Value then
  begin
    FKind := Value;
    QScrollBar.SetOrientation(TOrientation(FKind));
(*
    QScrollBar_setOrientation(Handle, Orientation(FKind));
*)
    if not (csLoading in ComponentState) then
      SetBounds(Left, Top, Height, Width);
  end;
end;

procedure TScrollBar.SetLargeChange(const Value: TScrollBarInc);
begin
  if Value <> LargeChange then
    QScrollBar.SetSteps(SmallChange, Value);
(*
    QRangeControl_setSteps(RangeControl, SmallChange, Value);
*)
end;

procedure TScrollBar.SetMax(Value: Integer);
begin
  SetParams(Position, Min, Value);
end;

procedure TScrollBar.SetMin(Value: Integer);
begin
  SetParams(Position, Value, Max);
end;

procedure TScrollBar.SetParams(APosition, AMin, AMax: Integer);
begin
  if AMax < AMin then
    raise EInvalidOperation.Create(SScrollBarRange);
  if APosition < AMin then APosition := AMin;
  if APosition > AMax then APosition := AMax;
  FParamsChanging := True;
  try
    if (Min <> AMin) or (Max <> AMax) then
      QScrollBar.SetRange(AMin, AMax);
(*
      QRangeControl_setRange(RangeControl, AMin, AMax);
*)
    if Position <> APosition then
    begin
      if Position <> APosition then
        QScrollBar.SetValue(APosition);
(*
        QRangeControl_setValue(RangeControl, APosition);
*)
      Enabled := True;
      DoScroll(scPosition);
    end;
  finally
    FParamsChanging := False;
  end;
end;

procedure TScrollBar.SetPosition(Value: Integer);
begin
  SetParams(Value, Min, Max);
end;

procedure TScrollBar.SetSmallChange(const Value: TScrollBarInc);
begin
  if Value <> SmallChange then
    QScrollBar.SetSteps(Value, LargeChange);
(*
    QRangeControl_SetSteps(RangeControl, Value, LargeChange);
*)
end;

procedure TScrollBar.SetTracking(const Value: Boolean);
begin
  if Track <> Value then
(*
    QScrollBar_setTracking(Handle, Value);
*)
end;

(*
procedure TScrollBar.SliderPressedHook;
begin
  if Track then
  try
    DoScroll(scTrack);
  except
    Application.HandleException(Self);
  end;
end;

procedure TScrollBar.SliderReleasedHook;
begin
  try
    DoScroll(scPosition);
  except
    Application.HandleException(Self);
  end;
end;
*)

procedure TScrollBar.TabStopChanged;
begin
  inherited TabStopChanged;
(*
  if TabStop then
    QWidget_setFocusPolicy(Handle, QWidgetFocusPolicy_ClickFocus)
  else
    QWidget_setFocusPolicy(Handle, QWidgetFocusPolicy_NoFocus)
*)
end;

function TScrollBar.QScrollBar: TQScrollBar;
begin
  Result := Handle as TQScrollBar;
end;

procedure TScrollBar.SetPageSize(Value: Integer);
begin
  { do nothing }
end;

function TScrollBar.GetPageSize: Integer;
begin
  Result := 0;
end;

procedure TScrollBar.ValueChangedHook(Sender: TObject);
begin
  try
    DoScroll(scTrack);
    if Position = Min then DoScroll(scTop)
    else if Position = Max then DoScroll(scBottom);
    Change;
  except
    Application.HandleException(Self);
  end;
end;

end.

