
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (clx/vi)      }
{ ViClasses.pas                                         }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Turbo Pascal Version 7.0                          }
{     Turbo Vision Unit                                 }
{     Copyright (c) 1992 Borland International          }
{     www: http://www.borland.com                       }
{   - Virtual Pascal v2.1                               }
{     Copyright (C) 1996-2000 vpascal.com               }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/ViClasses.pas,v 1.2 2002/10/27 22:25:16 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: ViClasses.pas,v $
  Revision 1.2  2002/10/27 22:25:16  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
{
  Version history:
    IWA 2001-05-24 Version 2.0.1.0 released
}

unit ViClasses;

{$X+}

interface

uses
  Classes, ViTypes;

{ Rectangle functions }

procedure RectAssign(var R: TRect; XA, YA, XB, YB: Integer);
procedure RectCopy(var R: TRect; Source: TRect);
procedure RectMove(var R: TRect; ADX, ADY: Integer);
procedure RectGrow(var R: TRect; ADX, ADY: Integer);
procedure RectIntersect(var RA: TRect; RB: TRect);
procedure RectUnion(var RA: TRect; RB: TRect);
function RectContains(R: TRect; P: TPoint): Boolean;
function RectEquals(RA: TRect; RB: TRect): Boolean;
function RectEmpty(R: TRect): Boolean;

{ Longint routines }

function LongMul(X, Y: Integer): Integer;

function LongDiv(X: Integer; Y: Integer): Integer;

function ISqr(X: Integer): Integer;

procedure MostEqualDivisors(N: Integer; var X, Y: Integer; FavorY: 
    Boolean);

implementation

function LongMul(X, Y: Integer): LongInt;
begin
  Result := X * Y;
end;

function LongDiv(X: Longint; Y: Integer): Integer;
begin
  Result := X div Y;
end;

function ISqr(X: Integer): Integer;
asm
                push    ebx
                mov     ecx,eax
                xor     ebx,ebx
              @@1:
                inc     ebx
                mov     eax,ebx
                imul    eax
                cmp     eax,ecx
                jle     @@1
                lea     eax,[ebx-1]
                pop     ebx
end;

procedure MostEqualDivisors(N: Integer; var X, Y: Integer; FavorY: 
    Boolean);
var
  I: Integer;
begin
  I := ISqr(N);
  if ((N mod I) <> 0) then
    if (N mod (I + 1)) = 0 then Inc(I);
  if I < (N div I) then I := N div I;
  if FavorY then
  begin
    X := N div I;
    Y := I;
  end
  else
  begin
    Y := N div I;
    X := I;
  end;
end;

function RectEmpty(R: TRect): Boolean;
begin
  Result := ((R.A.X = R.B.X) and (R.A.Y = R.B.Y));
end;

procedure RectAssign(var R: TRect; XA, YA, XB, YB: Integer);
begin
  R.A.X := XA;
  R.A.Y := YA;
  R.B.X := XB;
  R.B.Y := YB;
end;

procedure RectCopy(var R: TRect; Source: TRect);
begin
  R := Source;
end;

function RectEquals(RA: TRect; RB: TRect): Boolean;
begin
  Result := (RA.A.X = RB.A.X)
      and (RA.A.Y = RB.A.Y)
      and (RA.B.X = RB.B.X)
      and (RA.B.Y = RB.B.Y);
end;

function RectContains(R: TRect; P: TPoint): Boolean;
begin
  Result := (P.X >= R.A.X)
    and (P.X < R.B.X)
    and (P.Y >= R.A.Y)
    and (P.Y < R.B.Y)
end;

procedure RectUnion(var RA: TRect; RB: TRect);
begin
  if RA.A.X > RB.A.X then
    RA.A.X := RB.A.X;
  if RA.A.Y > RB.A.Y then
    RA.A.Y := RB.A.Y;
  if RA.B.X < RB.B.X then
    RA.B.X := RB.B.X;
  if RA.B.Y < RB.B.Y then
    RA.B.Y := RB.B.Y;
end;

procedure RectIntersect(var RA: TRect; RB: TRect);
begin
  if RA.A.X < RB.A.X then
    RA.A.X := RB.A.X;
  if RA.A.Y < RB.A.Y then
    RA.A.Y := RB.A.Y;
  if RA.B.X > RB.B.X then
    RA.B.X := RB.B.X;
  if RA.B.Y > RB.B.Y then
    RA.B.Y := RB.B.Y;
end;

procedure RectGrow(var R: TRect; ADX, ADY: Integer);
begin
    Dec(R.A.X, ADX);
    Inc(R.B.X, ADX);
    Dec(R.A.Y, ADY);
    Inc(R.B.Y, ADY);
end;

procedure RectMove(var R: TRect; ADX, ADY: Integer);
begin
    Inc(R.A.X, ADX);
    Inc(R.B.X, ADX);
    Inc(R.A.Y, ADY);
    Inc(R.B.Y, ADY);
end;

{ TRect }

end.
