
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (clx/vi)      }
{ ViEvents.pas                                          }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Turbo Pascal Version 7.0                          }
{     Turbo Vision Unit                                 }
{     Copyright (c) 1992 Borland International          }
{     www: http://www.borland.com                       }
{   - Virtual Pascal v2.1                               }
{     Copyright (C) 1996-2000 vpascal.com               }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/ViEventsClass.pas,v 1.4 2003/10/07 19:18:34 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: ViEventsClass.pas,v $
  Revision 1.4  2003/10/07 19:18:34  iwache
  Delay in TViEvents.GetKeyEvent (Sleep(1)) only applied
  for DWPL Windows mode.

  Revision 1.3  2002/10/27 22:25:17  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
{
  Version history:
    IWA 2001-05-24 Version 2.0.1.0 released
}

unit ViEventsClass;

interface

uses
  Windows, SysUtils, ViTypes, ViMouses;

type
  TViEvents = class(TObject)
  public
    procedure Done;
    procedure GetKeyEvent(var Event: TEvent);
    procedure GetMouseEvent(var Event: TEvent);
    procedure Init;
    procedure ProcessEvents;
  end;

function GetShiftState: Byte;

var
  ViEvents: TViEvents;

implementation

{$IFDEF WIN}
uses
  WinMouses;
{$ENDIF}
{$IFDEF WDOSX}
uses
  WDosMouses;
{$ENDIF}

function GetShiftState: Byte;
begin
  Result := KeyboardShiftState;
end;

procedure TViEvents.Done;
begin
  ViMouse.Hide;
end;

procedure TViEvents.GetKeyEvent(var Event: TEvent);
var
  I: Integer;

  // Keyboard scan codes
  const
    scSpace    = $39;  scIns      = $52;  scDel      = $53;
    scBack     = $0E;  scUp       = $48;  scDown     = $50;
    scLeft     = $4B;  scRight    = $4D;  scHome     = $47;
    scEnd      = $4F;  scPgUp     = $49;  scPgDn     = $51;
    scCtrlIns  = $92;  scCtrlDel  = $93;  scCtrlUp   = $8D;
    scCtrlDown = $91;  kbShift    = kbLeftShift + kbRightShift;

type
  KeyTransEntry = record
    Scan:  Byte;
    Shift: Byte;
    Code:  Word;
  end;

const
  KeyTranslateTable : array [1..15] of KeyTransEntry =
    (( Scan: scSpace   ; Shift: $08 ; Code: kbAltSpace    ), // Alt-Space
     ( Scan: scIns     ; Shift: $04 ; Code: kbCtrlIns     ), // Ctrl-Ins
     ( Scan: scCtrlIns ; Shift: $04 ; Code: kbCtrlIns     ), // Ctrl-Ins
     ( Scan: scIns     ; Shift: $01 ; Code: kbShiftIns    ), // Shift-Ins
     ( Scan: scIns     ; Shift: $02 ; Code: kbShiftIns    ), // Shift-Ins
     ( Scan: scIns     ; Shift: $03 ; Code: kbShiftIns    ), // Shift-Ins
     ( Scan: scDel     ; Shift: $04 ; Code: kbCtrlDel     ), // Ctrl-Del
     ( Scan: scCtrlDel ; Shift: $04 ; Code: kbCtrlDel     ), // Ctrl-Del
     ( Scan: scDel     ; Shift: $01 ; Code: kbShiftDel    ), // Shift-Del
     ( Scan: scDel     ; Shift: $02 ; Code: kbShiftDel    ), // Shift-Del
     ( Scan: scDel     ; Shift: $03 ; Code: kbShiftDel    ), // Shift-Del
     ( Scan: scBack    ; Shift: $09 ; Code: kbAltShiftBack), // Alt-Shift-Backspace
     ( Scan: scBack    ; Shift: $0A ; Code: kbAltShiftBack), // Alt-Shift-Backspace
     ( Scan: scBack    ; Shift: $0B ; Code: kbAltShiftBack), // Alt-Shift-Backspace
     ( Scan: scBack    ; Shift: $08 ; Code: kbAltBack    )); // Alt-Backspace
begin
{$IFDEF WIN}
  Sleep(1);  // this is only needed for Win9x/WinMe
{$ENDIF WIN}
  with Event do
    if KeyboardScanCode = 0 then
      What := evNothing
    else
    begin
      What := evKeyDown;
      CharCode := KeyboardCharCode;
      ScanCode := KeyboardScanCode;
      ShiftState := KeyboardShiftState;
      KeyboardScanCode := 0;
      for I := Low(KeyTranslateTable) to High(KeyTranslateTable) do
        with KeyTranslateTable[I] do
        begin
          if (Scan = ScanCode) and ((Shift and ShiftState) = Shift) then
          begin
            KeyCode := Code;
            Break;
          end;
        end;
      if (CharCode = #$E0) and (ScanCode in
        [scUp,scDown,scLeft,scRight,scIns,scDel,scHome,scEnd,scPgUp,scPgDn,
        Hi(kbCtrlHome), Hi(kbCtrlEnd)  , Hi(kbCtrlPgUp), Hi(kbCtrlPgDn),
        Hi(kbCtrlLeft), Hi(kbCtrlRight),    scCtrlUp,       scCtrlDown]) then
          CharCode := #0;      // Grey Keys
      if KeyCode = $E00D then  // Grey Enter
        KeyCode := kbEnter;
    end;
end;

procedure TViEvents.GetMouseEvent(var Event: TEvent);
begin
  ViMouse.GetEvent(Event);
end;

procedure TViEvents.Init;
begin
  ViMouse.Show;
end;

procedure TViEvents.ProcessEvents;
begin
  ViMouse.ProcessEvents;
end;


initialization
  ViEvents := TViEvents.Create;

finalization
  FreeAndNil(ViEvents);

end.
