
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (clx/vi)      }
{ ViScreens.pas                                         }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Turbo Pascal Version 7.0                          }
{     Turbo Vision Unit                                 }
{     Copyright (c) 1992 Borland International          }
{     www: http://www.borland.com                       }
{   - Virtual Pascal v2.1                               }
{     Copyright (C) 1996-2000 vpascal.com               }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/ViScreens.pas,v 1.6 2003/12/01 21:08:07 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: ViScreens.pas,v $
  Revision 1.6  2003/12/01 21:08:07  iwache
  Property BlinkMode to class TViScreen added. Thanks to
  Ingo for this nice feature. Method InitVideo now sets
  Blinkmode to False by default.

  Revision 1.5  2003/01/31 21:31:14  iwache
  Bug for multithreading fixed. Procedure TViScreen.WriteBuffer
  had a very big local var Scr: TScr (~8KByte). This is now a private
  field of TViScreen.

  Revision 1.4  2002/10/27 22:25:20  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
{
  Version history:
    IWA 2001-05-24 Version 2.0.1.0 released
}

unit ViScreens;

interface

{$IFDEF WIN}
uses
  SysUtils, WinScreens, ViTypes;
{$ENDIF}
{$IFDEF WDOSX}
uses
  SysUtils, WDosScreens, ViTypes;
{$ENDIF}

type
  TViScreen = class(TObject)
  private
    FScr: TScr;
    FStartupMode: Word;
    FMode: Word;
    FCursorVisible: Boolean;
    FCursorStyle: TCursorStyle;
    FVideoMirror: TVideoBuf;
    FCodePage: Integer;
    FBlinkMode: Boolean;
    procedure SetMode(Value: Word);
    function GetWidth: Byte;
    function GetHeight: Byte;
    function GetCursorLines: Integer;
    procedure SetCursorLines(Value: Integer);
    procedure SetCursorVisible(Value: Boolean);
    procedure SetCursorStyle(Value: TCursorStyle);
    procedure SetBlinkMode(Value: Boolean);
  public
    VideoBuffer: TVideoBuf;
    function GetHiRes: Boolean;
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    procedure DoneVideo;
    procedure InitVideo;
    procedure ShowCursor;
    procedure HideCursor;
    procedure BigCursor;
    procedure NormCursor;
    procedure WriteBuffer(X, Y, Count: Integer; var Source);
    procedure WriteScreen(X, Y, Index, Count: Integer);
    procedure SetCursor(X, Y: Byte);
    property Mode: Word read FMode write SetMode;
    property Width: Byte read GetWidth;
    property Height: Byte read GetHeight;
    property StartupMode: Word read FStartupMode;
    property CursorLines: Integer read GetCursorLines write SetCursorLines;
    property HiRes: Boolean read GetHiRes;
    property CursorVisible: Boolean read FCursorVisible write SetCursorVisible;
    property CursorStyle: TCursorStyle read FCursorStyle write SetCursorStyle;
    property CodePage: Integer read FCodePage;
    property BlinkMode: Boolean read FBlinkMode write SetBlinkMode;
  end;

var
  ViScreen: TViScreen;

implementation

constructor TViScreen.Create;
begin
  inherited Create;
  FStartupMode := $FFFF;
  FMode := FStartupMode;
  FCodePage := 437;
  FCursorVisible := True;
end;

destructor TViScreen.Destroy;
begin
  DoneVideo;
  inherited Destroy;
end;

procedure TViScreen.Clear;
begin
  ClrScr;
end;

procedure TViScreen.DoneVideo;
begin
  if (FStartupMode <> $FFFF) then
  begin
    if (FStartupMode <> FMode) then
      SetTextMode(FStartupMode);
    FStartupMode := $FFFF;
    Clear;
    NormCursor;
    ShowCursor;
    DoneCrt;
  end;
end;

procedure TViScreen.InitVideo;
begin
  if FStartupMode = $FFFF then
  begin
    InitCrt;
    FStartupMode := GetTextMode;
    FMode := FStartupMode;
    HideCursor;
    // switch off blink mode for background colors > $7F
    FBlinkMode := True;
    BlinkMode := False;
  end;
end;

procedure TViScreen.SetMode(Value: Word);
begin
  if FMode <> Value then
  begin
    SetTextMode(Value);
    FMode := GetTextMode;
    ShowCursor;
    HideCursor;
    FillChar(FVideoMirror[0], SizeOf(TVideoBuf), #0);
  end;
end;

function TViScreen.GetWidth: Byte;
begin
  Result := ScreenWidth;
end;

function TViScreen.GetHeight: Byte;
begin
  Result := ScreenHeight;
end;

function TViScreen.GetCursorLines: Integer;
begin
  Result := GetCursorSize;
end;

procedure TViScreen.SetCursorLines(Value: Integer);
begin
  SetCursorSize(Value);
end;

function TViScreen.GetHiRes: Boolean;
begin
  Result := (FMode = smFont8x8);
end;

procedure TViScreen.ShowCursor;
begin
  CursorVisible := True;
end;

procedure TViScreen.HideCursor;
begin
  CursorVisible := False;
end;

procedure TViScreen.BigCursor;
begin
  CursorStyle := csBig;
end;

procedure TViScreen.NormCursor;
begin
  CursorStyle := csNorm;
end;

procedure TViScreen.SetCursorVisible(Value: Boolean);
begin
  if FCursorVisible <> Value then
  begin
    FCursorVisible := Value;
    if FCursorVisible then
    begin
      CursorOn;
      if FCursorStyle = csBig then CursorBig;
    end
    else CursorOff;
  end;
end;

procedure TViScreen.SetCursorStyle(Value: TCursorStyle);
begin
  if FCursorStyle <> Value then
  begin
    FCursorStyle := Value;
    if FCursorVisible then
      if FCursorStyle = csNorm then CursorNorm else CursorBig;
  end;
end;

procedure TViScreen.WriteBuffer(X, Y, Count: Integer; var Source);
var
  I: Integer;
  SourceBuffer: PVideoBuf;
begin
  SourceBuffer := Addr(Source);
  FScr.X1 := X;
  FScr.Y1 := Y;
  FScr.X2 := X + Count - 1;
  FScr.Y2 := Y;
  for I := 1 to Count do
  begin
    FScr.Data[I - 1].AsciiChar := Chr(SourceBuffer[I - 1].Lo);
    FScr.Data[I - 1].Attributes := SourceBuffer[I - 1].Hi;
  end;
  WriteScr(FScr, False);
end;

procedure TViScreen.WriteScreen(X, Y, Index, Count: Integer);
var
  I: Integer;
  Delta: Integer;
begin
  Delta := 0;
  for I := 0 to Count -1 do
  begin
    if Word(VideoBuffer[Index + I]) <>
        Word(FVideoMirror[Index + I]) then Break;
    Inc(Delta);
  end;
  Dec(Count, Delta);
  Inc(X, Delta);
  Inc(Index, Delta);
  Move(VideoBuffer[Index], FVideoMirror[Index], Count shl 1);
  WriteBuffer(X, Y, Count, FVideoMirror[Index]);
end;

procedure TViScreen.SetCursor(X, Y: Byte);
begin
  GotoXY(X + 1, Y + 1);
end;

procedure TViScreen.SetBlinkMode(Value: Boolean);
begin
  if FBlinkMode <> Value then
  begin
    FBlinkMode := Value;
{$IFDEF WDOSX}
    asm
         MOV   AH, 10h
         MOV   AL, 03h
         MOV   BL, Value
         INT   10h
    end;
{$ENDIF}
  end;
end;

initialization
  ViScreen := TViScreen.Create;

finalization
  FreeAndNil(ViScreen);

end.

