
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (clx/vi)      }
{ ViStdDlg.pas                                          }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Turbo Pascal Version 7.0                          }
{     Turbo Vision Unit                                 }
{     Copyright (c) 1992 Borland International          }
{     www: http://www.borland.com                       }
{   - Virtual Pascal v2.1                               }
{     Copyright (C) 1996-2000 vpascal.com               }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/ViStdDlg.pas,v 1.9 2004/03/30 20:24:51 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: ViStdDlg.pas,v $
  Revision 1.9  2004/03/30 20:24:51  iwache
  Class TViFileDialog modified for connection
  with WQ layer.

  Revision 1.8  2003/12/01 21:16:06  iwache
  Compiler warnings fixed by include of the JEDI.INC file.
  Tidy up unneeded code. Workaraound for function
  DirectoryExists by use of ExcludeTrailingPathDelimiter added.
  DirectoryExists returns wrong result with trailing slash
  by use of original WDOSX API modules (this is not valid with
  the DWPL modified API modules).

  Revision 1.7  2003/11/19 18:45:04  iwache
  TViFileDialog and TViChDirDialog for Vi-Mode full implemented.
  Connection to the WQ layer still missing.

  Revision 1.6  2003/03/03 15:52:50  iwache
  TViFileDialog.Create
  - TViHistory.Create remarked for new interface
  TViChDirDialog.Create
  - TViHistory.Create remarked for new interface
  uses clause modified
  - ViHistList no longer needed

  Revision 1.5  2002/10/27 22:25:21  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
unit ViStdDlg;

{$INCLUDE ..\..\..\rtl\sys\JEDI.INC} // Pull in the JCL/J-VCL shared directives

{$V-,X+}

interface

uses
  SysUtils, Classes, ViTypes, ViUtils, ViClasses, ViViews, ViDialogs;

const

  { Commands }

  cmFileOpen = 800; { Returned from TFileDialog when Open pressed }
  cmFileReplace = 801; { Returned from TFileDialog when Replace pressed }
  cmFileClear = 802; { Returned from TFileDialog when Clear pressed }
  cmFileInit = 803; { Used by TFileDialog internally }
  cmChangeDirInternal = 804; { Used by TChDirDialog internally }
  cmRevert = 805; { Used by TChDirDialog internally }

  { Messages }

  cmFileFocused = 806; { A new file was focused in the TFileList }
  cmFileDoubleClicked { A file was selected in the TFileList }
  = 807;

type
  PathStr = string;
  DirStr = string;
  NameStr = string;
  ExtStr = string;

  { TViSearchRec }

  {  Record used to store directory information by TFileDialog }

  TViSearchRec = class
  public
    Time: Integer;
    Size: Integer;
    Attr: Integer;
    Name: TFileName;
  end;

type

  { TFileInputLine is a special input line that is used by      }
  { TFileDialog that will update its contents in response to a  }
  { cmFileFocused command from a TFileList.                     }

  TViFileInputLine = class(TViInputLine)
    constructor Create(var Bounds: TRect; AMaxLen: Integer);
    procedure HandleEvent(var Event: TEvent); override;
  end;

  { TSortedListBox is a TListBox that assumes it has a          }
  { TStoredCollection instead of just a TCollection.  It will   }
  { perform an incremental search on the contents.              }

  TViSortedListBox = class(TViListBox)
    SearchPos: Word;
    ShiftState: Byte;
    constructor Create(var Bounds: TRect; ANumCols: Word;
      AScrollBar: TViScrollBar);
    procedure HandleEvent(var Event: TEvent); override;
    function GetKey(var S: string): Pointer; virtual;
  public
    procedure AssignList(AList: TStrings); override;
  end;

  { TFileList is a TSortedList box that assumes it contains     }
  { a TFileCollection as its collection.  It also communicates  }
  { through broadcast messages to TFileInput and TInfoPane      }
  { what file is currently selected.                            }

  TViFileList = class(TViSortedListBox)
    constructor Create(var Bounds: TRect; AScrollBar: TViScrollBar);
    destructor Destroy; override;
    function DataSize: Word; virtual;
    procedure FocusItem(Item: Integer); override;
    function GetText(Item: Integer; MaxLen: Integer): string; override;
    function GetKey(var S: string): Pointer; override;
    procedure HandleEvent(var Event: TEvent); override;
    procedure ReadDirectory(AWildCard: PathStr); virtual;
  private
    procedure Clear;
  end;

  { TFileInfoPane is a TView that displays the information      }
  { about the currently selected file in the TFileList          }
  { of a TFileDialog.                                           }

  TViFileInfoPane = class(TViView)
    S: TViSearchRec;
    constructor Create(var Bounds: TRect);
    procedure Draw; override;
    procedure HandleEvent(var Event: TEvent); override;
  end;

  { TFileDialog is a standard file name input dialog            }

  TWildStr = PathStr;

const
  fdOkButton = $0001; { Put an OK button in the dialog }
  fdOpenButton = $0002; { Put an Open button in the dialog }
  fdReplaceButton = $0004; { Put a Replace button in the dialog }
  fdClearButton = $0008; { Put a Clear button in the dialog }
  fdHelpButton = $0010; { Put a Help button in the dialog }
  fdNoLoadDir = $0100; { Do not load the current directory }
  { contents into the dialog at Init. }
  { This means you intend to change the }
  { WildCard by using SetData or store }
  { the dialog on a stream. }

type
  TCloseRequestEvent = procedure(Sender: TObject; var CanClose: Boolean)
      of object;

  TViFileDialog = class(TViDialog)
  private
    FOnCloseRequest: TCloseRequestEvent;
  public
    OpenSaveButton: TViButton;  
    FileName: TViFileInputLine;
    FileList: TViFileList;
    WildCard: TWildStr;
    Directory: string;
    ViHistory: TViHistory;
    constructor Create(AWildCard: TWildStr; const ATitle, InputName: string;
      AOptions: Word; HistoryList: TStrings);
    destructor Destroy; override;
    function GetFileName: string;
    procedure HandleEvent(var Event: TEvent); override;
    function Valid(Command: Word): Boolean; override;
    procedure ReadDirectory;
    property OnCloseRequest: TCloseRequestEvent read FOnCloseRequest write
        FOnCloseRequest;
  end;

  { TDirEntry }

  TDirEntry = class
  public
    DisplayText: string;
    Directory: string;
  end;

  { TDirListBox displays a tree of directories for use in the }
  { TChDirDialog.                                             }

  TViDirListBox = class(TViListBox)
  private
    procedure Clear;
  public
    Dir: DirStr;
    Cur: Word;
    constructor Create(var Bounds: TRect; AScrollBar: TViScrollBar);
    destructor Destroy; override;
    function GetText(Item: Integer; MaxLen: Integer): string; override;
    procedure HandleEvent(var Event: TEvent); override;
    function IsSelected(Item: Integer): Boolean; override;
    procedure NewDirectory(var ADir: DirStr);
    procedure SetState(AState: Word; Enable: Boolean); override;
  end;

  { TChDirDialog is a standard change directory dialog.         }

const
  cdNormal = $0000; { Option to use dialog immediately }
  cdNoLoadDir = $0001; { Option to init the dialog to store on a stream }
  cdHelpButton = $0002; { Put a help button in the dialog }

type

  TViChDirDialog = class(TViDialog)
    DirInput: TViInputLine;
    DirList: TViDirListBox;
    OkButton: TViButton;
    ChDirButton: TViButton;
    constructor Create(const ATitle: string; AOptions: Word; HistoryList: TStrings);
    procedure HandleEvent(var Event: TEvent); override;
    function Valid(Command: Word): Boolean; override;
    procedure SetUpDialog;
  private
    function GetDirName: string;
    procedure SetDirName(const Value: string);
  public
    property DirName: string read GetDirName write SetDirName;
  end;

const

  CInfoPane = #30;

implementation

uses
  ViEventsClass, ViApp, ViMsgBox;

const
{$IFDEF COMPILER6_UP}
  {$WARN SYMBOL_PLATFORM OFF}
{$ENDIF}
  FindAttr = faReadOnly + faArchive;
{$IFDEF COMPILER6_UP}
  {$WARN SYMBOL_PLATFORM ON}
{$ENDIF}
  AllFiles = '*.*';

function DriveValid(Drive: Char): Boolean;
begin
  Result := DirectoryExists(Drive + ':\');
end;

procedure SplitFileName(const FullName: string; var Dir, Name, Ext: string);
begin
  Dir := ExtractFileDir(FullName);
  Name := ExtractFileName(FullName);
  Ext := ExtractFileExt(Name);
  Name := ChangeFileExt(Name, '');
end;

function ValidFileName(var FileName: PathStr): Boolean;
const
  IllegalChars = ';,=+<>|"[]';
var
  Dir: DirStr;
  Name: NameStr;
  Ext: ExtStr;

  { Contains returns true if S1 contains any characters in S2 }
  function Contains(S1, S2: string): Boolean;
  var
    I: Integer;
  begin
    Result := False;
    for I := 1 to Length(S2) do
      if Pos(S2[I], S1) <> 0 then
      begin
        Result := True;
        Break;
      end;
  end;

begin
  Result := True;
  SplitFileName(FileName, Dir, Name, Ext);
  if not ((Dir = '') or DirectoryExists(Dir)) or Contains(Name, IllegalChars)
      or Contains(Dir, IllegalChars) then Result := False;
end;

function IsWild(const S: string): Boolean;
begin
  IsWild := (Pos('?', S) > 0) or (Pos('*', S) > 0);
end;

{ TFileInputLine }

constructor TViFileInputLine.Create(var Bounds: TRect; AMaxLen: Integer);
begin
  inherited Create(Bounds, AMaxLen);
  EventMask := EventMask or evBroadcast;
end;

procedure TViFileInputLine.HandleEvent(var Event: TEvent);
begin
  inherited HandleEvent(Event);
  if (Event.What = evBroadcast) and (Event.Command = cmFileFocused) and
    (State and sfSelected = 0) and Assigned(Event.InfoPtr) then
  begin
    if TViSearchRec(Event.InfoPtr).Attr and faDirectory <> 0 then
      Data := TViSearchRec(Event.InfoPtr).Name + '\' +
        TViFileDialog(Owner).WildCard
    else
      Data := TViSearchRec(Event.InfoPtr).Name;
    DrawView;
  end;
end;

{ TSortedListBox }

constructor TViSortedListBox.Create(var Bounds: TRect; ANumCols: Word;
  AScrollBar: TViScrollBar);
begin
  inherited Create(Bounds, ANumCols, AScrollBar);
  SearchPos := 0;
  ShowCursor;
  SetCursor(1, 0);
end;

procedure TViSortedListBox.HandleEvent(var Event: TEvent);
var
  CurString, NewString: string;
  Value, OldPos, OldValue: Integer;
  (*
    K: Pointer;
    T: Boolean;
  *)
  function Equal(const S1, S2: string; Count: Word): Boolean;
  var
    I: Word;
  begin
    Equal := False;
    if (Length(S1) < Count) or (Length(S2) < Count) then
      Exit;
    for I := 1 to Count do
      if UpCase(S1[I]) <> UpCase(S2[I]) then
        Exit;
    Equal := True;
  end;

begin
  OldValue := Focused;
  inherited HandleEvent(Event);
  if (OldValue <> Focused) or ((Event.What = evBroadcast) and
    (Event.InfoPtr = @Self) and (Event.Command = cmReleasedFocus)) then
    SearchPos := 0;
  if Event.What = evKeyDown then
  begin
    if Event.CharCode <> #0 then
    begin
      Value := Focused;
      if Value < Range then
        CurString := GetText(Value, 255)
      else
        CurString := '';
      OldPos := SearchPos;
      if Event.KeyCode = kbBack then
      begin
        if SearchPos = 0 then
          Exit;
        Dec(SearchPos);
        if SearchPos = 0 then
          ShiftState := GetShiftState;
        SetLength(CurString, SearchPos);
      end
      else if (Event.CharCode = '.') then
        SearchPos := Pos('.', CurString)
      else
      begin
        Inc(SearchPos);
        if SearchPos = 1 then
          ShiftState := GetShiftState;
        SetLength(CurString, SearchPos);
        CurString[SearchPos] := Event.CharCode;
      end;
      (*
            K := GetKey(CurString);
            T := PSortedCollection(List)^.Search(K, Value);
      *)
      if Value < Range then
      begin
        if Value < Range then
          NewString := GetText(Value, 255)
        else
          NewString := '';
        if Equal(NewString, CurString, SearchPos) then
        begin
          if Value <> OldValue then
          begin
            FocusItem(Value);
            { Assumes ListControl will set the cursor to the first character }
            { of the sfFocused item }
            SetCursor(Cursor.X + SearchPos, Cursor.Y);
          end
          else
            SetCursor(Cursor.X + (SearchPos - OldPos), Cursor.Y);
        end
        else
          SearchPos := OldPos;
      end
      else
        SearchPos := OldPos;
      if (SearchPos <> OldPos) or (Event.CharCode in ['A'..'Z', 'a'..'z']) then
        ClearEvent(Event);
    end;
  end;
end;

function TViSortedListBox.GetKey(var S: string): Pointer;
begin
  GetKey := @S;
end;

procedure TViSortedListBox.AssignList(AList: TStrings);
begin
  inherited AssignList(AList);
  SearchPos := 0;
end;

{ TFileList }

constructor TViFileList.Create(var Bounds: TRect; AScrollBar: TViScrollBar);
begin
  inherited Create(Bounds, 2, AScrollBar);
end;

destructor TViFileList.Destroy;
begin
  Clear;
  inherited Destroy;
end;

function TViFileList.DataSize: Word;
begin
  DataSize := 0;
end;

procedure TViFileList.FocusItem(Item: Integer);
begin
  inherited FocusItem(Item);
  SendMessage(Owner, evBroadcast, cmFileFocused, List.Objects[Item]);
end;

var
  SR: TSearchRec = ();

function TViFileList.GetKey(var S: string): Pointer;
begin
  if (ShiftState and $03 <> 0) or ((S <> '') and (S[1] = '.')) then
    SR.Attr := faDirectory
  else
    SR.Attr := 0;
  SR.Name := UpperCase(S);
  GetKey := @SR;
end;

function TViFileList.GetText(Item: Integer; MaxLen: Integer): string;
var
  SearchRec: TViSearchRec;
begin
  SearchRec := List.Objects[Item] as TViSearchRec;
  Result := SearchRec.Name;
  if SearchRec.Attr and faDirectory <> 0 then
    Result := Result + '\';
end;

procedure TViFileList.HandleEvent(var Event: TEvent);
begin
  if (Event.What = evMouseDown) and (Event.Double) then
  begin
    Event.What := evCommand;
    Event.Command := cmOK;
    PutEvent(Event);
    ClearEvent(Event);
  end
  else
    inherited HandleEvent(Event);
end;

procedure TViFileList.ReadDirectory(AWildCard: PathStr);
const
  PrevDir = '..';
var
  S: TSearchRec;
  ViSearchRec: TViSearchRec;
  FileList: TStrings;
  Dir: DirStr;
  Name: NameStr;
  Ext: ExtStr;
  Event: TEvent;
  Tmp: PathStr;
begin
  AWildCard := ExpandFileName(AWildCard);
  SplitFileName(AWildCard, Dir, Name, Ext);
  Dir := IncludeTrailingPathDelimiter(Dir);
  FileList := TStringList.Create;
  try
    // list files
    if FindFirst(AWildCard, FindAttr, S) = 0 then
    try
      repeat
        if (S.Attr and faDirectory) = 0 then
        begin
          ViSearchRec := TViSearchRec.Create;
          try
            ViSearchRec.Time := S.Time;
            ViSearchRec.Size := S.Size;
            ViSearchRec.Attr := S.Attr;
            ViSearchRec.Name := S.Name;
            FileList.AddObject(S.Name, ViSearchRec);
            ViSearchRec := nil;
          finally
            ViSearchRec.Free;
          end;
        end;
      until FindNext(S) <> 0;
    finally
      FindClose(S);
    end;

    // list directories
    Tmp := Dir + AllFiles;
    if FindFirst(Tmp, FindAttr + faDirectory, S) = 0 then
    try
      repeat
        if (S.Attr and faDirectory <> 0) and (S.Name[1] <> '.') then
        begin
          ViSearchRec := TViSearchRec.Create;
          try
            ViSearchRec.Time := S.Time;
            ViSearchRec.Size := S.Size;
            ViSearchRec.Attr := S.Attr;
            ViSearchRec.Name := S.Name;
            FileList.AddObject(S.Name, ViSearchRec);
            ViSearchRec := nil;
          finally
            ViSearchRec.Free;
          end;
        end;
      until FindNext(S) <> 0;
    finally
      FindClose(S);
    end;

    // list up (..) directory
    if Length(Dir) > 3 then // was 4 before ?
    begin
      Tmp := Dir + AllFiles;
      if FindFirst(Tmp, FindAttr + faDirectory, S) = 0 then
      try
        repeat
          if (S.Attr and faDirectory <> 0) and (S.Name = PrevDir) then
          begin
            ViSearchRec := TViSearchRec.Create;
            try
              ViSearchRec.Time := S.Time;
              ViSearchRec.Size := S.Size;
              ViSearchRec.Attr := S.Attr;
              ViSearchRec.Name := S.Name;
              FileList.AddObject(S.Name, ViSearchRec);
              ViSearchRec := nil;
            finally
              ViSearchRec.Free;
            end;
          end;
        until FindNext(S) <> 0;
      finally
        FindClose(S);
      end;
    end;
    Clear;
    AssignList(FileList);
  finally
    FileList.Free;
  end;
  if List.Count > 0 then
  begin
    Event.What := evBroadcast;
    Event.Command := cmFileFocused;
    Event.InfoPtr := List.Objects[0];
    Owner.HandleEvent(Event);
  end;
end;

procedure TViFileList.Clear;
var
  I: Integer;
begin
  for I := 0 to List.Count - 1 do
    List.Objects[I].Free;
  List.Clear;
end;

{ TFileInfoPane }

constructor TViFileInfoPane.Create(var Bounds: TRect);
begin
  inherited Create(Bounds);
  Palette := CInfoPane;
  EventMask := EventMask or evBroadcast;
end;

procedure TViFileInfoPane.Draw;
var
  B: TDrawBuffer;
  Color: Word;
  Path: PathStr;
  Str: string;
  FileDateTime: TDateTime;
  DateStr: string;
  NameStr: string;
begin
  { Display path }
  with TViFileDialog(Owner) do
    Path := ExpandFileName(Directory + WildCard);
  Path := Copy(Path, 1, 45);
  Color := GetColor($01);
  MoveChar(B, ' ', Color, Size.X);
  MoveStr(B[1], Path, Color);
  WriteLine(0, 0, Size.X, 1, B);

  { Display file/directory details }
  MoveChar(B, ' ', Color, Size.X);
  if Assigned(S) then
  begin
    NameStr := Copy(S.Name, 1, 14);
    FileDateTime := FileDateToDateTime(S.Time);
    DateStr := FormatDateTime('mmm dd, yyyy hh:mmam/pm', FileDateTime);
    if S.Attr and faDirectory <> 0 then
      Str := Format('%-14s %9s %s', [NameStr, 'Directory', DateStr])
    else
      Str := Format('%-14s %9d %s', [NameStr, S.Size, DateStr]);
  end;
  
  { Fill in rest of rectangle }
  MoveChar(B, ' ', Color, Size.X);
  MoveStr(B[1], Str, Color);
  WriteLine(0, 1, Size.X, Size.Y - 1, B);
end;

procedure TViFileInfoPane.HandleEvent(var Event: TEvent);
begin
  inherited HandleEvent(Event);
  if (Event.What = evBroadcast) and (Event.Command = cmFileFocused)
    and Assigned(Event.InfoPtr) then
  begin
    S := TViSearchRec(Event.InfoPtr);
    DrawView;
    S := nil;
  end;
end;

{ TFileDialog }

constructor TViFileDialog.Create(AWildCard: TWildStr; const ATitle, InputName:
  string; AOptions: Word; HistoryList: TStrings);
var
  Control: TViView;
  R: TRect;
  Opt: Word;
begin
  RectAssign(R, 15, 1, 64, 20);
  inherited Create(R, ATitle, True);
  Options := Options or ofCentered;
  WildCard := AWildCard;

  RectAssign(R, 3, 3, 31, 4);
  FileName := TViFileInputLine.Create(R, 79);
  FileName.Data := WildCard;
  Insert(FileName);
  RectAssign(R, 2, 2, 3 + CStrLen(InputName), 3);
  Control := TViLabel.Create(R, InputName, FileName);
  Insert(Control);
  RectAssign(R, 31, 3, 34, 4);
  ViHistory := TViHistory.Create(R, FileName, HistoryList);
  Insert(ViHistory);

  RectAssign(R, 3, 14, 34, 15);
  Control := TViScrollBar.Create(R);
  Insert(Control);
  RectAssign(R, 3, 6, 34, 14);
  FileList := TViFileList.Create(R, TViScrollBar(Control));
  Insert(FileList);
  RectAssign(R, 2, 5, 8, 6);
  Control := TViLabel.Create(R, '~F~iles', FileList);
  Insert(Control);

  RectAssign(R, 35, 3, 46, 5);
  Opt := bfDefault;
  if AOptions and fdOpenButton <> 0 then
  begin
    OpenSaveButton := TViButton.Create(R, '~O~pen', cmFileOpen, Opt, True);
    Insert(OpenSaveButton);
    Opt := bfNormal;
    Inc(R.A.Y, 3);
    Inc(R.B.Y, 3);
  end;
  if AOptions and fdOkButton <> 0 then
  begin
    Insert(TViButton.Create(R, 'O~K~', cmFileOpen, Opt, True));
    Opt := bfNormal;
    Inc(R.A.Y, 3);
    Inc(R.B.Y, 3);
  end;
  if AOptions and fdReplaceButton <> 0 then
  begin
    Insert(TViButton.Create(R, '~R~eplace', cmFileReplace, Opt, True));
    Opt := bfNormal;
    Inc(R.A.Y, 3);
    Inc(R.B.Y, 3);
  end;
  if AOptions and fdClearButton <> 0 then
  begin
    Insert(TViButton.Create(R, '~C~lear', cmFileClear, Opt, True));
    Inc(R.A.Y, 3);
    Inc(R.B.Y, 3);
  end;
  Opt := bfNormal;
  Insert(TViButton.Create(R, 'Cancel', cmCancel, Opt, True));
  Inc(R.A.Y, 3);
  Inc(R.B.Y, 3);
  if AOptions and fdHelpButton <> 0 then
  begin
    Insert(TViButton.Create(R, 'Help', cmHelp, Opt, True));
    Inc(R.A.Y, 3);
    Inc(R.B.Y, 3);
  end;

  RectAssign(R, 1, 16, 48, 18);
  Control := TViFileInfoPane.Create(R);
  Insert(Control);

  SelectNext(False);

  if AOptions and fdNoLoadDir = 0 then
    ReadDirectory;
end;

destructor TViFileDialog.Destroy;
begin
  inherited Destroy;
end;

function TViFileDialog.GetFileName: string;
var
  Path: PathStr;
  Name: NameStr;
  Ext: ExtStr;
  TPath: PathStr;
  TName: NameStr;
  TExt: NameStr;

  function IsRelativePath(S: PathStr): Boolean;
  begin
    S := Trim(S);
    Result := (Length(S) > 0) and (S[1] <> '\')
      and (Length(S) > 1) and (S[2] <> ':');
  end;

  function NoWildChars(const S: string): string;
  begin
    Result := StringReplace(S, '?', '', [rfReplaceAll]);
    Result := StringReplace(Result, '*', '', [rfReplaceAll]);
  end;

begin
  Result := FileName.Data;
  if IsRelativePath(Result) then
    Result := ExpandFileName(Directory + Result)
  else
    Result := Trim(Result);
  SplitFileName(Result, Path, Name, Ext);
  Path := IncludeTrailingPathDelimiter(Path);
  if ((Name = '') or (Ext = ''))
    and not DirectoryExists(ExcludeTrailingPathDelimiter(Result)) then
  begin
    SplitFileName(WildCard, TPath, TName, TExt);
    if ((Name = '') and (Ext = '')) then
      Result := Path + TName + TExt
    else if Name = '' then
      Result := Path + TName + Ext
    else if Ext = '' then
    begin
      if IsWild(Name) then
        Result := Path + Name + TExt
      else
        Result := Path + Name + NoWildChars(TExt);
    end;
  end;
end;

procedure TViFileDialog.HandleEvent(var Event: TEvent);
begin
  inherited HandleEvent(Event);
  if Event.What = evCommand then
    case Event.Command of
      cmFileOpen, cmFileReplace, cmFileClear:
        begin
          EndModal(Event.Command);
          ClearEvent(Event);
        end;
    end;
end;

procedure TViFileDialog.ReadDirectory;
begin
  if Directory = '' then
    Directory := IncludeTrailingPathDelimiter(GetCurrentDir);
  FileList.ReadDirectory(Directory + WildCard);
end;

function TViFileDialog.Valid(Command: Word): Boolean;
var
  FName: PathStr;
  Dir: DirStr;
  Name: NameStr;
  Ext: ExtStr;

  function CheckDirectory(const S: PathStr): Boolean;
  begin
    if not DirectoryExists(ExcludeTrailingPathDelimiter(S)) then
    begin
      MessageBox(^C'Invalid drive or directory.', mfError + mfOkButton);
      FileName.Select;
      Result := False;
    end
    else
      Result := True;
  end;

begin
  if Command = 0 then
  begin
    Result := True;
    Exit;
  end
  else
    Result := False;
  if inherited Valid(Command) then
  begin
    FName := GetFileName;
    if (Command <> cmCancel) and (Command <> cmFileClear) then
    begin
      if IsWild(FName) then
      begin
        SplitFileName(FName, Dir, Name, Ext);
        Dir := IncludeTrailingPathDelimiter(Dir);
        if CheckDirectory(Dir) then
        begin
          Directory := Dir;
          WildCard := Name + Ext;
          if Command <> cmFileInit then
            FileList.Select;
          FileList.ReadDirectory(Directory + WildCard);
        end
      end
      else if DirectoryExists(ExcludeTrailingPathDelimiter(FName)) then
      begin
        if CheckDirectory(FName) then
        begin
          Directory := FName + '\';
          if Command <> cmFileInit then
            FileList.Select;
          FileList.ReadDirectory(Directory + WildCard);
        end
      end
      else
      begin
        if Assigned(OnCloseRequest) then
        begin
          Result := True;
          OnCloseRequest(Self, Result)
        end
        else if ValidFileName(FName) then
          Result := True
        else
        begin
          MessageBox('Invalid file name.', mfError + mfOkButton);
          Result := False;
        end;
      end;
    end
    else
      Result := True;
  end;
end;

{ TDirListBox }

const
  Drives: string = 'Drives';

constructor TViDirListBox.Create(var Bounds: TRect; AScrollBar: TViScrollBar);
begin
  inherited Create(Bounds, 1, AScrollBar);
  Dir := '';
end;

destructor TViDirListBox.Destroy;
begin
  Clear;
  inherited Destroy;
end;

function TViDirListBox.GetText(Item: Integer; MaxLen: Integer): string;
begin
  Result := TDirEntry(List.Objects[Item]).DisplayText;
end;

procedure TViDirListBox.HandleEvent(var Event: TEvent);
begin
  if (Event.What = evMouseDown) and (Event.Double) then
  begin
    Event.What := evCommand;
    Event.Command := cmChangeDirInternal;
    PutEvent(Event);
    ClearEvent(Event);
  end
  else
    inherited HandleEvent(Event);
end;

function TViDirListBox.IsSelected(Item: Integer): Boolean;
begin
  IsSelected := Item = Cur;
end;

procedure TViDirListBox.NewDirectory(var ADir: DirStr);
var
  DirList: TStrings;
  NewDir, Dirct: DirStr;
  C, OldC: Char;
  S, Indent: string;
  P: string;
  IsFirst: Boolean;
  SearchRec: TSearchRec;
  I: Integer;

  function NewDirEntry(const DisplayText, Directory: string): TDirEntry;
  begin
    Result := TDirEntry.Create;
    Result.DisplayText := DisplayText;
    Result.Directory := Directory;
  end;

  function GetCurDrive: Char;
  var
    Path: string;
  begin
    Path := GetCurrentDir;
    GetCurDrive := Path[1];
  end;

begin
  Dir := ADir;
  DirList := TStringList.Create;
  try
    DirList.AddObject('', NewDirEntry(Drives, Drives));
    if Dir = Drives then
    begin
      IsFirst := True;
      OldC := ' ';
      for C := 'A' to 'Z' do
      begin
        if (C < 'C') or DriveValid(C) then
        begin
          if OldC <> ' ' then
          begin
            if IsFirst then
            begin
              S := ldFirstDir + OldC;
              IsFirst := False;
            end
            else
              S := ldMiddleDir + OldC;
            DirList.AddObject('', NewDirEntry(S, OldC + ':\'));
          end;
          if C = GetCurDrive then
            Cur := DirList.Count;
          OldC := C;
        end;
      end;
      if OldC <> ' ' then
        DirList.AddObject('',
          NewDirEntry(ldLastDir + OldC, OldC + ':\'));
    end
    else
    begin
      Indent := ldIndentSize;
      NewDir := Dir;
      Dirct := Copy(NewDir, 1, 3);
      DirList.AddObject('', NewDirEntry(ldPathDir + Dirct, Dirct));
      NewDir := Copy(NewDir, 4, 255);
      while NewDir <> '' do
      begin
        I := Pos('\', NewDir);
        if I <> 0 then
        begin
          S := Copy(NewDir, 1, I - 1);
          Dirct := Dirct + S;
          DirList.AddObject('', NewDirEntry(Indent + ldPathDir + S, Dirct));
          NewDir := Copy(NewDir, I + 1, 255);
        end
        else
        begin
          Dirct := Dirct + NewDir;
          DirList.AddObject('', NewDirEntry(Indent + ldPathDir + NewDir,
              Dirct));
          NewDir := '';
        end;
        Indent := Indent + ldIndentSize;
        Dirct := Dirct + '\';
      end;
      Cur := DirList.Count - 1;
      IsFirst := True;
      NewDir := Dirct + AllFiles;

      // list sub directories
      if FindFirst(NewDir, FindAttr + faDirectory, SearchRec) = 0 then
      try
        repeat
          if (SearchRec.Attr and faDirectory <> 0)
              and (SearchRec.Name[1] <> '.') then
          begin
            if IsFirst then
            begin
              S := ldFirstDir;
              IsFirst := False;
            end
            else
              S := ldMiddleDir;
            DirList.AddObject('', NewDirEntry(Indent + S + SearchRec.Name,
                Dirct + SearchRec.Name));
          end;
        until FindNext(SearchRec) <> 0;
      finally
        FindClose(SearchRec);
      end;

      P := TDirEntry(DirList.Objects[DirList.Count - 1]).DisplayText;
      I := Pos(ldPathDir[0], P);
      if I = 0 then
      begin
        I := Pos(ldMiddleDir[1], P);
        if I <> 0 then
          P[I] := ldPathDir[0];
      end
      else
      begin
        P[I + 1] := ldPathDir[1];
        P[I + 2] := ldPathDir[1];
      end;
      TDirEntry(DirList.Objects[DirList.Count - 1]).DisplayText := P;
    end;
    Clear;
    AssignList(DirList);
  finally
    DirList.Free;
  end;
  FocusItem(Cur);
end;

procedure TViDirListBox.SetState(AState: Word; Enable: Boolean);
begin
  inherited SetState(AState, Enable);
  if AState and sfFocused <> 0 then
    TViChDirDialog(Owner).ChDirButton.MakeDefault(Enable);
end;

procedure TViDirListBox.Clear;
var
  I: Integer;
begin
  for I := 0 to List.Count - 1 do
    List.Objects[I].Free;
  List.Clear;
end;


{ TChDirDialog }

constructor TViChDirDialog.Create(const ATitle: string; AOptions: Word;
    HistoryList: TStrings);
var
  R: TRect;
  Control: TViView;
begin
  RectAssign(R, 16, 2, 64, 20);
  inherited Create(R, ATitle, True);

  Options := Options or ofCentered;

  RectAssign(R, 3, 3, 30, 4);
  DirInput := TViInputLine.Create(R, 68);
  Insert(DirInput);
  RectAssign(R, 2, 2, 17, 3);
  Control := TViLabel.Create(R, 'Directory ~n~ame', DirInput);
  Insert(Control);
  RectAssign(R, 30, 3, 33, 4);
  Control := TViHistory.Create(R, DirInput, HistoryList);
  Insert(Control);

  RectAssign(R, 32, 6, 33, 16);
  Control := TViScrollBar.Create(R);
  Insert(Control);
  RectAssign(R, 3, 6, 32, 16);
  DirList := TViDirListBox.Create(R, TViScrollBar(Control));
  Insert(DirList);
  RectAssign(R, 2, 5, 17, 6);
  Control := TViLabel.Create(R, 'Directory ~t~ree', DirList);
  Insert(Control);

  RectAssign(R, 35, 3, 45, 5);
  OkButton := TViButton.Create(R, 'O~K~', cmOK, bfDefault, True);
  Insert(OkButton);
  Inc(R.A.Y, 3);
  Inc(R.B.Y, 3);
  Insert(TViButton.Create(R, 'Cancel', cmCancel, bfNormal, True));
  Inc(R.A.Y, 3);
  Inc(R.B.Y, 3);
  ChDirButton := TViButton.Create(R, '~C~hdir', cmChangeDirInternal,
    bfNormal, True);
  Insert(ChDirButton);
  Inc(R.A.Y, 3);
  Inc(R.B.Y, 3);
  Insert(TViButton.Create(R, 'C~u~rrent', cmRevert, bfNormal, True));
  if AOptions and cdHelpButton <> 0 then
  begin
    Inc(R.A.Y, 3);
    Inc(R.B.Y, 3);
    Insert(TViButton.Create(R, 'Help', cmHelp, bfNormal, True));
  end;

  if AOptions and cdNoLoadDir = 0 then
    SetUpDialog;

  SelectNext(False);
end;

procedure TViChDirDialog.HandleEvent(var Event: TEvent);
var
  CurDir: DirStr;
  DirEntry: TDirEntry;
begin
  inherited HandleEvent(Event);
  case Event.What of
    evCommand:
      begin
        case Event.Command of
          cmRevert: GetDir(0, CurDir);
          cmChangeDirInternal:
            begin
              DirEntry := TDirEntry(DirList.List.Objects[DirList.Focused]);
              if (DirEntry.Directory = Drives)
                  or DriveValid(DirEntry.Directory[1]) then
                CurDir := DirEntry.Directory
              else Exit;
            end;
        else
          Exit;
        end;
        if (Length(CurDir) > 3) and (CurDir[Length(CurDir)] = '\') then
          CurDir := Copy(CurDir, 1, Length(CurDir) - 1);
        DirList.NewDirectory(CurDir);
        DirInput.Data := CurDir;
        DirInput.DrawView;
        DirList.Select;
        ClearEvent(Event);
      end;
  end;
end;

procedure TViChDirDialog.SetUpDialog;
var
  CurDir: DirStr;
begin
  if DirList <> nil then
  begin
    CurDir := GetDirName;
    if CurDir = '' then
      CurDir := GetCurrentDir;
    DirList.NewDirectory(CurDir);
    if (Length(CurDir) > 3) and (CurDir[Length(CurDir)] = '\') then
      CurDir := Copy(CurDir, 1, Length(CurDir) - 1);
    if DirInput <> nil then
    begin
      DirInput.Data := CurDir;
      DirInput.DrawView;
    end;
  end;
end;

function TViChDirDialog.Valid(Command: Word): Boolean;
var
  P: PathStr;
begin
  Valid := True;
  if Command = cmOk then
  begin
    P := ExpandFileName(DirInput.Data);
//    if (Length(P) > 3) and (P[Length(P)] = '\') then Dec(P[0]);
//    ChDir(P);
    if not DirectoryExists(ExcludeTrailingPathDelimiter(P)) then
    begin
      MessageBox('Invalid directory.', mfError + mfOkButton);
      Valid := False;
    end;
  end;
end;

function TViChDirDialog.GetDirName: string;
begin
  Result := ExpandFileName(DirInput.Data);
end;

procedure TViChDirDialog.SetDirName(const Value: string);
begin
  DirInput.Data := Value;
  SetUpDialog;
end;

end.

