
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (clx/vi)      }
{ ViThreads.pas                                         }
{                                                       }
{ for Delphi 4 - 7                                      }
{ Version 2.2                                           }
{ Copyright (c) 2003 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/ViThreads.pas,v 1.3 2004/03/30 20:24:11 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: ViThreads.pas,v $
  Revision 1.3  2004/03/30 20:24:11  iwache
  Memory leak in finalization section fixed.

  Revision 1.2  2003/02/05 20:13:19  iwache
  Now uses Windows for compatibility with Delphi 4 and 5.

  Revision 1.1  2003/02/04 17:25:20  iwache
  Thread synchronization mechanism for TurboVision layer added.

  ----------------------------------------------------------------------------
}
unit ViThreads;

{$IFDEF VER120}
{$DEFINE D45}
{$ENDIF}
{$IFDEF VER130}
{$DEFINE D45}
{$ENDIF}

interface

uses
  Windows, SysUtils, ViMessages, Threads;

implementation

{ Thread management routines }

const
  CM_EXECPROC = $8FFF;
  CM_DESTROYOBJECT = $8FFE;

type
{$IFDEF D45}
  PRaiseFrame = ^TRaiseFrame;
  TRaiseFrame = record
    NextRaise: PRaiseFrame;
    ExceptAddr: Pointer;
    ExceptObject: TObject;
    ExceptionRecord: PExceptionRecord;
  end;
{$ENDIF}

  TThreadObject = class(TObject)
  private
    FThreadCount: Integer;
    FHandle: THandle;
    procedure CMExecProc(var Msg: TMessage); message CM_EXECPROC;
    procedure CMDestroyObject(var Msg: TMessage); message CM_DESTROYOBJECT;
  public
    constructor Create;
    destructor Destroy; override;
    property ThreadCount: Integer read FThreadCount write FThreadCount;
    property Handle: THandle read FHandle;
  end;

  TProtectedThread = class(TThread)
  end;

var
//  ThreadLock: TRTLCriticalSection;
  ThreadObject: TThreadObject;

procedure AddThread;
begin
//  EnterCriticalSection(ThreadLock);
  try
    if not Assigned(ThreadObject) then
      ThreadObject := TThreadObject.Create;
    ThreadObject.ThreadCount := ThreadObject.ThreadCount + 1;
  finally
//    LeaveCriticalSection(ThreadLock);
  end;
end;

procedure RemoveThread;
begin
//  EnterCriticalSection(ThreadLock);
  try
    if Assigned(ThreadObject) then
      PostMessage(ThreadObject.Handle, CM_DESTROYOBJECT, 0, 0);
  finally
//    LeaveCriticalSection(ThreadLock);
  end;
end;

procedure ExecuteProcedure(Thread: TThread);
begin
  PostMessage(ThreadObject.Handle, CM_EXECPROC, 0, Longint(Thread));
end;

{************************************* TThreadObject **************************}

constructor TThreadObject.Create;
begin
  inherited Create;
  FHandle := RegisterHandleObject(Self);
end;

destructor TThreadObject.Destroy;
begin
  UnregisterHandleObject(FHandle);
  inherited Destroy;
end;

procedure TThreadObject.CMExecProc(var Msg: TMessage);
begin
  with TProtectedThread(Msg.LParam) do
  begin
    try
      FSynchronizeException := nil;
      FMethod;
    except
{$IFDEF D45}
      if RaiseList <> nil then
      begin
        FSynchronizeException := PRaiseFrame(RaiseList)^.ExceptObject;
        PRaiseFrame(RaiseList)^.ExceptObject := nil;
      end;
{$ELSE}
      FSynchronizeException := AcquireExceptionObject;
{$ENDIF}
    end;
    FSynchronizeDone := True;
  end;
end;

procedure TThreadObject.CMDestroyObject(var Msg: TMessage);
begin
//  EnterCriticalSection(ThreadLock);
  try
    Dec(FThreadCount);
    if ThreadCount = 0 then
      FreeAndNil(ThreadObject);
  finally
//    LeaveCriticalSection(ThreadLock);
  end;
end;

initialization
  Threads.AddThread := AddThread;
  Threads.RemoveThread := RemoveThread;
  Threads.ExecuteProcedure := ExecuteProcedure;

finalization
  Threads.AddThread := nil;
  Threads.RemoveThread := nil;
  Threads.ExecuteProcedure := nil;
  FreeAndNil(ThreadObject);
  
end.
