
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (clx/vi)      }
{ ViViews.pas                                           }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Turbo Pascal Version 7.0                          }
{     Turbo Vision Unit                                 }
{     Copyright (c) 1992 Borland International          }
{     www: http://www.borland.com                       }
{   - Virtual Pascal v2.1                               }
{     Copyright (C) 1996-2000 vpascal.com               }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/ViViews.pas,v 1.13 2003/12/01 20:58:00 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: ViViews.pas,v $
  Revision 1.13  2003/12/01 20:58:00  iwache
  Dialog result consts cmAbort, cmRetry and cmIgnore added.

  Revision 1.12  2003/11/19 18:48:18  iwache
  For Vi only mode TViScrollBar.Create has a
  new default parameter ViMode: Boolean = False.
  Set it to true if you like to use the class
  unconnected to WQ e.g. inside Vi dialogs.

  Revision 1.11  2003/03/03 15:49:16  iwache
  TViGroup:
  - function ExecMenuView added,
  - function SetCurrent to default TV behaviour
    modified (for TMenu now is ExecMenuView available)

  Revision 1.10  2003/01/23 19:20:59  iwache
  - TViView option ofViInternal = $0800 added. This is used
  for internal Vi objects, which are not connected to CLX
  components. This is needed to adapt the behaviour of
  ShowMessage or InputQuery identical to VisualCLX.
  - TViView.DragView  now handles also keys Ctrl-Up and
  Ctrl-Down to move a window by keyboard.

  Revision 1.9  2003/01/21 18:30:13  iwache
  DWPL-Bugs item #670579:
  Description:
  Edit Control on Panel can not get focus, and can
  not enter anything in it.
  Solution:
  Procedure TViGroup.SetCurrent modified.
  Local procedure SelectView modified;
  select view only when group is also selected.

  Revision 1.8  2003/01/20 22:11:11  iwache
  TViGroup.SetCurrent modified to avoid windows focus lost
  after using the main menu.

  Revision 1.7  2002/10/27 22:25:23  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
{
  Version history:
    IWA 2001-05-24 Version 2.0.1.0 released
}

unit ViViews;

{$X+}

interface

uses
  ViUtils, ViClasses, ViTypes, ViEventsClass;

const

{ TViView State masks }

  sfVisible = $0001;
  sfCursorVis = $0002;
  sfCursorIns = $0004;
  sfShadow = $0008;
  sfActive = $0010;
  sfSelected = $0020;
  sfFocused = $0040;
  sfDragging = $0080;
  sfDisabled = $0100;
  sfModal = $0200;
  sfDefault = $0400;
  sfExposed = $0800;

{ TViView Option masks }

  ofSelectable = $0001;
  ofTopSelect = $0002;
  ofFirstClick = $0004;
  ofFramed = $0008;
  ofPreProcess = $0010;
  ofPostProcess = $0020;
  ofBuffered = $0040;
  ofTileable = $0080;
  ofCenterX = $0100;
  ofCenterY = $0200;
  ofCentered = $0300;
  ofValidate = $0400;
  ofViInternal = $0800;  // internal Vi object, not connected to CLX
  ofVersion = $3000;
  ofVersion10 = $0000;
  ofVersion20 = $1000;

{ TViView GrowMode masks }

  gfGrowLoX = $01;
  gfGrowLoY = $02;
  gfGrowHiX = $04;
  gfGrowHiY = $08;
  gfGrowAll = $0F;
  gfGrowRel = $10;

{ TViView DragMode masks }

  dmDragMove = $01;
  dmDragGrow = $02;
  dmLimitLoX = $10;
  dmLimitLoY = $20;
  dmLimitHiX = $40;
  dmLimitHiY = $80;
  dmLimitAll = $F0;

{ TViView Help context codes }

  hcNoContext = 0;
  hcDragging = 1;

{ TScrollBar part codes }

  sbLeftArrow = 0;
  sbRightArrow = 1;
  sbPageLeft = 2;
  sbPageRight = 3;
  sbUpArrow = 4;
  sbDownArrow = 5;
  sbPageUp = 6;
  sbPageDown = 7;
  sbIndicator = 8;

{ TScrollBar options for TWindow.StandardScrollBar }

  sbHorizontal = $0000;
  sbVertical = $0001;
  sbHandleKeyboard = $0002;

{ TWindow Flags masks }

  wfMove = $01;
  wfGrow = $02;
  wfClose = $04;
  wfZoom = $08;

{ TWindow number constants }

  wnNoNumber = 0;

{ TWindow palette entries }

  wpBlueWindow = 0;
  wpCyanWindow = 1;
  wpGrayWindow = 2;

{ WCLX events }

  cmxBoundsChanged = 255;

{ Standard command codes }

  cmValid = 0;
  cmQuit = 1;
  cmError = 2;
  cmMenu = 3;
  cmClose = 4;
  cmZoom = 5;
  cmResize = 6;
  cmNext = 7;
  cmPrev = 8;
  cmHelp = 9;

  cmZoomed = 15; // signals change of zoom state after dragging

{ Application command codes }

  cmCut = 20;
  cmCopy = 21;
  cmPaste = 22;
  cmUndo = 23;
  cmClear = 24;
  cmTile = 25;
  cmCascade = 26;

{ TDialog standard commands }

  cmOK = 10;
  cmCancel = 11;
  cmYes = 12;
  cmNo = 13;
  cmDefault = 14;

  cmAbort = 16;
  cmRetry = 17;
  cmIgnore = 18;

{ Standard messages }

  cmReceivedFocus = 50;
  cmReleasedFocus = 51;
  cmCommandSetChanged = 52;

  cmFocusIn = 57;

{ TScrollBar messages }

  cmScrollBarChanged = 53;
  cmScrollBarClicked = 54;

{ TWindow select messages }

  cmSelectWindowNum = 55;

{ TListViewer messages }

  cmListItemSelected = 56;

{ Color palettes }

  CFrame = #1#1#2#2#3;
  CScrollBar = #4#5#5;
  CScroller = #6#7;
  CListViewer = #26#26#27#28#29;

  CBlueWindow = #8#9#10#11#12#13#14#15;
  CCyanWindow = #16#17#18#19#20#21#22#23;
  CGrayWindow = #24#25#26#27#28#29#30#31;

{ TDrawBuffer maximum view width }

  MaxViewWidth = 255;

type

{ Command sets }

  PCommandSet = ^TCommandSet;
  TCommandSet = set of Byte;

{ Color palette type }

  TPalette = string;

{ TDrawBuffer, buffer used by draw methods }

  TDrawBuffer = array[0..MaxViewWidth - 1] of Word;

{ forward references }

  TViGroup = class;

{ TViView }

  TEventFilterEvent = function(Sender: TObject; var Event: TEvent): Boolean
      of object;
  TChangedEvent = procedure(Sender: TObject) of object;
  TScrollEvent = procedure(Sender: TObject; PartCode: Integer) of object;
  TActivatedEvent = procedure(Sender: TObject; ID: Integer) of object;

  TViView = class(TObject)
  private
    FPalette: TPalette;
    FOnEvent: TEventFilterEvent;
    FBoundsChangedLock: Integer;
    procedure DrawCursor;
    procedure DrawHide(LastView: TViView);
    procedure DrawShow(LastView: TViView);
    procedure DrawUnderRect(var R: TRect; LastView: TViView);
    procedure DrawUnderView(DoShadow: Boolean; LastView: TViView);
    function MapColor(Color: Byte): Byte;
    function MapCPair(CPair: Word): Word;
    procedure WriteView(X, Y, W: Integer; var Buf);
    procedure BoundsChanged;
  protected
    function GetPalette: TPalette; virtual;
    procedure SetPalette(Value: TPalette); virtual;
    procedure ResetCursor; virtual;
    procedure LockBoundsChanged;
    procedure UnLockBoundsChanged;
  public
    Owner: TViGroup;
    Next: TViView;
    Origin: TPoint;
    Size: TPoint;
    Cursor: TPoint;
    GrowMode: Byte;
    DragMode: Byte;
    HelpCtx: Word;
    State: Word;
    Options: Word;
    EventMask: LongWord;
    constructor Create(var Bounds: TRect);
    destructor Destroy; override;
    procedure BlockCursor;
    procedure CalcBounds(var Bounds: TRect; Delta: TPoint); virtual;
    procedure ChangeBounds(var Bounds: TRect); virtual;
    procedure ClearEvent(var Event: TEvent);
    function CommandEnabled(Command: Word): Boolean;
    procedure DisableCommands(Commands: TCommandSet);
    procedure DragView(Event: TEvent; Mode: Byte;
      var Limits: TRect; MinSize, MaxSize: TPoint);
    procedure Draw; virtual;
    procedure DrawView;
    procedure EnableCommands(Commands: TCommandSet);
    procedure EndModal(Command: Word); virtual;
    function EventAvail: Boolean;
    function Execute: Word; virtual;
    function Exposed: Boolean;
    function Focus: Boolean; virtual;
    procedure GetBounds(var Bounds: TRect);
    procedure GetClipRect(var Clip: TRect);
    function GetColor(Color: Word): Word;
    procedure GetCommands(var Commands: TCommandSet);
    procedure GetEvent(var Event: TEvent); virtual;
    procedure GetExtent(var Extent: TRect);
    function GetHelpCtx: Word; virtual;
    function GetState(AState: Word): Boolean;
    procedure GrowTo(X, Y: Integer);
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure Hide;
    procedure HideCursor;
    procedure KeyEvent(var Event: TEvent);
    procedure Locate(var Bounds: TRect);
    procedure MakeFirst;
    procedure MakeGlobal(Source: TPoint; var Dest: TPoint);
    procedure MakeLocal(Source: TPoint; var Dest: TPoint);
    function MouseEvent(var Event: TEvent; Mask: Word): Boolean;
    function MouseInView(Mouse: TPoint): Boolean;
    procedure MoveTo(X, Y: Integer);
    function NextView: TViView;
    procedure NormalCursor;
    function Prev: TViView;
    function PrevView: TViView;
    procedure PutEvent(var Event: TEvent); virtual;
    procedure PutInFrontOf(Target: TViView);
    procedure Select;
    procedure SetBounds(var Bounds: TRect);
    procedure SetCommands(Commands: TCommandSet);
    procedure SetCmdState(Commands: TCommandSet; Enable: Boolean);
    procedure SetCursor(X, Y: Integer);
    procedure SetState(AState: Word; Enable: Boolean); virtual;
    procedure Show;
    procedure ShowCursor;
    procedure SizeLimits(var Min, Max: TPoint); virtual;
    function TopView: TViView;
    function Valid(Command: Word): Boolean; virtual;
    procedure WriteBuf(X, Y, W, H: Integer; var Buf);
    procedure WriteChar(X, Y: Integer; C: Char; Color: Byte;
      Count: Integer);
    procedure WriteLine(X, Y, W, H: Integer; var Buf);
    procedure WriteStr(X, Y: Integer; Str: string; Color: Byte);
    property Palette: TPalette read GetPalette write SetPalette;
    property OnEvent: TEventFilterEvent read FOnEvent write FOnEvent;
  end;

{ TFrame types }

  TTitleStr = string (*[80]*);

{ TViFrame }

  { Palette layout }
  { 1 = Passive frame }
  { 2 = Passive title }
  { 3 = Active frame }
  { 4 = Active title }
  { 5 = Icons }

  TViCustomFrame = class(TViView)
  private
    procedure FrameLine(var FrameBuf; Y, N: Integer; Color: Byte);
  public
    constructor Create(var Bounds: TRect);
    procedure SetState(AState: Word; Enable: Boolean); override;
    function Focus: Boolean; override;
  end;

  TViFrame = class(TViCustomFrame)
  private
    FrameMode: Word;
  public
    procedure Draw; override;
    procedure HandleEvent(var Event: TEvent); override;
  end;

{ ViScrollBar characters }

  TScrollChars = array[0..4] of Char;

{ TViScrollBar }

  { Palette layout }
  { 1 = Page areas }
  { 2 = Arrows }
  { 3 = Indicator }

  TViClientFrame = class(TViCustomFrame)
  private
    procedure FrameLine(var FrameBuf; Y, N: Integer; Color: Byte);
  public
    procedure Draw; override;
  end;

  TViScrollBar = class(TViView)
  private
    Chars: TScrollChars;
    FOnChange: TChangedEvent;
    FOnScroll: TScrollEvent;
    procedure DrawPos(Pos: Integer);
    function GetPos: Integer;
    function GetSize: Integer;
  protected
    procedure DoChange;
    procedure DoScroll(PartCode: Integer);
  public
    Value: Integer;
    Min: Integer;
    Max: Integer;
    PgStep: Integer;
    ArStep: Integer;
    constructor Create(var Bounds: TRect; ViMode: Boolean = False);
    procedure Draw; override;
    procedure HandleEvent(var Event: TEvent); override;
    procedure ScrollDraw; virtual;
    function ScrollStep(Part: Integer): Integer; virtual;
    procedure SetParams(AValue, AMin, AMax, APgStep, AArStep: Integer);
    procedure SetRange(AMin, AMax: Integer);
    procedure SetStep(APgStep, AArStep: Integer);
    procedure SetValue(AValue: Integer);
    procedure BoundsChanged;
    procedure SetOrientation(Value: TOrientation);
    property OnChange: TChangedEvent read FOnChange write FOnChange;
    property OnScroll: TScrollEvent read FOnScroll write FOnScroll;
  end;

{ TViScroller }

  { Palette layout }
  { 1 = Normal text }
  { 2 = Selected text }

  TViScroller = class(TViView)
  private
    DrawLock: Byte;
    DrawFlag: Boolean;
    procedure CheckDraw;
  public
    HScrollBar: TViScrollBar;
    VScrollBar: TViScrollBar;
    Delta: TPoint;
    Limit: TPoint;
    constructor Create(var Bounds: TRect; AHScrollBar,
      AVScrollBar: TViScrollBar);
    procedure ChangeBounds(var Bounds: TRect); override;
    procedure HandleEvent(var Event: TEvent); override;
    procedure ScrollDraw; virtual;
    procedure ScrollTo(X, Y: Integer);
    procedure SetLimit(X, Y: Integer);
    procedure SetState(AState: Word; Enable: Boolean); override;
  end;

{ TViListViewer }

  { Palette layout }
  { 1 = Active }
  { 2 = Inactive }
  { 3 = Focused }
  { 4 = Selected }
  { 5 = Divider }

  TViListViewer = class(TViView)
  private
    FOnChanged: TChangedEvent;
    FCapture: Boolean;
  protected
    procedure DoChanged;
  public
    HScrollBar: TViScrollBar;
    VScrollBar: TViScrollBar;
    NumCols: Integer;
    TopItem: Integer;
    Focused: Integer;
    Range: Integer;
    constructor Create(var Bounds: TRect; ANumCols: Word;
      AHScrollBar, AVScrollBar: TViScrollBar);
    procedure ChangeBounds(var Bounds: TRect); override;
    procedure Draw; override;
    procedure FocusItem(Item: Integer); virtual;
    function GetText(Item: Integer; MaxLen: Integer): string; virtual;
      abstract;
    function IsSelected(Item: Integer): Boolean; virtual;
    procedure HandleEvent(var Event: TEvent); override;
    procedure SelectItem(Item: Integer); virtual;
    procedure SetRange(ARange: Integer);
    procedure SetState(AState: Word; Enable: Boolean); override;
    procedure FocusItemNum(Item: Integer); virtual;
    property OnChanged: TChangedEvent read FOnChanged write FOnChanged;
  end;

{ Selection modes }

  SelectMode = (NormalSelect, EnterSelect, LeaveSelect);

{ TViGroup }

  TEachProcedure = procedure(View: TViView); stdcall;

  TViGroup = class(TViView)
  private
    Clip: TRect;
    LockFlag: Byte;
    procedure DrawSubViews(P, Bottom: TViView);
    function FirstMatch(AState: Word; AOptions: Word): TViView;
    function FindNext(Forwards: Boolean): TViView;
    procedure FreeBuffer;
    procedure GetBuffer;
    procedure InsertView(P, Target: TViView);
    procedure RemoveView(P: TViView);
  protected
    Buffer: PVideoBuf;
    procedure ResetCursor; override;
    function IndexOf(P: TViView): Integer;
    function At(Index: Integer): TViView;
  public
    Last: TViView;
    Current: TViView;
    Phase: (phFocused, phPreProcess, phPostProcess);
    EndState: Word;
    constructor Create(var Bounds: TRect);
    destructor Destroy; override;
    procedure SetCurrent(P: TViView; Mode: SelectMode);
    procedure ResetCurrent;
    procedure ChangeBounds(var Bounds: TRect); override;
    procedure Delete(P: TViView);
    procedure Draw; override;
    procedure EndModal(Command: Word); override;
    procedure EventError(var Event: TEvent); virtual;
    function ExecView(P: TViView): Word;
    function Execute: Word; override;
    function First: TViView;
    function FocusNext(Forwards: Boolean): Boolean;
    function GetHelpCtx: Word; override;
    procedure HandleEvent(var Event: TEvent); override;
    procedure Insert(P: TViView);
    procedure InsertBefore(P, Target: TViView);
    procedure Lock;
    procedure Redraw;
    procedure SelectNext(Forwards: Boolean);
    procedure SetState(AState: Word; Enable: Boolean); override;
    procedure Unlock;
    function Valid(Command: Word): Boolean; override;
    function ExecMenuView(P: TViView): Word;
  end;

{ TViWindow }

  { Palette layout }
  { 1 = Frame passive }
  { 2 = Frame active }
  { 3 = Frame icon }
  { 4 = ScrollBar page area }
  { 5 = ScrollBar controls }
  { 6 = Scroller normal text }
  { 7 = Scroller selected text }
  { 8 = Reserved }

  TViWindow = class(TViGroup)
  protected
    FPalettes: array[wpBlueWindow..wpGrayWindow] of string;
    function GetPalette: TPalette; override;
    procedure SetPalette(Value: TPalette); override;
  public
    Constraints: TRect;
    Flags: Byte;
    ZoomRect: TRect;
    Number: Integer;
    WindowPalette: Integer;
    Frame: TViFrame;
    Title: string;
    constructor Create(var Bounds: TRect; ATitle: TTitleStr; ANumber: Integer);
    destructor Destroy; override;
    procedure Close; virtual;
    function GetTitle: TTitleStr; virtual;
    procedure HandleEvent(var Event: TEvent); override;
    procedure InitFrame; virtual;
    procedure SetState(AState: Word; Enable: Boolean); override;
    procedure SizeLimits(var Min, Max: TPoint); override;
    function StandardScrollBar(AOptions: Word): TViScrollBar;
    procedure Zoom; virtual;
    procedure ShowMaximized; virtual;
    procedure ShowNormal; virtual;
  end;

  TViClientGroup = class(TViGroup)
  public
    procedure HandleEvent(var Event: TEvent); override;
  end;

{ Message dispatch function }

function SendMessage(Receiver: TViView; What, Command: Word; InfoPtr:
  Pointer): Pointer;

var

{ Event masks }

  PositionalEvents: Word = evMouse;
  FocusedEvents: Word = evKeyboard + evCommand;

{ Minimum window size }

  MinWinSize: TPoint = (X: 16; Y: 6);

{ Shadow definitions }

  ShadowSize: TPoint = (X: 2; Y: 1);
  ShadowAttr: Byte = $08;

{ Markers control }

  ShowMarkers: Boolean = False;

{ MapColor error return value }

  ErrorAttr: Byte = $CF;

{ True if the command set has changed since being set to false }

  CommandSetChanged: Boolean = False;

{ True if the cursor is currently hidden }

  CursorHidden: Boolean = True;

const

{ Characters used for drawing selected and default items in  }
{ monochrome color sets                                      }

  SpecialChars: array[0..5] of Char = (#175, #174, #26, #27, ' ', ' ');

{ Line drawing characters }

{$IFDEF LINUX}
  // In Linux, do not use high-ascii drawing characters

  ldFrameChars: array[0..31] of Char = #32#32#32#18#32#20#15#24
  + #32#17#19#21#16#23#22#25
    + #32#32#32#29#32#31#26#31
    + #32#28#30#30#27#31#30#32;
                      { '   A 3UA UA'A   E  III '; }

  ldMenuFrameChars: array[0..19] of Char = #32#15#19#16#32#32#18#19#17#32
  + #32#20#32#20#32#32#24#19#23#32;
                      { '  U  AU  3 3  A' '; }

  ldCloseWindow: array[0..4] of char = '[~'#07'~]';
  ldCloseClicked: array[0..4] of char = '[~*~]';
  ldMaximize: array[0..4] of char = '[~'#05'~]';
  ldBottomRight: array[0..3] of char = '~'#19#17'~';

  ldVerticalScroll: TScrollChars = #01#02#13#07#14;
  ldHorizontalScroll: TScrollChars = #03#04#13#07#14;

  ldHistoryDropDown: array[0..4] of char = #10'~'#02'~'#11;

  ldPathDir: array[0..2] of char = #18#19#22; // 'AA'
  ldFirstDir: array[0..2] of char = #18#22#19; //   'AA'
  ldMiddleDir: array[0..2] of char = #32#24#19; //   ' A'
  ldLastDir: array[0..2] of char = #32#18#19; //   ' A'
  ldIndentSize: string[10] = '  ';

  ldDesktopBackground: char = #14;
  ldSubMenuArrow: char = #04;

  ldVerticalBar: char = #20;
  ldHorizontalBar: char = #19;
  ldDblHorizontalBar: char = #30;
  ldBlockFull: char = #12;
  ldBlockBottom: char = #08;
  ldBlockTop: char = #09;
  ldRadioSelect: char = #07;
  ldModified: char = '*';
  ldLeftScroll: char = #03;
  ldRightScroll: char = #04;
{$ELSE}

  ldFrameChars: array[0..31] of Char = '         ɺ ͻ ';
{ Codepage = 437 :                     '          ϻ '; }
var
  ldMenuFrameChars: array[0..19] of Char = ' Ŀ       Ĵ ';
const
  ldCloseWindow: array[0..4] of char = '[~~]';
  ldCloseClicked: array[0..4] of char = '[~'#15'~]';
  ldMaximize: array[0..4] of char = '[~'#24'~]';
  ldBottomRight: array[0..3] of char = '~~';

  ldVerticalScroll: TScrollChars = #30#31#177#254#178;
  ldHorizontalScroll: TScrollChars = #17#16#177#254#178;

  ldHistoryDropDown: array[0..4] of char = #222'~'#25'~'#221;
//  ldHistoryDropDown: array[0..4] of char = '~<'#25'>~';

  ldPathDir: array[0..2] of char = '';
  ldFirstDir: array[0..2] of char = '';
  ldMiddleDir: array[0..2] of char = ' ';
  ldLastDir: array[0..2] of char = ' ';
  ldIndentSize: string[10] = '  ';

  ldDesktopBackground: char = #176; // 
  ldSubMenuArrow: char = #16; //

  ldVerticalBar: char = #179; // 
  ldHorizontalBar: char = #196; // 
  ldDblHorizontalBar: char = #205; // 
  ldBlockFull: char = #219; // 
  ldBlockBottom: char = #220; // 
  ldBlockTop: char = #223; // 
  ldRadioSelect: char = #7; //
  ldModified: char = #15;
  ldLeftScroll: char = #16;
  ldRightScroll: char = #17;
{$ENDIF}

implementation

uses
  ViDrivers, ViScreens, ViMouses, ViDialogs;

var
  OwnerGroup: TViGroup = nil;
  TheTopView: TViView = nil;

const

{ Bit flags to determine how to draw the frame icons }

  fmCloseClicked = $0001;
  fmZoomClicked = $0002;

var

{ Current command set. All but window commands are active by default }

  CurCommandSet: TCommandSet =
  [0..255] - [cmZoom, cmClose, cmResize, cmNext, cmPrev];

{ TViView }

constructor TViView.Create(var Bounds: TRect);
begin
  inherited Create;
  Owner := nil;
  State := sfVisible;
  SetBounds(Bounds);
  DragMode := dmLimitLoY;
  HelpCtx := hcNoContext;
  EventMask := evMouseDown + evKeyDown + evCommand;
end;

destructor TViView.Destroy;
begin
  Hide;
  if Owner <> nil then
    Owner.Delete(Self)
  else
    inherited Destroy;
end;

procedure TViView.BlockCursor;
begin
  SetState(sfCursorIns, True);
end;

procedure TViView.CalcBounds(var Bounds: TRect; Delta: TPoint);
var
  S, D: Integer;
(*
  Min, Max: TPoint;
*)

  procedure Grow(var I: Integer);
  begin
    if GrowMode and gfGrowRel = 0 then
      Inc(I, D)
    else
      I := (I * S + (S - D) shr 1) div (S - D);
  end;

  function Range(Val, Min, Max: Integer): Integer;
  begin
    if Val < Min then
      Range := Min
    else if Val > Max then
      Range := Max
    else
      Range := Val;
  end;

begin
  GetBounds(Bounds);
  S := Owner.Size.X;
  D := Delta.X;
  if GrowMode and gfGrowLoX <> 0 then Grow(Bounds.A.X);
  if GrowMode and gfGrowHiX <> 0 then Grow(Bounds.B.X);
  if Bounds.B.X - Bounds.A.X > MaxViewWidth then
    Bounds.B.X := Bounds.A.X + MaxViewWidth;
  S := Owner.Size.Y;
  D := Delta.Y;
  if GrowMode and gfGrowLoY <> 0 then Grow(Bounds.A.Y);
  if GrowMode and gfGrowHiY <> 0 then Grow(Bounds.B.Y);
(*
  SizeLimits(Min, Max);
  Bounds.B.X := Bounds.A.X + Range(Bounds.B.X - Bounds.A.X, Min.X, Max.X);
  Bounds.B.Y := Bounds.A.Y + Range(Bounds.B.Y - Bounds.A.Y, Min.Y, Max.Y);
*)  
end;

procedure TViView.ChangeBounds(var Bounds: TRect);
{
var
  E: TEvent;
}
begin
  SetBounds(Bounds);
  DrawView;
{
  E.What := evCommand;
  E.Command := cmxBoundsChanged;
  HandleEvent(E);
}  
end;

procedure TViView.ClearEvent(var Event: TEvent);
begin
  Event.What := evNothing;
  Event.InfoPtr := Self;
end;

function TViView.CommandEnabled(Command: Word): Boolean;
begin
  CommandEnabled := (Command > 255) or (Command in CurCommandSet);
end;

procedure TViView.DisableCommands(Commands: TCommandSet);
begin
  CommandSetChanged := CommandSetChanged or (CurCommandSet * Commands <> []);
  CurCommandSet := CurCommandSet - Commands;
end;

procedure TViView.DragView(Event: TEvent; Mode: Byte; var Limits: TRect;
  MinSize, MaxSize: TPoint);
var
  P, S: TPoint;
  SaveBounds: TRect;

  function Min(I, J: Integer): Integer;
  begin
    if I < J then
      Min := I
    else
      Min := J;
  end;

  function Max(I, J: Integer): Integer;
  begin
    if I > J then
      Max := I
    else
      Max := J;
  end;

  procedure MoveGrow(P, S: TPoint);
  var
    R: TRect;
  begin
    S.X := Min(Max(S.X, MinSize.X), MaxSize.X);
    S.Y := Min(Max(S.Y, MinSize.Y), MaxSize.Y);
    P.X := Min(Max(P.X, Limits.A.X - S.X + 1), Limits.B.X - 1);
    P.Y := Min(Max(P.Y, Limits.A.Y - S.Y + 1), Limits.B.Y - 1);
    if Mode and dmLimitLoX <> 0 then P.X := Max(P.X, Limits.A.X);
    if Mode and dmLimitLoY <> 0 then P.Y := Max(P.Y, Limits.A.Y);
    if Mode and dmLimitHiX <> 0 then P.X := Min(P.X, Limits.B.X - S.X);
    if Mode and dmLimitHiY <> 0 then P.Y := Min(P.Y, Limits.B.Y - S.Y);
    RectAssign(R, P.X, P.Y, P.X + S.X, P.Y + S.Y);
    Locate(R);
  end;

  procedure Change(DX, DY: Integer);
  begin
    if (Mode and dmDragMove <> 0) and (GetShiftState and $03 = 0) then
    begin
      Inc(P.X, DX);
      Inc(P.Y, DY);
    end
    else if (Mode and dmDragGrow <> 0) and (GetShiftState and $03 <> 0) then
    begin
      Inc(S.X, DX);
      Inc(S.Y, DY);
    end;
  end;

  procedure Update(X, Y: Integer);
  begin
    if Mode and dmDragMove <> 0 then
    begin
      P.X := X;
      P.Y := Y;
    end;
  end;

begin
  SetState(sfDragging, True);
  if Event.What = evMouseDown then
  begin
    if Mode and dmDragMove <> 0 then
    begin
      P.X := Origin.X - Event.Where.X;
      P.Y := Origin.Y - Event.Where.Y;
      repeat
        Inc(Event.Where.X, P.X);
        Inc(Event.Where.Y, P.Y);
        MoveGrow(Event.Where, Size);
      until not MouseEvent(Event, evMouseMove);
    end
    else
    begin
      P.X := Size.X - Event.Where.X;
      P.Y := Size.Y - Event.Where.Y;
      repeat
        Inc(Event.Where.X, P.X);
        Inc(Event.Where.Y, P.Y);
        MoveGrow(Origin, Event.Where);
      until not MouseEvent(Event, evMouseMove);
    end;
  end
  else
  begin
    GetBounds(SaveBounds);
    repeat
      P := Origin;
      S := Size;
      KeyEvent(Event);
      case Event.KeyCode and $FF00 of
        kbLeft: Change(-1, 0);
        kbRight: Change(1, 0);
        kbUp: Change(0, -1);
        kbDown: Change(0, 1);
        kbCtrlLeft: Change(-8, 0);
        kbCtrlRight: Change(8, 0);
        kbCtrlUp: Change(0, -4);
        kbCtrlDown: Change(0, 4);
        kbHome: Update(Limits.A.X, P.Y);
        kbEnd: Update(Limits.B.X - S.X, P.Y);
        kbPgUp: Update(P.X, Limits.A.Y);
        kbPgDn: Update(P.X, Limits.B.Y - S.Y);
      end;
      MoveGrow(P, S);
    until (Event.KeyCode = kbEnter) or (Event.KeyCode = kbEsc);
    if Event.KeyCode = kbEsc then Locate(SaveBounds);
  end;
  SetState(sfDragging, False);
end;

procedure TViView.Draw;
var
  B: TDrawBuffer;
begin
  MoveChar(B, ' ', GetColor(1), Size.X);
  WriteLine(0, 0, Size.X, Size.Y, B);
end;

procedure TViView.DrawCursor;
begin
  if State and sfFocused <> 0 then ResetCursor;
end;

procedure TViView.DrawHide(LastView: TViView);
begin
  DrawCursor;
  DrawUnderView(State and sfShadow <> 0, LastView);
end;

procedure TViView.DrawShow(LastView: TViView);
begin
  DrawView;
  if State and sfShadow <> 0 then DrawUnderView(True, LastView);
end;

procedure TViView.DrawUnderRect(var R: TRect; LastView: TViView);
begin
  RectIntersect(Owner.Clip, R);
  Owner.DrawSubViews(NextView, LastView);
  Owner.GetExtent(Owner.Clip);
end;

procedure TViView.DrawUnderView(DoShadow: Boolean; LastView: TViView);
var
  R: TRect;
begin
  GetBounds(R);
// AM: Always increase to include shadow rect, to make sure it is
// removed if this option is changed.
// if DoShadow then
  begin
    Inc(R.B.X, ShadowSize.X);
    Inc(R.B.Y, ShadowSize.Y);
  end;
  DrawUnderRect(R, LastView);
end;

procedure TViView.DrawView;
begin
  if Exposed then
  begin
    Draw;
    DrawCursor;
  end;
end;

procedure TViView.EnableCommands(Commands: TCommandSet);
begin
  CommandSetChanged := CommandSetChanged or
    (CurCommandSet * Commands <> Commands);
  CurCommandSet := CurCommandSet + Commands;
end;

procedure TViView.EndModal(Command: Word);
begin
  if TopView <> nil then TopView.EndModal(Command);
end;

function TViView.EventAvail: Boolean;
var
  Event: TEvent;
begin
  GetEvent(Event);
  if Event.What <> evNothing then PutEvent(Event);
  EventAvail := Event.What <> evNothing;
end;

procedure TViView.GetBounds(var Bounds: TRect);
begin
  Bounds.A.X := Origin.X;
  Bounds.A.Y := Origin.Y;
  Bounds.B.X := Origin.X + Size.X;
  Bounds.B.Y := Origin.Y + Size.Y;
end;

function TViView.Execute: Word;
begin
  Execute := cmCancel;
end;

function TViView.Exposed: Boolean;

  function Hidden(X, Y, W: Integer; Target: TViView): Boolean;

    function MemberHidden(X, Y, W: Integer; AView: TViView): Boolean;
    var
      AOwner: TViGroup;
      Temp: Integer;
    begin
      Result := False;
      repeat
        AView := AView.Next;
        if AView = Target then
        begin
          AOwner := AView.Owner;
          if Assigned(AOwner.Buffer) then Exit;
          Result := Hidden(X, Y, W, AOwner);
          Exit;
        end;
        if State and sfVisible = 0 then Continue;
        Temp := AView.Origin.Y;
        if Y < Temp then Continue;
        Temp := Temp + AView.Size.Y;
        if Y >= Temp then Continue;
        Temp := AView.Origin.X;
        if X < Temp then
        begin
          if W <= Temp then Continue;
          Temp := Temp + AView.Size.X;
          if W > Temp then
          begin
            if MemberHidden(X, Y, W, AView) then
              Continue else Exit;
          end;
          W := AView.Origin.X;
          Continue;
        end;
        Temp := Temp + AView.Size.X;
        if X >= Temp then Continue;
        X := Temp;
        if X >= W then
        begin
          Result := True;
          Exit;
        end;
      until False;
    end;

  var
    Owner: TViGroup;
    Temp: Integer;
  begin
    Result := False;
    Y := Y + Target.Origin.Y;
    Temp := Target.Origin.X;
    X := X + Temp;
    W := W + Temp;
    Owner := Target.Owner;
    if not Assigned(Owner) then Exit;
    if Y < Owner.Clip.A.Y then Exit;
    if Y >= Owner.Clip.A.Y then Exit;
    if X < Owner.Clip.A.X then
      X := Owner.Clip.A.X;
    if W > Owner.Clip.B.X then
      W := Owner.Clip.B.X;
    if X >= W then Exit;
    Result := MemberHidden(X, Y, W, Owner.Last);
  end;

var
  Y: Integer;
begin
  Result := False;
  if State and sfExposed = 0 then Exit;
  if (Size.X <= 0) or (Size.Y <= 0) then Exit;

  for Y := 0 to Size.Y - 1 do
    if not Hidden(0, Y, Size.X, Self) then
    begin
      Result := True;
      Exit;
    end;
end;

function TViView.Focus: Boolean;
begin
  Result := True;
  if State and (sfSelected + sfModal) = 0 then
  begin
    if Owner <> nil then
    begin
      Result := Owner.Focus;
      if Result then
        if ((Owner.Current = nil) or
          (Owner.Current.Options and ofValidate = 0) or
          (Owner.Current.Valid(cmReleasedFocus))) then
          Select
        else
          Result := False;
    end;
  end;
end;

procedure TViView.GetClipRect(var Clip: TRect);
begin
  GetBounds(Clip);
  if Owner <> nil then RectIntersect(Clip, Owner.Clip);
  RectMove(Clip, -Origin.X, -Origin.Y);
end;

function TViView.GetColor(Color: Word): Word;
begin
  Result := MapCPair(Color);
end;

procedure TViView.GetCommands(var Commands: TCommandSet);
begin
  Commands := CurCommandSet;
end;

procedure TViView.GetEvent(var Event: TEvent);
begin
  if Owner <> nil then Owner.GetEvent(Event);
end;

procedure TViView.GetExtent(var Extent: TRect);
begin
  Extent.A.X := 0;
  Extent.A.Y := 0;
  Extent.B := Size;
end;

function TViView.GetHelpCtx: Word;
begin
  if State and sfDragging <> 0 then
    GetHelpCtx := hcDragging
  else
    GetHelpCtx := HelpCtx;
end;

function TViView.GetState(AState: Word): Boolean;
begin
  GetState := State and AState = AState;
end;

procedure TViView.GrowTo(X, Y: Integer);
var
  R: TRect;
begin
  RectAssign(R, Origin.X, Origin.Y, Origin.X + X, Origin.Y + Y);
  Locate(R);
end;

procedure TViView.HandleEvent(var Event: TEvent);
begin
  if Assigned(OnEvent) then
    if OnEvent(Self, Event) then ClearEvent(Event);

  if Event.What = evMouseDown then
    if (State and (sfSelected + sfDisabled) = 0) and
      (Options and ofSelectable <> 0) then
      if not Focus or (Options and ofFirstClick = 0) then
        ClearEvent(Event);
end;

procedure TViView.Hide;
begin
  if State and sfVisible <> 0 then SetState(sfVisible, False);
end;

procedure TViView.HideCursor;
begin
  SetState(sfCursorVis, False);
end;

procedure TViView.KeyEvent(var Event: TEvent);
begin
  repeat GetEvent(Event)until Event.What = evKeyDown;
end;

procedure TViView.Locate(var Bounds: TRect);
var
  R: TRect;
  Min, Max: TPoint;

  function Range(Val, Min, Max: Integer): Integer;
  begin
    if Val < Min then
      Range := Min
    else if Val > Max then
      Range := Max
    else
      Range := Val;
  end;

begin
  SizeLimits(Min, Max);
  Bounds.B.X := Bounds.A.X + Range(Bounds.B.X - Bounds.A.X, Min.X, Max.X);
  Bounds.B.Y := Bounds.A.Y + Range(Bounds.B.Y - Bounds.A.Y, Min.Y, Max.Y);
  GetBounds(R);
  if not RectEquals(Bounds, R) then
  begin
    ChangeBounds(Bounds);
    if (Owner <> nil) and (State and sfVisible <> 0) then
    begin
      if State and sfShadow <> 0 then
      begin
        RectUnion(R, Bounds);
        Inc(R.B.X, ShadowSize.X);
        Inc(R.B.Y, ShadowSize.Y);
      end;
      DrawUnderRect(R, nil);
    end;
  end;
end;

procedure TViView.MakeFirst;
begin
  PutInFrontOf(Owner.First);
end;

procedure TViView.MakeGlobal(Source: TPoint; var Dest: TPoint);
var
  View: TViView;
begin
  View := Self;
  Dest.X := 0;
  Dest.Y := 0;
  repeat
    Inc(Dest.X, View.Origin.X);
    Inc(Dest.Y, View.Origin.Y);
    View := View.Owner;
  until View = nil;
  Inc(Dest.X, Source.X);
  Inc(Dest.Y, Source.Y);
end;

procedure TViView.MakeLocal(Source: TPoint; var Dest: TPoint);
var
  View: TViView;
begin
  View := Self;
  Dest.X := 0;
  Dest.Y := 0;
  repeat
    Inc(Dest.X, View.Origin.X);
    Inc(Dest.Y, View.Origin.Y);
    View := View.Owner;
  until View = nil;
  Dest.X :=Source.X - Dest.X;
  Dest.Y :=Source.Y - Dest.Y;
end;

function TViView.MouseEvent(var Event: TEvent; Mask: Word): Boolean;
begin
  repeat GetEvent(Event) until Event.What and (Mask or evMouseUp) <> 0;
  MouseEvent := Event.What <> evMouseUp;
end;

function TViView.MouseInView(Mouse: TPoint): Boolean;
var
  Extent: TRect;
begin
  MakeLocal(Mouse, Mouse);
  GetExtent(Extent);
  MouseInView := RectContains(Extent, Mouse);
end;

procedure TViView.MoveTo(X, Y: Integer);
var
  R: TRect;
begin
  RectAssign(R, X, Y, X + Size.X, Y + Size.Y);
  Locate(R);
end;

function TViView.NextView: TViView;
begin
  if Self = Owner.Last then
    NextView := nil
  else
    NextView := Next;
end;

procedure TViView.NormalCursor;
begin
  SetState(sfCursorIns, False);
end;

function TViView.Prev: TViView;
var
  View: TViView;
begin
  View := Self;
  repeat
    Result := View;
    View := View.Next;
  until View = Self;
end;

function TViView.PrevView: TViView;
begin
  if Self = Owner.First then
    PrevView := nil
  else
    PrevView := Prev;
end;

procedure TViView.PutEvent(var Event: TEvent);
begin
  if Owner <> nil then Owner.PutEvent(Event);
end;

procedure TViView.PutInFrontOf(Target: TViView);
var
  P, LastView: TViView;

  procedure MoveView;
  begin
    Owner.RemoveView(Self);
    Owner.InsertView(Self, Target);
  end;

begin
  if (Owner <> nil) and (Target <> Self) and (Target <> NextView) and
    ((Target = nil) or (Target.Owner = Owner)) then
    if State and sfVisible = 0 then
      MoveView
    else
    begin

      LastView := NextView;
      if LastView <> nil then
      begin
        P := Target;
        while (P <> nil) and (P <> LastView) do
          P := P.NextView;
        if P = nil then LastView := Target;
      end;
      State := State and not sfVisible;
      if LastView = Target then DrawHide(LastView);
      MoveView;
      State := State or sfVisible;
      if LastView <> Target then DrawShow(LastView);
      if Options and ofSelectable <> 0 then
      begin
        Owner.ResetCurrent;
        Owner.ResetCursor;
      end;
    end;
end;

procedure TViView.ResetCursor;
var
  X, Y: Integer;

  function CursorVisible: Boolean;
  var
    Target, AView: TViView;
    Temp: Integer;
  begin
    Result := False;
    if (not State and (sfVisible
        + sfCursorVis + sfFocused)) <> 0 then Exit;
    AView := Self;
    X := AView.Cursor.X;
    Y := AView.Cursor.Y;
    repeat // @@1
      if (Y < 0) or (Y >= AView.Size.Y)
          or (X < 0) or (X >= AView.Size.X) then Exit;
      Inc(X, AView.Origin.X);
      Inc(Y, AView.Origin.Y);
      Target := AView;
      AView := AView.Owner;
      if not Assigned(AView) then
      begin
        Result := True;
        Exit;
      end;
      if (AView.State and sfVisible) = 0 then Exit;
      AView := TViGroup(AView).Last;
      repeat // @@2
        AView := AView.Next;
        if AView = Target then Break;
        if (AView.State and sfVisible) <> 0 then Continue;
        Temp := AView.Origin.Y;
        if Y < Temp then Continue;
        Inc(Temp, AView.Size.Y);
        if Y >= Temp then Continue;
        Temp := AView.Origin.X;
        if X < Temp then Continue;
        Inc(Temp, AView.Size.X);
        if X >= Temp then Continue;
      until False;
      AView := AView.Owner;
    until False;
  end;

begin
  if not CursorVisible then
    ViScreen.CursorVisible := False
  else
  begin
    ViScreen.SetCursor(X, Y);
    if (State and sfCursorIns) <> 0 then
      ViScreen.CursorStyle := csBig
    else
      ViScreen.CursorStyle := csNorm;
    ViScreen.CursorVisible := True;
  end;
end;

procedure TViView.Select;
begin
  if Options and ofSelectable <> 0 then
    if Options and ofTopSelect <> 0 then
      MakeFirst
    else if Owner <> nil then
      Owner.SetCurrent(Self, NormalSelect);
end;

procedure TViView.SetBounds(var Bounds: TRect);
begin
  LockBoundsChanged;
  Origin.X := Bounds.A.X;
  Origin.Y := Bounds.A.Y;
  Size.X := Bounds.B.X - Origin.X;
  Size.Y := Bounds.B.Y - Origin.Y;
  UnlockBoundsChanged;
end;

procedure TViView.SetCmdState(Commands: TCommandSet; Enable: Boolean);
begin
  if Enable then
    EnableCommands(Commands)
  else
    DisableCommands(Commands);
end;

procedure TViView.SetCommands(Commands: TCommandSet);
begin
  CommandSetChanged := CommandSetChanged or (CurCommandSet <> Commands);
  CurCommandSet := Commands;
end;

procedure TViView.SetCursor(X, Y: Integer);
begin
  Cursor.X := X;
  Cursor.Y := Y;
  DrawCursor;
end;

procedure TViView.SetState(AState: Word; Enable: Boolean);
var
  Command: Word;
begin
  if Enable then
    State := State or AState
  else
    State := State and not AState;
  if Owner <> nil then
    case AState of
      sfVisible:
        begin
          if Owner.State and sfExposed <> 0 then
            SetState(sfExposed, Enable);
          if Enable then
            DrawShow(nil)
          else
            DrawHide(nil);
          if Options and ofSelectable <> 0 then Owner.ResetCurrent;
        end;
      sfCursorVis, sfCursorIns:
        DrawCursor;
      sfShadow:
        DrawUnderView(State and sfShadow <> 0, nil);
      sfFocused:
        begin
          ResetCursor;
          if Enable then
            Command := cmReceivedFocus
          else
            Command := cmReleasedFocus;
          SendMessage(Owner, evBroadcast, Command, Self);
          if Enable then
            SendMessage(Self, evCommand, cmFocusIn, Self);
        end;
    end;
end;

procedure TViView.Show;
begin
  if State and sfVisible = 0 then SetState(sfVisible, True);
end;

procedure TViView.ShowCursor;
begin
  SetState(sfCursorVis, True);
end;

procedure TViView.SizeLimits(var Min, Max: TPoint);
begin
  Min.X := 0;
  Min.Y := 0;
  if Owner <> nil then
    Max := Owner.Size
  else
  begin
    Max.X := MaxLongint;
    Max.Y := MaxLongint;
  end;
end;

function TViView.TopView: TViView;
var
  P: TViView;
begin
  if TheTopView = nil then
  begin
    P := Self;
    while (P <> nil) and (P.State and sfModal = 0) do
      P := P.Owner;
    TopView := P;
  end
  else
    TopView := TheTopView;
end;

function TViView.Valid(Command: Word): Boolean;
begin
  Valid := True;
end;

procedure TViView.WriteBuf(X, Y, W, H: Integer; var Buf);
var
  I: Integer;
  Buffer: PChar;
begin
  Dec(Y);
  Buffer := Addr(Buf);
  for I := 1 to H do
  begin
    WriteView(X, Y + I, W, Buffer^);
    Inc(Buffer, W shl 1);
  end;
end;

procedure TViView.WriteChar(X, Y: Integer; C: Char; Color: Byte; Count:
  Integer);
var
  Buffer: array[0..255] of Word;
begin
  Color := MapColor(Color);
  Count := Count and $FF;
  MoveChar(Buffer, C, Color, Count);
  WriteView(X, Y, Count, Buffer);
end;

procedure TViView.WriteLine(X, Y, W, H: Integer; var Buf);
var
  I: Integer;
begin
  Dec(Y);
  for I := 1 to H do
    WriteView(X, Y + I, W, Buf);
end;

procedure TViView.WriteStr(X, Y: Integer; Str: string; Color: Byte);
var
  Buffer: array[0..255] of Word;
  Count: Integer;
begin
  Color := MapColor(Color);
  Count := Length(Str);
  if Count > 256 then SetLength(Str, 256);
  MoveStr(Buffer, Str, Color);
  WriteView(X, Y, Count, Buffer);
end;

function TViView.MapColor(Color: Byte): Byte;
var
  Palette: TPalette;
  Curr: TViView;
begin
  Result := ErrorAttr;
  if Color = 0 then Exit;
  Curr := Self;
  while Curr <> nil do
  begin
    Palette := Curr.GetPalette;
    if Palette <> '' then
    begin
      if Color > Length(Palette) then Exit;
      Color := Byte(Palette[Color]);
      if Color = 0 then Exit;
    end;
    Curr := Curr.Owner;
  end;
  Result := Color;
end;

function TViView.MapCPair(CPair: Word): Word;
begin
  if WordRec(CPair).Hi <> 0 then
    WordRec(Result).Hi := MapColor(WordRec(CPair).Hi);
  WordRec(Result).Lo := MapColor(WordRec(CPair).Lo);
end;

procedure TViView.WriteView(X, Y, W: Integer; var Buf);
var
  SourceBuffer: PVideoBuf;
  BufOfs: Integer;
  CharOnly: Integer;

  procedure P10(X, Y, W: Integer; AView: TViView);
  var
    Target: TViView;

    procedure P40(X, Y, W: Integer; AView: TViView);
    var
      DestBuffer: PVideoBuf;

      procedure CopyToBuffer(X, Y, W: Integer; AView: TViView);
      var
        DestIndex: Integer;
        SourceIndex: Integer;
        Count: Integer;
      begin
        Count := W - X;
        DestIndex := Y * Byte(AView.Size.X) + X;
        SourceIndex := X - BufOfs;
        if CharOnly > 0 then
          MoveVideoBuf(DestBuffer[DestIndex], SourceBuffer[SourceIndex],
              ShadowAttr, Count)
        else
          Move(SourceBuffer[SourceIndex], DestBuffer[DestIndex],
              Count shl 1);
        if DestBuffer = Addr(ViScreen.VideoBuffer) then
          ViScreen.WriteScreen(X + 1, Y + 1, DestIndex, Count);
      end;

    var
      Owner: TViGroup;
    begin { P40 }
      Owner := AView.Owner;
      DestBuffer := Owner.Buffer;
      if Assigned(DestBuffer) then
      begin
        if DestBuffer = Addr(ViScreen.VideoBuffer) then
        begin
          ViMouse.ProcessEvents;
          if (Y = ViMouse.Where.Y)
              and (X <= ViMouse.Where.X)
              and (W > ViMouse.Where.X) then
          begin
            ViMouse.Hide;
            CopyToBuffer(X, Y, W, Owner);
            ViMouse.Show;
          end else CopyToBuffer(X, Y, W, Owner);
        end else CopyToBuffer(X, Y, W, Owner);
      end;
      if Owner.LockFlag <> 0 then Exit;
      P10(X, Y, W, Owner);
    end;

    procedure P20(X, Y, W: Integer; AView: TViView);
    var
      Temp, TempY: Integer;
      OldCharOnly: Integer;
      OldBufOfs: Integer;
    begin
      repeat
        AView := AView.Next;
        if AView = Target then
        begin
          P40(X, Y, W, AView);
          Exit;
        end;
        if AView.State and sfVisible = 0 then Continue;
        Temp := AView.Origin.Y;
        if Y < Temp then Continue;
        Temp := Temp + AView.Size.Y;
        if Y < Temp then
        begin
          Temp := AView.Origin.X;
          if X < Temp then
          begin
            if W <= Temp then Continue;
            OldCharOnly := CharOnly;
            OldBufOfs := BufOfs;
            P20(X, Y, Temp, AView);
            BufOfs := OldBufOfs;
            CharOnly := OldCharOnly;
            X := Temp;
          end;
          Temp := Temp + AView.Size.X;
          if X < Temp then
          begin
            if W <= Temp then Exit;
            X := Temp;
          end;
          if AView.State and sfShadow = 0 then Continue;
          TempY := AView.Origin.Y + ShadowSize.Y;
          if Y < TempY then Continue;
          Temp := Temp + ShadowSize.X;
          if X >= Temp then Continue;
          Inc(CharOnly);
          if W <= Temp then Continue;
          OldCharOnly := CharOnly;
          OldBufOfs := BufOfs;
          P20(X, Y, Temp, AView);
          BufOfs := OldBufOfs;
          CharOnly := OldCharOnly;
          X := Temp;
          Dec(CharOnly);
          Continue;
        end;
        if AView.State and sfShadow = 0 then Continue;
        Temp := Temp + ShadowSize.Y;
        if Y >= Temp then Continue;

        Temp := AView.Origin.X + ShadowSize.X;
        if X < Temp then
        begin
          if W <= Temp then Continue;
          OldCharOnly := CharOnly;
          OldBufOfs := BufOfs;
          P20(X, Y, Temp, AView);
          BufOfs := OldBufOfs;
          CharOnly := OldCharOnly;
          X := Temp;
        end;
        Temp := Temp + AView.Size.X;
        if X >= Temp then Continue;
        Inc(CharOnly);
        if W <= Temp then Continue;
        OldCharOnly := CharOnly;
        OldBufOfs := BufOfs;
        P20(X, Y, Temp, AView);
        BufOfs := OldBufOfs;
        CharOnly := OldCharOnly;
        X := Temp;
        Dec(CharOnly);
      until False;
    end;

  { P10 }
  var
    Temp: Integer;
    Owner: TViGroup;
  begin
    if AView.State and sfVisible = 0 then Exit;
    if not Assigned(AView.Owner) then Exit;
    Target := AView;
    Y := Y + AView.Origin.Y;
    Temp := AView.Origin.X;
    X := X + Temp;
    W := W + Temp;
    BufOfs := BufOfs + Temp;
    Owner := AView.Owner;
    if (Y < Owner.Clip.A.Y) or (Y >= Owner.Clip.B.Y) then Exit;
    if X < Owner.Clip.A.X then
      X := Owner.Clip.A.X;
    if W > Owner.Clip.B.X then
      W := Owner.Clip.B.X;
    if X < W then
      P20(X, Y, W, Owner.Last);
  end; { P10 }

begin
  SourceBuffer := Addr(Buf);
  BufOfs := X;
  W := W + X;
  CharOnly := 0;
  if (Y < 0) or (Y >= Size.Y) then Exit;
  if X < 0 then X := 0;
  if W > Size.X then W := Size.X;
  if X < W then P10(X, Y, W, Self);
end;

function TViView.GetPalette: TPalette;
begin
  Result := FPalette;
end;

procedure TViView.SetPalette(Value: TPalette);
begin
  if FPalette <> Value then
  begin
    FPalette := Value;
  end;
end;

procedure TViView.BoundsChanged;
var
  E: TEvent;
begin
  E.What := evCommand;
  E.Command := cmxBoundsChanged;
  HandleEvent(E);
end;

procedure TViView.LockBoundsChanged;
begin
  Inc(FBoundsChangedLock);
end;

procedure TViView.UnLockBoundsChanged;
begin
  Dec(FBoundsChangedLock);
  if FBoundsChangedLock = 0 then BoundsChanged;
end;


{ TFrame }

procedure TViFrame.Draw;
var
  CFrame, CTitle: Word;
  F, I, L, Width: Integer;
  B: TDrawBuffer;
  Title: TTitleStr;
  Min, Max: TPoint;
begin
  if State and sfDragging <> 0 then
  begin
    CFrame := $0505;
    CTitle := $0005;
    F := 0;
  end
  else if State and sfActive = 0 then
  begin
    CFrame := $0101;
    CTitle := $0002;
    F := 0;
  end
  else
  begin
    CFrame := $0503;
    CTitle := $0004;
    F := 9;
  end;
  CFrame := GetColor(CFrame);
  CTitle := GetColor(CTitle);
  Width := Size.X;
  FrameLine(B, 0, F, Byte(CFrame));
  if (TViWindow(Owner).Number <> wnNoNumber) and
    (TViWindow(Owner).Number < 10) then
  begin
    if TViWindow(Owner).Flags and wfZoom <> 0 then
      I := 7
    else
      I := 3;
    WordRec(B[Width - I]).Lo := TViWindow(Owner).Number + $30;
  end;
  if Owner <> nil then
    Title := TViWindow(Owner).GetTitle
  else
    Title := '';
  if Title <> '' then
  begin
    L := Length(Title);
    if L > Width - 10 then L := Width - 10;
    if L < 0 then L := 0;
    I := (Width - L) shr 1;
    MoveChar(B[I - 1], ' ', CTitle, 1);
    MoveBuf(B[I], Title[1], CTitle, L);
    MoveChar(B[I + L], ' ', CTitle, 1);
  end;
  if State and sfActive <> 0 then
  begin
    if TViWindow(Owner).Flags and wfClose <> 0 then
      if FrameMode and fmCloseClicked = 0 then
        MoveCStr(B[2], ldCloseWindow, CFrame)
      else
        MoveCStr(B[2], ldCloseClicked, CFrame);
    if TViWindow(Owner).Flags and wfZoom <> 0 then
    begin
      MoveCStr(B[Width - 5], ldMaximize, CFrame);
      Owner.SizeLimits(Min, Max);
      if FrameMode and fmZoomClicked <> 0 then
        WordRec(B[Width - 4]).Lo := 15
      else if (Owner.Size.X = Max.X) and (Owner.Size.Y = Max.Y) then
        WordRec(B[Width - 4]).Lo := 18;
    end;
  end;
  WriteLine(0, 0, Size.X, 1, B);
  for I := 1 to Size.Y - 2 do
  begin
    FrameLine(B, I, F + 3, Byte(CFrame));
    WriteLine(0, I, Size.X, 1, B);
  end;
  FrameLine(B, Size.Y - 1, F + 6, Byte(CFrame));
  if State and sfActive <> 0 then
    if TViWindow(Owner).Flags and wfGrow <> 0 then
      MoveCStr(B[Width - 2], ldBottomRight, CFrame);
  WriteLine(0, Size.Y - 1, Size.X, 1, B);
end;

procedure TViFrame.HandleEvent(var Event: TEvent);
var
  Mouse: TPoint;

  procedure DragWindow(Mode: Byte);
  var
    Limits: TRect;
    Min, Max: TPoint;
  begin
    Owner.Owner.GetExtent(Limits);
    Owner.SizeLimits(Min, Max);
    Owner.DragView(Event, Owner.DragMode or Mode, Limits, Min, Max);
    ClearEvent(Event);
  end;

begin
  inherited HandleEvent(Event);
  if Event.What = evMouseDown then
  begin
    MakeLocal(Event.Where, Mouse);
    if Mouse.Y = 0 then
    begin
      if (TViWindow(Owner).Flags and wfClose <> 0) and
        (State and sfActive <> 0) and (Mouse.X >= 2) and (Mouse.X <= 4) then
      begin
        repeat
          MakeLocal(Event.Where, Mouse);
          if (Mouse.X >= 2) and (Mouse.X <= 4) and (Mouse.Y = 0) then
            FrameMode := fmCloseClicked
          else
            FrameMode := 0;
          DrawView;
        until not MouseEvent(Event, evMouseMove + evMouseAuto);
        FrameMode := 0;
        if (Mouse.X >= 2) and (Mouse.X <= 4) and (Mouse.Y = 0) then
        begin
          Event.What := evCommand;
          Event.Command := cmClose;
          Event.InfoPtr := Owner;
          PutEvent(Event);
        end;
        ClearEvent(Event);
        DrawView;
      end
      else if (TViWindow(Owner).Flags and wfZoom <> 0) and
        (State and sfActive <> 0) and (Event.Double or
        (Mouse.X >= Size.X - 5) and
        (Mouse.X <= Size.X - 3)) then
      begin
        if not Event.Double then
          repeat
            MakeLocal(Event.Where, Mouse);
            if (Mouse.X >= Size.X - 5) and (Mouse.X <= Size.X - 3) and
              (Mouse.Y = 0) then
              FrameMode := fmZoomClicked
            else
              FrameMode := 0;
            DrawView;
          until not MouseEvent(Event, evMouseMove + evMouseAuto);
        FrameMode := 0;
        if ((Mouse.X >= Size.X - 5) and (Mouse.X <= Size.X - 3) and
          (Mouse.Y = 0)) or Event.Double then
        begin
          Event.What := evCommand;
          Event.Command := cmZoom;
          Event.InfoPtr := Owner;
          PutEvent(Event);
        end;
        ClearEvent(Event);
        DrawView;
      end
      else if TViWindow(Owner).Flags and wfMove <> 0 then
        DragWindow(dmDragMove);
    end
    else if (State and sfActive <> 0) and (Mouse.X >= Size.X - 2) and
      (Mouse.Y >= Size.Y - 1) then
      if TViWindow(Owner).Flags and wfGrow <> 0 then
        DragWindow(dmDragGrow);
  end;
end;

{ TScrollBar }

constructor TViScrollBar.Create(var Bounds: TRect; ViMode: Boolean = False);
begin
  inherited Create(Bounds);
  if ViMode then
    Options := Options or ofViInternal;
  Palette := CScrollBar;
  Value := 0;
  Min := 0;
  Max := 0;
  PgStep := 1;
  ArStep := 1;
  BoundsChanged;
end;

procedure TViScrollBar.Draw;
begin
  DrawPos(GetPos);
end;

procedure TViScrollBar.DrawPos(Pos: Integer);
var
  S: Integer;
  B: TDrawBuffer;
begin
  S := GetSize - 1;
  MoveChar(B[0], Chars[0], GetColor(2), 1);
  if Max = Min then
    MoveChar(B[1], Chars[4], GetColor(1), S - 1)
  else
  begin
    MoveChar(B[1], Chars[2], GetColor(1), S - 1);
    MoveChar(B[Pos], Chars[3], GetColor(3), 1);
  end;
  MoveChar(B[S], Chars[1], GetColor(2), 1);
  WriteBuf(0, 0, Size.X, Size.Y, B);
end;

function TViScrollBar.GetPos: Integer;
var
  R: Integer;
begin
  R := Max - Min;
  if R = 0 then
    GetPos := 1
  else
    GetPos := ((Value - Min) * (GetSize - 3) + R shr 1) div R + 1;
end;

function TViScrollBar.GetSize: Integer;
var
  S: Integer;
begin
  if Size.X = 1 then
    S := Size.Y
  else
    S := Size.X;
  if S < 3 then
    GetSize := 3
  else
    GetSize := S;
end;

procedure TViScrollBar.HandleEvent(var Event: TEvent);
var
  Tracking: Boolean;
  I, P, S, ClickPart: Integer;
  Mouse: TPoint;
  Extent: TRect;

  function GetPartCode: Integer;
  var
    Mark, Part: Integer;
  begin
    Part := -1;
    if RectContains(Extent, Mouse) then
    begin
      if Size.X = 1 then
        Mark := Mouse.Y
      else
        Mark := Mouse.X;
      if Mark = P then
        Part := sbIndicator
      else
      begin
        if Mark < 1 then
          Part := sbLeftArrow
        else if Mark < P then
          Part := sbPageLeft
        else if Mark < S then
          Part := sbPageRight
        else
          Part := sbRightArrow;
        if Size.X = 1 then Inc(Part, 4);
      end;
    end;
    GetPartCode := Part;
  end;

  procedure Clicked;
  begin
    SendMessage(Owner, evBroadcast, cmScrollBarClicked, Self);
  end;

begin
  case Event.What of
    evCommand, evMouseUp, evMouseMove:
      inherited HandleEvent(Event);
    evMouseDown:
      begin
        inherited HandleEvent(Event);
        if Event.What = evNothing then Exit;
        Clicked;
        MakeLocal(Event.Where, Mouse);
        GetExtent(Extent);
        RectGrow(Extent, 1, 1);
        P := GetPos;
        S := GetSize - 1;
        ClickPart := GetPartCode;
        if ClickPart <> sbIndicator then
        begin
          repeat
            MakeLocal(Event.Where, Mouse);
            if GetPartCode = ClickPart then
              SetValue(Value + ScrollStep(ClickPart));
          until not MouseEvent(Event, evMouseAuto);
        end
        else
        begin
          repeat
            MakeLocal(Event.Where, Mouse);
            Tracking := RectContains(Extent, Mouse);
            if Tracking then
            begin
              if Size.X = 1 then
                I := Mouse.Y
              else
                I := Mouse.X;
              if I <= 0 then I := 1;
              if I >= S then I := S - 1;
            end
            else
              I := GetPos;
            if I <> P then
            begin
              DrawPos(I);
              P := I;
            end;
          until not MouseEvent(Event, evMouseMove);
          if Tracking and (S > 2) then
          begin
            Dec(S, 2);
            SetValue(((P - 1) * (Max - Min) + S shr 1) div S + Min);
          end;
        end;
        ClearEvent(Event);
      end;
    evKeyDown:
      begin
        if Size.Y = 1 then
          case CtrlToArrow(Event.KeyCode) of
            kbLeft, kbRight, kbCtrlLeft, kbCtrlRight,
            kbHome, kbEnd: { do nothing }
          else
            inherited HandleEvent(Event);
          end
        else
          case CtrlToArrow(Event.KeyCode) of
            kbUp, kbDown, kbPgUp, kbPgDn,
            kbCtrlPgUp, kbCtrlPgDn: { do nothing }
          else
            inherited HandleEvent(Event);
          end;
        if State and sfVisible <> 0 then
        begin
          I := 0;
          ClickPart := sbIndicator;
          if Size.Y = 1 then
            case CtrlToArrow(Event.KeyCode) of
              kbLeft: ClickPart := sbLeftArrow;
              kbRight: ClickPart := sbRightArrow;
              kbCtrlLeft: ClickPart := sbPageLeft;
              kbCtrlRight: ClickPart := sbPageRight;
              kbHome: I := Min;
              kbEnd: I := Max;
            else
              Exit;
            end
          else
            case CtrlToArrow(Event.KeyCode) of
              kbUp: ClickPart := sbUpArrow;
              kbDown: ClickPart := sbDownArrow;
              kbPgUp: ClickPart := sbPageUp;
              kbPgDn: ClickPart := sbPageDown;
              kbCtrlPgUp: I := Min;
              kbCtrlPgDn: I := Max;
            else
              Exit;
            end;
          Clicked;
          if ClickPart <> sbIndicator then
          begin
            DoScroll(ClickPart);
            I := Value + ScrollStep(ClickPart);
          end;
          SetValue(I);
          ClearEvent(Event);
        end;
      end;
  end;
end;

procedure TViScrollBar.ScrollDraw;
begin
  SendMessage(Owner, evBroadcast, cmScrollBarChanged, Self);
end;

function TViScrollBar.ScrollStep(Part: Integer): Integer;
var
  Step: Integer;
begin
  if Part and 2 = 0 then
    Step := ArStep
  else
    Step := PgStep;
  if Part and 1 = 0 then
    ScrollStep := -Step
  else
    ScrollStep := Step;
end;

procedure TViScrollBar.SetParams(AValue, AMin, AMax, APgStep, AArStep:
  Integer);
var
  SValue: Integer;
begin
  if AMax < AMin then AMax := AMin;
  if AValue < AMin then AValue := AMin;
  if AValue > AMax then AValue := AMax;
  SValue := Value;
  if (SValue <> AValue) or (Min <> AMin) or (Max <> AMax) then
  begin
    Value := AValue;
    Min := AMin;
    Max := AMax;
    DrawView;
    if SValue <> AValue then
    begin
      ScrollDraw;
      DoChange;
    end;
  end;
  PgStep := APgStep;
  ArStep := AArStep;
end;

procedure TViScrollBar.SetRange(AMin, AMax: Integer);
begin
  SetParams(Value, AMin, AMax, PgStep, ArStep);
end;

procedure TViScrollBar.SetStep(APgStep, AArStep: Integer);
begin
  SetParams(Value, Min, Max, APgStep, AArStep);
end;

procedure TViScrollBar.SetValue(AValue: Integer);
begin
  SetParams(AValue, Min, Max, PgStep, ArStep);
end;

procedure TViScrollBar.BoundsChanged;
begin
  if Size.X = 1 then
    SetOrientation(orVertical)
  else
    SetOrientation(orHorizontal);
end;

procedure TViScrollBar.SetOrientation(Value: TOrientation);
begin
  case Value of
    orVertical:
    begin
      if Options and ofViInternal <> 0 then
        GrowMode := gfGrowLoX + gfGrowHiX + gfGrowHiY;
      Chars := ldVerticalScroll;
    end
  else
    if Options and ofViInternal <> 0 then
      GrowMode := gfGrowLoY + gfGrowHiX + gfGrowHiY;
    Chars := ldHorizontalScroll;
  end;
end;

procedure TViScrollBar.DoChange;
begin
  if Assigned(OnChange) then OnChange(Self);
end;

procedure TViScrollBar.DoScroll(PartCode: Integer);
begin
  if Assigned(OnScroll) then OnScroll(Self, PartCode);
end;



{ TScroller }

constructor TViScroller.Create(var Bounds: TRect; AHScrollBar,
  AVScrollBar: TViScrollBar);
begin
  inherited Create(Bounds);
  Palette := CScroller;
  Options := Options or ofSelectable;
  EventMask := EventMask or evBroadcast;
  HScrollBar := AHScrollBar;
  VScrollBar := AVScrollBar;
end;

procedure TViScroller.ChangeBounds(var Bounds: TRect);
begin
  SetBounds(Bounds);
  Inc(DrawLock);
  SetLimit(Limit.X, Limit.Y);
  Dec(DrawLock);
  DrawFlag := False;
  DrawView;
end;

procedure TViScroller.CheckDraw;
begin
  if (DrawLock = 0) and DrawFlag then
  begin
    DrawFlag := False;
    DrawView;
  end;
end;

procedure TViScroller.HandleEvent(var Event: TEvent);
begin
  inherited HandleEvent(Event);
  if (Event.What = evBroadcast) and (Event.Command = cmScrollBarChanged) and
    ((Event.InfoPtr = HScrollBar) or (Event.InfoPtr = VScrollBar)) then
    ScrollDraw;
end;

procedure TViScroller.ScrollDraw;
var
  D: TPoint;
begin
  if HScrollBar <> nil then
    D.X := HScrollBar.Value
  else
    D.X := 0;
  if VScrollBar <> nil then
    D.Y := VScrollBar.Value
  else
    D.Y := 0;
  if (D.X <> Delta.X) or (D.Y <> Delta.Y) then
  begin
    SetCursor(Cursor.X + Delta.X - D.X, Cursor.Y + Delta.Y - D.Y);
    Delta := D;
    if DrawLock <> 0 then
      DrawFlag := True
    else
      DrawView;
  end;
end;

procedure TViScroller.ScrollTo(X, Y: Integer);
begin
  Inc(DrawLock);
  if HScrollBar <> nil then HScrollBar.SetValue(X);
  if VScrollBar <> nil then VScrollBar.SetValue(Y);
  Dec(DrawLock);
  CheckDraw;
end;

procedure TViScroller.SetLimit(X, Y: Integer);
begin
  Limit.X := X;
  Limit.Y := Y;
  Inc(DrawLock);
  if HScrollBar <> nil then
    HScrollBar.SetParams(HScrollBar.Value, 0, X - Size.X, Size.X - 1,
      HScrollBar.ArStep);
  if VScrollBar <> nil then
    VScrollBar.SetParams(VScrollBar.Value, 0, Y - Size.Y, Size.Y - 1,
      VScrollBar.ArStep);
  Dec(DrawLock);
  CheckDraw;
end;

procedure TViScroller.SetState(AState: Word; Enable: Boolean);

  procedure ShowSBar(SBar: TViScrollBar);
  begin
    if (SBar <> nil) then
      if GetState(sfActive + sfSelected) then
        SBar.Show
      else
        SBar.Hide;
  end;

begin
  inherited SetState(AState, Enable);
  if AState and (sfActive + sfSelected) <> 0 then
  begin
    ShowSBar(HScrollBar);
    ShowSBar(VScrollBar);
  end;
end;

{ TListViewer }

constructor TViListViewer.Create(var Bounds: TRect; ANumCols: Word;
  AHScrollBar, AVScrollBar: TViScrollBar);
var
  ArStep, PgStep: Integer;
begin
  inherited Create(Bounds);
  Palette := CListViewer;
  Options := Options or (ofFirstClick + ofSelectable);
  EventMask := EventMask or evBroadcast;
  Range := 0;
  NumCols := ANumCols;
  Focused := 0;
  if AVScrollBar <> nil then
  begin
    if NumCols = 1 then
    begin
      PgStep := Size.Y - 1;
      ArStep := 1;
    end
    else
    begin
      PgStep := Size.Y * NumCols;
      ArStep := Size.Y;
    end;
    AVScrollBar.SetStep(PgStep, ArStep);
  end;
  if AHScrollBar <> nil then AHScrollBar.SetStep(Size.X div NumCols, 1);
  HScrollBar := AHScrollBar;
  VScrollBar := AVScrollBar;
end;

procedure TViListViewer.ChangeBounds(var Bounds: TRect);
begin
  inherited ChangeBounds(Bounds);
  if HScrollBar <> nil then
    HScrollBar.SetStep(Size.X div NumCols, 1);
  if VScrollBar <> nil then
    if NumCols > 1 then
      VScrollBar.SetStep(Size.Y * NumCols, Size.Y)
    else
      VScrollBar.SetStep(Size.Y - 1, 1);
end;

procedure TViListViewer.Draw;
var
  I, J, Item: Integer;
  NormalColor, SelectedColor, FocusedColor, Color: Word;
  ColWidth, CurCol, Indent: Integer;
  B: TDrawBuffer;
  Text: string;
  SCOff: Byte;
begin
  if State and (sfSelected + sfActive) = (sfSelected + sfActive) then
  begin
    NormalColor := GetColor(1);
    FocusedColor := GetColor(3);
    SelectedColor := GetColor(4);
  end
  else
  begin
    NormalColor := GetColor(2);
    SelectedColor := GetColor(4);
    FocusedColor := GetColor(3);
  end;
  if HScrollBar <> nil then
    Indent := HScrollBar.Value
  else
    Indent := 0;
  ColWidth := Size.X div NumCols + 1;
  for I := 0 to Size.Y - 1 do
  begin
    for J := 0 to NumCols - 1 do
    begin
      Item := J * Size.Y + I + TopItem;
      CurCol := J * ColWidth;
      if (State and (sfSelected + sfActive) = (sfSelected + sfActive)) and
        (Focused = Item) and (Range > 0) then
      begin
        Color := FocusedColor;
        SetCursor(CurCol + 1, I);
        SCOff := 0;
      end
      else if (Item < Range) and IsSelected(Item) then
      begin
        Color := SelectedColor;
        SCOff := 2;
      end
      else
      begin
        Color := NormalColor;
        SCOff := 4;
      end;
      MoveChar(B[CurCol], ' ', Color, ColWidth);
      if Item < Range then
      begin
        Text := GetText(Item, ColWidth + Indent);
        Text := Copy(Text, Indent, ColWidth);
        MoveStr(B[CurCol + 1], Text, Color);
        if ShowMarkers then
        begin
          WordRec(B[CurCol]).Lo := Byte(SpecialChars[SCOff]);
          WordRec(B[CurCol + ColWidth - 2]).Lo := Byte(SpecialChars[SCOff + 1]);
        end;
      end;
      MoveChar(B[CurCol + ColWidth - 1], ldVerticalBar, GetColor(5), 1);
    end;
    WriteLine(0, I, Size.X, 1, B);
  end;
end;

procedure TViListViewer.FocusItem(Item: Integer);
begin
  Focused := Item;
  if VScrollBar <> nil then VScrollBar.SetValue(Item);
  if Item < TopItem then
    if NumCols = 1 then
      TopItem := Item
    else
      TopItem := Item - Item mod Size.Y
  else if Item >= TopItem + (Size.Y * NumCols) then
    if NumCols = 1 then
      TopItem := Item - Size.Y + 1
    else
      TopItem := Item - Item mod Size.Y - (Size.Y * (NumCols - 1));
end;

procedure TViListViewer.FocusItemNum(Item: Integer);
begin
  if Item < 0 then
    Item := 0
  else if (Item >= Range) and (Range > 0) then
    Item := Range - 1;
  if Range <> 0 then FocusItem(Item);
end;

function TViListViewer.IsSelected(Item: Integer): Boolean;
begin
  IsSelected := Item = Focused;
end;

procedure TViListViewer.HandleEvent(var Event: TEvent);
const
  MouseAutosToSkip = 4;
var
  Mouse: TPoint;
  ColWidth: Word;
  OldItem, NewItem: Integer;
  Count: Word;
begin
  case Event.What of
    evCommand, evMouseUp, evMouseMove:
      inherited HandleEvent(Event);
    evMouseDown:
    begin
      FCapture := True;
      try
        ColWidth := Size.X div NumCols + 1;
        OldItem := Focused;
        MakeLocal(Event.Where, Mouse);
        if MouseInView(Event.Where) then
          NewItem := Mouse.Y + (Size.Y * (Mouse.X div ColWidth)) + TopItem
        else
          NewItem := OldItem;
        Count := 0;
        repeat
          inherited HandleEvent(Event);
          if NewItem <> OldItem then
          begin
            FocusItemNum(NewItem);
            DrawView;
          end;
          OldItem := NewItem;
          MakeLocal(Event.Where, Mouse);
          if MouseInView(Event.Where) then
            NewItem := Mouse.Y + (Size.Y * (Mouse.X div ColWidth)) + TopItem
          else
          begin
            if NumCols = 1 then
            begin
              if Event.What = evMouseAuto then Inc(Count);
              if Count = MouseAutosToSkip then
              begin
                Count := 0;
                if Mouse.Y < 0 then
                  NewItem := Focused - 1
                else if Mouse.Y >= Size.Y then
                  NewItem := Focused + 1;
              end;
            end
            else
            begin
              if Event.What = evMouseAuto then Inc(Count);
              if Count = MouseAutosToSkip then
              begin
                Count := 0;
                if Mouse.X < 0 then
                  NewItem := Focused - Size.Y
                else if Mouse.X >= Size.X then
                  NewItem := Focused + Size.Y
                else if Mouse.Y < 0 then
                  NewItem := Focused - Focused mod Size.Y
                else if Mouse.Y > Size.Y then
                  NewItem := Focused - Focused mod Size.Y + Size.Y - 1;
              end
            end;
          end;
        until not MouseEvent(Event, evMouseMove + evMouseAuto);
        FocusItemNum(NewItem);
        if Range > 0 then DoChanged;
        DrawView;
        if Event.Double and (Range > Focused) then SelectItem(Focused);
        inherited HandleEvent(Event);
        ClearEvent(Event);
      finally
        FCapture := False;
      end;
    end;
    evKeyDown:
    begin
      case CtrlToArrow(Event.KeyCode) of
        kbUp, kbDown, kbRight, kbLeft,
        kbPgDn, kbPgUp, kbHome, kbEnd,
        kbCtrlPgDn, kbCtrlPgUp: { do nothing };
      else
        inherited HandleEvent(Event);
      end;
      if (Event.CharCode = ' ') and (Focused < Range) then
      begin
        SelectItem(Focused);
        NewItem := Focused;
      end
      else
        case CtrlToArrow(Event.KeyCode) of
          kbUp: NewItem := Focused - 1;
          kbDown: NewItem := Focused + 1;
          kbRight:
            if NumCols > 1 then
              NewItem := Focused + Size.Y
            else
              Exit;
          kbLeft:
            if NumCols > 1 then
              NewItem := Focused - Size.Y
            else
              Exit;
          kbPgDn: NewItem := Focused + Size.Y * NumCols;
          kbPgUp: NewItem := Focused - Size.Y * NumCols;
          kbHome: NewItem := TopItem;
          kbEnd: NewItem := TopItem + (Size.Y * NumCols) - 1;
          kbCtrlPgDn: NewItem := Range - 1;
          kbCtrlPgUp: NewItem := 0;
        else
          Exit;
        end;
      FocusItemNum(NewItem);
      DrawView;
      ClearEvent(Event);
    end;
    evBroadcast:
    begin
      if Options and ofSelectable <> 0 then
        if (Event.Command = cmScrollBarClicked) and
          ((Event.InfoPtr = HScrollBar) or (Event.InfoPtr = VScrollBar)) then
          Select
        else if (Event.Command = cmScrollBarChanged) then
        begin
          if (VScrollBar = Event.InfoPtr) then
          begin
            FocusItemNum(VScrollBar.Value);
            DrawView;
            if (State and (sfSelected + sfActive) = (sfSelected + sfActive))
            and not FCapture then DoChanged;
          end
          else if (HScrollBar = Event.InfoPtr) then
            DrawView;
        end;
    end;
  end;
end;

procedure TViListViewer.SelectItem(Item: Integer);
begin
  SendMessage(Owner, evBroadcast, cmListItemSelected, Self);
end;

procedure TViListViewer.SetRange(ARange: Integer);
begin
  Range := ARange;
  if VScrollBar <> nil then
  begin
    if Focused > ARange then Focused := 0;
    VScrollbar.SetParams(Focused, 0, ARange - 1, VScrollBar.PgStep,
      VScrollBar.ArStep);
  end;
end;

procedure TViListViewer.SetState(AState: Word; Enable: Boolean);

  procedure ShowSBar(SBar: TViScrollBar);
  begin
    if (SBar <> nil) then
      if GetState(sfActive) and GetState(sfVisible) then
        SBar.Show
      else
        SBar.Hide;
  end;

begin
  inherited SetState(AState, Enable);
  if AState and (sfSelected + sfActive + sfVisible) <> 0 then
  begin
    ShowSBar(HScrollBar);
    ShowSBar(VScrollBar);
    DrawView;
  end;
end;

procedure TViListViewer.DoChanged;
begin
  if Assigned(OnChanged) then OnChanged(Self);
end;


{ TGroup }

constructor TViGroup.Create(var Bounds: TRect);
begin
  inherited Create(Bounds);
  Options := Options or (ofSelectable + ofBuffered);
  GetExtent(Clip);
  EventMask := $FFFFFFFF;
end;

destructor TViGroup.Destroy;
var
  P, T: TViView;
begin
  Hide;
  P := Last;
  if P <> nil then
  begin
    repeat
      P.Hide;
      P := P.Prev;
    until P = Last;
    repeat
      T := P.Prev;
      P.Free;
      P := T;
    until Last = nil;
  end;
  FreeBuffer;
  inherited Destroy;
end;

function TViGroup.At(Index: Integer): TViView;
var
  I: Integer;
begin
  Result := Last;
  for I := 1 to Index do
    Result := Result.Next;
end;

procedure TViGroup.ChangeBounds(var Bounds: TRect);
var
  D: TPoint;

  procedure DoCalcChange(P: TViView);
  var
    R: TRect;
  begin
    P.CalcBounds(R, D);
    P.ChangeBounds(R);
  end;

  procedure ForEachCalcChange;
  var
    Curr: TViView;
  begin
    Curr := Last;
    if not Assigned(Curr) then Exit;
    repeat
      Curr := Curr.Next;
      DoCalcChange(Curr);
    until Curr = Last;
  end;

begin
  LockBoundsChanged;
  try
    D.X := Bounds.B.X - Bounds.A.X - Size.X;
    D.Y := Bounds.B.Y - Bounds.A.Y - Size.Y;
    if (D.X or D.Y) = 0 then
    begin
      SetBounds(Bounds);
      DrawView;
    end
    else
    begin
      FreeBuffer;
      SetBounds(Bounds);
      GetExtent(Clip);
      GetBuffer;
      Lock;
      ForEachCalcChange;
      Unlock;
    end;
  finally
    UnLockBoundsChanged;
  end;
end;

procedure TViGroup.Delete(P: TViView);
var
  SaveState: Word;
begin
  SaveState := P.State;
  P.Hide;
  if P = Current then ResetCurrent;
  RemoveView(P);
  P.Owner := nil;
  P.Next := nil;
  if SaveState and sfVisible <> 0 then P.Show;
end;

procedure TViGroup.Draw;
begin
  if Buffer = nil then
  begin
    GetBuffer;
    if Buffer <> nil then
    begin
      Inc(LockFlag);
      Redraw;
      Dec(LockFlag);
    end;
  end;
  if Buffer <> nil then
    WriteBuf(0, 0, Size.X, Size.Y, Buffer^)
  else
  begin
    GetClipRect(Clip);
    Redraw;
    GetExtent(Clip);
  end;
end;

procedure TViGroup.DrawSubViews(P, Bottom: TViView);
begin
  if P <> nil then
    while P <> Bottom do
    begin
      P.DrawView;
      P := P.NextView;
    end;
end;

procedure TViGroup.EndModal(Command: Word);
begin
  if State and sfModal <> 0 then
    EndState := Command
  else
    inherited EndModal(Command);
end;

procedure TViGroup.EventError(var Event: TEvent);
begin
  if Owner <> nil then Owner.EventError(Event);
end;

function TViGroup.Execute: Word;
var
  E: TEvent;
begin
  repeat
    EndState := 0;
    repeat
      GetEvent(E);
      HandleEvent(E);
      if E.What <> evNothing then EventError(E);
    until EndState <> 0;
  until Valid(EndState);
  Execute := EndState;
end;

function TViGroup.ExecView(P: TViView): Word;
var
  SaveOptions: Word;
  SaveOwner: TViGroup;
  SaveTopView: TViView;
  SaveCurrent: TViView;
  SaveCommands: TCommandSet;
begin
  if P <> nil then
  begin
    SaveOptions := P.Options;
    SaveOwner := P.Owner;
    SaveTopView := TheTopView;
    SaveCurrent := Current;
    GetCommands(SaveCommands);
    TheTopView := P;
    P.Options := P.Options and not ofSelectable;
    P.SetState(sfModal, True);
    SetCurrent(P, EnterSelect);
    if SaveOwner = nil then Insert(P);
    ExecView := P.Execute;
    if SaveOwner = nil then Delete(P);
    SetCurrent(SaveCurrent, LeaveSelect);
    P.SetState(sfModal, False);
    P.Options := SaveOptions;
    TheTopView := SaveTopView;
    SetCommands(SaveCommands);
  end
  else
    ExecView := cmCancel;
end;

function TViGroup.First: TViView;
begin
  if Last = nil then
    First := nil
  else
    First := Last.Next;
end;

function TViGroup.FirstMatch(AState: Word; AOptions: Word): TViView;

  function Matches(P: TViView): Boolean;
  begin
    Matches := (P.State and AState = AState) and
      (P.Options and AOptions = AOptions);
  end;

begin
  Result := Last;
  if Result = nil then Exit;
  repeat
    Result := Result.Next;
    if Matches(Result) then Exit;
  until Result = Last;
  Result := nil;
end;

function TViGroup.FindNext(Forwards: Boolean): TViView;
var
  P: TViView;
begin
  FindNext := nil;
  if Current <> nil then
  begin
    P := Current;
    repeat
      if Forwards then
        P := P.Next
      else
        P := P.Prev;
    until ((P.State and (sfVisible + sfDisabled) = sfVisible) and
      (P.Options and ofSelectable <> 0)) or (P = Current);
    if P <> Current then FindNext := P;
  end;
end;

function TViGroup.FocusNext(Forwards: Boolean): Boolean;
var
  P: TViView;
begin
  P := FindNext(Forwards);
  FocusNext := True;
  if P <> nil then FocusNext := P.Focus;
end;

procedure TViGroup.FreeBuffer;
begin
  if (Options and ofBuffered <> 0) and (Buffer <> nil) then
  begin
    FreeMem(Buffer);
    Buffer := nil;
  end;
end;

{ Allocate a group buffer if the group is exposed and buffered }

procedure TViGroup.GetBuffer;
var
  MemSize: Integer;
begin
  if (State and sfExposed <> 0)
      and (Options and ofBuffered <> 0)
      and (Buffer <> nil) then
  begin
    MemSize := Size.X * Size.Y shl 1;
    GetMem(Buffer, MemSize);
  end;
end;

function TViGroup.GetHelpCtx: Word;
var
  H: Word;
begin
  H := hcNoContext;
  if Current <> nil then H := Current.GetHelpCtx;
  if H = hcNoContext then H := inherited GetHelpCtx;
  GetHelpCtx := H;
end;

procedure TViGroup.HandleEvent(var Event: TEvent);

  procedure DoHandleEvent(P: TViView);
  begin
    if (P = nil) or ((P.State and sfDisabled <> 0)
      and (Event.What and (PositionalEvents or FocusedEvents) <> 0)) then Exit;
    case Phase of
      phPreProcess:
        if P.Options and ofPreProcess = 0 then Exit;
      phPostProcess:
        if P.Options and ofPostProcess = 0 then Exit;
    end;
    if Event.What and P.EventMask <> 0 then P.HandleEvent(Event);
  end;

  procedure ForEachHandleEvent;
  var
    Curr: TViView;
  begin
    Curr := Last;
    if not Assigned(Curr) then Exit;
    repeat
      Curr := Curr.Next;
      if not Assigned(Curr) then Exit;
      DoHandleEvent(Curr);
    until Curr = Last;
  end;

  function FirstThatContainsMouse: TViView;
  begin
    Result := Last;
    if Result = nil then Exit;
    repeat
      Result := Result.Next;
      if (Result.State and sfVisible <> 0) and
          Result.MouseInView(Event.Where) then Exit;
    until Result = Last;
    Result := nil;
  end;


begin
  inherited HandleEvent(Event);
  if Event.What and FocusedEvents <> 0 then
  begin
    Phase := phPreProcess;
    ForEachHandleEvent;
    Phase := phFocused;
    DoHandleEvent(Current);
    Phase := phPostProcess;
    ForEachHandleEvent;
  end
  else
  begin
    Phase := phFocused;
    if (Event.What and PositionalEvents <> 0) then
      DoHandleEvent(FirstThatContainsMouse)
    else
      ForEachHandleEvent;
  end;
end;

function TViGroup.IndexOf(P: TViView): Integer;
var
  View: TViView;
begin
  Result := 0;
  View := Last;
  if not Assigned(View) then Exit;
  repeat
    Inc(Result);
    View := View.Next;
    if P = View then Exit;
  until View = Last;
  Result := 0;
end;

procedure TViGroup.Insert(P: TViView);
begin
  InsertBefore(P, First);
end;

procedure TViGroup.InsertBefore(P, Target: TViView);
var
  SaveState: Word;
begin
  if (P <> nil) and (P.Owner = nil) and
    ((Target = nil) or (Target.Owner = Self)) then
  begin
    if P.Options and ofCenterX <> 0 then
      P.Origin.X := (Size.X - P.Size.X) div 2;
    if P.Options and ofCenterY <> 0 then
      P.Origin.Y := (Size.Y - P.Size.Y) div 2;
    SaveState := P.State;
    P.Hide;
    InsertView(P, Target);
    if SaveState and sfVisible <> 0 then P.Show;
    if State and sfActive <> 0 then
      P.SetState(sfActive, True);
  end;
end;

procedure TViGroup.InsertView(P, Target: TViView);
begin
  P.Owner := Self;
  if Target <> nil then
  begin
    Target := Target.Prev;
    P.Next := Target.Next;
    Target.Next := P;
  end
  else
  begin
    if Last = nil then
      P.Next := P
    else
    begin
      P.Next := Last.Next;
      Last.Next := P;
    end;
    Last := P;
  end;
end;

procedure TViGroup.Lock;
begin
  if (Buffer <> nil) or (LockFlag <> 0) then Inc(LockFlag);
end;

procedure TViGroup.Redraw;
begin
  DrawSubViews(First, nil);
end;

procedure TViGroup.RemoveView(P: TViView);
var
  View: TViView;
  Prev: TViView;
begin
  View := Last;
  if not Assigned(View) then Exit;
  repeat
    Prev := View;
    View := View.Next;
    if P = View then Break;
    if View = Last then Exit;
  until False;
  Prev.Next := P.Next;
  if Last = P then
  begin
    if P.Next = P then
      Prev := nil;
    Last := Prev;
  end;
end;

procedure TViGroup.ResetCurrent;
begin
  SetCurrent(FirstMatch(sfVisible, ofSelectable), NormalSelect);
end;

procedure TViGroup.ResetCursor;
begin
  if Current <> nil then Current.ResetCursor;
end;

procedure TViGroup.SelectNext(Forwards: Boolean);
var
  P: TViView;
begin
  P := FindNext(Forwards);
  if P <> nil then P.Select;
end;

procedure TViGroup.SetCurrent(P: TViView; Mode: SelectMode);

  procedure SelectView(P: TViView; Enable: Boolean);
  begin
    { select view only when group is also selected }
    if (State and sfSelected <> 0) and (P <> nil) then
      P.SetState(sfSelected, Enable);
  end;

  procedure FocusView(P: TViView; Enable: Boolean);
  begin
    if (State and sfFocused <> 0) and (P <> nil) then
      P.SetState(sfFocused, Enable);
  end;

begin
  if Current <> P then
  begin
    Lock;
    FocusView(Current, False);
    if Mode <> EnterSelect then SelectView(Current, False);
//    if Mode = NormalSelect then SelectView(Current, False);
    if Mode <> LeaveSelect then SelectView(P, True);
    FocusView(P, True);
    Current := P;
    Unlock;
  end
  else
  begin
    Lock;
    if Mode <> LeaveSelect then SelectView(P, True);
    FocusView(P, True);
    Unlock;
  end;
end;

procedure TViGroup.SetState(AState: Word; Enable: Boolean);

  procedure ForEachSetState;
  var
    Curr: TViView;
  begin
    Curr := Last;
    if not Assigned(Curr) then Exit;
    repeat
      Curr := Curr.Next;
      Curr.SetState(AState, Enable);
    until Curr = Last;
  end;

  procedure ForEachExpose;
  var
    Curr: TViView;
  begin
    Curr := Last;
    if not Assigned(Curr) then Exit;
    repeat
      Curr := Curr.Next;
      if Curr.State and sfVisible <> 0 then
        Curr.SetState(sfExposed, Enable);
    until Curr = Last;
  end;

begin
  inherited SetState(AState, Enable);
  case AState of
    sfActive, sfDragging:
      begin
        Lock;
        ForEachSetState;
        Unlock;
      end;
    sfSelected:
      if Current <> nil then Current.SetState(sfSelected, Enable);
    sfFocused:
      if Current <> nil then Current.SetState(sfFocused, Enable);
    sfExposed:
      begin
        ForEachExpose;
        if not Enable then FreeBuffer;
      end;
  end;
end;

procedure TViGroup.Unlock;
begin
  if LockFlag <> 0 then
  begin
    Dec(LockFlag);
    if LockFlag = 0 then DrawView;
  end;
end;

function TViGroup.Valid(Command: Word): Boolean;

  function IsInvalid(P: TViView): Boolean;
  begin
    IsInvalid := not P.Valid(Command);
  end;

  function FirstThatIsInvalid: TViView;
  begin
    Result := Last;
    if Result = nil then Exit;
    repeat
      Result := Result.Next;
      if IsInvalid(Result) then Exit;
    until Result = Last;
    Result := nil;
  end;

begin
  Result := True;
  if Command = cmReleasedFocus then
  begin
    if (Current <> nil) and (Current.Options and ofValidate <> 0) then
      Result := Current.Valid(Command);
  end
  else
    Result := (FirstThatIsInvalid = nil);
end;

function TViGroup.ExecMenuView(P: TViView): Word;
var
  SaveOptions: Word;
  SaveOwner: TViGroup;
  SaveTopView: TViView;
  SaveCurrent: TViView;
  SaveCommands: TCommandSet;
begin
  if P <> nil then
  begin
    SaveOptions := P.Options;
    SaveOwner := P.Owner;
    SaveTopView := TheTopView;
    SaveCurrent := Current;
    GetCommands(SaveCommands);
    TheTopView := P;
    P.Options := P.Options and not ofSelectable;
    P.SetState(sfModal, True);
    SetCurrent(P, NormalSelect);
    if SaveOwner = nil then Insert(P);
    Result := P.Execute;
    if SaveOwner = nil then Delete(P);
    SetCurrent(SaveCurrent, NormalSelect);
    P.SetState(sfModal, False);
    P.Options := SaveOptions;
    TheTopView := SaveTopView;
    SetCommands(SaveCommands);
  end
  else
    Result := cmCancel;
end;


{ TWindow }

constructor TViWindow.Create(var Bounds: TRect; ATitle: TTitleStr;
  ANumber: Integer);
begin
  inherited Create(Bounds);
  FPalettes[wpBlueWindow] := CBlueWindow;
  FPalettes[wpCyanWindow] := CCyanWindow;
  FPalettes[wpGrayWindow] := CGrayWindow;
  State := State or sfShadow;
  Options := Options or (ofSelectable + ofTopSelect);
  GrowMode := gfGrowAll + gfGrowRel;
  Flags := wfMove + wfGrow + wfClose + wfZoom;
  Title := ATitle;
  Number := ANumber;
  WindowPalette := wpBlueWindow;
  InitFrame;
  if Frame <> nil then Insert(Frame);
  GetBounds(ZoomRect);

  Constraints.A := MinWinSize;
  Constraints.B.X := MaxLongint;
  Constraints.B.Y := MaxLongint;
end;

destructor TViWindow.Destroy;
begin
  inherited Destroy;
end;

procedure TViWindow.Close;
begin
  if Valid(cmClose) then Free;
end;

function TViWindow.GetPalette: TPalette;
begin
  GetPalette := FPalettes[WindowPalette];
end;

procedure TViWindow.SetPalette(Value: TPalette);
begin
  if FPalettes[WindowPalette] <> Value then
  begin
    FPalettes[WindowPalette] := Value;
  end;
end;

function TViWindow.GetTitle: TTitleStr;
begin
  GetTitle := Title;
end;

procedure TViWindow.HandleEvent(var Event: TEvent);
var
  Limits: TRect;
  Min, Max: TPoint;
begin
  inherited HandleEvent(Event);
  if (Event.What = evCommand) then
    case Event.Command of
      cmResize:
        if Flags and (wfMove + wfGrow) <> 0 then
        begin
          Owner.GetExtent(Limits);
          SizeLimits(Min, Max);
          DragView(Event, DragMode or (Flags and (wfMove + wfGrow)),
            Limits, Min, Max);
          ClearEvent(Event);
        end;
      cmClose:
        if (Flags and wfClose <> 0) and
          ((Event.InfoPtr = nil) or (Event.InfoPtr = Self)) then
        begin
          ClearEvent(Event);
          if State and sfModal = 0 then
            Close
          else
          begin
            Event.What := evCommand;
            Event.Command := cmCancel;
            PutEvent(Event);
            ClearEvent(Event);
          end;
        end;
      cmZoom:
        if (Flags and wfZoom <> 0) and
          ((Event.InfoPtr = nil) or (Event.InfoPtr = Self)) then
        begin
          Zoom;
          ClearEvent(Event);
        end;
    end
  else if Event.What = evKeyDown then
    case Event.KeyCode of
      kbTab:
        begin
          FocusNext(False);
          ClearEvent(Event);
        end;
      kbShiftTab:
        begin
          FocusNext(True);
          ClearEvent(Event);
        end;
    end
  else if (Event.What = evBroadcast) and (Event.Command = cmSelectWindowNum)
    and (Event.InfoInt = Number) and (Options and ofSelectable <> 0) then
  begin
    Select;
    ClearEvent(Event);
  end;
end;

procedure TViWindow.InitFrame;
var
  R: TRect;
begin
  GetExtent(R);
  Frame := TViFrame.Create(R);
end;

procedure TViWindow.SetState(AState: Word; Enable: Boolean);
var
  WindowCommands: TCommandSet;
  Event: TEvent;
begin
  inherited SetState(AState, Enable);
  if AState = sfSelected then
    SetState(sfActive, Enable);
  if (AState = sfSelected) or ((AState = sfExposed) and
    (State and sfSelected <> 0)) then
  begin
    WindowCommands := [cmNext, cmPrev];
    if Flags and (wfGrow + wfMove) <> 0 then
      WindowCommands := WindowCommands + [cmResize];
    if Flags and wfClose <> 0 then
      WindowCommands := WindowCommands + [cmClose];
    if Flags and wfZoom <> 0 then
      WindowCommands := WindowCommands + [cmZoom];
    if Enable then
      EnableCommands(WindowCommands)
    else
      DisableCommands(WindowCommands);
  end;
  if (AState = sfDragging) and not Enable then
  begin
    Event.What := evCommand;
    Event.Command := cmZoomed;
    Event.InfoPtr := Self;
    PutEvent(Event);
  end;
end;

function TViWindow.StandardScrollBar(AOptions: Word): TViScrollBar;
var
  R: TRect;
  S: TViScrollBar;
begin
  GetExtent(R);
  if AOptions and sbVertical = 0 then
    RectAssign(R, R.A.X + 2, R.B.Y - 1, R.B.X - 2, R.B.Y)
  else
    RectAssign(R, R.B.X - 1, R.A.Y + 1, R.B.X, R.B.Y - 1);
  S := TViScrollBar.Create(R);
  Insert(S);
  if AOptions and sbHandleKeyboard <> 0 then
    S.Options := S.Options or ofPostProcess;
  Result := S;
end;

procedure TViWindow.SizeLimits(var Min, Max: TPoint);
begin
  inherited SizeLimits(Min, Max);
  Min.X := MinWinSize.X;
  Min.Y := MinWinSize.Y;
  if (Constraints.A.X <> 0) and (Constraints.A.Y <> 0) then
    Min := Constraints.A;
  if (Constraints.B.X <> MaxLongint) and (Constraints.B.Y <> MaxLongint)
      and (Constraints.B.X <> 0) and (Constraints.B.Y <> 0) then
    Max := Constraints.B;
end;

procedure TViWindow.Zoom;
var
  R: TRect;
  Max, Min: TPoint;
begin
  SizeLimits(Min, Max);
  if (Size.X <> Max.X) or (Size.Y <> Max.Y) then
  begin
    GetBounds(ZoomRect);
    R.A.X := 0;
    R.A.Y := 0;
    R.B := Max;
    Locate(R);
  end
  else
    Locate(ZoomRect);
end;

procedure TViWindow.ShowMaximized;
var
  R: TRect;
  Max, Min: TPoint;
begin
  SizeLimits(Min, Max);
  if (Size.X <> Max.X) or (Size.Y <> Max.Y) then
  begin
    GetBounds(ZoomRect);
    R.A.X := 0;
    R.A.Y := 0;
    R.B := Max;
    Locate(R);
  end;
end;

procedure TViWindow.ShowNormal;
var
  Max, Min: TPoint;
begin
  SizeLimits(Min, Max);
  if (Size.X = Max.X) and (Size.Y = Max.Y) then
    Locate(ZoomRect);
end;




{ Message dispatch function }

function SendMessage(Receiver: TViView; What, Command: Word; InfoPtr:
  Pointer): Pointer;
var
  Event: TEvent;
begin
  Result := nil;
  if Receiver <> nil then
  begin
    Event.What := What;
    Event.Command := Command;
    Event.InfoPtr := InfoPtr;
    Receiver.HandleEvent(Event);
    if Event.What = evNothing then Result := Event.InfoPtr;
  end;
end;

procedure TViClientFrame.Draw;
var
  CFrame: Word;
  I: Integer;
  B: TDrawBuffer;
begin
  if State and sfDragging <> 0 then
  begin
    CFrame := $0505;
  end
  else if State and sfActive = 0 then
  begin
    CFrame := $0101;
  end
  else
  begin
    CFrame := $0503;
  end;
  CFrame := GetColor(CFrame);
  for I := 0 to Size.Y - 1 do
  begin
    FrameLine(B, I, 3, Byte(CFrame));
    WriteLine(0, I, Size.X, 1, B);
  end;
end;

procedure TViClientFrame.FrameLine(var FrameBuf; Y, N: Integer; Color: Byte);

  const
    InitFrame: array[0..17] of Byte =
      ($06, $0A, $0C, $05, $00, $05, $03, $0A, $09,
       $16, $1A, $1C, $15, $00, $15, $13, $1A, $19);

var
  FrameMask: array[0..MaxViewWidth - 1] of Byte;
  I, W, Temp: Integer;
  AView: TViView;
  MaskL, MaskH: Byte;
  SourceIndex, DestIndex: Integer;
  LeftBorder, RightBorder: Boolean;
begin
(*
  if ViScreen.CodePage <> 437 then
  begin
    ldFrameChars[23] := '';
    ldFrameChars[27] := '';
    ldFrameChars[29] := '';
    ldFrameChars[30] := '';
  end;
*)
  W := Size.X;
//  FrameMask[0] := InitFrame[N];
  FillChar(FrameMask[0], W, InitFrame[N + 1]);
//  FrameMask[W - 1] := InitFrame[N + 2];

  AView := Self.Owner.Last;
  W := W - 1;
  repeat
    AView := AView.Next;
    if AView = Self then Break;
    if ((AView.Options and ofFramed) = 0) or
        ((AView.State and sfVisible) = 0) then Continue;
    Temp := Y - AView.Origin.Y;
    MaskL := $05;
    MaskH := $00;
    if Temp < 0 then
    begin
      if Temp <> -1 then Continue;
      MaskL := $06;
      MaskH := $0A;
    end
    else
    begin
      if Temp > AView.Size.Y then Continue;
      if Temp = AView.Size.Y then
      begin
        MaskL := $03;
        MaskH := $0A;
      end;
    end;
    SourceIndex := AView.Origin.X;
    DestIndex := SourceIndex + AView.Size.X;
    LeftBorder := False;
    RightBorder := False;
    if SourceIndex < 1 then
    begin
      SourceIndex := 1;
      LeftBorder := True;
    end;
    if DestIndex > W then
    begin
      DestIndex := W;
      RightBorder := True;
    end;
    if SourceIndex >= DestIndex then Continue;
    if not LeftBorder then
      FrameMask[SourceIndex - 1] := FrameMask[SourceIndex - 1] or MaskL;
    MaskL := MaskL xor MaskH;
    if not RightBorder then
      FrameMask[DestIndex] := FrameMask[DestIndex] or MaskL;
    if MaskH = 0 then Continue;

    if LeftBorder then SourceIndex := 0;
    if RightBorder then DestIndex := W + 1;

    for I := 0 to DestIndex - SourceIndex  - 1 do
      FrameMask[SourceIndex + I] := FrameMask[SourceIndex + I] or MaskH;
  until False;
  for I := 0 to W do
    FrameMask[I] := Byte(ldFrameChars[FrameMask[I]]);
  MoveBuf(FrameBuf, FrameMask, Color, W + 1);
end;


procedure TViCustomFrame.FrameLine(var FrameBuf; Y, N: Integer; Color: Byte);

  const
    InitFrame: array[0..17] of Byte =
      ($06, $0A, $0C, $05, $00, $05, $03, $0A, $09,
       $16, $1A, $1C, $15, $00, $15, $13, $1A, $19);

var
  FrameMask: array[0..MaxViewWidth - 1] of Byte;
  I, W, Temp: Integer;
  AView: TViView;
  MaskL, MaskH: Byte;
  SourceIndex, DestIndex: Integer;
begin
(*
  if ViScreen.CodePage <> 437 then
  begin
    ldFrameChars[23] := '';
    ldFrameChars[27] := '';
    ldFrameChars[29] := '';
    ldFrameChars[30] := '';
  end;
*)
  W := Size.X;
  FrameMask[0] := InitFrame[N];
  FillChar(FrameMask[1], W - 2, InitFrame[N + 1]);
  FrameMask[W - 1] := InitFrame[N + 2];

  AView := Self.Owner.Last;
  W := W - 1;
  repeat
    AView := AView.Next;
    if AView = Self then Break;
    if ((AView.Options and ofFramed) = 0) or
        ((AView.State and sfVisible) = 0) then Continue;
    Temp := Y - AView.Origin.Y;
    MaskL := $05;
    MaskH := $00;
    if Temp < 0 then
    begin
      if Temp <> -1 then Continue;
      MaskL := $06;
      MaskH := $0A;
    end
    else
    begin
      if Temp > AView.Size.Y then Continue;
      if Temp = AView.Size.Y then
      begin
        MaskL := $03;
        MaskH := $0A;
      end;
    end;
    SourceIndex := AView.Origin.X;
    DestIndex := SourceIndex + AView.Size.X;
    if SourceIndex < 1 then SourceIndex := 1;
    if DestIndex > W then DestIndex := W;
    if SourceIndex >= DestIndex then Continue;
    FrameMask[SourceIndex - 1] := FrameMask[SourceIndex - 1] or MaskL;
    MaskL := MaskL xor MaskH;
    FrameMask[DestIndex] := FrameMask[DestIndex] or MaskL;
    if MaskH = 0 then Continue;

    for I := 0 to DestIndex - SourceIndex  - 1 do
      FrameMask[SourceIndex + I] := FrameMask[SourceIndex + I] or MaskH;
  until False;
  for I := 0 to W do
    FrameMask[I] := Byte(ldFrameChars[FrameMask[I]]);
  MoveBuf(FrameBuf, FrameMask, Color, W + 1);
end;

constructor TViCustomFrame.Create(var Bounds: TRect);
begin
  inherited Create(Bounds);
  Palette := CFrame;
  GrowMode := gfGrowHiX + gfGrowHiY;
  EventMask := EventMask or evBroadcast;
end;

procedure TViCustomFrame.SetState(AState: Word; Enable: Boolean);
begin
  inherited SetState(AState, Enable);
  if AState and (sfActive + sfDragging) <> 0 then DrawView;
end;

function TViCustomFrame.Focus: Boolean;
begin
  Result := False;
//  if State and (sfSelected + sfModal) = 0 then
//  begin
//    if Owner <> nil then
//    begin
//      Result := Owner.Focus;
//      if Result then
//        if ((Owner.Current = nil) or
//          (Owner.Current.Options and ofValidate = 0) or
//          (Owner.Current.Valid(cmReleasedFocus))) then
//          Select
//        else
//          Result := False;
//    end;
//  end;
end;


procedure TViClientGroup.HandleEvent(var Event: TEvent);

  function FirstThatContainsMouse(AGroup: TViGroup): TViView;
  begin
    Result := AGroup.Last;
    if Result = nil then Exit;
    repeat
      Result := Result.Next;
      if (Result.State and sfVisible <> 0) and
          Result.MouseInView(Event.Where) then
      begin
        if Result is TViClientGroup then
          Result := FirstThatContainsMouse(TViClientGroup(Result));
        Exit;
      end;
    until Result = AGroup.Last;
    Result := nil;
  end;

var
  AView: TViView;
begin
  if Event.What and PositionalEvents <> 0 then
  begin
    AView := FirstThatContainsMouse(Self);
    if Assigned(AView) then
      if ((AView.Options and ofSelectable) = 0)
          or (AView is TViInputLine) then
      begin
        AView.HandleEvent(Event);
        Exit;
      end;
  end;
  inherited HandleEvent(Event);
end;


end.

