
{*******************************************************}
{                                                       }
{ DWPL Delphi WDosX Core Library                        }
{ WDosTimers                                            }
{                                                       }
{ for Delphi 4, 5                                       }
{ Version 1.2                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/wdosx/WDosTimers.pas,v 1.2 2002/10/27 22:25:26 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosTimers.pas,v $
  Revision 1.2  2002/10/27 22:25:26  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
{
  Version history:
    IWA 2001-06-18 Version 1.2.4.1 released
    IWA 11/20/00 Version 1.2
    - RealNow
      Bug with GetMode in Win9x/Me solved:
        WinMe only emulates a 8253 without mode
        readback. Now Mode 2 is forced.
        (Thanks to Lszl Agcs)
    IWA 11/20/00 Version 1.1
    - TwdxCustomTimer
      New Property Messages: Boolean default True;
        When True, OnTimer events are generated.
        Set False to inhibit messages for fast
        interrupt timers (<10 ms) so the message queue
        will not overrun. Only OnTimerInterrupt events
        are generated.
}

unit WDosTimers;

interface

uses
  SysUtils, Classes, WDosMessages, WDosSyncs, WDosInterrupts,
  WDosResStrings;

type
  TIntFreq =(ifNone, if32768, if16384, if8192,
             if4096, if2048,  if1024,  if512,
             if256,  if128,   if64,    if32,
             if16,   if8,     if4,     if2);

const
  DInterval =1000;
  DEnabled =True;
  DIntFreq =if64;

type
  TViCustomTimer = class(TObject)
  private
    FHandle: THandle;
    FElapsed: Cardinal;
    FEnabled: Boolean;
    FInterval: Cardinal;
    FOnTimer: TNotifyEvent;
    function GetBaseInterval: LongInt; virtual; abstract;
    procedure SetEnabled(const Value: Boolean);
    procedure WMTimer(var Msg: TMessage); message WM_TIMER;
  protected
    procedure DoTimer;
    procedure DoTimerInterrupt;
    procedure LockIntTimer; virtual; abstract;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Process(aInterval: LongInt);
    property BaseInterval: LongInt read GetBaseInterval;
    property Enabled: Boolean read FEnabled write SetEnabled default DEnabled;
    property Interval: Cardinal read FInterval write FInterval
        default DInterval;
    property OnTimer: TNotifyEvent read FOnTimer write FOnTimer;
  end;

  TViTimer = class(TViCustomTimer)
  private
    function GetBaseInterval: LongInt; override;
  protected
    procedure LockIntTimer; override;
  public
    constructor Create;
    destructor Destroy; override;
  end;

  TViRtcTimer = class(TViCustomTimer)
  private
    function GetBaseInterval: LongInt; override;
    function GetIntFreq: TIntFreq;
    procedure SetIntFreq(Value: TIntFreq);
  protected
    procedure LockIntTimer; override;
  public
    constructor Create;
    destructor Destroy; override;
    property IntFreq: TIntFreq read GetIntFreq write SetIntFreq
        default DIntFreq;
  end;

  TCustomIntTimer = class (TObject)
  private
    FEnableCnt: Integer;
    FInterval: LongInt;
    FOverrun: Boolean;
    FRecursion: Boolean;
    FTimerList: TList;
    FInterruptSync: TInterruptSync;
  protected
    procedure BeforeDestroy; virtual; abstract;
    procedure DoClose; virtual; abstract;
    procedure DoOpen; virtual; abstract;
    procedure DoTimer;
    procedure RegisterTimer(aTimer: TViCustomTimer);
    procedure ResetOutput; virtual; abstract;
    procedure SetDisabled(Updating: Boolean);
    procedure UnRegisterTimer(aTimer: TViCustomTimer);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Disable;
    procedure Enable;
    function Enabled: Boolean;
    property Interval: LongInt read FInterval write FInterval;
    property Overrun: Boolean read FOverrun write FOverrun;
  end;

  TIntTimer = class (TCustomIntTimer)
  private
    FInt28: DpmiPmVector;
  protected
    constructor CreateInstance(Dummy: Integer =0);
    class function AccessInstance(Request: Integer): TIntTimer;
    procedure DoClose; override;
    procedure DoOpen; override;
  public
    constructor Create;
    destructor Destroy; override;
    class function Instance: TIntTimer;
    class procedure ReleaseInstance;
  end;

  TRtcIntTimer = class (TCustomIntTimer)
  private
    FInt70: DpmiPmVector;
    FIntFreq: TIntFreq;
    FStatA: Byte;
    FStatB: Byte;
    procedure SetIntFreq(const Value: TIntFreq);
  protected
    constructor CreateInstance(Dummy: Integer = 0);
    class function AccessInstance(Request: Integer): TRtcIntTimer;
    procedure DoClose; override;
    procedure DoOpen; override;
  public
    constructor Create;
    destructor Destroy; override;
    class function Instance: TRtcIntTimer;
    class procedure ReleaseInstance;
    property IntFreq: TIntFreq read FIntFreq write SetIntFreq default DIntFreq;
  end;

implementation

uses
  WDosPorts;

const
  RTCAdrPort  =$70;
  RTCDataPort =$71;
  IntControl  =$A1;
  StatusA     =$0A;
  StatusB     =$0B;
  StatusC     =$0C;

var
  InterruptSync: TInterruptSync;

function RTCRead(Adresse: Integer): Integer;
begin
  if (Adresse <0) or (Adresse >63) then
    RTCRead :=-1
  else
  begin
    Port[RTCAdrPort] :=Adresse;
    RTCRead :=Port[RTCDataPort];
  end;
end;

procedure RTCWrite(Adresse: Integer; Inhalt: Byte);
begin
  if (Adresse <0) or (Adresse >63) then
    Exit
  else
  begin
    Port[RTCAdrPort] :=Adresse;
    Port[RTCDataPort] :=Inhalt;
  end;
end;

procedure IntTime;
begin
  try
    TIntTimer.Instance.DoTimer;
  finally
    { enable interrupts }
    Port[$20]:=$20;
  end;
end;

procedure RtcTime;
begin
  try
    TRtcIntTimer.Instance.DoTimer;
  finally
    { enable interrupts }
    Port[$A0]:=$20;
    Port[$20]:=$20;
    RtcRead(StatusC);
  end;
end;

{
************************************ TViCustomTimer ***************************
}
constructor TViCustomTimer.Create;
begin
  inherited Create;
  FInterval :=DInterval;
  FHandle := RegisterHandleObject(Self);
  Enabled :=DEnabled;
end;

destructor TViCustomTimer.Destroy;
begin
  Enabled :=False;
  UnregisterHandleObject(FHandle);
  inherited Destroy;
end;

procedure TViCustomTimer.DoTimer;
begin
  if Assigned(FOnTimer) then FOnTimer(Self);
end;

procedure TViCustomTimer.DoTimerInterrupt;
begin
  PostMessage(FHandle, WM_TIMER, 0, 0);
end;

procedure TViCustomTimer.Process(aInterval: LongInt);
begin
  if not Enabled then Exit;
  Inc(FElapsed, aInterval);
  if FElapsed >Interval *1000 then
  begin
    DoTimerInterrupt;
    FElapsed :=FElapsed -Interval *1000;
    if FElapsed >Interval *1000 then FElapsed :=0;
  end;
end;

procedure TViCustomTimer.SetEnabled(const Value: Boolean);
begin
  if FEnabled <> Value then
  begin
    FEnabled := Value;
    FElapsed :=0;
    LockIntTimer;
  end;
end;

procedure TViCustomTimer.WMTimer(var Msg: TMessage);
begin
  DoTimer;
end;

{
*************************************** TViTimer ******************************
}
constructor TViTimer.Create;
begin
  inherited Create;
  TIntTimer.Instance.RegisterTimer(Self);
end;

destructor TViTimer.Destroy;
begin
  TIntTimer.Instance.UnRegisterTimer(Self);
  inherited Destroy;
end;

function TViTimer.GetBaseInterval: LongInt;
begin
  Result :=TIntTimer.Instance.Interval;
end;

procedure TViTimer.LockIntTimer;
begin
  if Enabled then
    TIntTimer.Instance.Enable
  else
    TIntTimer.Instance.Disable;
end;

{
************************************** TViRtcTimer ****************************
}
constructor TViRtcTimer.Create;
begin
  inherited Create;
  TRtcIntTimer.Instance.RegisterTimer(Self);
end;

destructor TViRtcTimer.Destroy;
begin
  TRtcIntTimer.Instance.UnRegisterTimer(Self);
  inherited Destroy;
end;

function TViRtcTimer.GetBaseInterval: LongInt;
begin
  Result :=TRtcIntTimer.Instance.Interval;
end;

procedure TViRtcTimer.LockIntTimer;
begin
  if Enabled then
    TRtcIntTimer.Instance.Enable
  else
    TRtcIntTimer.Instance.Disable;
end;

function TViRtcTimer.GetIntFreq: TIntFreq;
begin
  Result :=TRtcIntTimer.Instance.IntFreq;
end;

procedure TViRtcTimer.SetIntFreq(Value: TIntFreq);
begin
  TRtcIntTimer.Instance.IntFreq :=Value;
end;

{
************************************ TCustomIntTimer ***************************
}
constructor TCustomIntTimer.Create;
begin
  inherited Create;
  FInterruptSync := TInterruptSync.Create;
  FTimerList :=TList.Create;
end;

destructor TCustomIntTimer.Destroy;
begin
  FInterruptSync.Free;
  FTimerList.Free;
  inherited Destroy;
end;

procedure TCustomIntTimer.Disable;
begin
  Dec(FEnableCnt);
  if FEnableCnt = 0 then SetDisabled(True);
end;

procedure TCustomIntTimer.DoTimer;
var
  I: Integer;
begin
  if FRecursion then
  begin
    FOverrun :=True;
    Exit;
  end;
  FRecursion :=True;
  { update all timers }
  for I :=0 to FTimerList.Count -1 do
    TViCustomTimer(FTimerList[I]).Process(Interval);
  FRecursion :=False;
end;

procedure TCustomIntTimer.Enable;
begin
  Inc(FEnableCnt);
  if FEnableCnt = 1 then SetDisabled(False);
end;

function TCustomIntTimer.Enabled: Boolean;
begin
  Result := (FEnableCnt <> 0);
end;

procedure TCustomIntTimer.RegisterTimer(aTimer: TViCustomTimer);
begin
  FInterruptSync.Lock;
  try
    FTimerList.Add(aTimer);
  finally
    FInterruptSync.UnLock;
  end;
end;

procedure TCustomIntTimer.SetDisabled(Updating: Boolean);
begin
  if Updating then DoClose else DoOpen;
end;

procedure TCustomIntTimer.UnRegisterTimer(aTimer: TViCustomTimer);
begin
  FInterruptSync.Lock;
  try
    FTimerList.Remove(aTimer);
  finally
    FInterruptSync.UnLock;
  end;
end;

{
*************************************** TIntTimer ******************************
}
constructor TIntTimer.Create;
begin
  inherited Create;
  raise Exception.CreateFmt(SAccessThroughInstanceOnlyError, [ClassName]);
end;

constructor TIntTimer.CreateInstance(Dummy: Integer =0);
begin
  inherited Create;
  Interval :=54925; { timer tick x 1000 }
end;

destructor TIntTimer.Destroy;
begin
  if Enabled then DoClose;
  if AccessInstance(0) = Self then AccessInstance(2);
  inherited Destroy;
end;

var
  FIntInstance: TIntTimer = nil;

class function TIntTimer.AccessInstance(Request: Integer): TIntTimer;
begin
  case Request of
    0 : ;
    1 : if not Assigned(FIntInstance) then FIntInstance := CreateInstance;
    2 : FIntInstance := nil;
  else
    raise Exception.CreateFmt(SRequestAccessInstanceError, [Request]);
  end;
  Result := FIntInstance;
end;

procedure TIntTimer.DoClose;
begin
  SetIntVec(28, FInt28);
end;

procedure TIntTimer.DoOpen;
begin
  GetIntVec(28, FInt28);
  SetIntProc(28, @IntTime);
end;

class function TIntTimer.Instance: TIntTimer;
begin
  Result := AccessInstance(1);
end;

class procedure TIntTimer.ReleaseInstance;
begin
  AccessInstance(0).Free;
end;

{
************************************** TRtcIntTimer ****************************
}
constructor TRtcIntTimer.Create;
begin
  inherited Create;
  raise Exception.CreateFmt(SAccessThroughInstanceOnlyError, [ClassName]);
end;

constructor TRtcIntTimer.CreateInstance(Dummy: Integer = 0);
begin
  inherited Create;
  IntFreq :=DIntFreq;
end;

destructor TRtcIntTimer.Destroy;
begin
  if Enabled then DoClose;
  if AccessInstance(0) = Self then AccessInstance(2);
  inherited Destroy;
end;

var
  FRtcInstance: TRtcIntTimer = nil;

class function TRtcIntTimer.AccessInstance(Request: Integer): TRtcIntTimer;
begin
  case Request of
    0 : ;
    1 : if not Assigned(FRtcInstance) then FRtcInstance := CreateInstance;
    2 : FRtcInstance := nil;
  else
    raise Exception.CreateFmt(SRequestAccessInstanceError, [Request]);
  end;
  Result := FRtcInstance;
end;

procedure TRtcIntTimer.DoClose;

  var
    Reg: Byte;

begin
  { disable RTC interrupt on interrupt controller }
  Reg :=Port[IntControl];
  Reg :=Reg or $01;
  Port[IntControl] :=Reg;

  { restore old RTC status }
  RtcWrite(StatusA, FStatA);
  RtcWrite(StatusB, FStatB);

  { restore old interrupt vector }
  SetIntVec($70, FInt70);
end;

procedure TRtcIntTimer.DoOpen;

  var
    Reg: Byte;

begin
  { store old RTC status }
  FStatA :=RtcRead(StatusA);
  FStatB :=RtcRead(StatusB);

  { set interrupt frequence to new value }
  Reg :=FStatA and $F0;
  Reg :=Reg or Byte(FIntFreq);
  RtcWrite(StatusA, Reg);

  { enable cyclic interrupts }
  RtcWrite(StatusB, FStatB or $40);

  { redirect interrupt vector on own procedure }
  GetIntVec($70, FInt70);
  SetIntProc($70, @RtcTime);

  { enable RTC interrupt at interrupt controller }
  Reg :=Port[IntControl];
  Reg :=Reg and $FE;
  Port[IntControl] :=Reg;
end;

class function TRtcIntTimer.Instance: TRtcIntTimer;
begin
  Result := AccessInstance(1);
end;

class procedure TRtcIntTimer.ReleaseInstance;
begin
  AccessInstance(0).Free;
end;

procedure TRtcIntTimer.SetIntFreq(const Value: TIntFreq);
  
  const
    IntervalVal: array[TIntFreq] of LongInt =
      (   0,     31,     61,    122,
        244,    488,    977,   1953,
       3906,   7812,  15625,  31250,
      62500, 125000, 250000, 500000);
  
begin
  if FIntFreq <> Value then
  begin
    if Enabled then DoClose;
    FIntFreq := Value;
    Interval :=IntervalVal[FIntFreq];
    if Enabled then DoOpen;
  end;
end;

initialization
  InterruptSync := TInterruptSync.Create;
  TIntTimer.Instance;
  TRtcIntTimer.Instance;

finalization
  TIntTimer.ReleaseInstance;
  TRtcIntTimer.ReleaseInstance;
  FreeAndNil(InterruptSync);
end.
