{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/win/WinMessages.pas,v 1.3 2004/03/30 20:23:48 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WinMessages.pas,v $
  Revision 1.3  2004/03/30 20:23:48  iwache
  Memory leak bug fixed. New call to
  procedure FreeInstanceBlocks in
  the finalization section added.

  Revision 1.2  2002/10/27 22:25:26  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
unit WinMessages;

interface

uses
  Windows;

const
  WM_TIMER            = $0113;

type

{ Generic window message record }

  PMessage = ^TMessage;
  TMessage = packed record
    Msg: Cardinal;
    case Integer of
      0: (
        WParam: Longint;
        LParam: Longint;
        Result: Longint);
      1: (
        WParamLo: Word;
        WParamHi: Word;
        LParamLo: Word;
        LParamHi: Word;
        ResultLo: Word;
        ResultHi: Word);
  end;

  TWndMethod = procedure(var Message: TMessage) of object;

function ProcessMessage: Boolean;

function MakeObjectInstance(Method: TWndMethod): Pointer;

procedure FreeObjectInstance(ObjectInstance: Pointer);

function AllocateHWnd(Method: TWndMethod): HWND;

procedure DeallocateHWnd(Wnd: HWND);

implementation


function PeekMessage(var lpMsg: TMsg; hWnd: HWND;
    wMsgFilterMin, wMsgFilterMax, wRemoveMsg: UINT): BOOL; stdcall;
    external user32 name 'PeekMessageA';

function DispatchMessage(const lpMsg: TMsg): Longint; stdcall;
    external user32 name 'DispatchMessageA';

function TranslateMessage(const lpMsg: TMsg): BOOL; stdcall;
    external user32 name 'TranslateMessage';

function DefWindowProc(hWnd: HWND; Msg: UINT; wParam: WPARAM; lParam:
    LPARAM): LRESULT; stdcall;
    external user32 name 'DefWindowProcA';

function GetClassInfo(hInstance: HINST; lpClassName: PChar;
    var lpWndClass: TWndClass): BOOL; stdcall;
    external user32 name 'GetClassInfoA';

function RegisterClass(const lpWndClass: TWndClass): ATOM; stdcall;
    external user32 name 'RegisterClassA';

function UnregisterClass(lpClassName: PChar; hInstance: HINST): BOOL; stdcall;
    external user32 name 'UnregisterClassA';

function CreateWindowEx(dwExStyle: DWORD; lpClassName: PChar; lpWindowName:
    PChar; dwStyle: DWORD; X, Y, nWidth, nHeight: Integer; hWndParent: HWND;
    hMenu: HMENU; hInstance: HINST; lpParam: Pointer): HWND; stdcall;
    external user32 name 'CreateWindowExA';

function GetWindowLong(hWnd: HWND; nIndex: Integer): Longint; stdcall;
    external user32 name 'GetWindowLongA';

function SetWindowLong(hWnd: HWND; nIndex: Integer; dwNewLong: Longint): Longint; stdcall;
    external user32 name 'SetWindowLongA';


function ProcessMessage: Boolean;
var
  Msg: TMsg;
begin
  Result := False;
  if PeekMessage(Msg, 0, 0, 0, PM_REMOVE) then
  begin
    Result := True;
    TranslateMessage(Msg);
    DispatchMessage(Msg);
  end;
end;

const
  InstanceCount = 313;

{ Object instance management }

type
  PObjectInstance = ^TObjectInstance;
  TObjectInstance = packed record
    Code: Byte;
    Offset: Integer;
    case Integer of
      0: (Next: PObjectInstance);
      1: (Method: TWndMethod);
  end;

type
  PInstanceBlock = ^TInstanceBlock;
  TInstanceBlock = packed record
    Next: PInstanceBlock;
    Code: array[1..2] of Byte;
    WndProcPtr: Pointer;
    Instances: array[0..InstanceCount] of TObjectInstance;
  end;

var
  InstBlockList: PInstanceBlock;
  InstFreeList: PObjectInstance;

{ Standard window procedure }
{ In    ECX = Address of method pointer }
{ Out   EAX = Result }

function StdWndProc(Window: HWND; Message, WParam: Longint;
  LParam: Longint): Longint; stdcall; assembler;
asm
        XOR     EAX,EAX
        PUSH    EAX
        PUSH    LParam
        PUSH    WParam
        PUSH    Message
        MOV     EDX,ESP
        MOV     EAX,[ECX].Longint[4]
        CALL    [ECX].Pointer
        ADD     ESP,12
        POP     EAX
end;

{ Allocate an object instance }

function CalcJmpOffset(Src, Dest: Pointer): Longint;
begin
  Result := Longint(Dest) - (Longint(Src) + 5);
end;

function MakeObjectInstance(Method: TWndMethod): Pointer;
const
  BlockCode: array[1..2] of Byte = (
    $59,       { POP ECX }
    $E9);      { JMP StdWndProc }
  PageSize = 4096;
var
  Block: PInstanceBlock;
  Instance: PObjectInstance;
begin
  if InstFreeList = nil then
  begin
    Block := VirtualAlloc(nil, PageSize, MEM_COMMIT, PAGE_EXECUTE_READWRITE);
    Block^.Next := InstBlockList;
    Move(BlockCode, Block^.Code, SizeOf(BlockCode));
    Block^.WndProcPtr := Pointer(CalcJmpOffset(@Block^.Code[2], @StdWndProc));
    Instance := @Block^.Instances;
    repeat
      Instance^.Code := $E8;  { CALL NEAR PTR Offset }
      Instance^.Offset := CalcJmpOffset(Instance, @Block^.Code);
      Instance^.Next := InstFreeList;
      InstFreeList := Instance;
      Inc(Longint(Instance), SizeOf(TObjectInstance));
    until Longint(Instance) - Longint(Block) >= SizeOf(TInstanceBlock);
    InstBlockList := Block;
  end;
  Result := InstFreeList;
  Instance := InstFreeList;
  InstFreeList := Instance^.Next;
  Instance^.Method := Method;
end;

{ Free an object instance }

procedure FreeObjectInstance(ObjectInstance: Pointer);
begin
  if ObjectInstance <> nil then
  begin
    PObjectInstance(ObjectInstance)^.Next := InstFreeList;
    InstFreeList := ObjectInstance;
  end;
end;

{ Free all allocated InstanceBlocks }

procedure FreeInstanceBlocks;
var
  Next: PInstanceBlock;
begin
  while InstBlockList <> nil do
  begin
    Next := InstBlockList^.Next;
    VirtualFree(InstBlockList, 0, MEM_RELEASE);
    InstBlockList := Next;
  end;
end;

var
  UtilWindowClass: TWndClass = (
    style: 0;
    lpfnWndProc: @DefWindowProc;
    cbClsExtra: 0;
    cbWndExtra: 0;
    hInstance: 0;
    hIcon: 0;
    hCursor: 0;
    hbrBackground: 0;
    lpszMenuName: nil;
    lpszClassName: 'TPUtilWindow');

function AllocateHWnd(Method: TWndMethod): HWND;
var
  TempClass: TWndClass;
  ClassRegistered: Boolean;
begin
  UtilWindowClass.hInstance := HInstance;
{$IFDEF PIC}
  UtilWindowClass.lpfnWndProc := @DefWindowProc;
{$ENDIF}
  ClassRegistered := GetClassInfo(HInstance, UtilWindowClass.lpszClassName,
    TempClass);
  if not ClassRegistered or (TempClass.lpfnWndProc <> @DefWindowProc) then
  begin
    if ClassRegistered then
      UnregisterClass(UtilWindowClass.lpszClassName, HInstance);
    RegisterClass(UtilWindowClass);
  end;
  Result := CreateWindowEx(WS_EX_TOOLWINDOW, UtilWindowClass.lpszClassName,
    '', WS_POPUP {!0}, 0, 0, 0, 0, 0, 0, HInstance, nil);
  if Assigned(Method) then
    SetWindowLong(Result, GWL_WNDPROC, Longint(MakeObjectInstance(Method)));
end;

procedure DeallocateHWnd(Wnd: HWND);
var
  Instance: Pointer;
begin
  Instance := Pointer(GetWindowLong(Wnd, GWL_WNDPROC));
  DestroyWindow(Wnd);
  if Instance <> @DefWindowProc then FreeObjectInstance(Instance);
end;

initialization
  InstBlockList := nil;
  InstFreeList := nil;

finalization
  FreeInstanceBlocks;

end.
