
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (clx/vi)      }
{ WinMouses.pas                                         }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - CRT unit from Free Pascal RTL                     }
{     Copyright (c) 1999 by                             }
{     The Free Pascal Organization                      }
{     www: http://www.freepascal.org                    }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/win/WinMouses.pas,v 1.4 2002/10/27 22:25:26 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WinMouses.pas,v $
  Revision 1.4  2002/10/27 22:25:26  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
{
  Version history:
    IWA 2001-05-24 Version 2.0.1.0 released
}

unit WinMouses;

interface

procedure ShowMouse;

procedure HideMouse;

procedure ProcessConsoleEvents;

procedure InitMouse;

procedure DoneMouse;

var
  MouseButtonCount: Integer;
  MouseButtons: Integer;
  MouseWhereX: Integer;
  MouseWhereY: Integer;
  MouseTime: Cardinal;

  KeyboardShiftState: Byte;
  KeyboardScanCode: Byte;
  KeyboardCharCode: Char;
implementation

uses
  Windows, WinScreens, ViTypes;

procedure ShowMouse;
var
  Mode: DWord;
begin
  if GetConsoleMode(InHandle, Mode) then
  begin
    Mode := Mode or ENABLE_MOUSE_INPUT;
    SetConsoleMode(InHandle, Mode);
  end;
end;

procedure HideMouse;
var
  Mode: DWord;
begin
  if GetConsoleMode(InHandle, Mode) then
  begin
    Mode := Mode and not ENABLE_MOUSE_INPUT;
    SetConsoleMode(InHandle, Mode);
  end;
end;

procedure HandleMouseEvent(InputRecord: TInputRecord);
begin
  with InputRecord.Event.MouseEvent do
  begin
    MouseTime := GetTickCount;
    MouseButtons := dwButtonState;
    MouseWhereX := dwMousePosition.X;
    MouseWhereY := dwMousePosition.Y;
  end;
end;

var
  DoingNumChars: Boolean = False;
  DoingNumCode: Byte = 0;

procedure HandleKeyEvent(InputRecord: TInputRecord);
var
  ControlKeyState: LongWord;
  AltKey, CtrlKey, ShiftKey: Boolean;
  VirtualKeyCode: Word;

  procedure MapCharZero;

    const
      CtrlKeypadKeys: array[$47..$53] of Byte =(
          $77, $8D, $84, $8E, $73, $8F, $74, $4E,
          $75, $91, $76, $92, $93);

  begin
    if AltKey then
    begin
      case VirtualKeycode of
        VK_NUMPAD0..VK_NUMPAD9:
          begin
            DoingNumChars := True;
            DoingNumCode := Byte((DoingNumCode * 10)
                + (VirtualKeyCode - VK_NUMPAD0));
          end;
      end;
      case KeyboardScanCode of
        $02..$0D: Inc(KeyboardScanCode, $76);
        $3B..$44: Inc(KeyboardScanCode, $2D);
        $57..$58: Inc(KeyboardScanCode, $34);
        $47..$49, $4B, $4D, $4F..$53: Inc(KeyboardScanCode, $50);
        $1C: KeyboardScanCode := $A6;
        $35: KeyboardScanCode := $A4;
      end;
    end
    else if CtrlKey then
      case KeyboardScanCode of
        $0F: KeyboardScanCode := $94;
        $3B..$44: Inc(KeyboardScanCode, $23);
        $57..$58: Inc(KeyboardScanCode, $32);
        $35: KeyboardScanCode := $95;
        $37: KeyboardScanCode := $72;
        $47..$53: KeyboardScanCode := CtrlKeypadKeys[KeyboardScanCode];
      end
    else if ShiftKey then
      case KeyboardScanCode of
        $3B..$44: Inc(KeyboardScanCode, $19);
        $57..$58: Inc(KeyboardScanCode, $30);
      end
    else
      case KeyboardScanCode of
        $57..$58: Inc(KeyboardScanCode, $2E);
      end;
  end;

  procedure MapAltKeys;
  begin
    case KeyboardScanCode of
      8..12, 16, 27, 86:
      begin
        if Pos(KeyboardCharCode, '{[]}\@~|') = 0 then
          KeyboardCharCode := #0;
      end;
      else
      begin
        KeyboardCharCode := #0;
        case KeyboardScanCode of
          $02..$0D: Inc(KeyboardScanCode, $76);
          $0E: Dec(KeyboardScanCode, $06);
    (*
          $3B..$44: Inc(KeyboardScanCode, $2D);
          $57..$58: Inc(KeyboardScanCode, $34);
          $47..$49, $4B, $4D, $4F..$53: Inc(KeyboardScanCode, $50);
          $1C: KeyboardScanCode := $A6;
          $35: KeyboardScanCode := $A4;
    *)
        end;
      end;
    end;
  end;

  procedure MapCtrlKeys;
  begin
  end;

  procedure MapShiftKeys;
  begin
    if KeyboardScanCode = $0F then
      KeyboardCharCode := #0;
  end;

begin
  ControlKeyState := InputRecord.Event.KeyEvent.dwControlKeyState;

  KeyboardShiftState := 0;
  AltKey := ((ControlKeyState
      and (RIGHT_ALT_PRESSED or LEFT_ALT_PRESSED)) > 0);
  if AltKey then
    KeyboardShiftState := KeyboardShiftState or kbAltShift;
  CtrlKey := ((ControlKeyState
      and (RIGHT_CTRL_PRESSED or LEFT_CTRL_PRESSED)) > 0);
  if CtrlKey then
    KeyboardShiftState := KeyboardShiftState or kbCtrlShift;
  ShiftKey := ((ControlKeyState and SHIFT_PRESSED) > 0);
  if ShiftKey then
    KeyboardShiftState := KeyboardShiftState or kbLeftShift;
  if (ControlKeyState and NUMLOCK_ON) > 0 then
    KeyboardShiftState := KeyboardShiftState or kbNumState;
  if (ControlKeyState and SCROLLLOCK_ON) > 0 then
    KeyboardShiftState := KeyboardShiftState or kbScrollState;
  if (ControlKeyState and CAPSLOCK_ON) > 0 then
    KeyboardShiftState := KeyboardShiftState or kbCapsState;
  { TODO -oiwa : care of kbInsState key state }

  if InputRecord.Event.KeyEvent.bKeyDown then
  begin
    VirtualKeyCode := InputRecord.Event.KeyEvent.wVirtualKeyCode;

    if VirtualKeyCode in
        [VK_SHIFT, VK_MENU, VK_CONTROL, VK_CAPITAL,
        VK_NUMLOCK, VK_SCROLL] then Exit;
    KeyboardCharCode := InputRecord.Event.KeyEvent.AsciiChar;
    KeyboardScanCode := InputRecord.Event.KeyEvent.wVirtualScanCode;
    if KeyboardCharCode = #0 then
      MapCharZero
    else if AltKey then
      MapAltKeys
    else if CtrlKey then
      MapCtrlKeys
    else if ShiftKey then
      MapShiftKeys
    else (* nothing *);
  end
  else
  begin
  end;
end;

procedure ProcessConsoleEvents;
var
  I: Integer;
  EventCount, EventsRead: DWORD;
  InputRecord: TInputRecord;
begin
  EventCount := 0;
  GetNumberOfConsoleInputEvents(InHandle, EventCount);
  for I := 1 to EventCount do
  begin
    ReadConsoleInput(InHandle, InputRecord, 1, EventsRead);
    if InputRecord.EventType = KEY_EVENT then
      HandleKeyEvent(InputRecord)
    else if InputRecord.EventType = _MOUSE_EVENT then
      HandleMouseEvent(InputRecord);
  end;
end;

procedure InitMouse;
var
  AButtonCount: DWord;
  Mode: DWord;
begin
  if GetConsoleMode(InHandle, Mode) then
  begin
    Mode := Mode and not ENABLE_PROCESSED_INPUT
        and not ENABLE_MOUSE_INPUT;
    SetConsoleMode(InHandle, Mode);
  end;

  if not GetNumberOfConsoleMouseButtons(AButtonCount) then AButtonCount := 0;
  MouseButtonCount := AButtonCount;
end;

procedure DoneMouse;
begin
end;

end.

