
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (clx/vi)      }
{ WinScreens.pas                                        }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - CRT unit from Free Pascal RTL                     }
{     Copyright (c) 1999 by                             }
{     The Free Pascal Organization                      }
{     www: http://www.freepascal.org                    }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/clx/wq/vi/win/WinScreens.pas,v 1.7 2004/03/30 20:23:28 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WinScreens.pas,v $
  Revision 1.7  2004/03/30 20:23:28  iwache
  Memory leak in procedure DoneCrt fixed.

  Revision 1.6  2003/01/25 18:54:45  iwache
  Procedure InitCrt modified to set default US code page
  SetConsoleOutputCP(437) for correct display of frame
  border lines in DWPL Windows mode.

  Revision 1.5  2003/01/23 19:20:40  iwache
  - Source code in InitCrt beautified a little bit. Sorry Laszlo
  I could not resist :-)

  Revision 1.4  2003/01/23 11:27:43  lagocs
  screen buffer size is set to 80x25 during InitCrt in the native Win32 console version

  Revision 1.3  2002/10/27 22:25:26  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
unit WinScreens;

interface

uses
  Windows, SysUtils, ViTypes;

var
  TextAttr: Byte;
  WindMin, WindMax: Word;
  OutHandle, InHandle: THandle;
  ScreenWidth, ScreenHeight: Longint;
  CrtInited: Boolean;
  DefCursorSize: Longint;

procedure AssignCrt(var F: Text);
procedure SetTextMode(Mode: Integer);
procedure GotoXY(X, Y: Byte);
function WhereX: Byte;
function WhereY: Byte;
procedure ClrScr;
function GetCursorSize: Longint;
procedure SetCursorSize(Size: Longint);
procedure CursorOn;
procedure CursorOff;
procedure CursorBig;
procedure CursorNorm;
procedure InitCrt;
procedure DoneCrt;

procedure WriteScr(var Src: TScr; rCoordAttr: Boolean);
function GetTextMode: Word;

implementation

uses
  WinMouses;

var
  IsWindowsNT: Boolean;
  LastMode: Integer;

function GetScreenHeight: Longint;
var
  ConsoleInfo: TConsoleScreenBufferInfo;
begin
  Getconsolescreenbufferinfo(OutHandle, ConsoleInfo);
  Result := ConsoleInfo.dwSize.Y;
end;

function GetScreenWidth: Longint;
var
  ConsoleInfo: TConsoleScreenBufferInfo;
begin
  Getconsolescreenbufferinfo(OutHandle, ConsoleInfo);
  Result := ConsoleInfo.dwSize.X;
end;

var
  CursorSaveX, CursorSaveY: Longint;

procedure SetScreenCursor(X, Y: Longint);
var
  CurInfo: TCoord;
begin
  FillChar(CurInfo, Sizeof(CurInfo), 0);
  CurInfo.X := X - 1;
  CurInfo.Y := Y - 1;
  Setconsolecursorposition(OutHandle, CurInfo);
  CursorSaveX := X - 1;
  CursorSaveY := Y - 1;
end;

procedure GetScreenCursor(var X, Y: Longint);
begin
  X := CursorSaveX + 1;
  Y := CursorSaveY + 1;
end;

function WinMinX: Byte;
begin
  Result := (WindMin and $FF) + 1;
end;

function WinMinY: Byte;
begin
  Result := (WindMin shr 8) + 1;
end;

function WinMaxX: Byte;
begin
  Result := (WindMax and $FF) + 1;
end;

function WinMaxY: Byte;
begin
  Result := (WindMax shr 8) + 1;
end;

procedure GotoXY(X, Y: Byte);
begin
  if (X > 0) and (X <= WinMaxX - WinMinX + 1) and
    (Y > 0) and (Y <= WinMaxY - WinMinY + 1) then
  begin
    Inc(X, WinMinX - 1);
    Inc(Y, WinMinY - 1);
    SetScreenCursor(X, Y);
  end;
end;

procedure ClrScr;
var
  ClipRect: TSmallRect;
  SrcRect: TSmallRect;
  DestCoor: TCoord;
  CharInfo: TCharInfo;
begin
  CharInfo.Unicodechar := #32;
  CharInfo.attributes := TextAttr;
  SrcRect.Left := WinMinX - 1;
  SrcRect.Top := WinMinY - 1;
  SrcRect.Right := WinMaxX - 1;
  SrcRect.Bottom := WinMaxY - 1;
  ClipRect := SrcRect;
  if IsWindowsNT then
  begin
    DestCoor.X := -WinMaxX;
    DestCoor.Y := -WinMaxY;
    ScrollConsoleScreenBuffer(OutHandle, SrcRect, @ClipRect, DestCoor,
      CharInfo);
  end
  else
  begin
    DestCoor.X := WinMinX - 1;
    DestCoor.Y := WinMinY - (Succ((WinMaxY - WinMinY) div 2));
    ScrollConsoleScreenBuffer(OutHandle, SrcRect, @ClipRect, DestCoor,
      CharInfo);
    ScrollConsoleScreenBuffer(OutHandle, SrcRect, @ClipRect, DestCoor,
      CharInfo);
    ScrollConsoleScreenBuffer(OutHandle, SrcRect, @ClipRect, DestCoor,
      CharInfo);
  end;
  GotoXY(1, 1);
end;

function WhereX: Byte;
var
  X, Y: Longint;
begin
  GetScreenCursor(X, Y);
  Result := X - WinMinX + 1;
end;

function WhereY: Byte;
var
  X, Y: Longint;
begin
  GetScreenCursor(X, Y);
  Result := Y - WinMinY + 1;
end;

procedure RemoveLine(Y: Longint);
var
  ClipRect: TSmallRect;
  SrcRect: TSmallRect;
  DestCoor: TCoord;
  CharInfo: TCharInfo;
begin
  CharInfo.Unicodechar := #32;
  CharInfo.attributes := TextAttr;
  Y := WinMinY + Y - 1;
  SrcRect.Top := Y - 1;
  SrcRect.Left := WinMinX - 1;
  SrcRect.Right := WinMaxX - 1;
  SrcRect.Bottom := WinMaxY - 1;
  DestCoor.X := WinMinX - 1;
  DestCoor.Y := Y - 2;
  ClipRect := SrcRect;
  ScrollConsoleScreenBuffer(OutHandle, SrcRect, @ClipRect, DestCoor, CharInfo);
end;

function GetCursorSize: Longint;
var
  CursorInfo: TConsoleCursorInfo;
begin
  Getconsolecursorinfo(OutHandle, CursorInfo);
  Result := CursorInfo.dwSize;
end;

procedure SetCursorSize(Size: Longint);
var
  CursorInfo: TConsoleCursorInfo;
begin
  Getconsolecursorinfo(OutHandle, CursorInfo);
  CursorInfo.dwSize := Size;
  CursorInfo.bVisible := True;
  Setconsolecursorinfo(OutHandle, CursorInfo);
end;

procedure CursorOn;
begin
  SetCursorSize(DefCursorSize);
end;

procedure CursorOff;
var
  CursorInfo: TConsoleCursorInfo;
begin
  Getconsolecursorinfo(OutHandle, CursorInfo);
  CursorInfo.bVisible := False;
  Setconsolecursorinfo(OutHandle, CursorInfo);
end;

procedure CursorBig;
begin
  SetCursorSize(99);
end;

procedure CursorNorm;
begin
  SetCursorSize(12);
end;

var
  CurrX, CurrY: Longint;

procedure WriteChar(C: Char);
var
  Cell: TCharInfo;
  BufSize: TCoord;
  WritePos: TCoord;
  DestRect: TSmallRect;
begin
  case C of
    #10: Inc(CurrY);
    #13: CurrX := WinMinX;
    #08:
      if CurrX > WinMinX then Dec(CurrX);
    #07: ;
  else
    begin
      BufSize.X := 1;
      BufSize.Y := 1;
      WritePos.X := 0;
      WritePos.Y := 0;
      Cell.Unicodechar := Widechar(C);
      Cell.attributes := TextAttr;
      DestRect.Left := CurrX - 1;
      DestRect.Top := CurrY - 1;
      DestRect.Right := CurrX - 1;
      DestRect.Bottom := CurrY - 1;
      Writeconsoleoutput(OutHandle, @Cell, BufSize, WritePos, DestRect);
      Inc(CurrX);
    end;
  end;
  if CurrX > WinMaxX then
  begin
    CurrX := WinMinX;
    Inc(CurrY);
  end;
  while CurrY > WinMaxY do
  begin
    RemoveLine(1);
    Dec(CurrY);
  end;
end;

function CrtWrite(var F: TTextRec): Integer;
var
  I: Longint;
begin
  GetScreenCursor(CurrX, CurrY);
  for I := 0 to F.BufPos - 1 do
    WriteChar(F.Buffer[I]);
  SetScreenCursor(CurrX, CurrY);
  F.BufPos := 0;
  Result := 0;
end;

function CrtRead(var F: TTextRec): Integer;
begin
  F.BufPos := 0;
  F.BufEnd := 0;
  F.bufPtr[F.BufEnd] := #13;
  F.bufPtr[F.BufEnd + 1] := #10;
  Inc(F.BufEnd, 2);
  F.BufPos := 0;
  Result := 0;
end;

function CrtReturn(var F: TTextRec): Integer;
begin
  Result := 0;
end;

function CrtClose(var F: TTextRec): Integer;
begin
  F.Mode := fmClosed;
  Result := 0;
end;

function CrtOpen(var F: TTextRec): Integer;
begin
  if F.Mode = fmOutput then
  begin
    TTextRec(F).inoutfunc := @CrtWrite;
    TTextRec(F).Flushfunc := @CrtWrite;
  end
  else
  begin
    F.Mode := fmInput;
    TTextRec(F).inoutfunc := @CrtRead;
    TTextRec(F).Flushfunc := @CrtReturn;
  end;
  TTextRec(F).CloseFunc := @CrtClose;
  Result := 0;
end;

procedure AssignCrt(var F: Text);
begin
  Assign(F, '');
  TTextRec(F).OpenFunc := @CrtOpen;
end;

procedure InitCrt;
var
  CursorInfo: TConsoleCursorInfo;
  ConsoleInfo: TConsoleScreenBufferInfo;
  ScrSize: TCoord;
begin
  if CrtInited then Exit;
  OutHandle := GetStdHandle(STD_OUTPUT_HANDLE);
  InHandle := GetStdHandle(STD_INPUT_HANDLE);
  SetConsoleOutputCP(437); // set default US code page
  ScrSize.X := 80;
  ScrSize.Y := 25;
  SetConsoleScreenBufferSize(OutHandle, ScrSize);
  FillChar(CursorInfo, Sizeof(CursorInfo), 0);
  Getconsolecursorinfo(OutHandle, CursorInfo);
  DefCursorSize := CursorInfo.dwSize;
  FillChar(ConsoleInfo, Sizeof(ConsoleInfo), 0);
  Getconsolescreenbufferinfo(OutHandle, ConsoleInfo);
  CursorSaveX := ConsoleInfo.dwCursorPosition.X;
  CursorSaveY := ConsoleInfo.dwCursorPosition.Y;
  TextAttr := ConsoleInfo.wAttributes;
  ScreenWidth := GetScreenWidth;
  ScreenHeight := GetScreenHeight;
  LastMode := smCO80;
  WindMax := (ScreenWidth - 1) or ((ScreenHeight - 1) shl 8);
  AssignCrt(Output);
  Rewrite(Output);
  TTextRec(Output).Handle := OutHandle;
  AssignCrt(input);
  Reset(input);
  TTextRec(input).Handle := InHandle;
  CrtInited := True;
end;

procedure DoneCrt;
begin
  CrtInited := False;
  FlushConsoleInputBuffer(InHandle);
  CloseHandle(InHandle);
  CloseHandle(OutHandle);
end;

procedure WriteScr(var Src: TScr; rCoordAttr: Boolean);
var
  Rect: TSmallRect;
  BufSize, BufCoord: TCoord;
begin
  with Src do
  begin
    BufSize.X := 80;
    BufSize.Y := 25;
    BufCoord.X := 0;
    BufCoord.Y := 0;
    Rect.Left := X1 - 1;
    Rect.Top := Y1 - 1;
    Rect.Right := X2 - 1;
    Rect.Bottom := Y2 - 1;
    Writeconsoleoutput(OutHandle, @Src.Data[0], BufSize, BufCoord, Rect);
    if rCoordAttr then
    begin
      GotoXY(OldX, OldY);
      TextAttr := OldAttr;
    end;
  end;
end;

procedure SetTextMode(Mode: Integer);
var
  Coord: TCoord;
begin
  Coord.X := 80;
  if Mode and smFont8x8 > 0 then
    Coord.Y := 50
  else
    Coord.Y := 25;
  SetConsoleScreenBufferSize(OutHandle, Coord);
  LastMode := Mode;
  ScreenWidth := GetScreenWidth;
  ScreenHeight := GetScreenHeight;
end;

function GetTextMode: Word;
begin
  Result := LastMode;
end;

initialization
  IsWindowsNT := (Win32Platform = VER_PLATFORM_WIN32_NT);
  CrtInited := False;
end.

