
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ Threads.pas                                           }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.2.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   wwww:      http://www.dwp42.org                     }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Kylix and Delphi Cross-Platform Runtime Library   }
{     Copyright (C) 1996, 2001                          }
{     Borland Software Corporation                      }
{     www: http://www.borland.com                       }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}

{ *************************************************************************** }
{                                                                             }
{ Kylix and Delphi Cross-Platform Visual Component Library                    }
{                                                                             }
{ Copyright (c) 1997, 2001 Borland Software Corporation                       }
{                                                                             }
{ This file may be distributed and/or modified under the terms of the GNU     }
{ General Public License (GPL) version 2 as published by the Free Software    }
{ Software Foundation and appearing at http://www.borland.com/kylix/gpl.html. }
{                                                                             }
{ Licensees holding a valid Borland No-Nonsense License for this Software may }
{ use this file in accordance with such license, which appears in the file    }
{ license.txt that came with this Software.                                   }
{                                                                             }
{ *************************************************************************** }
{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/sys/SyncObjs.pas,v 1.1 2003/02/04 20:45:21 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: SyncObjs.pas,v $
  Revision 1.1  2003/02/04 20:45:21  iwache
  New unit SyncObjs added. Currently only TCriticalSection is implemented.
  TEvent need to be implemented in the future.

  ----------------------------------------------------------------------------
}

unit SyncObjs;

{$H+,X+}

{$IFDEF DOS}
  {$DEFINE DOS_OR_LINUX}
{$ENDIF}
{$IFDEF LINUX}
  {$DEFINE DOS_OR_LINUX}
{$ENDIF}

interface

uses
{$IFDEF MSWINDOWS}
  Windows,
{$IFDEF DOS}
  Threads,
{$ELSE}  
  Messages,
{$ENDIF DOS}
{$ENDIF}
{$IFDEF LINUX}
  Libc,
{$ENDIF}
  SysUtils,
  Classes;

type
{$IFNDEF MSWINDOWS}
  PSecurityAttributes = pointer;
{$ENDIF}

  TSynchroObject = class(TObject)
  public
    procedure Acquire; virtual;
    procedure Release; virtual;
  end;

  THandleObject = class(TSynchroObject)
{$IFDEF MSWINDOWS}
{$IFNDEF DOS}
  private
    FHandle: THandle;
    FLastError: Integer;
  public
    destructor Destroy; override;
    property LastError: Integer read FLastError;
    property Handle: THandle read FHandle;
{$ENDIF DOS}
{$ENDIF}
  end;

  TWaitResult = (wrSignaled, wrTimeout, wrAbandoned, wrError);

  TEvent = class(THandleObject)
  {$IFDEF LINUX}
  private
    FEvent: TSemaphore;
  {$ENDIF}
  public
    constructor Create(EventAttributes: PSecurityAttributes; ManualReset,
      InitialState: Boolean; const Name: string);
    function WaitFor(Timeout: LongWord): TWaitResult;
    procedure SetEvent;
    procedure ResetEvent;
  end;

  TSimpleEvent = class(TEvent)
  public
    constructor Create;
  end;

  TCriticalSection = class(TSynchroObject)
  protected
{$IFDEF DOS}
    FTaskID: Cardinal;
    FEnteredCount: Cardinal;
{$ELSE}
    FSection: TRTLCriticalSection;
{$ENDIF}
  public
    constructor Create;
    destructor Destroy; override;
    procedure Acquire; override;
    procedure Release; override;
    procedure Enter;
    procedure Leave;
  end;

implementation

{$IFDEF DOS}
const
  NO_TASK = $FFFFFFFF;
{$ENDIF}

{ TSynchroObject }

procedure TSynchroObject.Acquire;
begin
end;

procedure TSynchroObject.Release;
begin
end;

{ THandleObject }

{$IFDEF MSWINDOWS}
{$IFNDEF DOS}
destructor THandleObject.Destroy;
begin
  CloseHandle(FHandle);
  inherited Destroy;
end;
{$ENDIF DOS}
{$ENDIF}

{ TEvent }

constructor TEvent.Create(EventAttributes: PSecurityAttributes; ManualReset,
  InitialState: Boolean; const Name: string);
{$IFDEF MSWINDOWS}
{$IFNDEF DOS}
begin
  FHandle := CreateEvent(EventAttributes, ManualReset, InitialState, PChar(Name));
end;
{$ENDIF DOS}
{$ENDIF}
{$IFDEF LINUX}
var
   Value: Integer;
begin
   if InitialState then
     Value := 1
   else Value := 0;
   sem_init(FEvent, False, Value);
end;
{$ENDIF}
{$IFDEF DOS}
begin
  // not implemented for DWPL
  inherited Create;
end;
{$ENDIF DOS}


function TEvent.WaitFor(Timeout: LongWord): TWaitResult;
begin
{$IFDEF MSWINDOWS}
{$IFNDEF DOS}
  case WaitForSingleObject(Handle, Timeout) of
    WAIT_ABANDONED: Result := wrAbandoned;
    WAIT_OBJECT_0: Result := wrSignaled;
    WAIT_TIMEOUT: Result := wrTimeout;
    WAIT_FAILED:
      begin
        Result := wrError;
        FLastError := GetLastError;
      end;
  else
    Result := wrError;
  end;
{$ENDIF DOS}
{$ENDIF}
{$IFDEF LINUX}
  if Timeout = LongWord($FFFFFFFF) then
  begin
    sem_wait(FEvent);
    Result := wrSignaled;
  end else
    Result := wrError;
{$ENDIF}
{$IFDEF DOS}
  // not implemented for DWPL
  Result := wrError;
{$ENDIF DOS}
end;

procedure TEvent.SetEvent;
{$IFDEF MSWINDOWS}
{$IFNDEF DOS}
begin
  Windows.SetEvent(Handle);
end;
{$ENDIF DOS}
{$ENDIF}
{$IFDEF LINUX}
var
  I: Integer;
begin
  sem_getvalue(FEvent, I);
  if I = 0 then
    sem_post(FEvent);
end;
{$ENDIF}
{$IFDEF DOS}
begin
  // not implemented for DWPL
end;
{$ENDIF DOS}

procedure TEvent.ResetEvent;
begin
{$IFDEF MSWINDOWS}
{$IFNDEF DOS}
  Windows.ResetEvent(Handle);
{$ENDIF DOS}
{$ENDIF}
{$IFDEF LINUX}
//  All events on Linux are auto-reset
{$ENDIF}
{$IFDEF DOS}
  // not implemented for DWPL
{$ENDIF DOS}
end;

{ TSimpleEvent }

constructor TSimpleEvent.Create;
begin
  inherited Create(nil, True, False, '');
end;

{ TCriticalSection }

constructor TCriticalSection.Create;
begin
  inherited Create;
{$IFDEF DOS}
  FTaskID := NO_TASK;
  FEnteredCount := 0;
{$ELSE}
  InitializeCriticalSection(FSection);
{$ENDIF}
end;

destructor TCriticalSection.Destroy;
begin
{$IFNDEF DOS}
  DeleteCriticalSection(FSection);
{$ENDIF}
  inherited Destroy;
end;

procedure TCriticalSection.Acquire;
begin
{$IFDEF DOS}
  // If the Critical Section is in use by another task,
  // we wait for it to be released.
  while (FTaskID <> NO_TASK) and (FTaskID <> GetCurrentTask) do Yield;
  // We are ready to enter the Critical Section
  FTaskID := GetCurrentTask;
  Inc(FEnteredCount);
{$ELSE}
  EnterCriticalSection(FSection);
{$ENDIF}
end;

procedure TCriticalSection.Release;
begin
{$IFDEF DOS}
  if FTaskID = GetCurrentTask then
  begin
    Dec(FEnteredCount);
    if FEnteredCount = 0 then FTaskID := NO_TASK;
  end;
{$ELSE}
  LeaveCriticalSection(FSection);
{$ENDIF}
end;

procedure TCriticalSection.Enter;
begin
  Acquire;
end;

procedure TCriticalSection.Leave;
begin
  Release;
end;

end.
