
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ Threads.pas                                           }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.2.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   wwww:      http://www.dwp42.org                     }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - Kylix and Delphi Cross-Platform Runtime Library   }
{     Copyright (C) 1996, 2001                          }
{     Borland Software Corporation                      }
{     www: http://www.borland.com                       }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}

{ *************************************************************************** }
{                                                                             }
{ Kylix and Delphi Cross-Platform Runtime Library                             }
{                                                                             }
{ Copyright (c) 1995-2001 Borland Software Corporation                        }
{                                                                             }
{ This file may be distributed and/or modified under the terms of the GNU     }
{ General Public License (GPL) version 2 as published by the Free Software    }
{ Software Foundation and appearing at http://www.borland.com/kylix/gpl.html. }
{                                                                             }
{ Licensees holding a valid Borland No-Nonsense License for this Software may }
{ use this file in accordance with such license, which appears in the file    }
{ license.txt that came with this Software.                                   }
{                                                                             }
{ *************************************************************************** }
{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/sys/Threads.pas,v 1.4 2003/10/16 16:16:12 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: Threads.pas,v $
  Revision 1.4  2003/10/16 16:16:12  iwache
  Bug in TThread.Create fixed. The create constructor did
  always used the DefaultStackSize and not the actual one.

  Revision 1.3  2003/09/21 20:01:13  iwache
  DefaultStackSize extendet to 16384.

  Revision 1.2  2003/02/04 17:27:46  iwache
  Hack to avoid freeze of WinMe to TThread.Destroy added.
  This need now compiler option WDOSX.
  Support for thread synchronization added.

  Revision 1.1  2003/01/31 12:31:53  iwache
  New unit Threads.pas for cooperative multitasking added.
  This unit implements a TThread class, very similar to the
  native Delphi TThreads class. Many thanks to Diego De Marco

  ----------------------------------------------------------------------------
}

unit Threads;

interface

uses
  SysUtils, Classes;

const
  DefaultStackSize = $4000;

type
  EThread = class(Exception);

  TThreadMethod = procedure of object;
  TThreadPriority = (tpIdle, tpLowest, tpLower, tpNormal, tpHigher, tpHighest,
    tpTimeCritical);

  TThread = class
  private
    FHandle: THandle;
    FThreadID: THandle;
    FTerminated: Boolean;
    FSuspended: Boolean;
    FFreeOnTerminate: Boolean;
    FFinished: Boolean;
    FReturnValue: Integer;
    FOnTerminate: TNotifyEvent;
    FPriority: TThreadPriority;
    procedure CallOnTerminate;
    function GetPriority: TThreadPriority;
    procedure SetPriority(Value: TThreadPriority);
    procedure SetSuspended(Value: Boolean);
  protected
    FSynchronizeException: TObject;
    FSynchronizeDone: Boolean;
    FMethod: TThreadMethod;
    procedure DoTerminate; virtual;
    procedure Execute; virtual; abstract;
    procedure Synchronize(Method: TThreadMethod);
    property ReturnValue: Integer read FReturnValue write FReturnValue;
    property Terminated: Boolean read FTerminated;
  public
    constructor Create(CreateSuspended: Boolean;
        StackSize: Integer = DefaultStackSize);
    destructor Destroy; override;
    procedure Resume;
    procedure Suspend;
    procedure Terminate;
    function WaitFor: LongWord;
    property FreeOnTerminate: Boolean read FFreeOnTerminate
        write FFreeOnTerminate;
    property Handle: THandle read FHandle;
    property Priority: TThreadPriority read GetPriority write SetPriority;
    property Suspended: Boolean read FSuspended write SetSuspended;
    property ThreadID: THandle read FThreadID;
    property OnTerminate: TNotifyEvent read FOnTerminate write FOnTerminate;
  end;

{ Sleep suspends execution of the current task for the given time in
  milliseconds and switches execution to the next runnable task.
  The current task will resume execution when all runnable tasks have
  yielded or terminated and the given time is over. You have to call
  Sleep in the main loops of all of your threads Execute methods.
  Parameters:
    Milliseconds: Minimum time this task will be suspended, in milliseconds.
                  Please note that the task may end up being suspended for
                  much longer than the given time if other tasks don't yield
                  execution properly. }

procedure Sleep(Milliseconds: Cardinal);

{***************************** Task handling procedures ***********************}

type
  TTaskProc = procedure(Param: Pointer);
  TTaskStatus = (tsReady, tsRunning, tsPaused, tsDone);

{ CreateTask creates a new task.
  Parameters:
    TaskProc:     Procedure which represents the task. The procedure must have
                  the following prototype: procedure MyTask(Param: Pointer);
    Param:        Parameter to be passed to the new tak
    StackSize:    Stack size (in bytes) for the new task.
  Return Value:   the TaskID of the created task }

function CreateTask(TaskProc: TTaskProc; Param: Pointer; StackSize: Cardinal):
    Cardinal;

{ GetMainTask returns the TaskID for the main task }

function GetMainTask: Cardinal;

{ GetCurrentTask returns the TaskID for the current Task.
  Currently the main task's ID is always 0 (Zero). This may be changed
  in the future. }

function GetCurrentTask: Cardinal;

{ GetTaskStatus returns the current status of the given task.
    Parameters:
      TaskID:     Task for which status must be retrieved. }

function GetTaskStatus(TaskID: Cardinal; var Status: TTaskStatus): Boolean;

{ ResumeTask resumes execution of the given task.
    Parameters:
      TaskID:     Task to be resumed
    Return Value: True if the task was resumed,
                  False if it wasn't paused or didn't exist. }

function ResumeTask(TaskID: Cardinal): Boolean;

{ PauseTask suspends execution of the given task.
    Parameters:
      TaskID:     Task to be paused (suspended)
    Return Value: True if the task was paused,
                  False if it wasn't running or didn't exist. }

function PauseTask(TaskID: Cardinal): Boolean;

{ TerminateTask terminates execution of the given task.
    Parameters:
      TaskID:     Task to be terminated. If the task is currently runing
                  (i.e, is the current task), it will be terminated when
                  it first calls Yield() or Sleep().
    Return Value: True if the task was terminated,
                  False if it wasn't running
                  Please note that the main task CAN NOT be terminated. }

function TerminateTask(TaskID: Cardinal): Boolean;

{ Yield suspends execution of the current task, and switches execution
  to the next runnable task. The current task will resume execution when
  all runnable tasks have yielded or terminated. To be more Windows conform
  you can use Sleep(0) instead, which simply calls Yield. }

procedure Yield;

var
  AddThread: procedure;
  RemoveThread: procedure;
  ExecuteProcedure: procedure(Thread: TThread);

implementation

uses
  Windows;

{***************************** Task handling procedures ***********************}

const
  NO_TASK = $FFFFFFFF;

type
  PPointer = ^Pointer;
  PTaskEntry = ^TTaskEntry;
  TTaskEntry = record
    TaskID: Cardinal;
    EBX: Cardinal;
    ESI: Cardinal;
    EDI: Cardinal;
    EBP: Cardinal;
    ESP: Cardinal;
    RetAddr: Cardinal;
    Stack: Pointer;
    StackSize: Cardinal;
    Status: TTaskStatus;
  end;

var
  TaskList: TList;
  CurrentTask: Cardinal;
  LastTaskID: Integer = 0;
  // Internal variables used by "Yield".
  _TaskEntry: PTaskEntry;
  _i, _CurrentTaskIndex, _NewTaskIndex: Integer;
  _NewEBX, _NewESI, _NewEDI, _NewEBP, _NewESP, _NewRetAddr: Cardinal;
  _OldEBX, _OldESI, _OldEDI, _OldEBP, _OldESP, _OldRetAddr: Cardinal;

function CreateTaskID: Cardinal;
begin
  Inc(LastTaskID);
  Result := LastTaskID;
end;

function GetTaskInfo(TaskID: Cardinal; var TaskIndex: Integer): PTaskEntry;
begin
  TaskIndex := 0;
  while TaskIndex < TaskList.Count do
  begin
    Result := TaskList[TaskIndex];
    if Result.TaskID = TaskID then
      Exit;
    Inc(TaskIndex);
  end;
  TaskIndex := -1;
  Result := nil;
end;

function FindTask(TaskID: Cardinal): PTaskEntry;
var
  TaskIndex: Integer;
begin
  Result := GetTaskInfo(TaskID, TaskIndex);
end;

{ ProcEnd is called automatically when a task ends. It marks the task
  as "Terminated" and switches to the next available task. The task
  will be effectively freed when the main task executes. }

procedure ProcEnd;
begin
  // Mark the task as "terminated"
  FindTask(CurrentTask).Status := tsDone;
  Yield;
end;

{ ProcStart is a "stub" function used to start a new Task. In other words,
  all tasks start here, and this procedure actually calls the proper
  procedure task. }

procedure ProcStart; assembler;
asm
	pop  EAX	// Parameter
	pop  ECX	// Address of the routine to execute
	CALL ECX
	call ProcEnd
end;

procedure RemoveTask(Entry: PTaskEntry);
begin
  TaskList.Remove(Entry);
  FreeMem(Entry.Stack);
  Dispose(Entry);
end;

function GetMainTask: Cardinal;
begin
  Result := PTaskEntry(TaskList[0]).TaskID;
end;

function GetCurrentTask: Cardinal;
begin
  Result := CurrentTask;
end;

function GetTaskStatus(TaskID: Cardinal; var Status: TTaskStatus): Boolean;
var
  P: PTaskEntry;
begin
  P := FindTask(TaskID);
  Result := Assigned(P);
  if Result then
    Status := P.Status;
end;

function ResumeTask(TaskID: Cardinal): Boolean;
var
  P: PTaskEntry;
begin
  // Important: even the main task can be paused/resumed
  P := FindTask(TaskID);
  Result := Assigned(P) and (P.Status = tsPaused);
  if Result then
  begin
    if TaskID = GetCurrentTask then
      P.Status := tsRunning
    else
      P.Status := tsReady;
  end;
end;

function PauseTask(TaskID: Cardinal): Boolean;
var
  P: PTaskEntry;
begin
  // Important: even the main task can be paused/resumed
  P := FindTask(TaskID);
  Result := Assigned(P) and (P.Status in [tsRunning, tsReady]);
  if Result then
    P.Status := tsPaused;
end;

function TerminateTask(TaskID: Cardinal): Boolean;
var
  P: PTaskEntry;
begin
  // Important: the main task can not be terminated (but it can be paused)
  P := FindTask(TaskID);
  Result := Assigned(P) and Assigned(P.Stack)
    and (P.Status in [tsRunning, tsReady, tsPaused]);
  if Result then
    P.Status := tsDone;
end;

function CreateTask(TaskProc: TTaskProc; Param: Pointer; StackSize: Cardinal):
  Cardinal;
var
  P: PPointer;
  NewTask: PTaskEntry;
begin
  // Create a new task Entry
  New(NewTask);

  // Create the stack for the new task
  GetMem(NewTask.Stack, StackSize);
  NewTask.StackSize := StackSize;
  // "Push" the address of the new task procedure into the new stack
  P := Pointer(Cardinal(NewTask.Stack) + StackSize - 4);
  P^ := @TaskProc;
  // Also push the parameter
  Dec(P);
  P^ := Param;

  // Save information related to the new task
  NewTask.TaskID := CreateTaskID;
  NewTask.EBX := 0;
  NewTask.ESI := 0;
  NewTask.EDI := 0;
  NewTask.EBP := Cardinal(NewTask.Stack) + StackSize;
  NewTask.ESP := Cardinal(P);
  NewTask.RetAddr := Cardinal(@ProcStart);
    // The task will actually start executing "ProcStart"
  NewTask.Status := tsReady;

  // Done
  TaskList.Add(NewTask);
  Result := NewTask.TaskID;
end;

procedure Yield;
begin
  // Perform maintenance tasks (free terminated tasks, etc.). This is only
  // done if we're running in the main task.
  _TaskEntry := GetTaskInfo(CurrentTask, _CurrentTaskIndex);
  if _CurrentTaskIndex = 0 then
  begin
    _i := 0;
    while _i < TaskList.Count do
    begin
      _TaskEntry := TaskList[_i];
      if _TaskEntry.Status = tsDone then
        RemoveTask(_TaskEntry)
      else
        Inc(_i);
    end;
  end;
  // Search for a new task, starting at current task index
  _NewTaskIndex := -1;
  _i := _CurrentTaskIndex + 1;
  while _i < TaskList.Count do
  begin
    _TaskEntry := TaskList[_i];
    if _TaskEntry.Status = tsReady then
    begin
      _NewTaskIndex := _i;
      Break;
    end;
    Inc(_i);
  end;
  // No task found. Let's search from the beginning
  if _NewTaskIndex = -1 then
  begin
    _i := 0;
    while _i < _CurrentTaskIndex do
    begin
      _TaskEntry := TaskList[_i];
      if _TaskEntry.Status = tsReady then
      begin
        _NewTaskIndex := _i;
        Break;
      end;
    end;
    Inc(_i);
  end;
  // No task found?
  if _NewTaskIndex = -1 then
    Exit;

  // Save register status for the current task
  asm
	MOV     EAX,[ESP + 16]		// Ret Addr
        MOV     _OldRetAddr,EAX
	MOV     EAX,[ESP + 12]		// EBX
        MOV     _OldEBX,EAX
	MOV     EAX,[ESP +  8]		// ESI
	MOV     _OldESI,EAX
	MOV     EAX,[ESP +  4]		// EDI
	MOV     _OldEDI,EAX
	MOV     EAX, [ESP +  0]		// EBP
	MOV     _OldEBP, EAX
	MOV     EAX,ESP			// ESP
	ADD	EAX,20
	MOV     _OldESP,EAX

  end;
  _TaskEntry := TaskList[_CurrentTaskIndex];
  _TaskEntry.EBX := _OldEBX;
  _TaskEntry.ESI := _OldESI;
  _TaskEntry.EDI := _OldEDI;
  _TaskEntry.EBP := _OldEBP;
  _TaskEntry.ESP := _OldESP;
  _TaskEntry.RetAddr := _OldRetAddr;
  if _TaskEntry.Status = tsRunning then
    _TaskEntry.Status := tsReady;

  // Load the registers for the new task, execute it
  _TaskEntry := TaskList[_NewTaskIndex];
  _NewEBX := _TaskEntry.EBX;
  _NewESI := _TaskEntry.ESI;
  _NewEDI := _TaskEntry.EDI;
  _NewEBP := _TaskEntry.EBP;
  _NewESP := _TaskEntry.ESP;
  _NewRetAddr := _TaskEntry.RetAddr;
  _TaskEntry.Status := tsRunning;
  CurrentTask := _TaskEntry.TaskID;
  asm
	MOV	EAX, _NewRetAddr
	PUSH	_NewESP
	PUSH	_NewEBP
	PUSH	_NewEDI
	PUSH   	_NewESI
	PUSH   	_NewEBX
	POP	EBX
	POP	ESI
	POP	EDI
	POP	EBP
	POP	ESP
	PUSH	EAX
	RET
  end;
end;

{***************************** end of Task handling procedures ****************}

procedure Sleep(Milliseconds: Cardinal);
var
  Timeout: Int64;
  StartCount: Int64;
  CurrentCount: Int64;
begin
  if Milliseconds = 0 then
    Yield
  else
  begin
    StartCount := GetTickCount;
    Timeout := StartCount + Milliseconds;
    repeat
      Yield;
      CurrentCount := GetTickCount;
      if CurrentCount < StartCount then
        CurrentCount := CurrentCount + High(LongWord);
    until CurrentCount >= Timeout;
  end;
end;

procedure ThreadProc(AThread: Pointer);
var
  FreeThread: Boolean;
  Thread: TThread;
begin
  Thread := TThread(AThread);
  try
    Thread.Execute;
  finally
    FreeThread := Thread.FFreeOnTerminate;
    Thread.FFinished := True;
    Thread.DoTerminate;
    if FreeThread then Thread.Free;
  end;
end;

constructor TThread.Create(CreateSuspended: Boolean; StackSize: Integer);
begin
  inherited Create;
  if Assigned(AddThread) then AddThread;
  FThreadID := CreateTask(ThreadProc, Pointer(Self), StackSize);
  FHandle := FThreadID;
  FSuspended := CreateSuspended;
  if FSuspended then Suspend;
end;

destructor TThread.Destroy;
begin
{$IFDEF WDOSX}
  // hack to avoid freeze of WinMe
  asm
        MOV     AH,1
        INT     16H
  end;
{$ENDIF}  
  if not FFinished and not Suspended then
  begin
    Terminate;
    WaitFor;
  end;
  TerminateTask(FThreadID);
  inherited Destroy;
  if Assigned(RemoveThread) then RemoveThread;
end;

procedure TThread.CallOnTerminate;
begin
  if Assigned(FOnTerminate) then FOnTerminate(Self);
end;

procedure TThread.DoTerminate;
begin
  if Assigned(FOnTerminate) then Synchronize(CallOnTerminate);
end;

function TThread.GetPriority: TThreadPriority;
begin
  Result := FPriority;
end;

procedure TThread.Resume;
begin
  if FFinished then Exit;
  if ResumeTask(FThreadID) then FSuspended := False;
end;

procedure TThread.SetPriority(Value: TThreadPriority);
begin
  if Value <> FPriority then
    FPriority := Value;
end;

procedure TThread.SetSuspended(Value: Boolean);
begin
  if Value <> FSuspended then
    if Value then
      Suspend
    else
      Resume;
end;

procedure TThread.Suspend;
begin
  if FFinished then Exit;
  if PauseTask(FThreadID) then
    FSuspended := True;
end;

procedure TThread.Synchronize(Method: TThreadMethod);
begin
  FSynchronizeException := nil;
  FSynchronizeDone := False;
  FMethod := Method;
  // VisualCLX needs Synchronize for cooperative multitasking
  if Assigned(ExecuteProcedure) then
  begin
    ExecuteProcedure(Self);
    repeat
      Yield;
    until FSynchronizeDone or Terminated;
  end
  else
    Method;
  if Assigned(FSynchronizeException) then raise FSynchronizeException;
end;

procedure TThread.Terminate;
begin
  FTerminated := True;
end;

function TThread.WaitFor: LongWord;
Var
  Status: TTaskStatus;
begin
  if GetCurrentTask = FThreadID then
    raise EThread.Create('Thread cannot wait for its own termination');
  while GetTaskStatus(FThreadID, Status) and (Status <> tsDone) do
    Yield;
  Result := ReturnValue;
end;

var
  MainTask: PTaskEntry;

initialization
  TaskList := TList.Create;
  // Create the "main task" and mark it as "ready"
  New(MainTask);
  TaskList.Add(MainTask);
  MainTask.TaskID := CreateTaskID;
  MainTask.Stack := nil;
  MainTask.StackSize := 0;
  MainTask.Status := tsRunning;
  CurrentTask := MainTask.TaskID;

finalization
  // Free any task that wasn't terminated
  while TaskList.Count > 0 do
    RemoveTask(PTaskEntry(TaskList[0]));
  TaskList.Free;
end.
