{
  crt.pas: a Crt unit for Win32 console applications

  By Laszlo Agocs (alp@dwp42.org)

  Should work with all 32-bit Delphi versions.
  Some functions do nothing and some variables are ignored.
  WindMin and WindMax are not compatible with Turbo Pascal since they are now
  defined as longint instead of word (to support large screen buffers).
}

{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/tp/Crt.pas,v 1.10 2003/04/14 13:50:50 lagocs Exp $
  ----------------------------------------------------------------------------

  $Log: Crt.pas,v $
  Revision 1.10  2003/04/14 13:50:50  lagocs
  Fixed WDOSX-specific scrolling code in VertScroll, the last character
  in the window is now also copied.

  Revision 1.9  2003/01/28 11:15:07  lagocs
  KeyPressed and ReadKey use int 16h if running under WDOSX

  ----------------------------------------------------------------------------
}

unit crt;

interface

uses Windows;

const
  Black         = 0;
  Blue          = 1;
  Green         = 2;
  Cyan          = 3;
  Red           = 4;
  Magenta       = 5;
  Brown         = 6;
  LightGray     = 7;
  DarkGray      = 8;
  LightBlue     = 9;
  LightGreen    = 10;
  LightCyan     = 11;
  LightRed      = 12;
  LightMagenta  = 13;
  Yellow        = 14;
  White         = 15;
  Blink         = 128;

  BW40          = 0;
  CO40          = 1;
  BW80          = 2;
  CO80          = 3;
  Mono          = 7;
  Font8x8       = 256;

  C40           = CO40;
  C80           = CO80;

var
  shiftstate,cursorstate,mousex,mousey,mouseb:integer;
  mousedbl:boolean;
  textattr:byte;
  conwidth,conheight:integer;
  outhandle,inhandle:THandle;
  windmin,windmax:longint;
  CheckBreak,CheckEof,CheckSnow,DirectVideo:boolean;
  LastMode:word;

function InitCrt:boolean;
procedure QuitCrt;
procedure GetText(x1,y1,x2,y2:integer;dest:pointer;bufw:integer);
procedure PutText(x1,y1,x2,y2:integer;src:pointer;bufw:integer);
procedure MouseOn;
procedure MouseOff;
function KeyPressed:boolean;
function ReadKey:char;
procedure SetCursorSize(size:integer);
procedure CursorOn;
procedure CursorOff;
procedure CursorBig;
procedure GotoXY(x,y:integer);
function WhereX:integer;
function WhereY:integer;
procedure TextColor(color:byte);
procedure TextBackground(color:byte);
procedure HighVideo;
procedure LowVideo;
procedure NormVideo;
procedure ClrScr;
procedure FillVBuf(buf:pointer;x1,y1,x2,y2,bufw:integer;ch:char;attr:byte);
procedure FillArea(x1,y1,x2,y2:integer;ch:char;attr:byte);
procedure ClrEol;
procedure Window(x1,y1,x2,y2:integer);
procedure VertScroll(topy,n:integer);
procedure DelLineY(y:integer);
procedure DelLine;
procedure InsLine;
procedure Sound(hz:word);
procedure NoSound;
procedure TextMode(mode:integer);
procedure Delay(ms:word);
procedure AssignCrt(var f:text);

implementation

uses SysUtils;

var
  vbufc:pointer;
  vbuf:PByteArray;
  iswdosx,isnt:boolean;

function inithandles:boolean;
var
  sec:TSecurityAttributes;
begin
  result:=false;
  sec.nLength:=sizeof(sec);
  sec.lpSecurityDescriptor:=nil;
  sec.bInheritHandle:=true;
  outhandle:=CreateFile('CONOUT$',GENERIC_READ or GENERIC_WRITE,FILE_SHARE_READ or FILE_SHARE_WRITE,@sec,OPEN_EXISTING,0,0);
  if outhandle=INVALID_HANDLE_VALUE then
    exit;
  inhandle:=CreateFile('CONIN$',GENERIC_READ or GENERIC_WRITE,FILE_SHARE_READ or FILE_SHARE_WRITE,@sec,OPEN_EXISTING,0,0);
  if inhandle=INVALID_HANDLE_VALUE then begin
    CloseHandle(outhandle);
    exit;
  end;
  result:=true;
end;

procedure closehandles;
begin
  CloseHandle(outhandle);
  CloseHandle(inhandle);
end;

procedure GetText(x1,y1,x2,y2:integer;dest:pointer;bufw:integer);
var
  rect:TSmallRect;
  bufsize,bufcoord:TCoord;
  p1,p2:^byte;
  x,y:integer;
begin
  bufsize.X:=conwidth;
  bufsize.Y:=y2-y1+1;
  bufcoord.X:=0;
  bufcoord.Y:=0;
  rect.Left:=x1-1;
  rect.Top:=y1-1;
  rect.Right:=x2-1;
  rect.Bottom:=y2-1;
  ReadConsoleOutput(outhandle,vbufc,bufsize,bufcoord,rect);
  for y:=0 to y2-y1 do begin
    p1:=vbufc;
    inc(p1,y*conwidth*4);
    p2:=dest;
    inc(p2,y*bufw);
    for x:=0 to x2-x1 do begin
      p2^:=p1^;
      inc(p2);
      inc(p1,2);
      p2^:=p1^;
      inc(p2);
      inc(p1,2);
    end;
  end;
end;

procedure PutText(x1,y1,x2,y2:integer;src:pointer;bufw:integer);
var
  rect:TSmallRect;
  bufsize,bufcoord:TCoord;
  p1,p2:^byte;
  x,y:integer;
begin
  bufsize.X:=conwidth;
  bufsize.Y:=y2-y1+1;
  bufcoord.X:=0;
  bufcoord.Y:=0;
  rect.Left:=x1-1;
  rect.Top:=y1-1;
  rect.Right:=x2-1;
  rect.Bottom:=y2-1;
  for y:=0 to y2-y1 do begin
    p2:=vbufc;
    inc(p2,y*conwidth*4);
    p1:=src;
    inc(p1,y*bufw);
    for x:=0 to x2-x1 do begin
      p2^:=p1^;
      inc(p2);
      p2^:=0;
      inc(p2);
      inc(p1);
      p2^:=p1^;
      inc(p2);
      p2^:=0;
      inc(p2);
      inc(p1);
    end;
  end;
  WriteConsoleOutput(outhandle,vbufc,bufsize,bufcoord,rect);
end;

procedure MouseOn;
var
  mode:DWORD;
begin
  if GetConsoleMode(inhandle,mode) then begin
    mode:=mode or ENABLE_MOUSE_INPUT;
    SetConsoleMode(inhandle,mode);
  end;
end;

procedure MouseOff;
var
  mode:DWORD;
begin
  if GetConsoleMode(inhandle,mode) then begin
    mode:=mode and not ENABLE_MOUSE_INPUT;
    SetConsoleMode(inhandle,mode);
  end;
end;

const
  alt_tbl_a:array [0..37] of byte=(
    97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,
    112,113,114,115,116,117,118,119,120,121,122,
    48,49,50,51,52,53,54,55,56,57,45,43
  );
  alt_tbl_s:array [0..37] of byte=(
    30,48,46,32,18,33,34,35,23,36,37,38,50,49,24,25,16,19,
    31,20,22,47,17,45,21,44,
    129,120,121,122,123,124,125,126,127,128,130,131
  );
  ctrl_tbl_a:array [0..9] of byte=(
    71,72,73,75,77,79,80,81,82,82
  );
  ctrl_tbl_b:array [0..9] of byte=(
    119,141,132,115,116,117,145,118,146,147
  );

var
  scancode:integer=0;
  specialkey:boolean;
  altnumpad:integer=0;
  altnumpadcode:integer=0;

function KeyPressed:boolean;
var
  buf:TInputRecord;
  numevents,numread:DWORD;
  eventno,i,ch,vkc:integer;
begin
  if iswdosx then begin
    asm
      mov result,false
      mov ah,1
      int 16h
      jz @nokey
      cmp al,0
      jne @key_end
      cmp ah,0
      je @nokey
    @key_end:
      mov result,true
    @nokey:
    end;
    exit;
  end;
  if (scancode<>0) then begin
    result:=true;
    exit;
  end;
  result:=false;
  GetNumberOfConsoleInputEvents(inhandle,numevents);
  for eventno:=0 to numevents-1 do begin
    ReadConsoleInput(inhandle,buf,1,numread);
    if numread<>1 then
      break;
    if buf.EventType=KEY_EVENT then begin
      i:=buf.Event.KeyEvent.dwControlKeyState;
      shiftstate:=0;
      if (i and (RIGHT_ALT_PRESSED or LEFT_ALT_PRESSED))<>0 then
        shiftstate:=shiftstate or 8;
      if (i and (RIGHT_CTRL_PRESSED or LEFT_CTRL_PRESSED))<>0 then
        shiftstate:=shiftstate or 4;
      if (i and SHIFT_PRESSED)<>0 then
        shiftstate:=shiftstate or 2;
      if buf.Event.KeyEvent.bKeyDown then begin
        vkc:=buf.Event.KeyEvent.wVirtualKeyCode;
        if (vkc<>VK_SHIFT) and (vkc<>VK_MENU) and (vkc<>VK_CONTROL) and (vkc<>VK_CAPITAL) and (vkc<>VK_NUMLOCK) and (vkc<>VK_SCROLL) then begin
          result:=true;
          if buf.Event.KeyEvent.AsciiChar<>#0 then begin
            ch:=ord(buf.Event.KeyEvent.AsciiChar);
            specialkey:=false;
            scancode:=ch;
            if (shiftstate and 8)<>0 then begin
              if (ch>=65) and (ch<=90) then
                inc(ch,32);
              for i:=0 to 37 do
                if ch=alt_tbl_a[i] then begin
                  specialkey:=true;
                  scancode:=alt_tbl_s[i];
                  break;
                end;
            end;
          end else begin
            ch:=buf.Event.KeyEvent.wVirtualScanCode;
            specialkey:=true;
            if (ch=87) or (ch=88) then
              inc(ch,46);
            if (shiftstate and 8)<>0 then begin
              if (ch>=59) and (ch<=68) then
                inc(ch,45);
              if (ch=133) or (ch=134) then
                inc(ch,6);
              if (ch>=71) and (ch<=83) and (ch<>74) and (ch<>76) and (ch<>78) then
                inc(ch,80);
            end;
            if (shiftstate and 4)<>0 then begin
              if (ch>=59) and (ch<=68) then
                inc(ch,35);
              if (ch=133) or (ch=134) then
                inc(ch,4);
              for i:=0 to 9 do
                if ch=ctrl_tbl_a[i] then begin
                  ch:=ctrl_tbl_b[i];
                  break;
                end;
            end;
            if (shiftstate and 2)<>0 then begin
              if (ch>=59) and (ch<=68) then
                inc(ch,25);
              if (ch=133) or (ch=134) then
                inc(ch,2);
            end;
            scancode:=ch;
          end;
          if (shiftstate and 8)<>0 then begin
            if (vkc>=VK_NUMPAD0) and (vkc<=VK_NUMPAD9) then begin
              if altnumpad<=2 then begin
                inc(altnumpad);
                altnumpadcode:=altnumpadcode*10+vkc-VK_NUMPAD0;
              end;
              result:=false;
              scancode:=0;
              specialkey:=false;
            end;
          end;
        end;
      end else if buf.Event.KeyEvent.wVirtualKeyCode=VK_MENU then begin
        if altnumpad>0 then begin
          if altnumpadcode>0 then begin
            result:=true;
            scancode:=altnumpadcode;
            specialkey:=false;
            altnumpadcode:=0;
          end;
          altnumpad:=0;
        end;
      end;
    end else if buf.EventType=_MOUSE_EVENT then begin
      case buf.Event.MouseEvent.dwEventFlags of
        0:begin
          mousedbl:=false;
          mouseb:=buf.Event.MouseEvent.dwButtonState;
        end;
        1:begin
          mousedbl:=false;
          mousex:=buf.Event.MouseEvent.dwMousePosition.X;
          mousey:=buf.Event.MouseEvent.dwMousePosition.Y;
        end;
        DOUBLE_CLICK:begin
          mousedbl:=true;
        end;
      end;
    end;
  end;
end;

function ReadKey:char;
begin
  if iswdosx then begin
    if scancode=0 then asm
    @rk:
      mov ah,0
      int 16h
      cmp al,0
      je @speckey
      mov specialkey,false
      xor edx,edx
      mov dl,al
      mov scancode,edx
      jmp @key_end
    @speckey:
      cmp ah,0
      je @rk
      mov specialkey,true
      xor edx,edx
      mov dl,ah
      mov scancode,edx
    @key_end:
    end;
  end else begin
    while not keypressed do
      Sleep(1);
  end;
  if specialkey then begin
    result:=#0;
    specialkey:=false;
  end else begin
    result:=chr(scancode);
    scancode:=0;
  end;
end;

procedure SetCursorSize(size:integer);
var
  cursorinfo:TConsoleCursorInfo;
begin
  GetConsoleCursorInfo(outhandle,cursorinfo);
  cursorinfo.bVisible:=size>0;
  if size>0 then
    cursorinfo.dwSize:=size;
  SetConsoleCursorInfo(outhandle,cursorinfo);
end;

procedure CursorOn;
begin
  if cursorstate=1 then
    exit;
  setcursorsize(12);
  cursorstate:=1;
end;

procedure CursorOff;
begin
  if cursorstate=0 then
    exit;
  setcursorsize(0);
  cursorstate:=0;
end;

procedure CursorBig;
begin
  if cursorstate=2 then
    exit;
  setcursorsize(99);
  cursorstate:=2;
end;

procedure GotoXY(x,y:integer);
var
  cursorpos:TCoord;
begin
  cursorpos.X:=x-1+(windmin and $FFFF);
  cursorpos.Y:=y-1+(windmin shr 16);
  SetConsoleCursorPosition(outhandle,cursorpos);
end;

function wherexy:TCoord;
var
  info:TConsoleScreenBufferInfo;
begin
  GetConsoleScreenBufferInfo(outhandle,info);
  result:=info.dwCursorPosition;
end;

function WhereX:integer;
begin
  result:=wherexy.X+1-(windmin and $FFFF);
end;

function WhereY:integer;
begin
  result:=wherexy.Y+1-(windmin shr 16);
end;

procedure TextColor(color:byte);
begin
  textattr:=(textattr and $F0) or (color and $0F);
end;

procedure TextBackground(color:byte);
begin
  textattr:=(color shl 4) or (textattr and $0F);
end;

procedure HighVideo;
begin
  textcolor(textattr or $08);
end;

procedure LowVideo;
begin
  textcolor(textattr and $77);
end;

procedure NormVideo;
begin
  textattr:=7;
end;

procedure ClrScr;
var
  x,y,o:integer;
begin
  for y:=0 to conheight-1 do begin
    o:=y*conwidth*2;
    for x:=0 to conwidth-1 do begin
      vbuf^[o]:=32;
      inc(o);
      vbuf^[o]:=textattr;
      inc(o);
    end;
  end;
  puttext((windmin and $FFFF)+1,(windmin shr 16)+1,(windmax and $FFFF)+1,(windmax shr 16)+1,vbuf,conwidth*2);
  gotoxy(1,1);
end;

procedure FillVBuf(buf:pointer;x1,y1,x2,y2,bufw:integer;ch:char;attr:byte);
var
  x,y:integer;
  p:pbytearray;
begin
  p:=buf;
  for y:=y1 to y2 do
    for x:=x1 to x2 do begin
      p^[(x-x1)*2+(y-y1)*bufw]:=ord(ch);
      p^[(x-x1)*2+1+(y-y1)*bufw]:=attr;
    end;
end;

procedure FillArea(x1,y1,x2,y2:integer;ch:char;attr:byte);
begin
  fillvbuf(vbuf,x1,y1,x2,y2,conwidth*2,ch,attr);
  puttext(x1,y1,x2,y2,vbuf,conwidth*2);
end;

procedure ClrEol;
var
  x,y:integer;
begin
  x:=wherex;
  y:=wherey;
  fillarea(x,y,conwidth,y,' ',textattr);
end;

procedure Window(x1,y1,x2,y2:integer);
begin
  windmin:=(x1-1) or ((y1-1) shl 16);
  windmax:=(x2-1) or ((y2-1) shl 16);
  gotoxy(1,1);
end;

procedure VertScroll(topy,n:integer);
var
  cliprect,srcrect:TSmallRect;
  destpos:TCoord;
  charinfo:TCharInfo;
  y:integer;
  p:pbytearray;
begin
  if not iswdosx then begin
    charinfo.AsciiChar:=#32;
    charinfo.Attributes:=textattr;
    srcrect.Top:=topy-1;
    srcrect.Left:=(windmin and $FFFF);
    srcrect.Right:=(windmax and $FFFF);
    srcrect.Bottom:=(windmax shr 16);
    destpos.X:=(windmin and $FFFF);
    destpos.Y:=topy+n-1;
    cliprect:=srcrect;
    ScrollConsoleScreenBuffer(outhandle,srcrect,@cliprect,destpos,charinfo);
  end else begin //ScrollConsoleScreenBuffer seems to be broken in WDOSX 0.96
    p:=pointer($B8000);
    if n>0 then begin
      for y:=(windmax shr 16) downto topy-1 do begin
        if y+n<=(windmax shr 16) then
          move(p^[(windmin and $FFFF)*2+y*160],p^[(windmin and $FFFF)*2+(y+n)*160],((windmax and $FFFF)-(windmin and $FFFF)+1)*2);
      end;
      fillarea((windmin and $FFFF)+1,topy,(windmax and $FFFF)+1,topy,' ',textattr);
    end else if n<0 then begin
      for y:=topy-1 to (windmax shr 16) do begin
        if y+n>=topy-1 then
          move(p^[(windmin and $FFFF)*2+y*160],p^[(windmin and $FFFF)*2+(y+n)*160],((windmax and $FFFF)-(windmin and $FFFF)+1)*2);
      end;
      fillarea((windmin and $FFFF)+1,(windmax shr 16)+1,(windmax and $FFFF)+1,(windmax shr 16)+1,' ',textattr);
    end;
  end;
end;

procedure DelLineY(y:integer);
begin
  vertscroll(y+(windmin shr 16),-1);
end;

procedure DelLine;
begin
  delliney(wherey);
end;

procedure InsLine;
begin
  vertscroll(wherey+(windmin shr 16),1);
end;

procedure Sound(hz:word);
begin
  Beep;
end;

procedure NoSound;
begin
end;

procedure TextMode(mode:integer);
begin
end;

procedure Delay(ms:word);
var
  p:^cardinal;
  start:cardinal;
begin
  if not isnt then begin
    Sleep(ms);
  end else begin //this seems to be working with WDOSX under NT
    p:=pointer($46C);
    start:=p^;
    while (p^-start)*55<ms do ;
  end;
end;

procedure writechar(var x,y:integer;c:char);
var
  charinfo:TCharInfo;
  bufsize,destpos:TCoord;
  destrect:TSmallRect;
begin
  if c=#10 then
    inc(y)
  else if c=#13 then
    x:=1
  else begin
    bufsize.X:=1;
    bufsize.Y:=1;
    destpos.X:=0;
    destpos.Y:=0;
    charinfo.AsciiChar:=c;
    charinfo.Attributes:=textattr;
    destrect.Left:=x-1+(windmin and $FFFF);
    destrect.Top:=y-1+(windmin shr 16);
    destrect.Right:=destrect.Left;
    destrect.Bottom:=destrect.Top;
    WriteConsoleOutput(outhandle,@charinfo,bufsize,destpos,destrect);
    inc(x);
  end;
  if x+(windmin and $FFFF)>(windmax and $FFFF)+1 then begin
    inc(y);
    x:=1;
  end;
  while y+(windmin shr 16)>(windmax shr 16)+1 do begin
    delliney(1);
    dec(y);
  end;
end;

function crtread(var f:ttextrec):integer;
var
  x,y:integer;
  key:char;
begin
  x:=wherex;
  y:=wherey;
  f.bufpos:=0;
  f.bufend:=0;
  while true do begin
    if f.bufpos>f.bufend then
      f.bufend:=f.bufpos;
    gotoxy(x,y);
    key:=readkey;
    if key<>#0 then begin
      case key of
        #13:begin
          writechar(x,y,#13);
          writechar(x,y,#10);
          f.bufptr[f.bufend]:=#13;
          f.bufptr[f.bufend+1]:=#10;
          inc(f.bufend,2);
          break;
        end;
        #8:if f.bufpos>0 then begin
          dec(x);
          gotoxy(x,y);
          write(' ');
          gotoxy(x,y);
          dec(f.bufpos);
          dec(f.bufend);
        end;
        else if key>=#32 then begin
          if f.bufpos<f.bufsize-2 then begin
            f.buffer[f.bufpos]:=key;
            inc(f.bufpos);
            writechar(x,y,key);
          end;
        end;
      end;
    end else begin
      readkey;
    end;
  end;
  f.bufpos:=0;
  gotoxy(x,y);
  result:=0;
end;

function crtwrite(var f:ttextrec):integer;
var
  i,x,y:integer;
begin
  x:=wherex;
  y:=wherey;
  for i:=0 to f.bufpos-1 do
    writechar(x,y,f.buffer[i]);
  gotoxy(x,y);
  f.bufpos:=0;
  result:=0;
end;

function crtreturn(var f:ttextrec):integer;
begin
  result:=0;
end;

function crtclose(var f:ttextrec):integer;
begin
  f.mode:=fmclosed;
  result:=0;
end;

function crtopen(var f:ttextrec):integer;
begin
  if f.mode=fmoutput then begin
    ttextrec(f).inoutfunc:=@crtwrite;
    ttextrec(f).flushfunc:=@crtwrite;
  end else begin
    f.mode:=fminput;
    ttextrec(f).inoutfunc:=@crtread;
    ttextrec(f).flushfunc:=@crtreturn;
  end;
  ttextrec(f).closefunc:=@crtclose;
  result:=0;
end;

procedure AssignCrt(var f:text);
begin
  assign(f,'');
  ttextrec(f).openfunc:=@crtopen;
end;

function GetEnvironmentVariableA(name,buffer:pchar;size:longint):longint;stdcall;external 'kernel32.dll';

var
  oldinput,oldoutput:text;
  crtinited:boolean=false;

function InitCrt:boolean;
var
  info:TConsoleScreenBufferInfo;
  mode:DWORD;
  s:ansistring;
begin
  if crtinited then begin
    result:=true;
    exit;
  end;
  result:=false;
  if not inithandles then
    exit;
  if GetConsoleMode(inhandle,mode) then begin
    mode:=mode and not ENABLE_PROCESSED_INPUT;
    SetConsoleMode(inhandle,mode);
  end;
  GetConsoleScreenBufferInfo(outhandle,info);
  conwidth:=info.dwSize.X;
  conheight:=info.dwSize.Y;
  GetMem(vbuf,conwidth*conheight*2);
  if vbuf=nil then begin
    closehandles;
    exit;
  end;
  GetMem(vbufc,conwidth*conheight*sizeof(TCharInfo));
  if vbufc=nil then begin
    FreeMem(vbuf);
    closehandles;
    exit;
  end;
  iswdosx:=GetProcAddress(GetModuleHandle('kernel32.dll'),'Borland32')<>nil;
  isnt:=false;
  if iswdosx then begin
    setlength(s,255);
    if GetEnvironmentVariableA('OS',pchar(s),255)>0 then begin
      setlength(s,strlen(pchar(s)));
      if s='Windows_NT' then
        isnt:=true;
    end;
  end;
  mousex:=0;mousey:=0;mouseb:=0;mousedbl:=false;
  shiftstate:=0;cursorstate:=0;
  textattr:=7;
  windmin:=0;
  windmax:=(conwidth-1) or ((conheight-1) shl 16);
  checkbreak:=false;
  checkeof:=false;
  checksnow:=false;
  directvideo:=true;
  lastmode:=3;
  move(output,oldoutput,sizeof(text));
  move(input,oldinput,sizeof(text));
  assigncrt(output);
  rewrite(output);
  ttextrec(output).handle:=outhandle;
  assigncrt(input);
  reset(input);
  ttextrec(input).handle:=inhandle;
  mouseoff;
  cursoron;
  crtinited:=true;
  result:=true;
end;

procedure QuitCrt;
begin
  if not crtinited then
    exit;
  FreeMem(vbuf);
  FreeMem(vbufc);
  closehandles;
  move(oldinput,input,sizeof(text));
  move(oldoutput,output,sizeof(text));
  crtinited:=false;
end;

initialization
  InitCrt;

finalization
  QuitCrt;

end.
