
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ TCP/IP Stack for native DOS Packet drivers            }
{ WDosARP.pas                                           }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.2.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   wwww:      http://www.dwp42.org                     }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosARP.pas,v 1.3 2003/10/25 22:00:07 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosARP.pas,v $
  Revision 1.3  2003/10/25 22:00:07  iwache
  Minor changes in coding format added. No functional change.

  Revision 1.2  2003/06/11 20:57:13  iwache
  Debug sequences new formatted.

  Revision 1.1  2003/05/08 18:59:56  iwache
  TCP/IP Stack for native DOS Packet drivers by Diego De Marco added.

  ----------------------------------------------------------------------------
}
unit WDosARP;

{.$DEFINE DEBUG}

interface

uses
  Classes, WDosNetwork, SysUtils;

type
  { Routine for servicing ARP packets }
  TARP_Routine = procedure(Adapter: TNetworkAdapter;
    SourceEtherAddr: TEthernetAddress;
    Operation: WORD; PLEN: BYTE; SenderHA: TEthernetAddress;
    SenderPA: Pointer; TargetHA: TEthernetAddress; TargetPA: Pointer);

function ARP_AddProcessor(Adapter: TNetworkAdapter; ProtocolType: WORD; Routine:
  TARP_Routine): Boolean;
function ARP_RemoveProcessor(Adapter: TNetworkAdapter; ProtocolType: WORD):
  Boolean;
function ARP_Send(Adapter: TNetworkAdapter; DestEtherAddr: TEthernetAddress;
  ProtocolType, Operation: WORD; PLEN: BYTE; SenderHA: TEthernetAddress;
  SenderPA: Pointer; TargetHA: TEthernetAddress; TargetPA: Pointer): Boolean;

implementation

const
  TYPELEN_ARP = Swap(Word($0806));

type
  { ARP Packet }
  PARPPacket = ^TARPPacket;
  TARPPacket = packed record
    HardwareType: WORD;
    ProtocolType: WORD;
    HLen: BYTE;
    PLen: BYTE;
    Operation: WORD;
    Data: array[0..50] of BYTE;
  end;

  { This collection contains the protocols currently using ARP }
  PProtocolEntry = ^TProtocolEntry;
  TProtocolEntry = record
    Adapter: TNetworkAdapter;
    ProtocolType: WORD;
    Routine: TARP_Routine;
  end;

var
  ProtocolList: TList;

procedure ARP_ProcessPacket(Adapter: TNetworkAdapter; Packet: PEthernetPacket);
var
  i: Integer;
  PE: PProtocolEntry;
  ARPPacket: PARPPacket;
begin
  ARPPacket := @Packet^.Data;

{$IFDEF DEBUG}
  WriteLn('*ARP*');
  WriteLn(' Hardware Type          : ', IntToHex(Swap(ARPPacket^.HardwareType), 4));
  WriteLn(' Protocol Type          : ', IntToHex(Swap(ARPPacket^.ProtocolType), 4));
  WriteLn(' Operation              : ', IntToHex(Swap(ARPPacket^.Operation), 4));
  WriteLn(' HLEN                   : ', ARPPacket^.HLEN);
  WriteLn(' PLEN                   : ', ARPPacket^.PLEN);
{$ENDIF}

  for i := 0 to ProtocolList.Count - 1 do
  begin
    PE := ProtocolList[i];
    if (PE^.Adapter = Adapter) and (PE^.ProtocolType = ARPPacket^.ProtocolType)
      then
    begin
      PE^.Routine(
        Adapter,
        Packet^.Header.SourceAddr,
        ARPPacket^.Operation,
        ARPPacket^.PLEN,
        PEthernetAddress(@ARPPacket^.Data[0])^,
        @(ARPPacket^.Data[ARPPacket^.HLEN]),
        PEthernetAddress(@ARPPacket^.Data[ARPPacket^.HLEN + ARPPacket^.PLEN])^,
        @(ARPPacket^.Data[ARPPacket^.HLEN + ARPPacket^.PLEN + ARPPacket^.HLEN])
        );
    end;
  end;
end;

function ARP_AddProcessor(Adapter: TNetworkAdapter; ProtocolType: WORD; Routine:
  TARP_Routine): Boolean;
var
  i: Integer;
  Bound: Boolean;
  PE: PProtocolEntry;
begin
  { If we're not bound to the specified adapter, it's time to do it. }
  Bound := False;
  for i := 0 to ProtocolList.Count - 1 do
  begin
    PE := ProtocolList[i];
    if PE^.Adapter = Adapter then
    begin
      Bound := True;
      Break;
    end;
  end;
  if not Bound and not Adapter.AddProtocol(TYPELEN_ARP,
    @ARP_ProcessPacket) then
  begin
    ARP_AddProcessor := False;
    Exit;
  end;
  { Ok. We are bound. Let's move on }
  New(PE);
  PE^.Adapter := Adapter;
  PE^.ProtocolType := ProtocolType;
  PE^.Routine := Routine;
  ProtocolList.Add(PE);
  ARP_AddProcessor := True;
end;

function ARP_RemoveProcessor(Adapter: TNetworkAdapter; ProtocolType: WORD):
  Boolean;
var
  PE: PProtocolEntry;
  i, Index, Count: integer;
begin
  { Check to find the correct Entry. Also calculate how many other 		}
  { other entries are using this Adapter.                                 }
  Count := 0;
  Index := -1;
  for i := 0 to ProtocolList.Count - 1 do
  begin
    PE := ProtocolList[i];
    if (PE^.Adapter = Adapter) then
    begin
      Inc(Count);
      if (PE^.ProtocolType = ProtocolType) then
      begin
        Index := i;
        Break;
      end;
    end;
  end;
  { If we didn't found the entry, there's nothing to do.                  }
  if Index = -1 then
  begin
    ARP_RemoveProcessor := False;
    Exit;
  end;
  { If this is the only entry using the adapter, we also have to dettach  }
  { from the adapter.                                                     }
  if (Count = 1) and not Adapter.RemoveProtocol(TYPELEN_ARP,
    @ARP_ProcessPacket) then
  begin
    ARP_RemoveProcessor := False;
    Exit;
  end;
  { All done. Remove the entry.                                           }
  PE := ProtocolList[Index];
  ProtocolList.Remove(PE);
  Dispose(PE);
  ARP_RemoveProcessor := True;
end;

function ARP_Send(Adapter: TNetworkAdapter; DestEtherAddr: TEthernetAddress;
  ProtocolType, Operation: WORD; PLEN: BYTE; SenderHA: TEthernetAddress;
  SenderPA: Pointer; TargetHA: TEthernetAddress; TargetPA: Pointer): Boolean;
var
  ARPP: TARPPacket;
  ETP: TEthernetPacket;
  MyEtherAddr: TEthernetAddress;
begin
  { Armamos el Paquete ARP }
  ARPP.HardwareType := Swap(Word(1));
  ARPP.ProtocolType := ProtocolType;
  ARPP.HLen := 6;
  ARPP.PLen := PLEN;
  ARPP.Operation := Operation;
  Move(SenderHA, ARPP.Data[0], ARPP.HLEN);
  Move(SenderPA^, ARPP.Data[ARPP.HLEN], ARPP.PLEN);
  Move(TargetHA, ARPP.Data[ARPP.HLEN + ARPP.PLEN], ARPP.HLEN);
  Move(TargetPA^, ARPP.Data[ARPP.HLEN + ARPP.PLEN + ARPP.HLEN], ARPP.PLEN);
  { Armamos el Paquete Ethernet }
  Adapter.GetAddress(MyEtherAddr);
  Move(DestEtherAddr, ETP.Header.TargetAddr, SizeOf(TEthernetAddress));
  Move(MyEtherAddr, ETP.Header.SourceAddr, SizeOf(TEthernetAddress));
  ETP.Header.TypeLen := TYPELEN_ARP;
  Move(ARPP, ETP.Data, SizeOf(ARPP));
  { Enviamos el Paquete Ethernet }
  ARP_Send := Adapter.SendPkt(@ETP, 150);
end;

initialization
  ProtocolList := TList.Create;
  
finalization
  ProtocolList.Free;
end.

