
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ WDosCallbacks.pas                                     }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Laszlo Agocs                    }
{ e-mail: alp@dwp42.org                                 }
{ Portitions by Immo Wache                              }
{                                                       }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosCallbacks.pas,v 1.5 2002/10/27 22:53:12 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosCallbacks.pas,v $
  Revision 1.5  2002/10/27 22:53:12  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
{
  Version history:
    ALP 2002-03-02 Version 2.0.1 released (added wheel mouse support)
    ALP 2001-06-18 Version 1.2.2 released (fixed problem with mouse callbacks)
    ALP 2001-05-31 Version 1.2.1 released (BUGFIX: EBX was not saved)
    ALP 2001-05-31 Version 1.2.0 released
}

unit WDosCallbacks;

interface

const
  MOUSEACTION_MOVE=1;
  MOUSEACTION_LEFTDOWN=2;
  MOUSEACTION_LEFTUP=4;
  MOUSEACTION_RIGHTDOWN=8;
  MOUSEACTION_RIGHTUP=16;
  MOUSEACTION_MIDDOWN=32;
  MOUSEACTION_MIDUP=64;
  MOUSEACTION_WHEEL=128; //higher byte of lower word of 'b' = wheel movement

  MOUSEBUTTON_LEFT=1;
  MOUSEBUTTON_RIGHT=2;
  MOUSEBUTTON_MID=4;

type
  TKeyboardCallback=procedure(key:byte);
  TMouseCallback=procedure(x,y,b,action:integer);pascal;

var
  IsImmoPresent:boolean=False; // :-)
  MouseAvailable:boolean; //is a mouse available? (is int33h driver loaded?)
  MouseButtons:integer; //number of mouse buttons
  MouseWheel:boolean; //true if mouse has a wheel and the driver supports it

{
  Add/Remove a keyboard callback routine.
  The 'key' parameter of the callback procedure is the value of port[$60]
  The old keyboard handler is also called AFTER calling all callbacks.
  You can have 16 callbacks at once.
}
procedure AddKeyboardCallback(cb:TKeyboardCallback);
procedure RemoveKeyboardCallback(cb:TKeyboardCallback);

{
  This flushes the keyboard buffer, because the speaker will start beeping if
  if you don't process keys in the keyboard buffer with int16h or int21h.
}
procedure FlushKeyboardBuffer;

{
  Add/Remove a mouse callback routine. The bits of 'mask' specify when the
  callback routine needs to be called (use the mouseaction_XXX constants).
  The 'x' and 'y' parameter of the callback procedure is the mouse position
  (which is not shifted right by the correct values!), 'b' is the button state,
  'action' indicates the reason of the callback call (one bit is set).
  You can have 16 callbacks at once.
}
procedure AddMouseCallback(cb:TMouseCallback;mask:integer);
procedure RemoveMouseCallback(cb:TMouseCallback);

implementation

uses WDosPorts,WDosDpmi,WDosInterrupts;

var
  oldkbdint:DpmiPmVector;
  kbdcb:array [1..16] of TKeyboardCallback;

procedure newkbdint;
var
  i,key:byte;
begin
  key:=port[$60];
  for i:=1 to 16 do if assigned(kbdcb[i]) then kbdcb[i](key);
  CallInterrupt(oldkbdint);
end;

procedure InitKeyboard;
begin
  FillChar(kbdcb,sizeof(kbdcb),0);
  GetIntVec($09,oldkbdint);
  SetIntProc($09,newkbdint);
end;

procedure DoneKeyboard;
begin
  SetIntVec($09,oldkbdint);
end;

procedure AddKeyboardCallback(cb:TKeyboardCallback);
var
  i:integer;
begin
  for i:=1 to 16 do if not assigned(kbdcb[i]) then begin
    kbdcb[i]:=cb;
    break;
  end;
end;

procedure RemoveKeyboardCallback(cb:TKeyboardCallback);
var
  i:integer;
begin
  for i:=1 to 16 do if @kbdcb[i]=@cb then begin
    kbdcb[i]:=nil;
    break;
  end;
end;

function crt_keypressed:boolean;
begin
  asm
    mov ah,1
    mov result,false
    int 16h
    jz @1
    mov result,true
  @1:
  end;
end;

procedure FlushKeyboardBuffer;
begin
  while crt_keypressed do asm
    xor ax,ax
    int 16h
  end;
end;

const
  MouseStackSize=$1000;

var
  mousecb:array [1..16] of record
    proc:TMouseCallback;
    mask:integer;
  end;
  DataSel:word;
  MouseStack:array [0..MouseStackSize-1] of byte;
  _ax,_bx,_cx,_dx:longint;

procedure mousehandler;assembler;
asm
  push ds
  mov ds,cs:DataSel
  and eax,0000FFFFh
  and ebx,0000FFFFh
  and ecx,0000FFFFh
  and edx,0000FFFFh
  mov _ax,eax
  mov _bx,ebx
  mov _cx,ecx
  mov _dx,edx
  push gs
  push fs
  push es
  mov edx,ss
  mov eax,esp
  push ds
  pop ss
  mov ecx,offset MouseStack
  add ecx,MouseStackSize
  mov esp,ecx
  push edx
  mov edx,ds
  mov fs,edx
  mov es,edx
  push eax

  cld
  lea esi,mousecb
  mov ecx,16
@cb_loop:
  lodsd
  mov edx,eax
  cmp edx,0
  je @cb_next
  lodsd
  and eax,_ax
  jz @cb_notnow
  push ecx
  push esi
  push _cx
  push _dx
  push _bx
  push _ax
  call edx
  pop esi
  pop ecx
@cb_notnow:
  sub esi,4
@cb_next:
  add esi,4
  loop @cb_loop

  db 0Fh,0B2h,24h,24h //lss esp,[esp]
  pop es
  pop fs
  pop gs
  pop ds
  retf
end;

procedure InstallMouseHandler(mask:integer;proc:pointer);
begin
  asm
    push ebx
    push es
    mov ecx,mask
    mov edx,proc
    mov eax,12
    push cs
    pop es
    int 33h
    pop es
    pop ebx
  end;
end;

procedure InitMouse;
var
  _bx:word;
begin
  asm
    push ebx
    mov DataSel,ds
    mov MouseAvailable,True
    mov ax,0
    int 33h
    mov _bx,bx
    cmp ax,0
    jne @end
    mov MouseAvailable,False
  @end:
    pop ebx
  end;
  if MouseAvailable then begin
    case _bx of
      $0000:MouseButtons:=3; //other than two (???)
      $FFFF:MouseButtons:=2; //two (???)
      else MouseButtons:=_bx;
    end;
    MouseWheel:=false;
    //wheel support is based on wheelapi.txt of ctmouse 2.0a1
    asm
      push ebx
      mov ax,3fh
      int 33h
      cmp ax,574dh
      jne @not_wm
      and bx,1
      and cx,1
      cmp bx,1
      jne @not_wm
      cmp cx,1
      jne @not_wm
      mov MouseWheel,true
    @not_wm:
      pop ebx
    end;
    FillChar(mousecb,sizeof(mousecb),0);
    InstallMouseHandler($7f,@mousehandler);
  end;
end;

procedure DoneMouse;
begin
  InstallMouseHandler(0,@mousehandler);
  asm
    push ebx
    mov ax,0
    int 33h
    pop ebx
  end;
end;

procedure AddMouseCallback(cb:TMouseCallback;mask:integer);
var
  i:integer;
begin
  for i:=1 to 16 do if not assigned(mousecb[i].proc) then begin
    mousecb[i].proc:=cb;
    mousecb[i].mask:=mask;
    break;
  end;
end;

procedure RemoveMouseCallback(cb:TMouseCallback);
var
  i:integer;
begin
  for i:=1 to 16 do if @mousecb[i].proc=@cb then begin
    mousecb[i].proc:=nil;
    mousecb[i].mask:=0;
    break;
  end;
end;

initialization
  InitKeyboard;
  InitMouse;

finalization
  DoneMouse;
  DoneKeyboard;

end.
