
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ WDosDpmi.pas                                          }
{                                                       }
{ for Delphi 4, 5, 6                                    }
{ Version 2.0                                           }
{ Copyright (c) 2001 by Immo Wache                      }
{ e-mail: immo.wache@dwp42.org                          }
{ www: http://www.dwp42.org                             }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{ this software contains portions from:                 }
{                                                       }
{   - WDOSX DOS Extender                                }
{     Copyright (c) 1996, 1998, Michael Tippach         }
{     e-mail: mtippach@gmx.net                          }
{     www: http://www.wuschel.demon.co.uk               }
{     Realeased under the terms of the WDOSX            }
{     license agreement.                                }
{                                                       }
{ Please inform you about special copyrights for these  }
{ portions on the apropriate web pages.                 }
{                                                       }
{*******************************************************}
{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosDpmi.pas,v 1.3 2002/10/27 22:53:12 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosDpmi.pas,v $
  Revision 1.3  2002/10/27 22:53:12  iwache
  CVS tags "Header" and "Log" added.

  ----------------------------------------------------------------------------
}
unit WDosDpmi;

interface

uses
  WDosCheck;

const
  DPMIError   = -1;
  DPMISuccess = 0;

type
  { structure for Real-Mode register, ref. to DPMI specification }
  DpmiRmRegisters = packed record
    case Integer of
    0: (EDI, ESI, EBP, EXX, EBX, EDX, ECX, EAX: LongInt;
        Flags, ES, DS, FS, GS, IP, CS, SP, SS: Word);
    1: (DI, DIH, SI, SIH, BP, BPH, XX, XXH: Word;
        case Integer of
        0: (BX, BXH, DX, DXH, CX, CXH, AX, AXH: Word);
        1: (BL, BH, BLH, BHH, DL, DH, DLH, DHH,
            CL, CH, CLH, CHH, AL, AH, ALH, AHH: Byte))
  end;

  DpmiDescriptor = Integer;
  DpmiStatus = Integer;

  RealModePtr = record
    RmOffset: Word;
    RmSegment: Word;
  end;

function DpmiFreeDescriptor(Selector: DpmiDescriptor): DpmiStatus;
    register; assembler;
function DpmiRmSelector(Segment: Integer; var Selector:
    DpmiDescriptor): DpmiStatus; register; assembler;
function DpmiGetRmVector(IntNo: Integer; var RmProc: RealModePtr): DpmiStatus;
    register; assembler;
function DpmiRealModeInt(IntNo: Integer; var Regs: DpmiRmRegisters):
    DpmiStatus; register; assembler;
function DpmiAllocDOSMem(NumPara: Integer; var Segment: Integer;
    var Selector: DpmiDescriptor): DpmiStatus; register; assembler;
function DpmiFreeDOSMem(Selector: DpmiDescriptor): DpmiStatus; register;
    assembler;
function DpmiRealModeFarCall(Reset: Boolean; var Regs: DpmiRmRegisters;
    StackSize: Integer): DpmiStatus; register; assembler;
function DpmiRealModeIntCall(Reset: Boolean; var Regs: DpmiRmRegisters;
    StackSize: Integer): DpmiStatus; register; assembler;

procedure MoveFar(const Source; const SourceSel: DpmiDescriptor;
    var Dest; const DestSel: DpmiDescriptor; Count: Integer);

implementation

const
  DPMIINT = $31;

function DpmiFreeDescriptor(Selector: DpmiDescriptor): DpmiStatus;
    register; assembler;
asm
        PUSH    EBX
        MOV     EBX,EAX
        MOV     EAX,1
        INT     DPMIINT
        SBB     EAX,EAX
        POP     EBX
end;

function DpmiRmSelector(Segment: Integer; var Selector:
    DpmiDescriptor): DpmiStatus; register; assembler;
asm
        PUSH    EBX
        MOV     EBX,EAX
        MOV     EAX,2
        PUSH    EDX
        INT     DPMIINT
        POP     EBX
        MOV     [EBX],EAX
        SBB     EAX,EAX
        POP     EBX
end;

function DpmiGetRmVector(IntNo: Integer; var RmProc: RealModePtr): DpmiStatus;
    register; assembler;
asm
        PUSH    EBX
        MOV     EBX,EAX
        MOV     EAX,200H
        PUSH    EDX
        INT     DPMIINT
        POP     EBX
        MOV     RealModePtr [EBX].RmOffset,DX
        MOV     RealModePtr [EBX].RmSegment,CX
        SBB     EAX,EAX
        POP     EBX
end;

function DpmiRealModeInt(IntNo: Integer; var Regs: DpmiRmRegisters):
    DpmiStatus; register; assembler;
asm
        PUSH    EBX
        PUSH    EDI
        MOV     EBX,EAX
        MOV     EDI,EDX
        MOV     EAX,300H
        XOR     ECX,ECX
        INT     DPMIINT
        SBB     EAX,EAX
        POP     EDI
        POP     EBX
end;

function DpmiAllocDOSMem(NumPara: Integer; var Segment: Integer;
    var Selector: DpmiDescriptor): DpmiStatus; register; assembler;
asm
        PUSH    EBX
        MOV     EBX,EAX
        MOV     EAX,100H
        PUSH    EDX
        INT     DPMIINT
        POP     EBX
        MOV     [EBX],EAX
        MOV     [ECX],EDX
        SBB     EAX,EAX
        POP     EBX
end;

function DpmiFreeDOSMem(Selector: DpmiDescriptor): DpmiStatus; register;
    assembler;
asm
        MOV     EDX,EAX
        MOV     EAX,101H
        INT     DPMIINT
        SBB     EAX,EAX
end;

function DpmiRealModeFarCall(Reset: Boolean; var Regs: DpmiRmRegisters;
    StackSize: Integer): DpmiStatus; register; assembler;
asm
        PUSH    EBX
        PUSH    EDI
        XOR     EBX,EBX
        CMP     EBX,EAX
        JE      @@00
        INC     BH
  @@00:
        MOV     EDI,EDX
        MOV     EAX,301H
        INT     DPMIINT
        SBB     EAX,EAX
        POP     EDI
        POP     EBX
end;


function DpmiRealModeIntCall(Reset: Boolean; var Regs: DpmiRmRegisters;
    StackSize: Integer): DpmiStatus; register; assembler;
asm
        PUSH    EBX
        PUSH    EDI
        XOR     EBX,EBX
        CMP     EBX,EAX
        JE      @@00
        INC     BH
  @@00:
        MOV     EDI,EDX
        MOV     EAX,302H
        INT     DPMIINT
        SBB     EAX,EAX
        POP     EDI
        POP     EBX
end;

procedure MoveFar(const Source; const SourceSel: DpmiDescriptor;
    var Dest; const DestSel: DpmiDescriptor; Count: Integer);
asm
        PUSH    DS
        PUSH    ES
        PUSH    ESI
        PUSH    EDI
        MOV     DS, SourceSel
        MOV     ES, DestSel
        MOV     ESI,Source
        MOV     EDI,Dest
        MOV     ECX,Count
        MOV     EDX,ECX
        AND     EDX,3
        SHR     ECX,1
        SHR     ECX,1
        CLD
        REP     MOVSD
        MOV     ECX,EDX
        REP     MOVSB
        POP     EDI
        POP     ESI
        POP     ES
        POP     DS
end;

end.
