
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ TCP/IP Stack for native DOS Packet drivers            }
{ WDosIP.pas                                            }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.2.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   wwww:      http://www.dwp42.org                     }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosIP.pas,v 1.4 2003/06/11 20:59:54 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosIP.pas,v $
  Revision 1.4  2003/06/11 20:59:54  iwache
  Procedure WDosIPInitalization and
  procedure WDosIPFinalization added to solve
  a units initialization problem with WDosSock.
  Now WDosSock calls these procedures.

  Revision 1.3  2003/05/11 20:01:37  iwache
  New parameter in function GetAdapterFromAddress: AnyAddr: Boolean = False
  This parameter is for allow to find an Adapter with IP=0.0.0.0 which is needed
  to let DHCP work correctly. This fix a but with wrong ARP response telegrams.
  AnyAddr is only used in procedure IP_ProcessPacket.

  Revision 1.2  2003/05/11 15:05:05  iwache
  Function GetAdapterFromAddress modified to make it work
  with DHCP protocol. Now finds an adapter when the adapter
  IP address is INADDR_ANY but the target IP is different.

  Revision 1.1  2003/05/08 18:59:57  iwache
  TCP/IP Stack for native DOS Packet drivers by Diego De Marco added.

  ----------------------------------------------------------------------------
}
unit WDosIP;

{.$DEFINE DEBUG}

interface

uses
  SysUtils, Classes, WDosClock, WDosNetwork, WDosSock, WDosARP, Threads;

type
  { 4 Byte IP address}
  PIPAddress = ^TIPAddress;
  TIPAddress = Cardinal;

  { Rutina para procesar paquetes IP }
  TIP_Routine = procedure(SourceIP, TargetIP: TIPAddress;
    var Data; DataLen: Word);

  TIPAdapterConfig = record
    DHCP: Boolean;
    IPAddress: TIPAddress;
    NetMask: TIPAddress;
    Gateway: TIPAddress;
  end;

  PRouteEntry = ^TRouteEntry;
  TRouteEntry = record
    Destination: TIPAddress;
    Mask: TIPAddress;
    Gateway: TIPAddress;
    LocalIP: TIPAddress;
    Cost: Integer;
  end;

  TAddressEntry = class
    // Basic parameters
    IPAddress: TIPAddress; // IP Address
    NetMask: TIPAddress; // Net Mask
    Adapter: TNetworkAdapter; // Adapter
    // DHCP-related parameters
    DHCPInUse: Boolean; // Are we using DHCP for this address ?
    DHCPServer: TIPAddress; // DHCP Server used to get it
    DHCPT1: Int64; // Renewal Time
    DHCPT2: Int64; // Rebinding Time
    DHCPT3: Int64; // End-of-lease Time
  end;

  { General IP Management }
function IP_Init(Adapter: TNetworkAdapter; Config: TIPAdapterConfig): Cardinal;
function IP_Deinit(IPHandle: Cardinal): Boolean;
function IP_GetInfo(Index: Integer; var Entry: TAddressEntry): Boolean;
function IP_AddProcessor(Protocol: BYTE; Routine: TIP_Routine): Boolean;
function IP_RemoveProcessor(Protocol: BYTE): Boolean;
function IP_Checksum(Data: Pointer; DataCnt: WORD; PIP: pointer; PIPCnt: WORD):
  WORD;
function IP_Send(LocalIP, TargetIP: TIPAddress; Protocol: BYTE; var Data;
  DataLen: WORD): Integer;
{ Conversion Routines }
function IP_AddrToString(IP: TIPAddress): string;
function IP_StringToAddr(StrAddr: string): TIPAddress;
{ Routing Management }
function IP_AddRoute(Destination, Mask, Gateway, LocalIP: TIPAddress; Cost:
  Integer): Boolean;
function IP_DeleteRoute(Destination, Mask, Gateway, LocalIP: TIPAddress):
  Boolean;
function IP_GetRoute(Index: Integer; var Entry: TRouteEntry): Boolean;
function IP_GetRouteForTarget(TargetIP: TIPAddress; var LocalIP, Gateway:
  TIPAddress): Boolean;
{ DHCP Management }
function IP_GetDHCPOption(Option: BYTE; var Buffer; var BuffLen: Integer):
  Boolean;
function IP_GetIPByHandle(IPHandle: Cardinal): TIPAddress;

procedure WDosIPInitalization;

procedure WDosIPFinalization;

implementation

uses
  WDosICMP, WDosUDP, WDosTCP, WDosDHCP;

const
  TYPELEN_IP = Swap(Word($0800));

type
  PIPHeader = ^TIPHeader;
  TIPHeader = packed record
    VersHLen: BYTE;
    ServiceType: BYTE;
    TotalLength: WORD;
    ID: WORD;
    FlagsFrag1: BYTE;
    Frag2: BYTE;
    LifeTime: BYTE;
    Protocol: BYTE;
    Checksum: WORD;
    SourceIP: TIPAddress;
    TargetIP: TIPAddress;
  end;

  TARPCacheEntry = record
    Expires: Int64;
    IPAddress: TIPAddress;
    EthernetAddress: TEthernetAddress;
  end;

  TIPProcessors = record
    Protocol: BYTE;
    Routine: TIP_Routine;
  end;

  TFragment = record
    Offset: WORD;
    Data: Pointer;
    DataLen: Integer;
  end;

  TDatagram = record
    SourceIP: TIPAddress;
    TargetIP: TIPAddress;
    Id: WORD;
    Protocol: BYTE;
    FragCount: BYTE;
    TotalLength: WORD;
    ReceivedLength: WORD;
    Timeout: Int64;
    Fragments: array[1..40] of TFragment;
  end;

var
  RouteList: TList;
  AddressList: TList;
  CheckingDHCP: Boolean;
  Datagrams: array[1..10] of TDatagram;
  Processors: array[1..10] of TIPProcessors;
  ARPCacheList: array[1..100] of TARPCacheEntry;

procedure IP_ProcessPacket(Adapter: TNetworkAdapter; Packet: PEthernetPacket);
  forward;

{**************************************************************************}
{                                                                          }
{                   Manejo de Adaptadores de Red para IP                   }
{                                                                          }
{**************************************************************************}

function IP_GetIPByHandle(IPHandle: Cardinal): TIPAddress;
var
  AE: TAddressEntry;
begin
  Result := 0;
  AE := TAddressEntry(IPHandle);
  if AddressList.IndexOf(AE) = -1 then
    Exit;
  Result := AE.IPAddress;
end;

function GetAdapterFromAddress(IPAddress: TIPAddress;
    AnyAddr: Boolean = False): TNetworkAdapter;
var
  i: Integer;
  AE: TAddressEntry;
begin
  for i := 0 to AddressList.Count - 1 do
  begin
    AE := AddressList[i];
    if (AE.IPAddress = IPAddress)
        or (AnyAddr and (AE.IPAddress = INADDR_ANY)) then
    begin
      GetAdapterFromAddress := AE.Adapter;
      Exit;
    end;
  end;
  GetAdapterFromAddress := nil;
end;

procedure IP_CheckDHCPEntry(var Entry: TAddressEntry);
var
  i: Integer;
  DW: Cardinal;
  Timeout: Int64;
  RE: PRouteEntry;
  Gateway: TIPAddress;
begin
  // If the entry is still valid, there's nothing to do
  if (Entry.DHCPT1 > GetMSCounter) and (Entry.DHCPT2 > GetMSCounter)
    and (Entry.DHCPT3 > GetMSCounter) then
    Exit;
  // Es hora de renovar nuestro IP
  if Entry.IPAddress = 0 then
  begin
    // Este adaptador perdi su direccin IP.
    // Hay que nogociarla de cero.
    if DHCP_ConfigAdapter(Entry.Adapter, Entry.IPAddress,
      Entry.NetMask, Gateway, Entry.DHCPServer, DW) then
    begin
      Timeout := DW;
      Timeout := Timeout * 1000;
      Entry.DHCPT1 := GetMSCounter + (Timeout div 2);
      Entry.DHCPT2 := GetMSCounter + ((Timeout div 8) * 7);
      Entry.DHCPT3 := GetMSCounter + Timeout;
      (* WriteLn(Entry.DHCPT1);
         WriteLn(Entry.DHCPT2);
         WriteLn(Entry.DHCPT3); *)
        // Agregamos la ruta de loopback para el adaptador
      IP_AddRoute(Entry.IPAddress, $FFFFFFFF, INADDR_LOOPBACK, INADDR_LOOPBACK,
        0);
      // Agregamos la ruta para la red local del adaptador
      IP_AddRoute(Entry.IPAddress and Entry.NetMask, Entry.NetMask,
        Entry.IPAddress, Entry.IPAddress, 0);
      // Agregamos la ruta del gateway
      if Gateway <> 0 then
        IP_AddRoute(0, 0, Gateway, Entry.IPAddress, 9999);
      {            WriteLn('Se volvi a conseguir un IP');}
         { Notification: we got a new IP address }
      Net_NotifyNetChanged;
    end
    else
    begin
      Entry.IPAddress := 0;
      Entry.DHCPT1 := GetMSCounter + MaxInt;
      Entry.DHCPT2 := GetMSCounter + MaxInt;
      Entry.DHCPT3 := GetMSCounter + 60000; // Reintentar en 1 minuto
      //          WriteLn('NO se volvi a conseguir un IP');
    end;
  end
  else
  begin
    // Intentamos renovar la direccin IP
    if DHCP_RenewIP(Entry.Adapter, Entry.IPAddress,
      Entry.DHCPServer, DW) then
    begin
      Timeout := DW;
      Timeout := Timeout * 1000;
      Entry.DHCPT1 := GetMSCounter + (Timeout div 2);
      Entry.DHCPT2 := GetMSCounter + ((Timeout div 8) * 7);
      Entry.DHCPT3 := GetMSCounter + Timeout;
      //          WriteLn('Se renov el IP');
      Exit;
    end;
    // No se pudo renovar la direccin IP. En qu punto estbamos ?
    if Entry.DHCPT1 < GetMSCounter then
    begin
      // Fallamos al renovar en T1
      Entry.DHCPT1 := GetMSCounter + MaxInt;
      //          WriteLn('No se renov el IP en T1');
    end
    else if Entry.DHCPT2 < GetMSCounter then
    begin
      // Fallamos al renovar en T2
      Entry.DHCPT2 := GetMSCounter + MaxInt;
      //          WriteLn('No se renov el IP en T2');
    end
    else if Entry.DHCPT3 < GetMSCounter then
    begin
      // Fallamos al renovar en T3. El IP ya no es vlido.
      // Eliminamos los items de la tabla de ruteo que corresponden a
      // este IP
      i := 0;
      while i < RouteList.Count do
      begin
        RE := RouteList[i];
        if (RE^.Destination = Entry.IPAddress) or
          (RE^.LocalIP = Entry.IPAddress) then
        begin
          // Hay que eliminar este item
          RouteList[i];
          Dispose(RE);
        end
        else
        begin
          Inc(i);
        end;
      end;
      // Ponemos esta direccin en cero.
      Entry.DHCPT1 := GetMSCounter + MaxInt;
      Entry.DHCPT2 := GetMSCounter + MaxInt;
      Entry.DHCPT3 := GetMSCounter + 60000; // Reintentar en 1 minuto
      Entry.IPAddress := 0;
      //          WriteLn('Se perdi el IP');
         { Notification: we lost one IP address }
      Net_NotifyNetChanged;
    end;
  end;
end;

procedure IP_CheckDHCP;
var
  i: Integer;
  AE: TAddressEntry;
begin
  CheckingDHCP := True;
  for i := 0 to AddressList.Count - 1 do
  begin
    AE := AddressList[i];
    if AE.DHCPInUse then
      IP_CheckDHCPEntry(AE);
  end;
  CheckingDHCP := False;
end;

{**************************************************************************}
{                                                                          }
{                          Manejo de ARP para IP                           }
{                                                                          }
{**************************************************************************}

function GetEthernetAddress(IPAddress: TIPAddress;
  var EthernetAddress: TEthernetAddress): Boolean;
(*
  Esta rutina busca en el Cache de direcciones de hardware la direccin de
  un IP dado. Si la direccin no est en el Cache, devuelve False.
*)
var
  i: integer;
begin
  GetEthernetAddress := True;
  for i := Low(ARPCacheList) to High(ARPCacheList) do
  begin
    if (ARPCacheList[i].IPAddress = IPAddress)
      and (ARPCacheList[i].Expires > GetMSCounter) then
    begin
      EthernetAddress := ARPCacheList[i].EthernetAddress;
      Exit;
    end;
  end;
  GetEthernetAddress := False;
end;

procedure SetEthernetAddress(IPAddress: TIPAddress;
  EthernetAddress: TEthernetAddress);
(*
  Esta rutina actualiza el Cache de direcciones de hardware. Si el Cache
  est lleno, se elige una entrada al azar entre las usadas, y se utiliza
  esa. Si la entrada ya est en la lista, no se hace nada.
*)
var
  i: integer;
  CacheEntry: Integer;
begin
  { Si la direccin de Hardware es nula o es broadcast, la ignoramos }
  if Ethernet_Compare(EthernetAddress, ETH_BROADCASTADDRESS) then
    Exit;
  if Ethernet_Compare(EthernetAddress, ETH_NULLADDRESS) then
    Exit;
  { Si la direccin IP es nula o es broadcast, la ignoramos }
  if (IPAddress = 0) then
    Exit;
  if (IPAddress = INADDR_BROADCAST) then
    Exit;
  { Si la entrada ya est en el Cache, no hacemos nada }
  if GetEthernetAddress(IPAddress, EthernetAddress) then
    Exit;
  { Si hay lugar en el Cache, agregamos la entrada. Si no, elegimos }
{ una entrada al azar.                                            }
  CacheEntry := Random(High(ARPCacheList)) + 1;
  for i := Low(ARPCacheList) to High(ARPCacheList) do
  begin
    if (ARPCacheList[i].Expires < GetMSCounter) then
    begin
      CacheEntry := i;
      Break;
    end;
  end;
  { Actualizamos la entrada en cuestin. Es vlida por 2 minutos }
  ARPCacheList[CacheEntry].IPAddress := IPAddress;
  ARPCacheList[CacheEntry].EthernetAddress := EthernetAddress;
  ARPCacheList[CacheEntry].Expires := GetMSCounter + 120000;
end;

procedure ProcessARP(Adapter: TNetworkAdapter;
  SourceEtherAddr: TEthernetAddress; Operation: WORD;
  PLEN: BYTE; SenderHA: TEthernetAddress; SenderPA: Pointer;
  TargetHA: TEthernetAddress; TargetPA: Pointer);
(*
  Recibimos un mensaje ARP de IP. Tenemos que ver si es una pregunta
  o una respuesta. Si es una pregunta para nosotros, la contestamos.
  En cualquier caso, anotamos las direcciones IP de los participantes,
  para actualizar nuestra tablita de IPs.
*)
var
  AA: TNetworkAdapter;
  MyEtherAddr: TEthernetAddress;
begin
(*
  { Debug }
  if Swap(Operation) = 1 then
  begin
    WriteLn('  *ARP-IP* (El equipo ',
    IP_AddrToString(PIPAddress(SenderPA)^), ' busca al equipo ',
    IP_AddrToString(PIPAddress(TargetPA)^), ')');
  end;
*)

{$IFDEF DEBUG}
  WriteLn(' Operation              : ', Swap(Operation));
  WriteLn(' Sender Hardware Address: ', Ethernet_AddrToString(SenderHA));
  WriteLn(' Target Hardware Address: ', Ethernet_AddrToString(TargetHA));
  WriteLn(' Sender IP Address      : ', IP_AddrToString(PIPAddress(SenderPA)^));
  WriteLn(' Target IP Address      : ', IP_AddrToString(PIPAddress(TargetPA)^));
{$ENDIF}

  { Revisamos si hay algo que configurar con el DHCP }
  if not CheckingDHCP then
    IP_CheckDHCP;

  { Actualizamos las direcciones de hardware de los participantes }
  SetEthernetAddress(PIPAddress(SenderPA)^, SenderHA);
  SetEthernetAddress(PIPAddress(TargetPA)^, TargetHA);

  { Si es una pregunta ARP para nosotros, la contestamos }
  if (Swap(Word(Operation)) = 1) then
  begin
    AA := GetAdapterFromAddress(PIPAddress(TargetPA)^);
    if AA <> nil then
    begin
      AA.GetAddress(MyEtherAddr);
      ARP_Send(Adapter, SourceEtherAddr, TYPELEN_IP, Swap(Word(2)), 4,
        MyEtherAddr, TargetPA, SenderHA, SenderPA);
    end;
  end;
end;

function GetHardwareAddress(TargetIP, LocalIP: TIPAddress;
  var EthernetAddress: TEthernetAddress): Boolean;
(*
  Queremos saber la direccin Ethernet de un equipo con un IP dado. Para
  eso, emitimos una peticin ARP a todos los equipos de nuestra red
  (broadcast) y esperamos una respuesta. Si al final de la espera, el IP
  solicitado est en nuestra tabla de IPs, lo hemos conseguido. De lo
  contrario, devolvemos FALSE.
*)
var
  Timeout: Int64;
  Adapter: TNetworkAdapter;
  MyEthernetAddress: TEthernetAddress;
begin
  Result := False;
  { Revisamos para ver si la direccin Ethernet est en Cache }
  if GetEthernetAddress(TargetIP, EthernetAddress) then
  begin
    Result := True;
    Exit;
  end;
  Adapter := GetAdapterFromAddress(LocalIP);
  Adapter.GetAddress(MyEthernetAddress);
  { Enviamos la solicitud ARP }
  if not ARP_Send(Adapter, ETH_BROADCASTADDRESS, TYPELEN_IP, Swap(Word(1)), 4,
    MyEthernetAddress, @LocalIP, ETH_NULLADDRESS, @TargetIP) then
    Exit;
  { Esperamos el resultado }
  Timeout := GetMSCounter + 500; // 1/2 segundo
  while (GetMSCounter < Timeout) do
  begin
    NET_Work_Low;
    if GetEthernetAddress(TargetIP, EthernetAddress) then
    begin
      Result := True;
      Exit;
    end
    else
    begin
      Yield;
    end;
  end;
  { Timeout: Nadie contest nuestro ARP }
end;

{**************************************************************************}
{                                                                          }
{              Inicializacin y Deinicializacin de IP                     }
{                                                                          }
{**************************************************************************}

function BindAdapter(Adapter: TNetworkAdapter): Boolean;
var
  i: Integer;
  Bound: Boolean;
  AE: TAddressEntry;
begin
  BindAdapter := False;
  { Check to see if the specified adapter is already bound              }
  Bound := False;
  for i := 0 to AddressList.Count - 1 do
  begin
    AE := AddressList[i];
    if AE.Adapter = Adapter then
    begin
      Bound := True;
      Break;
    end;
  end;
  { If the adapter is not bound, we have to bind it now                 }
  if not Bound then
  begin
    { Bind to IP }
    if not Adapter.AddProtocol(TYPELEN_IP, @IP_ProcessPacket) then
      Exit;
    { Bind to ARP }
    if not ARP_AddProcessor(Adapter, TYPELEN_IP, @ProcessARP) then
    begin
      Adapter.RemoveProtocol(TYPELEN_IP, @IP_ProcessPacket);
      Exit;
    end;
  end;
  { All Done }
  BindAdapter := True;
end;

function IP_Init(Adapter: TNetworkAdapter; Config: TIPAdapterConfig): Cardinal;
{ This function is intended to install the IP protocol over the given 	  }
{ adapter. If DHCP is specified in the configuration record, the supplied }
{ IP address, netmask and gateway are ignored, and they're obtained (if   }
{ possible) from a DHCP server.                                           }
{ If DHCP is not used, the Adapter is assigned the IP address, netmask,   }
{ and Gateway indicated by the corresponding fields in the configuration  }
{ record.                                                                 }
var
  AE: TAddressEntry;
begin
  IP_Init := 0;
  { Validamos los parmetros de entrada                                 }
  if not Config.DHCP and
    ((Config.IPAddress = 0) or (Config.NetMask = 0)) then
    Exit;
  { Bind this adapter to ARP and IP			    					  }
  if (Adapter <> nil) and not BindAdapter(Adapter) then
    Exit;
  { Registramos esta direccin IP en la lista de direcciones            }
  AE := TAddressEntry.Create;
  AddressList.Add(AE);
  AE.Adapter := Adapter;
  AE.DHCPInUse := Config.DHCP;
  { Intentamos obtener una direccin IP por DHCP }
  if Config.DHCP then
  begin
    AE.DHCPInUse := True;
    AE.IPAddress := 0;
    AE.DHCPT1 := 0;
    AE.DHCPT2 := 0;
    AE.DHCPT3 := 0;
    { Intentamos obtener el IP ahora (puede fallar). Si falla no   }
    { importa: el DHCP intentar peridicamente obtener un IP para }
    { esta direccin.                                              }
    IP_CheckDHCP;
  end
  else
  begin
    AE.IPAddress := Config.IPAddress;
    AE.NetMask := Config.NetMask;
    { Add a loopback-route for the new IP }
    IP_AddRoute(AE.IPAddress, $FFFFFFFF, INADDR_LOOPBACK, INADDR_LOOPBACK, 0);
    { Add a route for the local network of the new adapter }
    IP_AddRoute(AE.IPAddress and Config.NetMask, Config.NetMask, AE.IPAddress,
      AE.IPAddress, 0);
    { Add a default route for the Gateway, if there's one }
    if Config.Gateway <> 0 then
    begin
      IP_AddRoute(0, 0, Config.Gateway, Config.IPAddress, 9999);
    end;
  end;
  { Notification: we have a new protocol bound to the adapter }
  Net_NotifyNetChanged;
  { Done }
  IP_Init := Cardinal(AE);
end;

function IP_Deinit(IPHandle: Cardinal): Boolean;
var
  Index: Integer;
  Found: Boolean;
  AE: TAddressEntry;
  Adapter: TNetworkAdapter;
begin
  AE := TAddressEntry(IPHandle);

  { Search for the indicated IPHandle }
  Index := AddressList.IndexOf(AE);
  if Index = -1 then
  begin
    Result := False;
    Exit;
  end;

  { if the Address was obtained via DHCP, we have to release it }
  if (AE.DHCPInUse) and (AE.IPAddress <> 0) then
  begin
    DHCP_Release(AE.Adapter, AE.IPAddress, AE.DHCPServer);
  end;

  { Deallocate the Entry and remove it from the list }
  Adapter := AE.Adapter;
  AddressList.Remove(AE);
  AE.Free;

  { If the adapter is no longer in use, we have to unbind it          }
  if Adapter <> nil then
  begin
    Found := False;
    for Index := 0 to AddressList.Count - 1 do
    begin
      AE := AddressList[Index];
      if (AE.Adapter = Adapter) then
      begin
        Found := True;
        Break;
      end;
    end;
    if not Found then
    begin
      Adapter.RemoveProtocol(TYPELEN_IP, @IP_ProcessPacket);
      ARP_RemoveProcessor(Adapter, TYPELEN_IP);
    end;
  end;

  { Done }
  IP_Deinit := True;
end;

function IP_GetInfo(Index: Integer; var Entry: TAddressEntry): Boolean;
begin
  if (Index >= 0) and (Index < AddressList.Count) then
  begin
    Entry := AddressList[Index];
    Result := True;
  end
  else
  begin
    Result := False;
  end;
end;

{**************************************************************************}
{                                                                          }
{              Rutinas varias de conversin de direcciones IP              }
{                                                                          }
{**************************************************************************}

function IP_AddrToString(IP: TIPAddress): string;

  function IStr(i: Integer): string;
  var
    s: string;
  begin
    Str(i, s);
    IStr := s;
  end;

var
  Arr: packed array[0..3] of byte;
begin
  Move(IP, Arr, 4);
  IP_AddrToString :=
    IStr(Arr[0]) + '.' + IStr(Arr[1]) + '.' +
    IStr(Arr[2]) + '.' + IStr(Arr[3]);
end;

function IP_GetDHCPOption(Option: BYTE; var Buffer; var BuffLen: Integer):
  Boolean;
var
  i: Integer;
  AE: TAddressEntry;
begin
  { Search for an adapter with DHCP installed }
  for i := 0 to AddressList.Count - 1 do
  begin
    AE := TAddressEntry(AddressList[i]);
    if AE.DHCPInUse then
    begin
      { Found it !. Get the requested option }
      Result := DHCP_GetInfo(AE.Adapter, AE.IPAddress, AE.DHCPServer, Option,
        Buffer, BuffLen);
      Exit;
    end;
  end;
  Result := False;
end;

function IP_StringToAddr(StrAddr: string): TIPAddress;
var
  Part: string;
  IPAddress: TIPAddress;
  i, j, k, Code: Integer;
  Arr: array[0..3] of BYTE;
begin
  IP_StringToAddr := 0;
  for i := 0 to 3 do
  begin
    if i = 3 then
    begin
      Part := StrAddr;
    end
    else
    begin
      j := Pos('.', StrAddr);
      if j = 0 then
        Exit;
      Part := Copy(StrAddr, 1, j - 1);
      StrAddr := Copy(StrAddr, j + 1, 255);
    end;
    { Validamos esta parte de la direccin IP }
    Val(Part, k, Code);
    if (Code <> 0) or (k < 0) or (k > 255) then
      Exit;
    { Seguimos }
    Arr[i] := k;
  end;
  Move(Arr, IPAddress, 4);
  IP_StringToAddr := IPAddress;
end;

function IP_CompareNetwork(IP1, IP2, NetMask: TIPAddress): Boolean;
{ Esta funcin devuelve True si las direcciones P1 y P2 estn en la }
{ misma subred, dada la mscara NetMask.                            }
begin
  { Enmascaramos el IP 1 }
  IP1 := IP1 and NetMask;
  { Enmascaramos el IP 2 }
  IP2 := IP2 and NetMask;
  { Comparamos los resultados }
  IP_CompareNetwork := (IP1 = IP2);
end;

{**************************************************************************}
{                                                                          }
{           Rutinas de procesamiento de datagramas IP entrantes            }
{                                                                          }
{**************************************************************************}

function IP_Checksum(Data: Pointer; DataCnt: WORD; PIP: pointer; PIPCnt: WORD):
  WORD;
type
  PBYTE = ^BYTE;
var
  pw: ^WORD;
  loop: word;
  x, w: word;
  csum: Cardinal;
begin
  IP_Checksum := 0;
  csum := 0;
  if DataCnt = 0 then
    exit;
  loop := DataCnt shr 1;

  if pip <> nil then
  begin
    pw := pip;
    for x := 1 to (PipCnt div 2) do
    begin
      csum := csum + pw^;
      inc(pw);
    end;
  end;

  pw := data;
  for x := 1 to loop do
  begin
    csum := csum + pw^;
    inc(pw);
  end;

  if (DataCnt mod 2 = 1) then
  begin
    w := PBYTE(pw)^;
    csum := csum + w;
  end;

  csum := (csum mod $10000) + (csum div $10000);
  csum := csum + (csum shr 16);
  IP_Checksum := Word(not csum);
end;

function GetPacketData(P: PIPHeader; var DataLen: Word): Pointer;
var
  HLen: BYTE;
begin
  HLen := (P^.VersHLen and $0F) * 4; { En bytes }
  GetPacketData := PChar(@P^) + HLen;
  DataLen := Swap(Word(P^.TotalLength)) - HLen;
end;

procedure IP_ProcessDatagram(Protocol: BYTE; SourceIP, TargetIP: TIPAddress;
  var Data; DataLen: WORD);
var
  i: BYTE;
begin
  { Revisamos el Tipo de Protocolo, y Ejecutamos el Procesador que }
  { corresponda                                                    }
  for i := Low(Processors) to High(Processors) do
  begin
    if Processors[i].Protocol = Protocol then
    begin
{$IFDEF DEBUG}
      WriteLn('Packet from ', IP_AddrToString(SourceIP), ' to ',
        IP_AddrToString(TargetIP), ' processed');
{$ENDIF}
      Processors[i].Routine(SourceIP, TargetIP, Data, DataLen);
    end;
  end;
end;

procedure IP_AddFragment(ID: WORD; Protocol: BYTE;
  SourceIP, TargetIP: TIPAddress; var Data; DataLen, Offset: WORD;
  Last: Boolean);
var
  i, dg, dgfree, FragPos: integer;
begin
  { Nos aseguramos de que exista memoria para almacenar el nuevo fragmento }
//  if MaxAvail < (DataLen + 1024) then Exit;
{ Buscamos el Datagrama (si existe) }
  dg := 0;
  dgfree := 0;
  for i := Low(Datagrams) to High(Datagrams) do
  begin
    if (Datagrams[i].FragCount = 0) and (dgfree = 0) then
      dgfree := i;
    if (Datagrams[i].Id = Id) and (SourceIP = Datagrams[i].SourceIP) then
    begin
      dg := i;
      break;
    end;
  end;
  if dg + dgfree = 0 then
    Exit;
  { Si es necesario, agregamos un nuevo datagrama }
  if (dg = 0) and (dgfree <> 0) then
  begin
    dg := dgfree;
    Datagrams[dg].SourceIP := SourceIP;
    Datagrams[dg].TargetIP := TargetIP;
    Datagrams[dg].Id := ID;
    Datagrams[dg].Protocol := Protocol;
    Datagrams[dg].TotalLength := 0; { an no lo sabemos }
    Datagrams[dg].ReceivedLength := 0; { an no se recibi nada }
    Datagrams[dg].FragCount := 0;
    Datagrams[dg].Timeout := GetMSCounter + 2000; { 2 segundos }
    FillChar(Datagrams[dg].Fragments, SizeOf(Datagrams[dg].Fragments), 0);
  end;
  { Buscamos la posicin del nuevo fragmento en el datagrama }
  FragPos := 1;
  for i := Datagrams[dg].FragCount downto 1 do
  begin
    if (Offset > Datagrams[dg].Fragments[i].Offset) then
    begin
      FragPos := i + 1;
      Break;
    end;
  end;
  { Corremos los fragmentos para agregar el nuevo fragmento }
  Inc(Datagrams[dg].FragCount);
  for i := Datagrams[dg].FragCount downto FragPos + 1 do
  begin
    Datagrams[dg].Fragments[i] := Datagrams[dg].Fragments[i - 1];
  end;
  { Agregamos el nuevo fragmento }
  Datagrams[dg].Fragments[FragPos].Offset := Offset;
  GetMem(Datagrams[dg].Fragments[FragPos].Data, DataLen);
  Move(Data, Datagrams[dg].Fragments[FragPos].Data^, Datalen);
  Datagrams[dg].Fragments[FragPos].DataLen := DataLen;
  { Actualizamos los campos del datagrama }
  Inc(Datagrams[dg].ReceivedLength, DataLen);
  if Last then
    Datagrams[dg].TotalLength := (Offset * 8) + DataLen;
end;

procedure IP_CheckDatagrams;
var
  i, f: Integer;
  PData, PP: PChar;
begin
  { Procesamos los datagramas completos }
  for i := Low(Datagrams) to High(Datagrams) do
  begin
    if (Datagrams[i].FragCount > 0) and
      (Datagrams[i].ReceivedLength = Datagrams[i].TotalLength) then
    begin
      { Armamos el Datagrama final }
      if GetMemCheck(Pointer(PData), Datagrams[i].TotalLength) then
      begin
        PP := PData;
        for f := 1 to Datagrams[i].FragCount do
        begin
          Move(Datagrams[i].Fragments[f].Data^, PP^,
            Datagrams[i].Fragments[f].DataLen);
          FreeMem(Datagrams[i].Fragments[f].Data,
            Datagrams[i].Fragments[f].DataLen);
          Inc(PP, Datagrams[i].Fragments[f].DataLen);
        end;
        { Ahora que tenemos el datagrama armado, lo procesamos }
        IP_ProcessDatagram(Datagrams[i].Protocol,
          Datagrams[i].SourceIP, Datagrams[i].TargetIP, PData^,
          Datagrams[i].TotalLength);
        { Eliminanos el datagrama recin creado }
        FreeMem(PData, Datagrams[i].TotalLength);
        { Eliminamos el datagrama usado }
        FillChar(Datagrams[i], SizeOf(Datagrams[i]), 0);
      end;
    end;
  end;
  { Eliminanos los datagramas incompletos }
  for i := Low(Datagrams) to High(Datagrams) do
  begin
    if (Datagrams[i].Timeout < GetMSCounter) then
    begin
      { Liberamos la memoria de los fragmentos }
      for f := 1 to Datagrams[i].FragCount do
      begin
        FreeMem(Datagrams[i].Fragments[f].Data,
          Datagrams[i].Fragments[f].DataLen);
      end;
      { Eliminamos el datagrama usado }
      FillChar(Datagrams[i], SizeOf(Datagrams[i]), 0);
    end;
  end;
end;

procedure IP_ProcessFragment(ID: WORD; Protocol: BYTE;
  SourceIP, TargetIP: TIPAddress; var Data; DataLen, Offset: WORD;
  Last: Boolean);
begin
  IP_AddFragment(ID, Protocol, SourceIP, TargetIP, Data, DataLen, Offset, Last);
end;

function IP_IsBroadcast(TargetIP: TIPAddress): Boolean;
var
  IPArr: packed array[1..4] of BYTE;
begin
  Move(TargetIP, IPArr, 4);
  Result := (IPArr[1] = 255) or (IPARR[2] = 255)
    or (IPArr[3] = 255) or (IPArr[4] = 255);
end;

procedure IP_ProcessPacket(Adapter: TNetworkAdapter; Packet: PEthernetPacket);
var
  Flags: BYTE;
  PData: PChar;
  DataLen: Word;
  Pkt: PIPHeader;
  //	FragmentLen: WORD
  FragmentOffset: WORD;
begin
  Pkt := @Packet^.Data;

  { Revisamos si hay algo que configurar con el DHCP }
  if not CheckingDHCP then
    IP_CheckDHCP;

  { Revisamos la versin de IP. Slo aceptamos la versin 4 }
  if (Pkt^.VersHLen shr 4) <> 4 then
    Exit;

{$IFDEF DEBUG}
  WriteLn('IP fragment from ', IP_AddrToString(Pkt^.SourceIP),
      ' to ', IP_AddrToString(Pkt^.TargetIP), ' Protocol ',
      IntToHex(Pkt.Protocol, 2));
{$ENDIF}

  { Si el paquete no es para nosotros, lo ignoramos }
  if not IP_IsBroadcast(Pkt^.TargetIP) and
    (GetAdapterFromAddress(Pkt^.TargetIP, True) = nil) then
  begin
{$IFDEF DEBUG}
    WriteLn('IP packet rejected: no valid IP or broadcast IP');
{$ENDIF}
    Exit;
  end;

  { Tomamos los Datos del Paquete }
  Flags := Pkt^.FlagsFrag1 shr 5;
  FragmentOffset := Pkt^.Frag2 + (Pkt^.FlagsFrag1 and $1F) * $100;
  //	FragmentLen := Swap(Word(Pkt^.TotalLength));
  PData := GetPacketData(Pkt, DataLen);

  { Debug }
(*  WriteLn(' HLEN:            ', Pkt^.VersHLen and $0F);
WriteLn(' Protocol:        ', Pkt^.Protocol);
WriteLn(' Identification:  ', Pkt^.Id);
WriteLn(' Flags:           ', Flags);
WriteLn(' Fragment Offset: ', FragmentOffset);
WriteLn(' Fragment Length: ', FragmentLen);
WriteLn(' Checksum:        ', IntToHex(Pkt^.Checksum, 4));
Pkt^.Checksum := 0;
WriteLn(' Checksum Recal.: ', IntToHex(IP_Checksum(@Pkt^, SizeOf(TIPHeader), nil, 0), 4));
*)

  { Cacheamos la direccin IP del que nos enva el paquete }
//	SetEthernetAddress(Pkt^.SourceIP, Packet^.Header.SourceAddr);
  { Si este paquete es todo el datagrama, lo procesamos directamente }
  if ((Flags and 1) = 0) and (FragmentOffset = 0) then
  begin
    IP_ProcessDatagram(Pkt^.Protocol, Pkt^.SourceIP, Pkt^.TargetIP,
      PData^, DataLen);
  end
  else
  begin
    { El paquete es slo un fragmento del datagrama. Lo agregamos a }
       { lista de fragmentos.                                          }
    IP_CheckDatagrams;
    IP_ProcessFragment(Pkt^.Id, Pkt^.Protocol, Pkt^.SourceIP,
      Pkt^.TargetIP, PData^, DataLen, FragmentOffset, (Flags and 1) = 0);
    IP_CheckDatagrams;
  end;
end;

function IP_AddProcessor(Protocol: BYTE; Routine: TIP_Routine): Boolean;
var
  i: integer;
begin
  IP_AddProcessor := False;
  IP_RemoveProcessor(Protocol);
  for i := Low(Processors) to High(Processors) do
  begin
    if not Assigned(Processors[i].Routine) then
    begin
      Processors[i].Protocol := Protocol;
      Processors[i].Routine := Routine;
      IP_AddProcessor := True;
      Exit;
    end;
  end;
end;

function IP_RemoveProcessor(Protocol: BYTE): Boolean;
var
  i: integer;
begin
  IP_RemoveProcessor := False;
  for i := Low(Processors) to High(Processors) do
  begin
    if (Processors[i].Protocol = Protocol) then
    begin
      Processors[i].Protocol := 0;
      Processors[i].Routine := nil;
      IP_RemoveProcessor := True;
      Exit;
    end;
  end;
end;

{**************************************************************************}
{                                                                          }
{                    Rutinas de envo de datagramas IP                     }
{                                                                          }
{**************************************************************************}

function IP_AddRoute(Destination, Mask, Gateway, LocalIP: TIPAddress; Cost:
  Integer): Boolean;
var
  P: PRouteEntry;
  i, Index: Integer;
begin
  { Primero buscamos en qu lugar insertar la nueva ruta }
  Index := -1;
  for i := 0 to RouteList.Count - 1 do
  begin
    P := RouteList[i];
    if P^.Cost > Cost then
    begin
      Index := i;
      break;
    end;
  end;
  { Agregamos la ruta a la lista }
  New(P);
  P^.Destination := Destination;
  P^.Mask := Mask;
  P^.LocalIP := LocalIP;
  P^.Gateway := Gateway;
  P^.Cost := Cost;
  if Index = -1 then
    RouteList.Add(P)
  else
    RouteList.Insert(Index, P);
  { Listo }
  Result := True;
end;

function IP_DeleteRoute(Destination, Mask, Gateway, LocalIP: TIPAddress):
  Boolean;
var
  i: Integer;
  P: PRouteEntry;
begin
  i := 0;
  Result := False;
  while i < RouteList.Count do
  begin
    P := RouteList[i];
    if (P^.Destination = Destination) and (P^.Mask = Mask)
      and (P^.LocalIP = LocalIP) and (P.Gateway = Gateway) then
    begin
      RouteList.Delete(i);
      Dispose(P);
      Result := True;
    end
    else
    begin
      Inc(i);
    end;
  end;
end;

function IP_GetRoute(Index: Integer; var Entry: TRouteEntry): Boolean;
var
  P: PRouteEntry;
begin
  if Index < RouteList.Count then
  begin
    P := RouteList[Index];
    Entry := P^;
    Result := True;
  end
  else
  begin
    Result := False;
  end;
end;

function IP_GetRouteForTarget(TargetIP: TIPAddress; var LocalIP, Gateway:
  TIPAddress): Boolean;
var
  i: Integer;
  RE: PRouteEntry;
begin
  { Si no tenemos ninguna direccin IP configurada, nos vamos }
  if RouteList.Count = 0 then
  begin
    IP_GetRouteForTarget := False;
    Exit;
  end;
  { Primero recorremos la tabla de ruteo para encontrar una ruta }
  { apropiada }
  for i := 0 to RouteList.Count - 1 do
  begin
    RE := RouteList[i];
    if (RE^.Destination <> 0) and (IP_CompareNetwork(RE^.Destination,
      TargetIP, RE^.Mask)) then
    begin
      LocalIP := RE^.LocalIP;
      Gateway := RE^.Gateway;
      IP_GetRouteForTarget := True;
      Exit;
    end;
  end;
  { No encontramos una ruta vlida. Buscamos entonces un gateway por }
  { default }
  for i := 0 to RouteList.Count - 1 do
  begin
    RE := RouteList[i];
    if (RE^.Destination = 0) then
    begin
      LocalIP := RE^.LocalIP;
      Gateway := RE^.Gateway;
      IP_GetRouteForTarget := True;
      Exit;
    end;
  end;
  { No hay una ruta adecuada, ni tampoco un gateway default }
  IP_GetRouteForTarget := False;
end;

function IP_Snd(
  Adapter: TNetworkAdapter;
  DestEth: TEthernetAddress;
  Protocol: BYTE;
  LocalIP, TargetIP: TIPAddress;
  var Data;
  DataLen: Integer
  ): Integer;
const
  IP_MAXDATA = 1480;
type
  TTrama = packed record
    EthHeader: TEthernetHeader;
    IPHeader: TIPHeader;
    IPData: packed array[0..IP_MAXDATA - 1] of BYTE;
  end;
var
  PData: PChar;
  Trama: TTrama;
  FragLen: WORD;
  FragOffs: WORD;
  LocalEth: TEthernetAddress;
begin
  Adapter.GetAddress(LocalEth);
  { Debug }
 (*  WriteLn('Enviando un datagrama desde ', Ethernet_AddrToString(LocalEth),
    ' (', IP_AddrToString(LocalIP), ') hacia ', Ethernet_AddrToString(DestEth),
    ' (', IP_AddrToString(TargetIP), ') con ', DataLen, ' bytes.');
 *)
  { Armamos los datos fijos de la trama Ethernet, que se repiten en   }
  { Todos los fragmentos.                                             }
  Trama.EthHeader.TargetAddr := DestEth;
  Trama.EthHeader.SourceAddr := LocalEth;
  Trama.EthHeader.TypeLen := TYPELEN_IP;
  { Armamos los datos fijos de la cabecera IP, que se repiten en      }
  { todos los fragmentos.                                             }
  Trama.IPHeader.VersHLen := $45; { VERS = 4; HLEN = 5 }
  Trama.IPHeader.ServiceType := 0;
  Trama.IPHeader.ID := Random($FFFF) + 1;
  Trama.IPHeader.LifeTime := 128;
  Trama.IPHeader.Protocol := Protocol;
  Trama.IPHeader.SourceIP := LocalIP;
  Trama.IPHeader.TargetIP := TargetIP;
  { Componemos los fragmentos, metindolos directamente en la trama   }
  { Ethernet.                                                         }
  PData := @Data;
  FragOffs := 0;
  while DataLen > 0 do
  begin
    { Modificamos los flags y dems de este fragmento }
    Trama.IPHeader.FlagsFrag1 := FragOffs shr 8;
    Trama.IPHeader.Frag2 := FragOffs and $FF;
    if DataLen > IP_MAXDATA then
    begin
      Trama.IPHeader.FlagsFrag1 :=
        Trama.IPHeader.FlagsFrag1 or 32; { More fragments }
      FragLen := IP_MAXDATA;
    end
    else
    begin
      FragLen := DataLen;
    end;
    Trama.IPHeader.TotalLength := Swap(Word(FragLen + SizeOf(TIPHeader)));
    Trama.IPHeader.Checksum := 0;
    Trama.IPHeader.Checksum := IP_Checksum(@Trama.IPHeader, SizeOf(TIPHeader),
      nil, 0);
    { Agregamos los datos }
    Move(PData^, Trama.IPData, FragLen);
    { Enviamos el Paquete Ethernet }
    Adapter.SendPkt(@Trama, SizeOf(TEthernetHeader) + SizeOf(TIPHeader) +
      FragLen);
    { Nos preparamos para el Fragmento siguiente }
    Inc(FragOffs, FragLen div 8);
    Inc(PData, FragLen);
    Dec(DataLen, FragLen);
  end;
  IP_Snd := 0;
end;

function IP_Send(LocalIP, TargetIP: TIPAddress; Protocol: BYTE; var Data;
  DataLen: WORD): Integer;
var
  i: Integer;
  AE: TAddressEntry;
  Adapter: TNetworkAdapter;
  DestEth: TEthernetAddress;
  Gateway, TempIP: TIPAddress;
begin
  { Si no tenemos ningn IP, no hay nada que hacer }
 //	if AddressList.Count = 0 then begin
 //		IP_Send := WSAENETUNREACH;
 //		Exit;
 //	end;

  { Si es una direccin de broadcast, hay que salir por los IPs         }
  { que correspondan (que pueden ser ms de uno). Si no es un broadcast }
  { tenemos que elegir un adaptador/gateway y sacar el paquete por l.  }
  if IP_IsBroadcast(TargetIP) then
  begin
    { The destination IP is the Broadcast Address. We have to   }
    { cycle through all the addresses                           }
    for i := 0 to AddressList.Count - 1 do
    begin
      AE := AddressList[i];
      Adapter := AE.Adapter;
      if Assigned(Adapter) then
        IP_Snd(Adapter, ETH_BROADCASTADDRESS,
          Protocol, AE.IPAddress, TargetIP, Data, DataLen);
    end;
    Result := 0;
  end
  else
  begin
    { El destino no es un broadcast. Buscamos una ruta para sacar el  }
    { datagrama.                                                      }
    if not IP_GetRouteForTarget(TargetIP, TempIP, Gateway) then
    begin
      IP_Send := WSAENETUNREACH;
      Exit;
    end;
    if LocalIP = 0 then
      LocalIP := TempIP
    else
    begin
      if TempIP = Gateway then
        Gateway := LocalIP;
    end;
    { Si el IP local resulta ser el LOOPBACK, procesamos el datagrama }
    { en forma interna                                                }
    if LocalIP = INADDR_LOOPBACK then
    begin
      IP_ProcessDatagram(Protocol, INADDR_LOOPBACK, INADDR_LOOPBACK,
        Data, DataLen);
      IP_Send := 0;
      Exit;
    end;
    {  If we must go through a gateway in order to reach the	}
    { destination, we have to direct our datagram to the        }
    { hardware address of the gateway. Otherwise, the datagram  }
    { is directed to the hardware address of the destination    }
    { itself.                                                   }
    Adapter := GetAdapterFromAddress(LocalIP);
    if (LocalIP = Gateway) then
    begin
      if not GetHardwareAddress(TargetIP, LocalIP, DestEth) then
      begin
        { We couldn't determine the destination's hardware  }
        { address. }
        IP_Send := WSAEHOSTDOWN;
        Exit;
      end;
    end
    else
    begin
      if not GetHardwareAddress(Gateway, LocalIP, DestEth) then
      begin
        { We couldn't determine the gateway's hardware      }
        { address.                                          }
        IP_Send := WSAEHOSTDOWN;
        Exit;
      end;
    end;
    IP_Send := IP_Snd(Adapter, DestEth, Protocol, LocalIP, TargetIP, Data,
      DataLen);
  end;
end;

var
  RE: PRouteEntry;

procedure WDosIPInitalization;
begin
  //
  CheckingDHCP := False;
  // Address List
  AddressList := TList.Create;
  // Route List
  RouteList := TList.Create;
  //
  FillChar(ARPCacheList, SizeOf(ARPCacheList), 0);
  FillChar(Processors, SizeOf(Processors), 0);
  FillChar(Datagrams, SizeOf(Datagrams), 0);
  // Inicializamos ICMP
  ICMP_Init;
  // Inicializamos UDP
  UDP_Init;
  // Inicializamos el TCP
  TCP_Init;
  // Armamos la tabla de ruteo bsica, conteniendo slo el ruteo
  // hacia el LOOPBACK.
  IP_AddRoute(IP_StringToAddr('127.0.0.0'), IP_StringToAddr('255.0.0.0'),
    INADDR_LOOPBACK, INADDR_LOOPBACK, 0);
end;

procedure WDosIPFinalization;
begin
  { Delete Addresses & Address List }
  while AddressList.Count > 0 do
  begin
    IP_Deinit(Cardinal(AddressList[0]));
  end;
  AddressList.Free;

  { Delete Routes & Route Table }
  while RouteList.Count > 0 do
  begin
    RE := RouteList[0];
    Dispose(RE);
    RouteList.Delete(0);
  end;
  RouteList.Free;
end;

end.

