
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ TCP/IP Stack for native DOS Packet drivers            }
{ WDosNE2000.pas                                        }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.2.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   wwww:      http://www.dwp42.org                     }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosNE2000.pas,v 1.3 2003/09/22 18:46:30 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosNE2000.pas,v $
  Revision 1.3  2003/09/22 18:46:30  iwache
  Bug in TNE2000Adapter.Poll fixed. The pushf
  command breaks the proper stack structure.
  CLI is not needed in this procedure. This
  is fixed.
  Comments are translated from Diego Di Marco
  from Spanish to English. Thanks for it.

  Revision 1.2  2003/08/15 21:07:24  iwache
  Missing type PWord for D5 compatibility added.

  Revision 1.1  2003/05/08 18:59:57  iwache
  TCP/IP Stack for native DOS Packet drivers by Diego De Marco added.

  ----------------------------------------------------------------------------
}
unit WDosNE2000;

interface

uses
  Crt, WDosPorts, WDosNetwork, WDosClock;

type
  TNE2000PacketEntry = packed record
    Size: WORD;
    Packet: TEthernetPacket;
  end;

  { Adapter Object }
  TNE2000Adapter = class(TNetworkAdapter)
  private
    PORT_COMMAND: WORD;
    PORT_DATACONFIG: WORD;
    PORT_RBCR0: WORD;
    PORT_RBCR1: WORD;
    PORT_RXCONFIG: WORD;
    PORT_TXPAGESTART: WORD;
    PORT_TXCONFIG: WORD;
    PORT_RXPSTART: WORD;
    PORT_RXBOUNDARY: WORD;
    PORT_RXPSTOP: WORD;
    PORT_ISR: WORD;
    PORT_IMR: WORD;
    PORT_RXCURR: WORD;
    PORT_RSAR0: WORD;
    PORT_TBCR0: WORD;
    PORT_TBCR1: WORD;
    {}
    nextpkt: WORD;
    eeprom: array[0..$1F] of byte;
    {}
    FBuffer: array[0..31] of TNE2000PacketEntry;
    RecvBuffer: array[0..2047] of BYTE;
    SendBuffer: array[0..2047] of BYTE;

    UltimaVez: Int64;
    UltimoBoundary: byte;
  protected
    procedure Poll;
    procedure InitBoard;
    procedure NICToPC(page, buffwhere: word);
    procedure PCToNIC(page, fromwhere: word);
    procedure CheckStatus; override;
    function GetPacket(var Packet: TEthernetPacket): Boolean; override;
    function IntAddProtocol(TypeLen: WORD): Boolean; override;
    function IntRemoveProtocol(TypeLen: WORD): Boolean; override;
  public
    function GetAddress(var Addr: TEthernetAddress): Boolean; override;
    function SendPkt(Buffer: Pointer; BufLen: Word): Boolean; override;
    constructor Create(IOBase: WORD; Int: BYTE);
    destructor Destroy; override;
  end;

implementation

type
  PWord  = ^Word;

const
  NE_IMR = $00; { Interupt mask register value }

function ReadPort(nPort: WORD): BYTE;
begin
  asm
        nop
        nop
        nop
        nop
        nop
        nop
        nop
        nop
        nop
  end;
  ReadPort := Port[nPort];
  asm
        nop
        nop
        nop
        nop
        nop
        nop
        nop
        nop
        nop
  end;
end;

function ReadPortW(nPort: WORD): WORD;
begin
  asm
        nop
        nop
        nop
        nop
        nop
        nop
        nop
        nop
        nop
  end;
  ReadPortW := PortW[nPort];
  asm
        nop
        nop
        nop
        nop
        nop
        nop
        nop
        nop
        nop
  end;
end;

procedure WritePort(nPort: WORD; Value: BYTE);
begin
  asm
        nop
        nop
        nop
        nop
        nop
        nop
        nop
        nop
        nop
  end;
  Port[nPort] := Value;
  asm
        nop
        nop
        nop
        nop
        nop
        nop
        nop
        nop
        nop
  end;
end;

procedure WritePortW(nPort, Value: WORD);
begin
  asm
        nop
        nop
        nop
        nop
        nop
        nop
        nop
        nop
        nop
  end;
  PortW[nPort] := Value;
  asm
        nop
        nop
        nop
        nop
        nop
        nop
        nop
        nop
        nop
  end;
end;

constructor TNE2000Adapter.Create(IOBase: WORD; Int: BYTE);
begin
  { Variables }
  FillChar(FBuffer, SizeOf(FBuffer), 0);

  { All pages }
  PORT_COMMAND := IOBase;
  { Page 0 }
  PORT_RXPSTART := IOBase + $01;
  PORT_RXPSTOP := IOBase + $02;
  PORT_RXBOUNDARY := IOBase + $03;
  PORT_TXPAGESTART := IOBase + $04;
  PORT_TBCR0 := IOBase + $05;
  PORT_TBCR1 := IOBase + $06;
  PORT_ISR := IOBase + $07;
  PORT_RSAR0 := IOBase + $08;
  PORT_RBCR0 := IOBase + $0A;
  PORT_RBCR1 := IOBase + $0B;
  PORT_RXCONFIG := IOBase + $0C;
  PORT_TXCONFIG := IOBase + $0D;
  PORT_DATACONFIG := IOBase + $0E;
  PORT_IMR := IOBase + $0F;
  { Page 1 }
  PORT_RXCURR := IOBase + $07;

  inherited Create;
end;

destructor TNE2000Adapter.Destroy;
begin
  WritePort(PORT_COMMAND, $21); { offline }
  WritePort(PORT_IMR, 0); { no int's }
  inherited Destroy;
end;

procedure TNE2000Adapter.CheckStatus;
begin
  InitBoard;
  FStatus := asOk;
end;

procedure TNE2000Adapter.InitBoard;
var
  i: Integer;
begin
  NextPkt := $4100;
  {}
  WritePort(PORT_COMMAND, $21); { nic page 0, stop }
  Delay(10);
  WritePort(PORT_DATACONFIG, $59); { Data Configuration }
  WritePort(PORT_RBCR0, 0); { remote byte count, low byte }
  WritePort(PORT_RBCR1, 0); { remote byte count, high byte }
  WritePort(PORT_RXCONFIG, $20); { Receive in monitor mode }
  WritePort(PORT_TXCONFIG, 2); { loopback }
  { Read eeprom information }
  nictopc(0, 0);
  for i := 0 to $1F do
    eeprom[i] := RecvBuffer[i];
  WritePort(PORT_DATACONFIG, $59); { Data Configuration }
  WritePort(PORT_RXPSTART, $40); { rx page start = $4000}
  WritePort(PORT_RXBOUNDARY, $40); { rx boundary = $4000 }
  WritePort(PORT_RXPSTOP, $79); { rx page stop = $79FF }
  WritePort(PORT_ISR, $FF); { Clear pending interrupts }
  WritePort(PORT_IMR, NE_IMR); { Enable Interrupts }
  { Page 1 }
  { Program the Ethernet Address }
  WritePort(PORT_COMMAND, $61); { nic page 1 }
  WritePort(PORT_COMMAND + $1, eeprom[0]);
  WritePort(PORT_COMMAND + $2, eeprom[2]);
  WritePort(PORT_COMMAND + $3, eeprom[4]);
  WritePort(PORT_COMMAND + $4, eeprom[6]);
  WritePort(PORT_COMMAND + $5, eeprom[8]);
  WritePort(PORT_COMMAND + $6, eeprom[10]);
  { Program multicast address }
  WritePort(PORT_COMMAND + $8, $FF);
  WritePort(PORT_COMMAND + $9, $FF);
  WritePort(PORT_COMMAND + $A, $FF);
  WritePort(PORT_COMMAND + $B, $FF);
  WritePort(PORT_COMMAND + $C, $FF);
  WritePort(PORT_COMMAND + $D, $FF);
  WritePort(PORT_COMMAND + $E, $FF);
  WritePort(PORT_COMMAND + $F, $FF);
  { RX Current Page }
  WritePort(PORT_RXCURR, $41); { rx current page = $4100 }
  { Page 0 }
  WritePort(PORT_COMMAND, $22); { Start! }
  WritePort(PORT_TXCONFIG, 0); { tx configuration }
  WritePort(PORT_RXCONFIG, $0C); { accept unicast and broadcast }
end;

procedure TNE2000Adapter.Poll;
var
  i: Integer;
  Ok: Boolean;
  Count, Size, blocks: WORD;
  bv, btemp, ctemp, dist: BYTE;
begin
  { why was the IRQ triggered ? }
  bv := ReadPort(PORT_ISR);

  { Overflow }
  if (bv and $10) = $10 then
  begin
	{ When there's an overflow, we need to reset the board }
    WriteLn('Overflow');
    WriteLn('------------------------------------------------------------');
    InitBoard;
    Exit;
  end;

  { RXOK }
  if (bv and 1) = 1 then
  begin

    repeat
      nictopc(nextpkt, 0);
      Size := WORD(WORD(RecvBuffer[3]) shl 8) + RecvBuffer[2];
      blocks := (Size div 256) + 1;

	  { Is there anything in the board's queue? }
      WritePort(PORT_COMMAND, $62); { nic page 1 }
      ctemp := ReadPort(PORT_RXCURR);
      WritePort(PORT_COMMAND, $22); { nic page 0 }
      btemp := ReadPort(PORT_RXBOUNDARY);

	  { Calculate distance from "write" pointer to "read" pointer }
	  if ctemp >= btemp then
		dist := ctemp - btemp
	  else
	  begin
		dist := ($78 - btemp) + (ctemp - $40) + 1;
	  end;

	  if dist < 2 then
		break;

	  { If any pointer has an incorrect value, dump on the screen }
	  if (btemp < $40) or (btemp > $78) or (ctemp < $40)
		or (ctemp > $78) {or (ftemp < $40) or (ftemp > $78)}
	  or (Size > 1518) or (size < 54) then
	  begin
		WriteLn('*************************************************');
		Writeln('Dist=', dist, ' Size=', size, ' blocks=', blocks);
		WriteLn('B=', btemp, ' RcvBf[1]=', RecvBuffer[1], ' C=', ctemp);
		if (Dist > 7) then
		begin
		  writeln('[RESET because of incorrect pointer distance]');
		  InitBoard;
		  break;
		end;
		if UltimoBoundary = btemp then
		begin
		  if GetMSCounter - UltimaVez > 10 then
		  begin
			writeln('[RESET because of timeout]');
			InitBoard;
			break;
		  end;
		end
		else
		begin
		  UltimaVez := GetMSCounter;
		  UltimoBoundary := btemp;
		end;
		break;
	  end;

	  { Does the packet fit in the calculated distance }
	  Ok := blocks < dist;

	  { If there's nothing to read, we quit }
	  if not Ok then
		Break;

	  { clear the received bit }
	  WritePort(PORT_ISR, 1);

	  { See what we have to read. First of all, advance the         }
      { boundary.                                                   }
      Inc(btemp);
      if btemp > $78 then
        btemp := $40;
      WritePort(PORT_RXBOUNDARY, btemp);

	  { Advance NextPkt }
      inc(nextpkt, 256);
      if nextpkt > $7800 then
        nextpkt := $4000;

	  { Read other blocks }
      Count := 256;
      while Count < Size do
      begin
        nictopc(nextpkt, count);
		{ Advance boundary }
		inc(btemp);
		if btemp > $78 then
		  btemp := $40;
		WritePort(PORT_RXBOUNDARY, btemp);
		{ Advance NextPkt }
		inc(nextpkt, 256);
		if nextpkt > $7800 then
		  nextpkt := $4000;
		{ Increase read bytes }
		inc(Count, 256);
	  end;

	  { We need to update the boundary, to set it to the value }
	  { calculated before                                      }
	  if (RecvBuffer[1] < $40) or (RecvBuffer[1] > $78) then
	  begin
		writeln('RecvBuffer[1] fuera de rango: ', RecvBuffer[1]);
		writeln('btemp=', btemp, ' ctemp=', ctemp, ' size=', size);
	  end
	  else
	  begin
		btemp := RecvBuffer[1];
		nextpkt := Word(btemp) shl 8;
		dec(btemp);
		if btemp < $40 then
		  btemp := $40;
		WritePort(PORT_RXBOUNDARY, btemp);
	  end;

	  { Copy the received packet to the packet list }
	  for i := Low(FBuffer) to High(FBuffer) do
	  begin
		if FBuffer[i].Size = 0 then
		begin
		  FBuffer[i].Size := Size - 4;
          Move(RecvBuffer[4], FBuffer[i].Packet, Size - 4);
          break;
        end;
      end;

    until False;

  end;

  { TXOK }
  if (bv and 2) = 2 then
  begin
    WritePort(PORT_ISR, 2); { clear transmitted bit }
  end;

  { RXerr }
  if (bv and 4) = 4 then
  begin
    WritePort(PORT_ISR, 4); { Clear rx error bit }
  end;

  { TXERR }
  if (bv and 8) = 8 then
  begin
    WritePort(PORT_ISR, 8); { clear tx error bit }
  end;
end;

function TNE2000Adapter.GetPacket(var Packet: TEthernetPacket): Boolean;
var
  i: integer;
begin
  Poll;
  for i := Low(FBuffer) to High(FBuffer) do
  begin
    if FBuffer[i].Size > 0 then
    begin
      Move(FBuffer[i].Packet, Packet, FBUffer[i].Size);
      FBuffer[i].Size := 0;
      Result := True;
      Exit;
    end;
  end;
  Result := False;
end;

function TNE2000Adapter.IntAddProtocol(TypeLen: WORD): Boolean;
begin
  Result := True;
end;

function TNE2000Adapter.IntRemoveProtocol(TypeLen: WORD): Boolean;
begin
  Result := True;
end;

function TNE2000Adapter.GetAddress(var Addr: TEthernetAddress): Boolean;
var
  i: Integer;
begin
  i := 0;
  repeat
    Addr[i div 2] := eeprom[i];
    Inc(i, 2);
  until i > $0A;
  Result := True;
end;

function TNE2000Adapter.SendPkt(Buffer: Pointer; BufLen: Word): Boolean;
var
  i: WORD;
begin
  Move(Buffer^, SendBuffer, BufLen);
  if BufLen < 64 then
    BufLen := 64;
  // #iwa-2003-09-16: may be this pushf should be removed too
  asm
       pushf
       cli
  end;
  WritePort(PORT_ISR, $40 + 2 + 8);
    { Clear "remote DMA complete" and "Transfer completed" }
  WritePort(PORT_TBCR0, Lo(BufLen)); { Size }
  WritePort(PORT_TBCR1, Hi(BufLen));
  for i := 0 to (BufLen div 256) do
    pctonic($7A00 + (i shl 8), i shl 8);
  WritePort(PORT_TXPAGESTART, $7A);
  WritePort(PORT_COMMAND, $26); { start transmission }
  { Wait for the packet to be sent }
  for i := 1 to 500 do
    if (ReadPort(PORT_COMMAND) and 4) = 0 then
      break;
  asm
       popf
  end;
  Result := True;
end;

procedure TNE2000Adapter.NICToPC(Page: Word; buffwhere: Word);
var
  P: PChar;
  i: Integer;
begin
  WritePortW(PORT_RSAR0, Page);
  WritePortW(PORT_RBCR0, 256);
  WritePort(PORT_COMMAND, $0A); { remote read }
  P := @RecvBuffer[buffwhere];
  i := 0;
  while i < 128 do
  begin
    PWord(P)^ := ReadPortW(PORT_COMMAND + $10);
    Inc(P, 2);
    Inc(i);
  end;
  WritePort(PORT_ISR, $40); { clear DMA transfer complete flag }
end;

procedure TNE2000Adapter.PCToNIC(Page, FromWhere: word);
var
  P: PChar;
  i: Integer;
begin
  WritePortW(PORT_RSAR0, page);
  WritePortW(PORT_RBCR0, 256);
  WritePort(PORT_COMMAND, $12); { remote write }
  P := @SendBuffer[FromWhere];
  i := 0;
  while (i < 128) do
  begin
    WritePortW(PORT_COMMAND + $10, PWord(P)^);
    Inc(P, 2);
    Inc(i);
  end;
  WritePort(PORT_ISR, $40); { clear DMA transfer complete flag }
end;

end.

