
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ TCP/IP Stack for native DOS Packet drivers            }
{ WDosNetBIOS.pas                                       }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.2.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   wwww:      http://www.dwp42.org                     }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosNetBIOS.pas,v 1.3 2003/10/25 22:00:07 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosNetBIOS.pas,v $
  Revision 1.3  2003/10/25 22:00:07  iwache
  Minor changes in coding format added. No functional change.

  Revision 1.2  2003/08/15 21:09:19  iwache
  Missing type PWord for D5 compatibility added.
  Warning switch off removed for compatibility too.
  This warnings are useless for automatic DWPL
  build process.

  Revision 1.1  2003/05/08 18:59:57  iwache
  TCP/IP Stack for native DOS Packet drivers by Diego De Marco added.

  ----------------------------------------------------------------------------
}
unit WDosNetBIOS;

interface

uses
  SysUtils, WDosUtil, WDosClock, WDosSock, WDosIP, WDosNetwork;

type
  TNetBIOSName = array[1..34] of char;

function NetBIOS_GetIPAddress(Name: AnsiString): TIPAddress;
procedure NetBIOS_MakeName(Name: AnsiString; _Type: BYTE; var NBName:
  TNetBIOSName);

implementation

const
  NS_PORT = 137;

type
  PWord  = ^Word;
  PDWORD = ^DWORD;

  TNBNameHeader = packed record
    TransactionID: WORD;
    Flags: WORD;
    QuestionCount: WORD;
    AnswerCount: WORD;
    AuthCount: WORD;
    AdditionalCount: WORD;
  end;

  TNBNameQuery = packed record
    Header: TNBNameHeader;
    NetBIOSName: TNetBIOSName;
    QuestionType: WORD;
    QuestionClass: WORD;
  end;

  TIPEntry = packed record
    Flags: WORD;
    IPAddress: TIPAddress;
  end;

  TNBNameAnswer = packed record
    Header: TNBNameHeader;
    NetBIOSName: TNetBIOSName;
    ReplyType: WORD;
    ReplyClass: WORD;
    TTL: DWORD;
    DataLen: WORD;
    IPAddress: array[1..10] of TIPEntry;
  end;

  TNBRecord = packed record
    Header: TNBNameHeader;
    Data: array[1..8192] of char;
  end;

  TIPArray = array[1..100] of TIPAddress;

  TNBCacheEntry = record
    NetBIOSName: string[15];
    IPAddress: TIPAddress;
    Timeout: Int64;
  end;

var
  NetBIOS_Started: Boolean;
  NetBIOS_Cache: array[1..100] of TNBCacheEntry;

function Cache_Get(Name: AnsiString): TIPAddress;
var
  i: integer;
begin
  Name := UpperCase(Name);
  for i := Low(NetBIOS_Cache) to High(NetBIOS_Cache) do
  begin
    if NetBIOS_Cache[i].Timeout < GetMSCounter then
      NetBIOS_Cache[i].NetBIOSName := '';
    if NetBIOS_Cache[i].NetBIOSName = Name then
    begin
      Result := NetBIOS_Cache[i].IPAddress;
      Exit;
    end;
  end;
  Result := 0;
end;

procedure Cache_Set(Name: AnsiString; IP: TIPAddress);
var
  i, entry: Integer;
begin
  // Find an unused entry in the cache
  entry := -1;
  Name := UpperCase(Name);
  for i := Low(NetBIOS_Cache) to High(NetBIOS_Cache) do
  begin
    if NetBIOS_Cache[i].Timeout < GetMSCounter then
      NetBIOS_Cache[i].NetBIOSName := '';
    if NetBIOS_Cache[i].NetBIOSName = '' then
    begin
      entry := i;
      break;
    end;
  end;
  if entry = -1 then
    entry := Random(High(NetBIOS_Cache)) + 1;
  NetBIOS_Cache[entry].NetBIOSName := Name;
  NetBIOS_Cache[entry].IPAddress := IP;
  NetBIOS_Cache[entry].Timeout := GetMSCounter + 60000; // 1 minute
end;

procedure NetBIOS_MakeName(Name: AnsiString; _Type: BYTE; var NBName:
  TNetBIOSName);
var
  i: Integer;
begin
  while Length(Name) < 15 do
    Name := Name + ' ';
  SetLength(Name, 15);
  Name := Name + Chr(_Type);
  FillChar(NBName, SizeOf(NBName), 0);
  NBName[1] := #32;
  for i := 1 to 16 do
  begin
    NBName[2 * i] := Chr(65 + Ord(Name[i]) shr 4);
    NBName[2 * i + 1] := Chr(65 + Ord(Name[i]) and $0F);
  end;
end;

function GetIPAddresses(Resp: TNBRecord; var IPs: TIPArray): Integer;
var
  P, PData: PChar;
  nResp, Len, i: Integer;
  RRType, RRClass, DataLen: WORD;
begin
  Result := 0;
  P := @Resp.Data;
  for nResp := 1 to Swap(Resp.Header.AnswerCount) do
  begin
    // Skip the Netbios Names
    repeat
      Len := Byte(P[0]);
      Inc(P);
      //          WriteLn('Len: ', Len);
      if Len = 0 then
        Break;
      Inc(P, Len);
    until False;
    // Read RRType, RRClass, TTL, and DataLen
    RRType := Swap(PWORD(P)^);
    Inc(P, 2);
    RRClass := Swap(PWORD(P)^);
    Inc(P, 2);
    DataLen := Swap(PWORD(P)^);
    Inc(P, 2);
    //      WriteLn(RRType, ' ', RRClass, ' ', TTL, ' ', DataLen);
            // Read Data
    if (RRType = $20) and (RRClass = 1) then
    begin
      PData := P;
      for i := 1 to DataLen div 6 do
      begin
        Inc(Result);
        Inc(PData, 2); // Skip Flags
        Move(PData^, IPs[Result], 4);
        //              WriteLn(IP_AddrToString(IPs[Result]));
        Inc(PData, 4);
      end;
    end;
    // Skip Data
    Inc(P, DataLen);
  end;
end;

function SendReceive(HSocket: Integer; const Query: TNBNameQuery; var Response:
  TNBRecord): Boolean;
var
  i: integer;
  RcvLen: DWORD;
  Timeout: Int64;
begin
  // 3 retries, with one second between each
  for i := 1 to 3 do
  begin
    send(HSocket, (@Query)^, SizeOf(Query), 0);
    Timeout := GetMSCounter + 1000;
    while GetMSCounter < Timeout do
    begin
      ioctlsocket(HSocket, FIONREAD, @RcvLen);
      if (RcvLen > 0) then
      begin
        FillChar(Response, SizeOf(Response), 0);
        recv(HSocket, Response, RcvLen, 0);
        if (Response.Header.TransactionID = Query.Header.TransactionID)
          and (Response.Header.AnswerCount > 0) then
        begin
          Result := True;
          Exit;
        end;
      end;
    end;
  end;
  Result := False;
end;

function GetLocalIP(const IPs: TIPArray; Count: Integer): TIPAddress;
var
  i, j: Integer;
  Entry: TAddressEntry;
begin
  for i := 1 to Count do
  begin
    Result := IPs[i];
    j := 0;
    while IP_GetInfo(j, Entry) do
    begin
      if (Entry.IPAddress and Entry.NetMask) =
        (Result and Entry.NetMask) then
        Exit;
      inc(j);
    end;
  end;
  Result := 0;
end;

function NetBIOS_GetIPAddress(Name: AnsiString): TIPAddress;
var
  SockAddr: TSockAddr;
  Query: TNBNameQuery;
  Response: TNBRecord;
  IPAddresses: TIPArray;
  HSocket, Cnt: Integer;
begin
  Name := UpperCase(Name);
  { First, try the Cache }
  Result := Cache_Get(Name);
  if Result <> 0 then
    Exit;
  { Next, go to the network }
  Result := 0;
  HSocket := socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
  if HSocket <= 0 then
    Exit;
  { Connect to the "NetBIOS Name Services" port }
  SockAddr.sin_family := AF_INET;
  SockAddr.sin_addr.S_addr := INADDR_BROADCAST;
  SockAddr.sin_port := Swap(Word(NS_PORT));
  if Connect(HSocket, SockAddr, SizeOf(SockAddr)) = SOCKET_ERROR then
  begin
    CloseSocket(HSocket);
    Exit;
  end;
  { Create Name Query }
  FillChar(Query, SizeOf(Query), 0);
  Query.Header.TransactionID := Random($FFFF);
  Query.Header.Flags := Swap(WORD($0110)); { No recursion, broadcast }
  Query.Header.QuestionCount := Swap(WORD(1));
  NetBIOS_MakeName(Name, 0, Query.NetBIOSName);
  Query.QuestionType := Swap(WORD($0020)); { NetBIOS resource record }
  Query.QuestionClass := Swap(WORD($0001)); { Internet Class }
  { Send the request, and wait for a reply }
  if SendReceive(HSocket, Query, Response) then
  begin
    { Process Response }
    Cnt := GetIPAddresses(Response, IPAddresses);
    { Return a local IP address is possible, since the }
    { remote machine may have multiple IP addresses.   }
    if Cnt > 0 then
    begin
      Result := GetLocalIP(IPAddresses, Cnt);
      if Result = 0 then
        Result := IPAddresses[1];
      Cache_Set(Name, Result);
    end;
  end;
  { Done }
  CloseSocket(HSocket);
end;

function IsOurNameQuery(const Query: TNBRecord): Boolean;
var
  P: PChar;
  nQuery: Integer;
  NBName: TNetBIOSName;
begin
  Result :=
    ((Swap(Query.Header.Flags) and $F800) = $0000) // Name Query
  and (Swap(Query.Header.QuestionCount) > 0);
  if not Result then
    Exit;
  // Check the Queries
  NetBIOS_MakeName(UpperCase(Net_GetHostName), 0, NBName);
  P := @Query.Data;
  for nQuery := 1 to Swap(Query.Header.QuestionCount) do
  begin
    if Copy(StrPas(P), 2, 16) = Copy(StrPas(@NBName), 2, 16) then
    begin
      Result := True;
      Exit;
    end;
  end;
  Result := False;
end;

var
  NameSock: Integer;

procedure NameAnswerProc;
var
  i: Integer;
  Buffer: TNBRecord;
  Reply: TNBNameAnswer;
  Entry: TAddressEntry;
  DataLen, RcvLen: DWORD;
begin
  ioctlsocket(NameSock, FIONREAD, @RcvLen);
  if RcvLen > 0 then
  begin
    recv(NameSock, Buffer, RcvLen, 0);
    if Net_GetHostName = '' then
      Exit;
    { If they're asking for our IP address, we reply }
    if IsOurNameQuery(Buffer) then
    begin
      Move(Buffer, Reply, SizeOf(Reply));
      Reply.Header.Flags := Swap(WORD($8500)); // Auth. response
      Reply.Header.QuestionCount := Swap(WORD(0));
      Reply.Header.AnswerCount := Swap(WORD(1));
      Reply.Header.AuthCount := Swap(WORD(0));
      Reply.Header.AdditionalCount := Swap(WORD(0));
      Reply.TTL := SwapDWORD(300000);
      DataLen := 0;
      i := 0;
      while IP_GetInfo(i, Entry) do
      begin
        Inc(i);
        Inc(DataLen, 6);
        Reply.IPAddress[i].Flags := 0;
        Reply.IPAddress[i].IPAddress := Entry.IPAddress;
      end;
      Reply.DataLen := Swap(Word(DataLen));
      { Calculate the final size of the packet to be sent }
      DataLen := SizeOf(Reply.Header) + SizeOf(Reply.NetBIOSName) +
        SizeOf(Reply.ReplyType) + SizeOf(Reply.ReplyClass) +
        SizeOf(Reply.TTL) + SizeOf(Reply.DataLen) + DataLen;
      { Send the packet }
      send(NameSock, Reply, DataLen, 0);
    end;
  end;
end;

function NetBIOS_Init: Boolean;
var
  SockAddr: TSockAddr;
begin
  if not NetBIOS_Started then
  begin
    FillChar(NetBIOS_Cache, SizeOf(NetBIOS_Cache), 0);
    NameSock := socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
    { Bind to the "NetBIOS Name Services" port }
    SockAddr.sin_family := AF_INET;
    SockAddr.sin_addr.S_addr := INADDR_ANY;
    SockAddr.sin_port := Swap(Word(NS_PORT));
    bind(NameSock, SockAddr, SizeOf(SockAddr));
    { Connect to the "NetBIOS Name Services" port }
    SockAddr.sin_family := AF_INET;
    SockAddr.sin_addr.S_addr := INADDR_BROADCAST;
    SockAddr.sin_port := Swap(Word(NS_PORT));
    connect(NameSock, SockAddr, SizeOf(SockAddr));
    { Add Idle Proc }
    Net_AddIdleProc(NameAnswerProc);
    { Done }
    NetBIOS_Started := True;
  end;
  Result := True;
end;

function NetBIOS_Deinit: Boolean;
begin
  if NetBIOS_Started then
  begin
    if NameSock <> SOCKET_ERROR then
      CloseSocket(NameSock);
    Net_RemoveIdleProc(NameAnswerProc);
    NetBIOS_Started := False;
  end;
  Result := True;
end;

procedure NetChanged;
begin
  { The network changed. We should try to restart NetBIOS }
  NetBIOS_DeInit;
  NetBIOS_Init;
end;

initialization
  NetBIOS_Started := False;
  Net_AddNetChangedProc(NetChanged);
  
finalization
  NetBIOS_Deinit;
  Net_RemoveNetChangedProc(NetChanged);
end.

