
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ TCP/IP Stack for native DOS Packet drivers            }
{ WDosUDP.pas                                           }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.2.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   wwww:      http://www.dwp42.org                     }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosUDP.pas,v 1.3 2003/10/25 22:00:07 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosUDP.pas,v $
  Revision 1.3  2003/10/25 22:00:07  iwache
  Minor changes in coding format added. No functional change.

  Revision 1.2  2003/06/11 20:55:56  iwache
  New function UDP_Select added for compatibility
  with unit Sockets.pas.

  Revision 1.1  2003/05/08 18:59:57  iwache
  TCP/IP Stack for native DOS Packet drivers by Diego De Marco added.

  ----------------------------------------------------------------------------
}
unit WDosUDP;

interface

uses
  Classes, WDosNetwork, WDosSock, WDosIP;

function UDP_Init: Boolean;
function UDP_Deinit: Boolean;

{ Manejo de Sockets UDP }
function UDP_Socket(Domain, SocketType, Protocol: Integer): Pointer;
function UDP_Bind(Sock: Pointer; var Addr; AddrLen: Integer): Integer;
function UDP_Connect(Sock: Pointer; var Addr; Addrlen: Integer): Integer;
function UDP_Send(Sock: Pointer; var Data; DataLen, Flags: Integer): Integer;
function UDP_SendTo(Sock: Pointer; var Data; DataLen, Flags: Integer; var Addr;
  AddrLen: Integer): Integer;
function UDP_Recv(Sock: Pointer; var Data; DataLen, Flags: Integer): Integer;
function UDP_RecvFrom(Sock: Pointer; var Data; DataLen, Flags: Integer; var
  AddrFrom; var FromLen: Integer): Integer;
function UDP_Close(Sock: Pointer): Integer;
function UDP_GetLastError: Integer;
function UDP_ioctlsocket(Sock: Pointer; Option: Integer; d: Pointer): Integer;
function UDP_Listen(Sock: Pointer; MaxConnect: Integer): Integer;
function UDP_Accept(Sock: Pointer; var Addr; var Addrlen: Integer): Pointer;
function UDP_GetSocketName(Sock: Pointer; var Addr; var Addrlen: Integer):
  Integer;
function UDP_GetPeerName(Sock: Pointer; var Addr; var Addrlen: Integer):
  Integer;
function UDP_Shutdown(Sock: Pointer; How: Integer): Integer;
function UDP_GetSocketOptions(Sock: Pointer; Level, OptName: Integer; var
  OptVal; var OptLen: Integer): Integer;
function UDP_SetSocketOptions(Sock: Pointer; Level, OptName: Integer; var
  OptVal; OptLen: Integer): Integer;
function UDP_Select(Sock: Pointer; var ReadReady, WriteReady, ExceptFlag:
    Boolean): Integer;

implementation

type
  PUDPHeader = ^TUDPHeader;
  TUDPHeader = packed record
    SourcePort: WORD;
    TargetPort: WORD;
    MessageLength: WORD;
    Checksum: WORD;
  end;

  TUDPBuff = record
    RemoteIP: TIPAddress;
    RemotePort: Integer;
    BufferData: AnsiString;
  end;

  TUDPSocket = class
    LocalIP: TIPAddress;
    LocalPort: WORD;
    RemoteIP: TIPAddress;
    RemotePort: WORD;
    BufferCount: Integer;
    Buffers: array[1..10] of TUDPBuff;
  end;

var
  SockList: TList;
  LastError: Integer;

procedure ProcessIP(SourceIP, TargetIP: TIPAddress;
  var Data; DataLen: Word);
var
  i: Integer;
  Hdr: PUDPHeader;
  SockObj: TUDPSocket;
begin
  if DataLen <= SizeOf(TUDPHeader) then
    Exit;
  Hdr := @Data;
  if Hdr.TargetPort = 0 then
    Exit;
  { Lleg un datagrama UDP. Si algn socket lo quiere, lo colocamos }
  { en el buffer del socket correspondiente. Si no, lo descartamos. }
  for i := 0 to SockList.Count - 1 do
  begin
    SockObj := TUDPSocket(SockList[i]);
    if (SockObj.LocalPort = Hdr.TargetPort) then
    begin
      { Este es el socket que atiende este puerto local }
      if SockObj.BufferCount < High(SockObj.Buffers) then
      begin
        Inc(SockObj.BufferCount);
        with SockObj.Buffers[SockObj.BufferCount] do
        begin
          RemoteIP := SourceIP;
          RemotePort := Hdr.SourcePort;
          Dec(DataLen, SizeOf(TUDPHeader));
          SetLength(BufferData, DataLen);
          Move(PChar(@Data)[SizeOf(TUDPHeader)],
            BufferData[1], DataLen);
        end;
      end;
      Break;
    end;
  end;
(*
  { Debug }
  WriteLn('  *UDP*');
  WriteLn('   Source Port   : ', Swap(Hdr.SourcePort));
  WriteLn('   Target Port   : ', Swap(Hdr.TargetPort));
  WriteLn('   Message Length: ', Swap(Hdr.MessageLength));
  WriteLn('   Checksum      : ', IntToHex(Hdr.Checksum);
  Write  ('   Data          : ');
  { Tomamos los Datos }
  Dec(DataLen, SizeOf(TUDPHeader));
  PData := @Data;
  Inc(PData, SizeOf(TUDPHeader));
  for i := 0 to DataLen - 1 do Write(PData[i]);
  WriteLn('(fin)');
*)
end;

function GetFreePort: Integer;
var
  Resu: WORD;
  i, Prt: Integer;
begin
  for Prt := 1024 to 8192 do
  begin
    Resu := Prt;
    for i := 0 to SockList.Count - 1 do
    begin
      if TUDPSocket(SockList[i]).LocalPort = Prt then
      begin
        Resu := 0;
        break;
      end;
    end;
    if Resu <> 0 then
    begin
      GetFreePort := Swap(Resu);
      Exit;
    end;
  end;
  GetFreePort := 0;
end;

function UDP_Init: Boolean;
var
  PS: TProtocolStruct;
begin
  Result := IP_AddProcessor(IPPROTO_UDP, @ProcessIP);
  {}
  PS.Domain := PF_INET;
  PS.SocketType := SOCK_DGRAM;
  PS.Protocol := IPPROTO_IP;
  PS.FNSocket := UDP_Socket;
  PS.FNClose := UDP_Close;
  PS.FNGetLastError := UDP_GetLastError;
  PS.FNSend := UDP_Send;
  PS.FNSendTo := UDP_SendTo;
  PS.FNRecv := UDP_Recv;
  PS.FNRecvFrom := UDP_RecvFrom;
  PS.FNBind := UDP_Bind;
  PS.FNListen := UDP_Listen;
  PS.FNAccept := UDP_Accept;
  PS.FNConnect := UDP_Connect;
  PS.FNShutdown := UDP_Shutdown;
  PS.FNGetSocketName := UDP_GetSocketName;
  PS.FNGetPeerName := UDP_GetPeerName;
  PS.FNSetSockOpt := UDP_SetSocketOptions;
  PS.FNGetSockOpt := UDP_GetSocketOptions;
  PS.FNioctlsocket := UDP_ioctlsocket;
  PS.FNSelect := UDP_Select;
  Result := Result and AddProtocol(PS);
  PS.Protocol := IPPROTO_UDP;
  Result := Result and AddProtocol(PS);
end;

function UDP_GetLastError: Integer;
begin
  UDP_GetLastError := LastError;
end;

function UDP_Deinit: Boolean;
begin
  UDP_Deinit := IP_RemoveProcessor(IPPROTO_UDP);
end;

function UDP_Socket(Domain, SocketType, Protocol: Integer): Pointer;
var
  SockObj: TUDPSocket;
begin
  SockObj := TUDPSocket.Create;
  SockObj.LocalIP := 0;
  SockObj.LocalPort := 0;
  SockObj.RemoteIP := 0;
  SockObj.RemotePort := 0;
  SockObj.BufferCount := 0;
  SockList.Add(SockObj);
  Result := SockObj;
end;

function UDP_Bind(Sock: Pointer; var Addr; AddrLen: Integer): Integer;
var
  InAddr: TSockAddr;
  SockObj: TUDPSocket;
begin
  UDP_Bind := SOCKET_ERROR;
  SockObj := TUDPSocket(Sock);
  InAddr := TSockAddr(Addr);
  { Check the given Address }
  if InAddr.sin_family <> AF_INET then
  begin
    LastError := WSAEFAULT;
    Exit;
  end;
  { If it's necessary, we choose a free local port for the socket }
  if InAddr.sin_port = 0 then
    InAddr.sin_port := GetFreePort;
  { Bind the socket to the local IP and Address }
  SockObj.LocalIP := InAddr.sin_addr.S_addr;
  SockObj.LocalPort := InAddr.sin_port;
  { Done }
  UDP_Bind := 0;
end;

function UDP_Connect(Sock: Pointer; var Addr; Addrlen: Integer): Integer;
var
  InAddr: TSockAddr;
  SockObj: TUDPSocket;
begin
  SockObj := TUDPSocket(Sock);
  InAddr := TSockAddr(Addr);
  { Check the given Address }
  if InAddr.sin_family <> AF_INET then
  begin
    UDP_Connect := SOCKET_ERROR;
    LastError := WSAEFAULT;
    Exit;
  end;
  { If the socket is not bound to a local port, now it's the time to do it }
  if SockObj.LocalPort = 0 then
    SockObj.LocalPort := GetFreePort;
  { Bind the socket to the remote peer }
  SockObj.RemoteIP := InAddr.sin_addr.S_addr;
  SockObj.RemotePort := InAddr.sin_port;
  { Done }
  UDP_Connect := 0;
end;

function UDP_SendTo(Sock: Pointer; var Data; DataLen, Flags: Integer; var Addr;
  AddrLen: Integer): Integer;
var
  Resp: Integer;
  P: PUDPHeader;
  InAddr: TSockAddr;
  SockObj: TUDPSocket;
begin
  SockObj := TUDPSocket(Sock);
  InAddr := TSockAddr(Addr);
  { Create de UDP datagram }
  GetMem(P, DataLen + SizeOf(TUDPHeader));
  P.SourcePort := SockObj.LocalPort;
  P.TargetPort := InAddr.sin_port;
  P.MessageLength := Swap(Word(DataLen + SizeOf(TUDPHeader)));
  P.Checksum := 0;
  Move(Data, PChar(P)[SizeOf(TUDPHeader)], DataLen);
  { Send the UDP datagram }
  Resp := IP_Send(SockObj.LocalIP, InAddr.sin_addr.S_addr, IPPROTO_UDP, P^,
    DataLen + SizeOf(TUDPHeader));
  if Resp = 0 then
  begin
    Result := DataLen;
  end
  else
  begin
    Result := SOCKET_ERROR;
    LastError := Resp;
  end;
  FreeMem(P, DataLen + SizeOf(TUDPHeader));
end;

function UDP_Send(Sock: Pointer; var Data; DataLen, Flags: Integer): Integer;
var
  Addr: TSockAddr;
  SockObj: TUDPSocket;
begin
  SockObj := TUDPSocket(Sock);
  Addr.sin_family := AF_INET;
  Addr.sin_port := SockObj.RemotePort;
  Addr.sin_addr.S_addr := SockObj.RemoteIP;
  Result := UDP_SendTo(Sock, Data, DataLen, Flags, Addr, SizeOf(Addr));
end;

function UDP_RecvFrom(Sock: Pointer; var Data; DataLen, Flags: Integer; var
  AddrFrom; var FromLen: Integer): Integer;
var
  SockObj: TUDPSocket;
  i, BuffSize: Integer;
begin
  SockObj := TUDPSocket(Sock);

  { If there's nothing to be read, we return zero }
  if (SockObj.BufferCount = 0) then
  begin
    UDP_RecvFrom := 0;
    LastError := 0;
    Exit;
  end;

  { If the buffer supplied for is not large enough to hold the address, }
  { we return SOCKET_ERROR, setting LastError to WSAEMSGSIZE.           }
  if (FromLen <> SizeOf(TSockAddr)) then
  begin
    UDP_RecvFrom := SOCKET_ERROR;
    LastError := WSAEMSGSIZE;
    Exit;
  end;

  { If the supplied buffer is not large enough, we return SOCKET_ERROR, }
  { setting LastError to WSAEMSGSIZE.                                   }
  BuffSize := Length(SockObj.Buffers[1].BufferData);
  if DataLen < BuffSize then
  begin
    UDP_RecvFrom := SOCKET_ERROR;
    LastError := WSAEMSGSIZE;
    Exit;
  end;

  { Read data from the first available datagram, and copy it to the     }
  { user-supplied buffer.                                               }
  Move(SockObj.Buffers[1].BufferData[1], Data, BuffSize);
  TSockAddr(AddrFrom).sin_family := AF_INET;
  TSockAddr(AddrFrom).sin_addr.S_addr := SockObj.Buffers[1].RemoteIP;
  TSockAddr(AddrFrom).sin_port := SockObj.Buffers[1].RemotePort;

  { If necesary, we remove the datagram from the buffer queue           }
  if (Flags and MSG_PEEK) <> MSG_PEEK then
  begin
    SockObj.Buffers[1].BufferData := '';
    Dec(SockObj.BufferCount);
    for i := 1 to SockObj.BufferCount do
    begin
      SockObj.Buffers[i] := SockObj.Buffers[i + 1];
    end;
  end;

  { Done }
  UDP_RecvFrom := BuffSize;
  LastError := 0;
end;

function UDP_Recv(Sock: Pointer; var Data; DataLen, Flags: Integer): Integer;
var
  FromLen: Integer;
  AddrFrom: TSockAddr;
begin
  FromLen := SizeOf(AddrFrom);
  UDP_Recv := UDP_RecvFrom(Sock, Data, DataLen, Flags, AddrFrom, FromLen);
end;

function UDP_Close(Sock: Pointer): Integer;
var
  SockObj: TUDPSocket;
begin
  SockObj := TUDPSocket(Sock);
  SockList.Remove(SockObj);
  SockObj.Free;
  UDP_Close := 0;
  LastError := 0;
end;

function UDP_ioctlsocket(Sock: Pointer; Option: Integer; d: Pointer): Integer;
var
  dw: Cardinal;
  SockObj: TUDPSocket;
begin
  NET_Work_Low;
  UDP_ioctlsocket := 0;
  SockObj := TUDPSocket(Sock);
  case Option of
    FIONREAD:
      begin
        if (SockObj.BufferCount > 0) then
          dw := Length(SockObj.Buffers[1].BufferData)
        else
        begin
          dw := 0;
        end;
        Move(dw, d^, 4);
      end;
  else
    begin
      UDP_ioctlsocket := SOCKET_ERROR;
      LastError := WSAEINVAL;
    end;
  end;
end;

function UDP_Listen(Sock: Pointer; MaxConnect: Integer): Integer;
begin
  UDP_Listen := SOCKET_ERROR;
  LastError := WSAEOPNOTSUPP;
end;

function UDP_Accept(Sock: Pointer; var Addr; var Addrlen: Integer): Pointer;
begin
  UDP_Accept := nil;
  LastError := WSAEOPNOTSUPP;
end;

function UDP_Select(Sock: Pointer; var ReadReady, WriteReady, ExceptFlag:
    Boolean): Integer;
begin
  UDP_Select := SOCKET_ERROR;
  LastError := WSAEOPNOTSUPP;
end;

function UDP_GetSocketName(Sock: Pointer; var Addr; var Addrlen: Integer):
  Integer;
var
  SockObj: TUDPSocket;
begin
  SockObj := TUDPSocket(Sock);
  { Check to see if there's enough room in the destination buffer }
  if AddrLen < SizeOf(TSockAddr) then
  begin
    UDP_GetSocketName := SOCKET_ERROR;
    LastError := WSAEMSGSIZE;
    Exit;
  end;
  { Copy the local IP and Port to the destination buffer }
  TSockAddr(Addr).sin_family := AF_INET;
  TSockAddr(Addr).sin_addr.S_addr := SockObj.LocalIP;
  TSockAddr(Addr).sin_port := SockObj.LocalPort;
  AddrLen := SizeOf(TSockAddr);
  { Done }
  UDP_GetSocketName := 0;
end;

function UDP_GetPeerName(Sock: Pointer; var Addr; var Addrlen: Integer):
  Integer;
var
  SockObj: TUDPSocket;
begin
  SockObj := TUDPSocket(Sock);
  { Check to see if there's enough room in the destination buffer }
  if AddrLen < SizeOf(TSockAddr) then
  begin
    UDP_GetPeerName := SOCKET_ERROR;
    LastError := WSAEMSGSIZE;
    Exit;
  end;
  { Check to see if the socket is connected }
  if (SockObj.RemoteIP = 0) or (SockObj.RemotePort = 0) then
  begin
    UDP_GetPeerName := SOCKET_ERROR;
    LastError := WSAENOTCONN;
    Exit;
  end;
  { Copy the remote IP and Port to the destination buffer }
  TSockAddr(Addr).sin_family := AF_INET;
  TSockAddr(Addr).sin_addr.S_addr := SockObj.RemoteIP;
  TSockAddr(Addr).sin_port := SockObj.RemotePort;
  AddrLen := SizeOf(TSockAddr);
  { Done }
  UDP_GetPeerName := 0;
end;

function UDP_Shutdown(Sock: Pointer; How: Integer): Integer;
begin
  UDP_Shutdown := SOCKET_ERROR;
  LastError := WSAEINVAL;
end;

function UDP_GetSocketOptions(Sock: Pointer; Level, OptName: Integer; var
  OptVal; var OptLen: Integer): Integer;
begin
  UDP_GetSocketOptions := SOCKET_ERROR;
  LastError := WSAEINVAL;
end;

function UDP_SetSocketOptions(Sock: Pointer; Level, OptName: Integer; var
  OptVal; OptLen: Integer): Integer;
begin
  UDP_SetSocketOptions := SOCKET_ERROR;
  LastError := WSAEINVAL;
end;

initialization
  LastError := 0;
  SockList := TList.Create;

finalization
  { If there's any unclosed socket, close it now }
  while SockList.Count > 0 do
    UDP_Close(SockList[0]);
  { Free the socket collection }
  SockList.Free;
end.

