
{*******************************************************}
{                                                       }
{ DWPL Delphi Cross Platform Core Library (rtl)         }
{ TCP/IP Stack for native DOS Packet drivers            }
{ WDosUtil.pas                                          }
{                                                       }
{ for Delphi 4-7                                        }
{ Version 2.2.0                                         }
{ Copyright (c) 2003 by Diego De Marco                  }
{ e-mail: ddemarco@dpsautomation.com                    }
{                                                       }
{ Contributor(s):                                       }
{   Immo Wache immo.wache@dwp42.org                     }
{   wwww:      http://www.dwp42.org                     }
{                                                       }
{ This file may be distributed and/or modified under    }
{ the terms of the GNU General Public License (GPL)     }
{ version 2 as published by the Free Software           }
{ Foundation and appearing at                           }
{ http://www.dwp42.org/license/gpl.html.                }
{                                                       }
{*******************************************************}

{
  $Header: /cvsroot/dwpl/dwpl/source/rtl/wdosx/WDosUtil.pas,v 1.3 2003/10/25 22:04:37 iwache Exp $
  ----------------------------------------------------------------------------

  $Log: WDosUtil.pas,v $
  Revision 1.3  2003/10/25 22:04:37  iwache
  Conditional compilation settings for compatibilty
  with Delphi 5 added by use of the JEDI.INC include file.
  Procedure call Randomize in initialization section removed.

  Revision 1.2  2003/08/15 21:07:24  iwache
  Missing type PWord for D5 compatibility added.

  Revision 1.1  2003/05/08 18:59:57  iwache
  TCP/IP Stack for native DOS Packet drivers by Diego De Marco added.

  ----------------------------------------------------------------------------
}
unit WDosUtil;

{$INCLUDE WDOSX.INC}

{$IFDEF COMPILER6_UP}
  {$WARN SYMBOL_PLATFORM OFF}
{$ENDIF}

interface

uses
  Windows, Crt, SysUtils, WDOSDPMI;

{**************************************************************************
                        Constantes Pblicas
**************************************************************************}
const
  CR = #13;
  LF = #10;
  CRLF = CR + LF;
  fsFromBeginning = 0;
  fsFromEnd = 2;
  {**************************************************************************
                              Tipos Pblicos
  **************************************************************************}
type
  DWORD = Cardinal;
  PScreenBuffer = ^TScreenBuffer;
  TScreenBuffer = array[0..7999] of char;

  {**************************************************************************
                              Variables Pblicas
  **************************************************************************}

  {**************************************************************************
                              Funciones Pblicas
  **************************************************************************}
  {A}

  {B}
function BlockWriteEx(FileName: AnsiString; Offset: Integer; const Data:
  AnsiString): Boolean;

{C}
procedure ClearError;
function CurDir: string;

{D}
function Day(Fecha: TDateTime): Word;
function Del(Wildcard: string): integer;

{E}
function ExtMkDir(Dir: string): Boolean;

{F}
function FExt(s: string): string;
function FName(s: string): string;
function FormatShortDate(Fecha: TDateTime): string;
function FormatTime(Hora: TDateTime; ConSegundos: Boolean): string;
function FPath(s: string): string;
function FileToString(const Archivo: AnsiString): AnsiString;
function FileTruncate(Handle, Length: Integer): Boolean;

{G}
function GoodDir(s: string): string;
function GoodFileName(s, ext: string): string;

{H}
function Hour(d: TDateTime): Word;
function HexToInt(HexStr: AnsiString): DWORD;

{I}
function IsDir(s: string): Boolean;
function IStr(i: LongInt; ancho: integer): string;
function IVal(s: string): LongInt;

{J}

{L}

{M}
function Max(a, b: LongInt): LongInt;
function MemComp(p, q: PChar; largo: Word): Boolean;
procedure MemMove(Dest, Src: PChar; Count: Word);
function MemPos(p, q: PChar; largo: Word): PChar;
function Month(Fecha: TDateTime): Word;
procedure MoveToMyOwnDir;
function Minute(d: TDateTime): Word;

{P}
function PadC(s: string; ancho: Integer; caracter: char): string;
function PadL(s: string; ancho: Integer; caracter: char): string;
function PadR(s: string; ancho: Integer; caracter: char): string;
function ParseParamByNumber(Texto: AnsiString; NumParam: Integer; const
  Separador: AnsiString): string;
function Power(a, b: LongInt): LongInt;

{R}
function Replicate(caracter: char; veces: Integer): string;
function StrRight(s: string; cuantos: Integer): string;
function RStr(i: Real; Ancho, Dec: Integer): string;
function RVal(s: string): Real;
procedure RestoreScreen(Scrn: TScreenBuffer);

{S}
function Space(veces: Integer): string;
function StrLeft(s: string; cuantos: Integer): string;
function Second(d: TDateTime): Word;
procedure SaveScreen(var Scrn: TScreenBuffer);
function SwapDWORD(D: DWORD): DWORD;
function StringToFile(const Datos, Archivo: AnsiString): Boolean;

{T}
function TempFile: string;

{U}

{W}

{X}

{Y}
function Year(Fecha: TDateTime): Word;

{XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX}
implementation
{XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX}
{**************************************************************************
     Variables Pblicas para la Unidad
**************************************************************************}
const
  Numeros: set of char = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '.',
    '-'];

var
  VideoBase: Pointer;

  {A}

  {B}

function BlockWriteEx(FileName: AnsiString; Offset: Integer; const Data:
  AnsiString): Boolean;
var
  H: LongInt;
  i: Integer;
  Copia: AnsiString;
begin
  if Length(Data) = 0 then
  begin
    Result := True;
    Exit;
  end;

  // Intentamos grabar un mximo de 20 veces
  for i := 1 to 20 do
  begin
    // Grabamos
    H := FileOpen(FileName, fmOpenReadWrite);
    if H = -1 then
    begin
      Result := False;
      Exit;
    end;
    FileSeek(H, Offset, fsFromBeginning);
    if FileWrite(H, Data[1], Length(Data)) <> Length(Data) then
    begin
      FileClose(H);
      Result := False;
      Exit;
    end;
    FileClose(H);
    // Leemos para comprobar
    H := FileOpen(FileName, fmOpenReadWrite);
    FileSeek(H, Offset, fsFromBeginning);
    SetLength(Copia, Length(Data));
    FileRead(H, Copia[1], Length(Data));
    FileClose(H);
    if Copia = Data then
    begin
      Result := True;
      Exit;
    end;
  end;
  Result := False;
end;

{C}

procedure ClearError;
begin
  if IOResult <> 0 then
    ;
end;

function CurDir: string;
var
  s: string;
begin
  GetDir(0, s);
  CurDir := GoodDir(s);
end;

{D}

function Day(Fecha: TDateTime): Word;
var
  y, m, d: WORD;
begin
  DecodeDate(Fecha, y, m, d);
  Day := d;
end;

function Del(wildcard: string): integer;
var
  camino: string;
  Archivo: TSearchRec;
  Resu, Count: Integer;
begin
  Count := 0;
  Camino := FPath(wildcard);
  Resu := FindFirst(WildCard, faArchive, Archivo);
  while Resu = 0 do
  begin
    if DeleteFile(Camino + Archivo.name) then
      Inc(Count);
    Resu := FindNext(Archivo);
  end;
  FindClose(Archivo);
  Del := Count;
end;

{E}

function ExtMkDir(Dir: string): Boolean;
var
  a: string;
  s: string;
  n: Integer;
begin
  s := Trim(Dir);
  a := '';
  repeat
    n := Pos('\', s);
    if n = 0 then
      a := a + s + ' '
    else
      a := a + StrLeft(s, n);
    s := Copy(s, n + 1, 255);
{$I-}
    MkDir(StrLeft(a, Length(a) - 1));
{$I+}
    ClearError;
  until n = 0;
  ExtMkDir := IsDir(Dir);
end;

{F}

function FName(s: string): string;
begin
  FName := ExtractFileName(s);
end;

function FExt(s: string): string;
begin
  FExt := ExtractFileExt(s);
end;

function FormatShortDate(Fecha: TDateTime): string;
var
  y, m, d: WORD;
begin
  DecodeDate(Fecha, y, m, d);
  FormatShortDate :=
    PadL(IntToStr(d), 2, '0') + '/' +
    PadL(IntToStr(m), 2, '0') + '/' +
    PadL(IntToStr(y), 4, '0');
end;

function FormatTime(Hora: TDateTime; ConSegundos: Boolean): string;
var
  ss: string[20];
  h, m, s, ms: WORD;
begin
  DecodeTime(Hora, h, m, s, ms);
  ss :=
    PadL(IntToStr(h), 2, '0') + ':' +
    PadL(IntToStr(m), 2, '0');
  if ConSegundos then
    ss := ss + ':' +
      PadL(IntToStr(s), 2, '0');
  FormatTime := ss;
end;

function FPath(s: string): string;
begin
  FPath := GoodDir(ExtractFilePath(s));
end;

function FileToString(const Archivo: AnsiString): AnsiString;
var
  Handle: Integer;
begin
  Result := '';
  Handle := FileOpen(Archivo, fmOpenread);
  if (Handle < 0) then
    Exit;
  SetLength(Result, FileSeek(Handle, 0, 2));
  if Length(Result) > 0 then
  begin
    FileSeek(Handle, 0, 0);
    FileRead(Handle, Result[1], Length(Result));
  end;
  FileClose(Handle);
end;

function FileTruncate(Handle, Length: Integer): Boolean;
begin
  Result :=
    (FileSeek(Handle, Length, 0) = Length)
    and SetEndOfFile(Handle);
end;

{G}

function GoodDir(s: string): string;
begin
  if s = '' then
    GoodDir := ''
  else
  begin
    s := Trim(s);
    if StrRight(s, 1) = '\' then
      GoodDir := s
    else
      GoodDir := s + '\';
  end;
end;

function GoodFileName(s, ext: string): string;
begin
  if s = '' then
    GoodFileName := ''
  else if Pos('.', s) = 0 then
  begin
    if Pos('.', ext) = 0 then
      ext := '.' + ext;
    GoodFileName := s + ext;
  end
  else
  begin
    GoodFileName := s;
  end;
end;

{H}

function Hour(d: TDateTime): Word;
var
  h, m, s, ms: WORD;
begin
  DecodeTime(d, h, m, s, ms);
  Hour := h;
end;

function HexToInt(HexStr: AnsiString): DWORD;

  function DigitoHexa(H: Char): Integer;
  begin
    case UpCase(h) of
      '0'..'9': Result := Ord(h) - 48;
      'A'..'F': Result := Ord(h) - 55;
    else
      Result := -1;
    end;
  end;

var
  d, i: Integer;
  Factor: DWORD;
begin
  Result := 0;
  Factor := 1;
  for i := Length(HexStr) downto 1 do
  begin
    // El dgito que estamos analizando es vlido ?
    d := DigitoHexa(HexStr[i]);
    if d = -1 then
    begin
      Result := 0;
      Exit;
    end;
    // Sumamos al total
    Result := Result + (Factor * WORD(d));
    // Pasamos al dgito siguiente
    Factor := Factor * $10;
  end;
end;

function Minute(d: TDateTime): Word;
var
  h, m, s, ms: WORD;
begin
  DecodeTime(d, h, m, s, ms);
  Minute := m;
end;

function Second(d: TDateTime): Word;
var
  h, m, s, ms: WORD;
begin
  DecodeTime(d, h, m, s, ms);
  Second := s;
end;

{I}

function IsDir(s: string): Boolean;
var
  f: file;
begin
  Assign(f, GoodDir(s) + TempFile);
{$I-}
  ReWrite(f);
  Close(f);
  Erase(f);
{$I+}
  IsDir := (IOResult = 0);
end;

function IStr(i: LongInt; ancho: integer): string;
var
  s: string;
begin
  if i = 0 then
    IStr := Space(ancho)
  else
  begin
    Str(i, s);
    if ancho = 0 then
      IStr := s
    else
      IStr := PadL(s, ancho, ' ');
  end;
end;

function IVal(s: string): LongInt;
var
  i: Byte;
  Code: integer;
  Resu: LongInt;
  ss: string;
begin
  s := Trim(s);
  ss := '';
  for i := 1 to Length(s) do
  begin
    if s[i] in numeros then
      ss := ss + s[i]
    else
      break;
  end;
  Val(ss, resu, code);
  IVal := resu;
end;

{J}

{L}

{M}

function Max(a, b: LongInt): LongInt;
begin
  if a > b then
    Max := a
  else
    Max := b;
end;

function MemComp(p, q: PChar; largo: Word): Boolean;
var
  i: Word;
begin
  MemComp := true;
  for i := 0 to largo - 1 do
    if p[i] <> q[i] then
    begin
      MemComp := False;
      break;
    end;
end;

procedure MemMove(Dest, Src: PChar; Count: Word);
begin
  Move(Src^, Dest^, Count);
end;

function MemPos(p, q: PChar; largo: Word): PChar;
var
  i: Word;
  Resp: PChar;
begin
  i := 0;
  Resp := nil;
  while (Resp = nil) and (i < largo) do
  begin
    Resp := StrPos(p + i, q);
    i := i + StrLen(p) + 1;
  end;
  MemPos := resp;
end;

function Month(Fecha: TDateTime): Word;
var
  y, m, d: WORD;
begin
  DecodeDate(Fecha, y, m, d);
  Month := m;
end;

procedure MoveToMyOwnDir;
var
  s: string;
begin
  s := FPath(ParamStr(0));
  if StrRight(s, 1) = '\' then
    s := StrLeft(s, Length(s) - 1);
  if Copy(ParamStr(0), 2, 1) = ':' then
    ChDir(s); {Me cambio al Directorio}
end;

{P}

function PadC(s: string; ancho: Integer; caracter: char): string;
begin
  s := PadL(s, (ancho + Length(s)) div 2, caracter);
  PadC := PadR(s, ancho, caracter);
end;

function PadL(s: string; ancho: Integer; caracter: char): string;
begin
  if Length(s) <= ancho then
  begin
    s := Replicate(caracter, ancho - Length(s)) + s
  end
  else
  begin
    s := StrRight(s, ancho);
  end;
  PadL := s;
end;

function PadR(s: string; ancho: Integer; caracter: char): string;
begin
  if Length(s) <= ancho then
  begin
    s := s + replicate(caracter, ancho - Length(s))
  end
  else
  begin
    s := StrLeft(s, ancho);
  end;
  PadR := s;
end;

function ParseParamByNumber(Texto: AnsiString; NumParam: Integer; const
  Separador: AnsiString): string;
var
  n, i: Integer;
begin
  n := 1;
  Result := '';
  while n <> NumParam do
  begin
    i := Pos(Separador, Texto);
    if i = 0 then
      Exit;
    Texto := TrimLeft(Copy(Texto, i + Length(Separador), Length(Texto)));
    Inc(n);
  end;
  i := Pos(Separador, Texto);
  if i > 0 then
    Texto := Copy(Texto, 1, i - 1);
  Result := Trim(Texto);
end;

function Power(a, b: LongInt): LongInt;
var
  cont, resp: LongInt;
begin
  if b < 0 then
  begin
    Power := 0;
  end
  else if b = 0 then
  begin
    Power := 1;
  end
  else
  begin
    resp := 1;
    for cont := 1 to b do
    begin
      resp := resp * a;
    end;
    power := resp;
  end;
end;

{R}

function Replicate(caracter: char; veces: Integer): string;
begin
  if veces = 0 then
  begin
    Result := '';
  end
  else
  begin
    SetLength(Result, veces);
    FillChar(Result[1], Length(Result), caracter);
  end;
end;

procedure RestoreScreen(scrn: TScreenBuffer);
begin
  Move(Scrn, VideoBase^, SizeOf(Scrn));
end;

function RStr(i: Real; Ancho, Dec: Integer): string;
var
  s: string;
begin
  if i = 0 then
    RStr := Space(Ancho)
  else
  begin
    if Ancho = 0 then
      RStr := IStr(Trunc(i), 0)
    else
    begin
      Str(i: Ancho: Dec, s);
      RStr := s;
    end;
  end;
end;

function RVal(s: string): Real;
var
  i: Byte;
  Resu: Real;
  ss: string;
  Code: integer;
begin
  s := Trim(s);
  ss := '';
  for i := 1 to Length(s) do
  begin
    if s[i] in numeros then
      ss := ss + s[i]
    else
      break;
  end;
  Val(ss, Resu, code);
  RVal := Resu;
end;

{S}

function Space(veces: Integer): string;
begin
  Space := Replicate(' ', veces);
end;

procedure SaveScreen(var Scrn: TScreenBuffer);
begin
  Move(VideoBase^, Scrn, SizeOf(Scrn));
end;

function StrLeft(s: string; cuantos: Integer): string;
begin
  if Cuantos <= 0 then
    StrLeft := ''
  else
  begin
    if Length(s) > cuantos then
      s := Copy(s, 1, cuantos);
    StrLeft := s;
  end;
end;

function StrRight(s: string; cuantos: Integer): string;
begin
  StrRight := Copy(s, Length(s) - Cuantos + 1, 255);
end;

function SwapDWORD(D: DWORD): DWORD;
var
  r1, r2: packed array[1..4] of BYTE;
begin
  Move(D, r1, 4);
  r2[1] := r1[4];
  r2[2] := r1[3];
  r2[3] := r1[2];
  r2[4] := r1[1];
  Move(r2, D, 4);
  SwapDWORD := D;
end;

function StringToFile(const Datos, Archivo: AnsiString): Boolean;
const
  BLOCKSIZE = 8192;
var
  Bloque: AnsiString;
  Handle, Offset: Integer;
begin
  Result := False;
  // Intentamos crear el archivo de salida
  Handle := FileCreate(Archivo);
  if (Handle <= 0) then
    Exit;
  FileClose(Handle);
  // Vamos copiando los datos dentro del archivo
  Offset := 0;
  while Offset < Length(Datos) do
  begin
    Bloque := Copy(Datos, Offset + 1, BLOCKSIZE);
    if not BlockWriteEx(Archivo, Offset, Bloque) then
      Exit;
    Inc(Offset, BLOCKSIZE);
  end;
  // Listo
  Result := True;
end;

{T}

function TempFile: string;
var
  cont, n: byte;
  s: string;
begin
  s := '';
  for cont := 1 to 8 do
  begin
    n := Random(9);
    s := s + chr(48 + n);
  end;
  TempFile := s;
end;

{X}

{Y}

function Year(Fecha: TDateTime): Word;
var
  y, m, d: WORD;
begin
  DecodeDate(Fecha, y, m, d);
  Year := y;
end;

{XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
El Cuerpo de la Base de la Unidad
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX}
var
  Regs: DpmiRmRegisters;
  
initialization
  DpmiRealModeInt($11, Regs);
  if (Regs.ax and 48) = 48 then
    VideoBase := Pointer($B0000)
  else
    VideoBase := Pointer($B8000);

finalization
end.

